/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        emul_offload.h

    Purpose:

        This file contains the declarations for structures used in 
        emul_offload.c
--*/
#ifdef BE_EMULATION

typedef struct _EMUL_OFFLOAD_PARAMS
{
    struct {
        unsigned short  dst_hi16;
        unsigned int    dst_lo32;
        unsigned int    src_hi32;
        unsigned short  src_lo16;
    }eth;

    struct {
        unsigned int    src;
        unsigned int    dst;
        unsigned int    ttl;
        unsigned int    id;
    }ip;

    struct {
        unsigned int    srcport;
        unsigned int    dstport;
        unsigned int    window_size;
        unsigned int    hdr_len;
        unsigned long   seq;
        unsigned long   ack;
        unsigned int    flags;
    }tcp;

    struct {
        unsigned int    hde;
        unsigned int    dde;
    }iscsi;
}
EMUL_OFFLOAD_PARAMS, *PEMUL_OFFLOAD_PARAMS;

#define OPCODE_EMUL_OFFLOAD 0x90
#define SUBSYSTEM_EMUL_OFFLOAD 0x90

ACIT_UEXACT32
emul_init_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
emul_hwi_init(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT16
emul_get_cid(
    PACIT_CONTROLLER_WS     pcontroller,
    PEMUL_OFFLOAD_PARAMS    offload_params);

BESTATUS
emul_post_mcc_wrb (
    PACIT_CONTROLLER_WS     pcontroller, 
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    MCC_WRB_CQE_CALLBACK    callback_ptr);

ACIT_UEXACT32
emul_mgmt_init_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT16           connection_handle,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           cq_id,
    ACIT_UEXACT16           defq_id,
    PPHYS_ADDR              ptemplate_address,
    ACIT_UEXACT32           sync);

MGMT_STATUS
emul_mgmt_process_offload(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

#define MAX_R2T_DESC        (16)
#define R2T_DESC_SIZE       (32*BE_PAGE_SIZE)
#define R2T_ELEMENT_SIZE    (16)

#endif
