/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        acitmain.h

    Purpose:

        Header file for declaring all data structures, definitions and 
        functions exported by ACIT library for internal use.
        

--*/

#if defined(NUMBERED_BUILD) || defined(BRANCH_BUILD)
    #define ENABLE_VERSION_CHECK   1
#else
    #define ENABLE_VERSION_CHECK   0
#endif

#define MIN_ETO     (20)
#define MIN_LDTO    (0)
#define MAX_ETO     (3600)
#define MAX_LDTO    (3600)

/* Flow Control definitions */
#define CONTROLLER_BUSY_IO_HANDLE                       (1U << 0)
#define CONTROLLER_BUSY_IO_SGL_HANDLE                   (1U << 1)
#define CONTROLLER_BUSY_EH_SGL_HANDLE                   (1U << 2)
#define CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS            (1U << 3)
#define CONTROLLER_BUSY_ALL_SESSIONS_RECOVERY           (1U << 4)
#define CONTROLLER_BUSY_LINK_DOWN                       (1U << 5)
#define CONTROLLER_BUSY_LDTO_EXPIRED                    (1U << 6)
#define CONTROLLER_BUSY_UE                              (1U << 7)

#define TARGET_BUSY_ISCSI_WINDOW                        (1U << 0)
#define TARGET_BUSY_ISCSI_IO_CREDITS                    (1U << 1)
#define TARGET_BUSY_SESSION_RECOVERY                    (1U << 2)
#define TARGET_BUSY_SESSION_LOGOUT                      (1U << 3)
#define TARGET_BUSY_TMF_RESET_IN_PROGRESS               (1U << 4)
#define TARGET_BUSY_WRB_HANDLE                          (1U << 5)
#define TARGET_BUSY_NEW_CONNECTION_ADD                  (1U << 6)
#define TARGET_BUSY_REMOVE_CONNECTION                   (1U << 7)
#define TARGET_BUSY_ETO_EXPIRED                         (1U << 8)
#define TARGET_BUSY_UNTAGGED_CMD                        (1U << 9)
#define TARGET_BUSY_FLUSH_UNTAGGED_CMD                  (1U << 10)

#define CONNECTION_BUSY_WRB_HANDLE                      (1U << 0)

// needed when failover is not transparent to the driver
#ifdef NO_ARM_PORT_FAILOVER
#define CONNECTION_BUSY_PORT_A_DOWN                     (1U << 1)
#define CONNECTION_BUSY_PORT_B_DOWN                     (1U << 2)
#endif

#define LUN_BUSY_QUEUE_FULL                             (1U << 0)/* Not used */
#define LUN_BUSY_RESET_WAITING                          (1U << 1)
#define LUN_BUSY_RESET_WAIT_FOR_RESOURCE                (1U << 2)
#define LUN_BUSY_RESET_IN_PROGRESS                      (1U << 3)
#define LUN_BUSY_RESET_SUBMITTED                        (1U << 4)

/* Debug flags available as config parameter */
#define ACIT_DBG_FLAGS_DUMP_PENDING_CMDS                (1U << 0)
#define ACIT_DBG_FLAGS_DUMP_CMD_HISTORY                 (1U << 1)
#define ACIT_DBG_FLAGS_VALIDATE_PENDING_IOS             (1U << 2)

/* This defines what queues the handles must be purged from */
#define LUN_QUEUE                                       0
#define TARGET_QUEUE                                    1
#define CONTROLLER_QUEUE                                2

#define ACIT_IS_CONTROLLER_LINK_DOWN(pcontroller, port)   \
    ((pcontroller->link_down[port] == ACIT_TRUE))

#define ACIT_IS_CONTROLLER_LINK_UP(pcontroller, port)   \
    ((pcontroller->link_down[port] == ACIT_FALSE))

#define STRUCT_FIELD_OFFSET(struct_type,field_name) \
            ((ACIT_PTR)((PACIT_VOID)(&(((struct_type *)0)->field_name))))

#define STRUCT_BASE(struct_type, field_name, field_ptr) \
            ((struct_type *)((ACIT_UEXACT8 *)(field_ptr)\
            - ((PACIT_UEXACT8)(&(((struct_type *)0)->field_name)))))

#define CURRENT_TICK(pc)    pc->counters.current_tick_count

#define LINK_INVALID    (ACIT_NULL)

/*
 * Flags common across all handles
 */
#define FLAG_INVALIDATE_PENDING                     (1U << 30)
#define FLAG_INVALIDATE_COMPLETE                    (1U << 29)

/*
 * Define flags for ACIT_TMF_HANDLE.
 */
#define FLAG_TMFH_TMF_STARTED                       (1U << 0)
#define FLAG_TMFH_TMF_TIMEDOUT                      (1U << 1)

/*
 * Define flags for ACIT_IO_HANDLE.
 */
#define FLAG_IOH_IO_FREE                            (1U << 0)
#define FLAG_IOH_IO_SUBMITTED                       (1U << 1)
#define FLAG_IOH_IO_COMPLETION                      (1U << 2)
#define FLAG_IOH_TMF_ABORT                          (1U << 3)
#define FLAG_IOH_CQ_ERROR                           (1U << 4)
#define FLAG_IOH_REJECTED_BY_TARGET                 (1U << 5)
#define FLAG_IOH_TMF_ABORT_QUEUED                   (1U << 6)
#define FLAG_IOH_INVALIDATE_PENDING                 (FLAG_INVALIDATE_PENDING)
#define FLAG_IOH_INVALIDATE_COMPLETE                (FLAG_INVALIDATE_COMPLETE)

#define ACIT_ALIGN(val, mask) (val = (val+mask) & ~mask)

STATIC INLINE
ACIT_UEXACT32
ap_cmdh_flag_from_inv_status(
    ACIT_UEXACT32           inv_status)
{
    switch (inv_status)
    {
    case ACIT_STATUS_SUCCESS:
    case ACIT_STATUS_INVALID_PARAMETER:
        return FLAG_INVALIDATE_COMPLETE;
    }

    OSM_ASSERT(0);
    return FLAG_INVALIDATE_COMPLETE;
}

/*
 * Define flags for ACIT_TARGET_WS
 */
#define EVENT_TGT_CXN_EVENT_PENDING                 (1U << 0) 
#define EVENT_TGT_WS_INIT_PENDING                   (1U << 1)
#define EVENT_TGT_REOPEN_PENDING                    (1U << 2)
#define EVENT_TGT_LOGOUT_PENDING                    (1U << 3)
#define EVENT_TGT_ETO_IN_PROGRESS                   (1U << 4)
#define EVENT_TGT_PING_WAIT                         (1U << 5)
#define EVENT_TGT_RECOVERY_LOGOUT_PENDING           (1U << 6)

/* Target cleanup flags */
#define TGT_CLEANUP_FLAG_FOR_TMF_RESET              (1U << 1)
#define TGT_CLEANUP_FLAG_FOR_RECOVERY               (1U << 2)
#define TGT_CLEANUP_FLAG_FOR_LOGOUT                 (1U << 3)

#define TGT_FLAG_SKIP_TIMEOUT_TICKS_COUNTDOWN       (1U << 0)
#define TGT_FLAG_TMF_QUEUE_FULL                     (1U << 1)
#define TGT_FLAG_OFFLOAD_RETRY_EXCEEDED             (1U << 2)

/*
#define FLAG_TGT_UPLOAD_COMPLETE                    (1U << 12)
#define FLAG_TGT_WS_INIT_PENDING                    (1U << 13)
#define FLAG_TGT_REOPEN_PENDING                     (1U << 14)
*/
#define FLAG_TGT_ALL                                (0xffffffff)

#define ACIT_IS_TARGET_INITIALIZED(ptgt)            (ptgt->initialized)

#define ACIT_SET_TARGET_SUBSTATE(ptgt, substate)    {                                 \
                                                        ptgt->sub_state = substate;   \
                                                        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("0x%p Target substate set to 0x%x \n"), ptgt, substate)); \
                                                    }
#define ACIT_IS_TARGET_SUBSTATE(ptgt, substate)     ((ptgt->sub_state == substate))
#define ACIT_IS_TARGET_BUSY(ptgt, bit)              ((((PACIT_TARGET_WS) ptgt)->busy & bit))

#define TARGET_SUB_STATE_FREE                       (0)
#define TARGET_SUB_STATE_TMF_RESET_WAIT             (1U << 0)   //0x01
#define TARGET_SUB_STATE_TMF_TIMEOUT                (1U << 1)   //0x02
#define TARGET_SUB_STATE_TMF_RESET_DONE             (1U << 2)   //0x04
#define TARGET_SUB_STATE_ETO_IN_PROGRESS            (1U << 3)   //0x08
#define TARGET_SUB_STATE_TMF_RESET_START            (1U << 4)   //0x10
#define TARGET_SUB_STATE_INVALIDATION_UPLOAD_IN_PROGRESS (1U << 5)  //0x20
#define TARGET_SUB_STATE_LOGOUT_PENDING             (1U << 6)   //0x40
#define TARGET_SUB_STATE_CONX_UPLOAD_COMPLETE       (1U << 7)   //0x80
#define TARGET_SUB_STATE_REOPEN_IN_PROGRESS         (1U << 8)   //0x100
#define TARGET_SUB_STATE_TMF_ERROR                  (1U << 9)   //0x200
#define TARGET_SUB_STATE_IO_ERROR                   (1U << 10)  //0x400
#define TARGET_SUB_STATE_ASYNC_MSG_ERROR            (1U << 11)  //0x800
#define TARGET_SUB_STATE_PING_WAIT_TIMEOUT          (1U << 12)  //0x1000
#define TARGET_SUB_STATE_IO_RETRIES_EXCEEDED        (1U << 13)  //0x2000
#define TARGET_SUB_STATE_REJECT_MSG                 (1U << 14)  //0x4000
#define TARGET_SUB_STATE_INVALID_CQ_ERR             (1U << 15)  //0x8000
#define TARGET_SUB_STATE_ASYNC_MSG_LOGOUT           (1U << 16)  //0x10000
#define TARGET_SUB_STATE_OFFLOAD_ERR                (1U << 17)  //0x20000

/*
 * Define flags for ACIT_LUN_WS
#define FLAG_LUN_TMF_CLEANUP_WAIT                   (1U << 0)
 */

/*
 * Retry limits during polling. All units are in 100s of usecs
 */
/* 2 secs */
#define ACIT_ARM_INIT_MAX_RETRIES                   (2*10*1000)
/* 60 secs */
#define ACIT_ARM_READY_MAX_RETRIES                  (60*10*1000)

#if ACIT_CFG_SAVE_CMD_STATE
    #define ACIT_MAX_CMDS                               (128)
#endif

/* helper macros - misc */
#define ACIT_MIN(x, y)                          (((x) < (y))?(x):(y))
#define ACIT_MAX(x, y)                          (((x) > (y))?(x):(y))

#define TIME_SPENT(pc, tick)                    ((CURRENT_TICK(pc) - tick)/pc->params.ticks_per_second)
#define TICKS_SPENT(pc, tick)                   (CURRENT_TICK(pc) - tick)
#define TICKS_PER_SEC(pc)                       (pc->params.ticks_per_second)

/* macros for bit operations */
#define ACIT_BIT_SET(Value, Bit)                ((Value) |= (Bit))
#define ACIT_BIT_CLR(Value, Bit)                ((Value) &= ~(Bit))
#define ACIT_IS_BIT_SET(Value, Bit)             ((Value) & (Bit))
#define ACIT_IS_BIT_CLR(Value, Bit)             (!((Value) & (Bit)))

#define ACIT_BIT_SET_IF(Value, Bit, Condition)  if((Condition)) ((Value) |= (Bit))
#define ACIT_BIT_CLR_IF(Value, Bit, Condition)  if((Condition)) ((Value) &= ~(Bit))

#define GET_TMF_WS(ptmf_key) \
            ((PACIT_TMF_WS) ptmf_key->acit_key)
#define GET_IO_WS(pio_key) \
            ((PACIT_IO_WS) pio_key->acit_key)
#define GET_LUN_WS(plun_key) \
            ((PACIT_LUN_WS) plun_key->acit_key)
#define GET_TARGET_WS(ptarget_key) \
            ((PACIT_TARGET_WS) ptarget_key->acit_key)
#define GET_CONTROLLER_WS(pcontroller_key) \
            ((PACIT_CONTROLLER_WS) pcontroller_key->acit_key)


/*
 * 'Safe' mode of above GET WS macros, checks for valid key
 */
#define GET_LUN_WS_SAFE(plun_key) \
            ((plun_key)?((PACIT_LUN_WS) plun_key->acit_key):ACIT_NULL)
#define GET_TARGET_WS_SAFE(ptarget_key) \
            ((ptarget_key)?((PACIT_TARGET_WS) ptarget_key->acit_key):ACIT_NULL)

#define GET_SESSION_HANDLE(pc, pt) \
        pc->ptarget_table->target_entry[pt->target_table_index].session_handle

/*
 * ETO helper macros
 */
#define ACIT_IS_ETO_VALID(eto) (eto >= 0 && eto <= ACIT_CFG_MAX_ETO)
#define ACIT_IS_ETO_ENABLED(pt) (pt->eto.original_ticks > MIN_ETO)

/*
 * SCSI Command Opcodes
 */
#define ACIT_SCSI_CMD_TUR                           (0x00)
#define ACIT_SCSI_CMD_FORMAT                        (0x04)
#define ACIT_SCSI_CMD_INQUIRY                       (0x12)
#define ACIT_SCSI_CMD_REPORTLUNS                    (0xa0)

/*
 * Other SCSI Definition
 */
#define ACIT_SCSI_CMD_INQUIRY_EVPD_BIT              (0x01)

#define ACIT_PORT_NOT_APPLICABLE                    (0xFF)

static INLINE
UEXACT32
ios_index_from_status(ACIT_UEXACT32 status)       
{
    ACIT_UEXACT32   index = 0;
    if(!status)
    {
        index = 0;
    }
    else
    {
        index = status - ACIT_IO_STATUS_ERROR + 1;
    }

    OSM_ASSERT(index <= 9);
    return (index);
}

typedef struct _ACIT_DC_LIST
{
    struct _ACIT_DC_LIST  *next;
    struct _ACIT_DC_LIST  *prev;
} ACIT_DC_LIST, *PACIT_DC_LIST;


#if ACIT_CFG_SAVE_CMD_STATE
typedef struct _ACIT_CMD_STATE
{
    ACIT_UEXACT32               itt;
    ACIT_UEXACT32               cmd_sn;
} ACIT_CMD_STATE, *PACIT_CMD_STATE;
#endif


/*
 * This is for a BE specific requirement wherein the first SG
 * in the request structure points to sense data and header. This
 * should ideally come from HWI IO WS area for I/Os. Since we don't
 * have one, and to handle TMFs and Logouts, we declare it in the
 * ACIT IO WS for I/Os and the TMF/Logout handles for TMF and Logouts
 * respectively.
 */

#define ACIT_SENSE_INFO_SIZE        256
#define ACIT_ISCSI_PDU_HEADER_SIZE  64

typedef struct _ACIT_CMD_BHS
{
    union {
        ACIT_UEXACT8        iscsi_hdr[ACIT_ISCSI_PDU_HEADER_SIZE];
    }u1;

    union {
        // The plus 2 below is to hold the sense info length that gets
        // DMA'ed by RxULP
        ACIT_UEXACT8        sense_info[ACIT_SENSE_INFO_SIZE + 2];
    }u2;

}ACIT_CMD_BHS, *PACIT_CMD_BHS;

#define ACIT_CMD_BHS_LEN   sizeof(ACIT_CMD_BHS)

typedef struct _ACIT_NONCMD_BHS
{
    ACIT_UEXACT8        iscsi_hdr[ACIT_ISCSI_PDU_HEADER_SIZE];
}ACIT_NONCMD_BHS, *PACIT_NONCMD_BHS;

#define ACIT_NON_CMD_BHS_LEN   sizeof(ACIT_NONCMD_BHS)

typedef struct OSM_TARGET_NOTIFICATION
{
    union {
        ACIT_UEXACT32   flags;

        struct {
            ACIT_UEXACT32   add : 1;
            ACIT_UEXACT32   remove : 1;
            ACIT_UEXACT32   busy : 1;
            ACIT_UEXACT32   ready : 1;
            ACIT_UEXACT32   missing : 1;
            ACIT_UEXACT32   reconnected : 1;
            ACIT_UEXACT32   remove_pending : 1;
            ACIT_UEXACT32   reserved : 25;
        }bits;
    }u;

}OSM_TARGET_NOTIFICATION, *POSM_TARGET_NOTIFICATION;

/*
 * IO_HANDLE used to keep ITT around even when the 
 * original I/O has been returned to OS.
 */

typedef struct _ACIT_IO_HANDLE
{
    ACIT_DC_LIST                controller_link;
    ACIT_DC_LIST                target_link;
    ACIT_DC_LIST                lun_link;

    ACIT_DC_LIST                tmf_link;   /* Abort IO pending for this I/O */

    struct _WRB_HANDLE          *pwrb_handle;
    struct _SGL_HANDLE          *psgl_handle;
    struct _ISCSI_CONNECTION    *pconnection_handle;

    struct _ACIT_IO_WS          *pio;
    struct _ACIT_TMF_WS         *ptmf_ws;   /* OSM reference to abort */
    struct _ACIT_TARGET_WS      *ptarget;
    struct _ACIT_LUN_WS         *plun;
    ACIT_UEXACT16               unused;
    ACIT_UEXACT16               cri;
    ACIT_UEXACT32               flags;
    ACIT_UEXACT32               cmd_sn;
    ACIT_UEXACT32               itt;
    ACIT_UEXACT16               header_len;
    ACIT_UEXACT8                is_write;
    ACIT_UEXACT8                is_untagged;
    ACIT_UEXACT32               exp_xfer_length;
#if ACIT_CFG_INTERRUPT_REDIRECTION
    ACIT_UEXACT32               message_id;
    ACIT_UEXACT32               cq_id;
#endif
    ACIT_LUN                    lun;
    ACIT_UEXACT32               starting_tick_count;
    ACIT_UEXACT32               inv_start_tick;

} ACIT_IO_HANDLE, *PACIT_IO_HANDLE;    


typedef struct _ACIT_TMF_HANDLE
{
    ACIT_DC_LIST                controller_link;
    ACIT_DC_LIST                target_link;
    ACIT_DC_LIST                lun_link;

    struct _WRB_HANDLE          *pwrb_handle;
    struct _SGL_HANDLE          *psgl_handle;

    struct _ACIT_TMF_WS        *ptmf;           /* Valid only for TMFs from OSMs */
    struct _ACIT_IO_HANDLE     *pio_handle;     /* Valid only for abort I/O */
    ACIT_UEXACT32               tmf_function;   /* used to send the actual TMF function */
    struct _ACIT_TARGET_WS     *ptarget;
    struct _ACIT_LUN_WS        *plun;
    ACIT_UEXACT32               available_ticks;
    ACIT_UEXACT32               starting_tick_count;
    ACIT_UEXACT32               flags;

    ACIT_UEXACT32               cmd_sn;
    ACIT_UEXACT16               cri;
    ACIT_UEXACT16               port_num;       /* Filled by iSCSI when sending out TMF */

    ACIT_NONCMD_BHS             noncmd_bhs;
    ACIT_BUS_ADDRESS            bhs_pa;
    ACIT_UEXACT32               bhs_len;
    ACIT_UEXACT32               itt;
    ACIT_UEXACT16               header_len;

} ACIT_TMF_HANDLE, *PACIT_TMF_HANDLE;  

typedef struct _ACIT_LOGOUT_HANDLE
{
    ACIT_DC_LIST                controller_link;
    ACIT_DC_LIST                target_link;
    struct _WRB_HANDLE          *pwrb_handle;
    struct _SGL_HANDLE          *psgl_handle;

    struct _ACIT_TARGET_WS     *ptarget;
    ACIT_UEXACT32               available_ticks;
    ACIT_UEXACT32               starting_tick_count;
    ACIT_UEXACT32               cmd_sn;
    ACIT_UEXACT32               flags;
    ACIT_UEXACT32               reason_code;

    ACIT_NONCMD_BHS             noncmd_bhs;
    ACIT_BUS_ADDRESS            bhs_pa;
    ACIT_UEXACT32               bhs_len;
    ACIT_UEXACT32               itt;
    ACIT_UEXACT16               header_len;
    ACIT_UEXACT16               cri;

} ACIT_LOGOUT_HANDLE, *PACIT_LOGOUT_HANDLE;

typedef struct _ACIT_NOP_HANDLE
{
    ACIT_DC_LIST                controller_link;
    ACIT_DC_LIST                target_link;
    struct _WRB_HANDLE          *pwrb_handle;
    struct _SGL_HANDLE          *psgl_handle;

    struct _ACIT_TARGET_WS     *ptarget;
    ACIT_UEXACT32               available_ticks;
    ACIT_UEXACT32               starting_tick_count;
    ACIT_UEXACT32               cmd_sn;

    ACIT_NONCMD_BHS             noncmd_bhs;
    ACIT_BUS_ADDRESS            bhs_pa;
    ACIT_UEXACT32               bhs_len;
    ACIT_LUN                    optional_lun;
    ACIT_UEXACT32               ttt;
    ACIT_UEXACT32               itt;
    ACIT_UEXACT32               issued_itt;
    ACIT_UEXACT16               header_len;
    ACIT_UEXACT16               cri;
    ACIT_UEXACT32               flags;

} ACIT_NOP_HANDLE, *PACIT_NOP_HANDLE;

#define ACIT_INVH_TYPE_IO       (1 << 0)

typedef struct _ACIT_INVALIDATE_HANDLE
{
    ACIT_DC_LIST                controller_link;

    ACIT_UEXACT8                flags;
    ACIT_UEXACT8                num_invs;
    ACIT_UEXACT16               index;

    ACIT_DC_LIST                list_header;
    ACIT_UEXACT32               submit_tick;
    ACIT_UEXACT32               complete_tick;

} ACIT_INVALIDATE_HANDLE, *PACIT_INVALIDATE_HANDLE;

typedef struct _ACIT_TMF_WS
{
    PACIT_TMF_KEY               ptmf_key;      /* TMF Key for this TMF */
    PACIT_TMF_REQUEST           ptmf_request;
    ACIT_DC_LIST                controller_link;

    ACIT_DC_LIST                busy_io_list;   /* List of all I/Os to be aborted */
                                                /* Used in aborting multiple I/Os */
    PACIT_VOID                  piscsi_ws;
    PACIT_VOID                  pmgmt_ws;
    ACIT_UEXACT32               original_ticks;
    ACIT_UEXACT32               available_ticks;
    ACIT_UEXACT32               tmf_status;
    ACIT_UEXACT32               pad;

} ACIT_TMF_WS, *PACIT_TMF_WS;

typedef struct _ACIT_IO_WS
{
    PACIT_IO_KEY                pio_key;            /* I/O Key for this I/O */
    PACIT_IO_REQUEST            pio_request;        /* Back pointer to the request */

    /*
     * The following 3 links are used to queue I/Os internally
     * in ACIT. They are queued along io_queued_list of the
     * respective workspaces
     */
    ACIT_DC_LIST                controller_link;
    ACIT_DC_LIST                target_link;
    ACIT_DC_LIST                lun_link;

    PACIT_IO_HANDLE             pio_handle;
    PACIT_VOID                  piscsi_ws;
    PACIT_VOID                  pmgmt_ws;

#if ACIT_IO_PERFORMANCE
    ACIT_U64                    io_start_time;
    ACIT_U64                    io_end_time;
#endif
    
    ACIT_UEXACT32               retry_count; // is reset during session recovery
    ACIT_UEXACT32               io_retries; // total, across all session recoveries

    ACIT_UEXACT32               original_ticks;
    ACIT_SEXACT32               available_ticks;
    ACIT_SEXACT32               internal_available_ticks;
    ACIT_UEXACT32               io_status;

    ACIT_BUS_ADDRESS            bhs_pa;
    ACIT_CMD_BHS                cmd_bhs;
    ACIT_UEXACT32               bhs_len;
    ACIT_UEXACT32               queued;

} ACIT_IO_WS, *PACIT_IO_WS;

typedef struct _ACIT_LUN_WS
{
    PACIT_LUN_KEY               plun_key;
    ACIT_DC_LIST                io_handle_busy_list;
    ACIT_DC_LIST                io_queued_list; /* ACIT internal queue of IO_WS */
    ACIT_DC_LIST                tmf_handle_busy_list;
    ACIT_DC_LIST                io_pending_reset_list;

    ACIT_DC_LIST                tmf_link;       /* Queue of LUNs waiting for TMF resources */
    ACIT_DC_LIST                tmf_abort_list; /* List of all I/O aborts waiting for TMF resources */

    PACIT_VOID                  piscsi_ws;
    PACIT_VOID                  pmgmt_ws;
#if ACIT_LUN_PERFORMANCE
    ACIT_U64                    io_started;
    ACIT_U64                    io_completed;
    ACIT_U64                    io_pending;
    ACIT_U64                    io_aborted;
    ACIT_U64                    lun_resets;
    ACIT_U64                    max_response_time;
    ACIT_U64                    min_response_time;
#endif /* ACIT_LUN_PERFORMANCE */
    ACIT_UEXACT32               flags;
    ACIT_UEXACT32               busy;
    ACIT_UEXACT32               dev_type;
    ACIT_UEXACT32               retry_io;

    struct {
        ACIT_UEXACT32               io_status[10];
    } stats;

    struct {

        ACIT_UEXACT32           num_io_invs;
        ACIT_UEXACT32           queued_tmfs;
        ACIT_UEXACT32           lun_reset_queued;

    } counters;

    struct _ACIT_TMF_REQUEST    *plun_reset_tmf_request;

    struct {

        ACIT_UEXACT32           tmf_lun_resets;
        ACIT_UEXACT32           io_busy_during_lun_rst_timeout;
        ACIT_UEXACT32           io_queued_during_lun_rst_timeout;
        ACIT_UEXACT32           io_rstpend_during_lun_rst_timeout;
        ACIT_UEXACT32           tmf_err_stats[9];

    } eh_stats;


} ACIT_LUN_WS, *PACIT_LUN_WS;

#define GET_TARGET_DOMAIN(pc, pt) (pc->ptarget_table->target_entry[pt->target_table_index].domain)

typedef struct _ACIT_TARGET_WS
{
    PACIT_TARGET_KEY            ptarget_key;
    ACIT_DC_LIST                io_handle_busy_list;
    ACIT_DC_LIST                io_queued_list; /* ACIT internal queue of IO_WS */
    ACIT_DC_LIST                io_pending_reset_list;

    ACIT_DC_LIST                tmf_handle_busy_list;/* TMFs pending with h/w */
    ACIT_DC_LIST                tmf_link;       /* Queue of targets waiting for TMF resources */
    ACIT_DC_LIST                tmf_lun_list;   /* List of all LUNs waiting for TMF resources */
    
    ACIT_DC_LIST                logout_handle_list; /* logout handle pending with h/w */
    ACIT_DC_LIST                nop_handle_list;/* Queue of NOP handles pending with h/w */

    ACIT_DC_LIST                event_link;
    struct _ISCSI_TARGET_WS     *piscsi_ws;
    struct _MGMT_TARGET_WS      *pmgmt_ws;
    PACIT_VOID                  plast_hiber_io;
#if ACIT_TARGET_PERFORMANCE
    ACIT_U64                    started_ios;
    ACIT_U64                    completed_ios;
    ACIT_U64                    pending_ios;
    ACIT_U64                    aborted_ios;
    ACIT_U64                    target_resets;
    ACIT_U64                    max_response_time;
    ACIT_U64                    min_response_time;
#endif /* ACIT_TARGET_PERFORMANCE */
    struct _ACIT_CONTROLLER_WS  *pcontroller;
    ACIT_UEXACT8                cleanup_flag;
    ACIT_UEXACT8                flags;
    ACIT_UEXACT8                event_flag;
    ACIT_UEXACT8                eto_status;
    OSM_TARGET_NOTIFICATION     osm_notification;
    ACIT_UEXACT32               busy;
    ACIT_UEXACT32               sub_state;
    ACIT_UEXACT32               target_table_index;

    struct {

        ACIT_UEXACT32               pending_ios;
        ACIT_UEXACT32               num_io_invs;

        ACIT_UEXACT32               queued_tmfs;
        ACIT_UEXACT32               tmf_queued_luns;
        ACIT_UEXACT32               missing_since;

        // Counter to monitor successive offload failures
        ACIT_UEXACT32               offload_err;

    } counters;

    struct {
        // valid if target is missing, counting down
        ACIT_UEXACT32               ticks_counter;
        
        // eto value in ticks for this target, mainly for display. This is used
        // when starting the ETO timer.
        ACIT_UEXACT32               original_ticks;

        // This is the ETO value adjusted in case ETO is set to 0. This is used
        // when setting command timeouts.
        ACIT_UEXACT32               adjusted_ticks;

        // max amount of time that can be given for a command based on its ETO
        // value. Pre-calculated and saved so that this calculation need not be
        // done at runtime
        ACIT_UEXACT32               rto_ticks;
    } eto;

    ACIT_UEXACT32               ping_wait_ticks;

    struct _ACIT_TMF_REQUEST    *ptarget_reset_tmf_request;/* Target Reset TMF request from OSM */
    ACIT_UEXACT32               target_reset_queued;

    ACIT_UEXACT32               timeout_factor;

    struct {

        // This will be used when ETO timers are to be started.
        ACIT_UEXACT32           tmf_timeout;

        // Updated when a TMF timeout is triggerred. And then used when
        // starting ETO timers
        ACIT_UEXACT32           io_resp_missing_since;
     
    }recovery_reason;

    // As per the new ITO scheme, the following is not required

    /*
     * last_cmd_ticks - defines the number of ticks taken to complete the
     * last command. 
     * Depending on the target response, this value can steadily increase
     * upto eto/3. However, if there is a MPIO configuration, then this
     * will not be changed.
     * This value will NEVER go below default_cmd_ticks.
     */
    // ACIT_UEXACT32               last_cmd_ticks;
    ACIT_UEXACT32               initialized;

    struct { 
          
        ACIT_UEXACT32               ssn_digest_errors;
        ACIT_UEXACT32               ssn_cxn_timeout_errors;
        ACIT_UEXACT32               ssn_format_errors;

        ACIT_UEXACT32               tmf_target_resets;
        ACIT_UEXACT32               recoveries;

        ACIT_UEXACT32               ping_wait_timeouts;
        ACIT_UEXACT32               tmf_timeouts;
        ACIT_UEXACT32               tmf_internal_timeouts;
        ACIT_UEXACT32               tmf_errors;
        ACIT_UEXACT32               ioretry_exceeded;
        ACIT_UEXACT32               rejects;
        ACIT_UEXACT32               async_logouts;
        ACIT_UEXACT32               async_reopens;
        ACIT_UEXACT32               async_param_renego;
        ACIT_UEXACT32               io_status_errors;
        ACIT_UEXACT32               unknown_defpdu;

        ACIT_UEXACT32               io_busy_during_lun_rst_timeout;
        ACIT_UEXACT32               io_queued_during_lun_rst_timeout;
        ACIT_UEXACT32               io_rstpend_during_lun_rst_timeout;

        ACIT_UEXACT32               io_busy_during_tgt_rst_timeout;
        ACIT_UEXACT32               io_queued_during_tgt_rst_timeout;
        ACIT_UEXACT32               io_rstpend_during_tgt_rst_timeout;

        ACIT_UEXACT32               reopen_mismatch;

        ACIT_UEXACT32               tmf_err_stats[9];

        ACIT_UEXACT32               tmf_abort_timeouts;
        ACIT_UEXACT32               tmf_lunrst_timeouts;
        ACIT_UEXACT32               tmf_tgtrst_timeouts;

        ACIT_UEXACT32               tmf_abort_internal_timeouts;
        ACIT_UEXACT32               tmf_lunrst_internal_timeouts;
        ACIT_UEXACT32               tmf_tgtrst_internal_timeouts;

        ACIT_UEXACT32               reopen_issued;
        ACIT_UEXACT32               reopen_success;
        ACIT_UEXACT32               reopen_error;
        ACIT_UEXACT32               offloads;
        ACIT_UEXACT32               armfw_async_reopens;

    } eh_stats;

    struct {
        ACIT_UEXACT32               last_cmdsn_tick;
        ACIT_UEXACT32               last_async_tick;
        ACIT_UEXACT32               last_cqerr_tick;

        ACIT_UEXACT32               untagged_cmd;
        ACIT_UEXACT32               io_status[10];

        // max num_cq can ever be is 8
        ACIT_UEXACT32               io_completion[ACIT_CFG_MAX_PARTITIONS];
        // last completion will be used by timer
        ACIT_UEXACT32               last_completion[ACIT_CFG_MAX_PARTITIONS];

    } stats;

} ACIT_TARGET_WS, *PACIT_TARGET_WS;

/*
 * Memory descriptors are of 2 kinds - fragmented and
 * non-fragmented. Since ACIT supports OSMs that cannot
 * allocate contiguous memory beyond ACIT_CFG_MAX_CONTIGUOUS_MEM,
 * some descriptors may get split into multiple descriptors
 */

typedef enum _AP_MEM_ENUM {

     /* ACIT main descriptors - non-fragmented*/
    AP_MEM_VASE

     // non-cached
    ,AP_MEM_LOGOUT_HANDLE
    ,AP_MEM_TMF_HANDLE
    ,AP_MEM_NOP_HANDLE
    ,AP_MEM_INV_HANDLE
    ,AP_MEM_TARGET_KEY
    ,AP_MEM_TARGET_TABLE

 /* iSCSI main descriptors - non-fragmented*/
    ,ISCSI_MEM_GLOBAL_HEADER

 /* MGMT main descriptors - non-fragmented*/

     // non-cached
    ,MGMT_MEM_IOCTL_HANDLE
    ,MGMT_MEM_IOCTL_BUFFER

    /* HWI main descriptors - non-fragmented*/

    // cached
    ,HWI_MEM_ADDN_CONTEXT
    ,HWI_MEM_ASYNC_PDU_CONTEXT

    // non-cached - non-fragmented
    ,HWI_MEM_MAILBOX
    ,HWI_MEM_HDRS
#ifdef BE_EMULATION
    ,HWI_MEM_EMUL
    ,HWI_MEM_ZEROB
    ,HWI_MEM_PT
    ,HWI_MEM_JELL
    ,HWI_MEM_R2TARRAY0
    ,HWI_MEM_R2TARRAY1
    ,HWI_MEM_R2TARRAY2
    ,HWI_MEM_R2TARRAY3
    ,HWI_MEM_R2TARRAY4
    ,HWI_MEM_R2TARRAY5
    ,HWI_MEM_R2TARRAY6
    ,HWI_MEM_R2TARRAY7
    ,HWI_MEM_R2TARRAY8
    ,HWI_MEM_R2TARRAY9
    ,HWI_MEM_R2TARRAY10
    ,HWI_MEM_R2TARRAY11
    ,HWI_MEM_R2TARRAY12
    ,HWI_MEM_R2TARRAY13
    ,HWI_MEM_R2TARRAY14
    ,HWI_MEM_R2TARRAY15

    ,EMUL_MEM_OFFLOAD_PARAMS
#endif
    ,HWI_MEM_CQ          
    ,HWI_MEM_EQ          
    ,HWI_MEM_MCC         
    ,HWI_MEM_MCC_CQ      
    ,HWI_MEM_ASYNC_HEADER_BUF
    ,HWI_MEM_ASYNC_HEADER_HANDLE
    ,HWI_MEM_ASYNC_HEADER_RING
    ,HWI_MEM_ASYNC_DATA_BUF
    ,HWI_MEM_ASYNC_DATA_HANDLE
    ,HWI_MEM_ASYNC_DATA_RING
    ,HWI_MEM_SGLH
    ,HWI_MEM_MCC_CONTEXT

    ,AP_MEM_MAX
   
}AP_MEM_ENUM, *PAP_MEM_ENUM;

#define GET_MEM_DESC_VA(pc, idx)    pc->mem_descriptors[idx].virtual_address;
#define GET_MEM_DESC_PA64(pc, idx)  pc->mem_descriptors[idx].bus_address.u.a64.address

/* If there is a reminder from a/b, add 1 to the dividend else return dividend */
#define ACIT_CEIL(a, b)     ((a+b-1)/b)

#define AP_IOH_DESC_SZ      (sizeof(struct _ACIT_IO_HANDLE)* BE_MAX_IOS)
#define AP_MEM_NUM_IOH      ACIT_CEIL(AP_IOH_DESC_SZ, ACIT_CFG_MAX_CONTIGUOUS_MEM)

#define ACIT_MEM_MAX        (AP_MEM_MAX + AP_MEM_NUM_IOH + ISCSI_MEM_NUM_CXN +  \
                             HWI_MEM_NUM_FRAG + HWI_MEM_NUM_WRB +               \
                             HWI_MEM_NUM_WRBH)

typedef struct _ACIT_VAR_DESC_INDEX
{
    struct {
        ACIT_UEXACT8                ioh;
        ACIT_UEXACT8                cxn;
        ACIT_UEXACT8                wrb;
        ACIT_UEXACT8                frag;
        ACIT_UEXACT8                wrbh;
    }num;

    struct {
        ACIT_UEXACT8                ioh;
        ACIT_UEXACT8                cxn;
        ACIT_UEXACT8                wrb;
        ACIT_UEXACT8                frag;
        ACIT_UEXACT8                wrbh;
    }range;

    struct {
        ACIT_UEXACT8                frag;
    }run_time;

    ACIT_UEXACT8                acit_mem_max;
}
ACIT_VAR_DESC_INDEX, *PACIT_VAR_DESC_INDEX;

//
// Interrupt redirection support macros. Depending on whether FW supports the new
// feature, different MGMT structure will be used
//
#define FW_SUPPORTS_INTERRUPT_REDIRECTION(pc)   (pc->addi_attribs.interrupt_redirection)
#define DRVR_SUPPORTS_INTERRUPT_REDIRECTION(pc) (pc->params.interrupt_redirection)

typedef struct _ACIT_CONTROLLER_WS
{
    PACIT_CONTROLLER_KEY        pcontroller_key;
    ACIT_OSM_API_TABLE          osm_api_table;

    // BladeEngine 2 uses Bar0 through Bar3
    ACIT_REGISTER_HANDLE        bar0_handle;      /* PCI CFG mapped in memory space. */
    ACIT_REGISTER_HANDLE        bar1_handle;      /* CSRs in memory space */
    ACIT_REGISTER_HANDLE        bar2_handle;      /* Doorbells in memory space. */

    ACIT_BUS_ADDRESS            bar0_pa;
    ACIT_BUS_ADDRESS            bar1_pa;
    ACIT_BUS_ADDRESS            bar2_pa;

    PACIT_VOID                  bar0_va;
    PACIT_VOID                  bar1_va;
    PACIT_VOID                  bar2_va;

    ACIT_DC_LIST                io_handle_free_list;         /* Free IOH */
    ACIT_DC_LIST                io_handle_busy_list;         /* IOH submitted to hardware */

    ACIT_DC_LIST                inv_handle_free_list;
    ACIT_DC_LIST                inv_handle_busy_list;

    ACIT_DC_LIST                io_pending_reset_list;

    ACIT_DC_LIST                logout_handle_free_list;
    ACIT_DC_LIST                logout_handle_busy_list;

    ACIT_DC_LIST                io_queued_list;   /* ACIT internal queue of IO_WS */

    ACIT_DC_LIST                tmf_handle_free_list;/* Free list of TMF handles */
    ACIT_DC_LIST                tmf_handle_busy_list;/* Busy list of TMF handles */

    ACIT_DC_LIST                tmf_target_list;     /* List of all targets waiting for TMF resources */

    ACIT_DC_LIST                tmf_queued_list;     /* used only for the purpose of 
                                                        timing out TMFs in internal queue*/

    ACIT_DC_LIST                target_event_list;

    ACIT_DC_LIST                nop_handle_free_list;/* Free list of NOP handles */
    ACIT_DC_LIST                nop_handle_busy_list;/* Busy list of NOP handles */

    struct _ISCSI_CONTROLLER_WS *piscsi_ws;
    struct _MGMT_CONTROLLER_WS  *pmgmt_ws;
    struct _HWI_CONTROLLER_WS   *phwi_ws;

    PACIT_TARGET_TABLE          ptarget_table;

    ACIT_UEXACT32               busy;

    struct {

        ACIT_UEXACT16           flush_queued_ios;
        ACIT_UEXACT16           flush_queued_tmfs;

        ACIT_UEXACT8            asic_revision;
        ACIT_UEXACT8            interrupt_enabled;
        ACIT_UEXACT8            operating_mem_level;
        ACIT_UEXACT8            invalidation_ticks;

        ACIT_UEXACT8            unused;
        ACIT_UEXACT8            discovery_in_progress;
        ACIT_UEXACT8            discovery_type;
        ACIT_UEXACT8            defer_invalidations;

        ACIT_UEXACT8            init_type;
        ACIT_UEXACT8            soft_reset;
        ACIT_UEXACT8            api_type;
        ACIT_UEXACT8            arm_async_registered;

        ACIT_UEXACT8            asic_generation;
        ACIT_UEXACT8            iscsi_features;

    } state;

    ACIT_FW_CONFIG              fw_config;
    ACIT_FW_ADDI_ATTRIBS        addi_attribs;
    ACIT_COUNTERS               counters;

    ACIT_PARAMETERS             params;

    //ACIT_VAR_DESC_INDEX         var_desc_index;
    ACIT_UEXACT32               ioh_mem_index;

    /*
     * This points to the end of ACIT_CONTROLLER_WS
     */
    PACIT_MEM_DESCRIPTOR        mem_descriptors;

    ACIT_VAR_DESC_INDEX         acit_var_desc_index;
    
    PACIT_IO_HANDLE             pio_handle_base[AP_MEM_NUM_IOH];
    struct _ACIT_LOGOUT_HANDLE  *plogout_handle_base;
    PACIT_TMF_HANDLE            ptmf_handle_base;
    PACIT_NOP_HANDLE            pnop_handle_base;
    PACIT_INVALIDATE_HANDLE     pinv_handle_base;
        
    struct {

        ACIT_UEXACT32           acit_controller_ws_size;
      
        ACIT_UEXACT32           iscsi_controller_ws_size;
        ACIT_UEXACT32           iscsi_target_ws_size;
        ACIT_UEXACT32           iscsi_lun_ws_size;
        ACIT_UEXACT32           iscsi_io_ws_size;
        ACIT_UEXACT32           iscsi_tmf_ws_size;
      
        ACIT_UEXACT32           mgmt_controller_ws_size;
        ACIT_UEXACT32           mgmt_target_ws_size;
        ACIT_UEXACT32           mgmt_lun_ws_size;
        ACIT_UEXACT32           mgmt_io_ws_size;
        ACIT_UEXACT32           mgmt_tmf_ws_size;

        ACIT_UEXACT32           hwi_controller_ws_size;
        ACIT_UEXACT32           hwi_target_ws_size;
        ACIT_UEXACT32           hwi_lun_ws_size;
        ACIT_UEXACT32           hwi_io_ws_size;
        ACIT_UEXACT32           hwi_tmf_ws_size;

    } ws_sizes;

#if ACIT_CFG_VASE_ENABLED
    DBG_MASTER_BLOCK            vase_master_block;
#endif
    ACIT_UEXACT32               vase_mem_needed;

    PACIT_TARGET_KEY            ptarget_key;
    ACIT_UEXACT32               ld_timeout_ticks[MAX_PHYS_PORTS]; /* LDTO on each port */
    ACIT_UEXACT32               link_down[MAX_PHYS_PORTS];  /* Link down status of each port */
    
#if ACIT_CFG_SAVE_CMD_STATE
    ACIT_UEXACT16               last_cmd;
    ACIT_UEXACT16               num_cmds;
    ACIT_UEXACT32               unused_mem;
    ACIT_CMD_STATE              cmd_state[ACIT_MAX_CMDS];
#endif

    ACIT_CONTROLLER_EH_STATS    eh_stats;
    ACIT_CONTROLLER_STATS       stats;

    struct {

        ACIT_UEXACT32           enabled;

    } dic;

    struct {

        ACIT_CONTROLLER_INT_STATS   interrupt_stats;
        ACIT_CONTROLLER_DIC_COUNTERS dic;

    } partition[ACIT_CFG_MAX_PARTITIONS];

#ifdef BE_EMULATION
    ACIT_UEXACT32   ifid[2];
    struct _EMUL_OFFLOAD_PARAMS * poffload_params;
#endif

    MGMT_PCIE_INFO pcie_info;

    struct {

        ACIT_UEXACT64           last_io_completed;
        ACIT_UEXACT32           last_ioctl_completed;
        ACIT_UEXACT16           ue_recovery_started;
        ACIT_UEXACT16           reinit_after_ue;

    } ue_detection;

} ACIT_CONTROLLER_WS, *PACIT_CONTROLLER_WS;

#define ACIT_UPDATE_INT_COUNT(pc,msg)               pc->partition[msg].dic.interrupts++


#define ACIT_UPDATE_EQ_MAJ_COMP_INT_COUNT(pc, pid)   pc->partition[pid].interrupt_stats.eq_maj_ints++
#define ACIT_UPDATE_CQ_INT_COUNT(pc, pid)            pc->partition[pid].interrupt_stats.cq_ints++
#define ACIT_UPDATE_DFT_PDU_INT_COUNT(pc, pid)       pc->partition[pid].interrupt_stats.dft_pdu_ints++
#define ACIT_UPDATE_DRIVER_MSG_INT_COUNT(pc, pid)    pc->partition[pid].interrupt_stats.drvr_msg_ints++
#define ACIT_UPDATE_ISCSI_CMD_INT_COUNT(pc, pid)     pc->partition[pid].interrupt_stats.iscsi_cmd_ints++
#define ACIT_UPDATE_CQ_PROCESSED_COUNT(pc, pid)      pc->partition[pid].interrupt_stats.iscsi_cq_processed++

#define ACIT_INC_IOS_SUBMITTED(pc)              {pc->stats.ios_submitted++; \
                                                 if((pc->stats.ios_submitted - pc->stats.ios_completed) > pc->stats.max_queue_depth) \
                                                    pc->stats.max_queue_depth = (ACIT_UEXACT32)(pc->stats.ios_submitted - pc->stats.ios_completed);}

#define ACIT_INC_IOS_COMPLETED(pc)              {pc->stats.ios_completed++;}
#define ACIT_UPDATE_IOSTATUS_COUNT(pc, ios)     pc->stats.io_status[ios_index_from_status(ios)]++;
#define ACIT_UPDATE_TARGET_IOSTATUS_COUNT(pt, ios) pt->stats.io_status[ios_index_from_status(ios)]++;
#define ACIT_UPDATE_LUN_IOSTATUS_COUNT(pl, ios) pl->stats.io_status[ios_index_from_status(ios)]++;
#define ACIT_UPDATE_ASYNC_HDR_POST_COUNT(pc, count)    pc->stats.async_hdr_posted+=count
#define ACIT_UPDATE_ASYNC_DATA_POST_COUNT(pc, count)   pc->stats.async_data_posted+=count

#define ACIT_UPDATE_TXULP_DB_COUNT(pc)          pc->stats.txulp_db++

#define ACIT_UPDATE_LAST_IO_TICK(pc)            pc->stats.last_io_tick = CURRENT_TICK(pc);
#define ACIT_UPDATE_LAST_ISR_TICK(pc, pid)      pc->partition[pid].interrupt_stats.last_isr_tick = CURRENT_TICK(pc);

#define ACIT_UPDATE_MAX_EQE_PER_ISR(pc, pid) {                                                               \
                                            if(pc->partition[pid].dic.num_eq_processed > pc->partition[pid].interrupt_stats.max_eqe_per_isr)    \
                                                pc->partition[pid].interrupt_stats.max_eqe_per_isr = pc->partition[pid].dic.num_eq_processed;   \
                                        }

#define ACIT_UPDATE_FLUSH_IO_COUNTERS(pc, nf, nq) {                                             \
                                                    if(nf > pcontroller->stats.max_num_flushed) \
                                                        pc->stats.max_num_flushed = nf;         \
                                                    if(nq > pcontroller->stats.max_num_queued)  \
                                                        pc->stats.max_num_queued = nq;          \
                                                  }

#define ACIT_UPDATE_CQE_PER_ISR(pc, c)      {                                                   \
                                                if(c > pc->stats.max_cqe_per_isr)               \
                                                    pc->stats.max_cqe_per_isr = c;              \
                                             }

#if 0

#define ACIT_UPDATE_EQ_MAJ_COMP_INT_COUNT(pc, pid)
#define ACIT_UPDATE_CQ_INT_COUNT(pc, pid)
#define ACIT_UPDATE_DFT_PDU_INT_COUNT(pc, pid)
#define ACIT_UPDATE_DRIVER_MSG_INT_COUNT(pc, pid)
#define ACIT_UPDATE_ISCSI_CMD_INT_COUNT(pc, pid)
#define ACIT_UPDATE_CQ_PROCESSED_COUNT(pc, pid)

#define ACIT_INC_IOS_SUBMITTED(pc)
#define ACIT_INC_IOS_COMPLETED(pc)
#define ACIT_UPDATE_IOSTATUS_COUNT(pc, ios)
#define ACIT_UPDATE_TARGET_IOSTATUS_COUNT(pt, ios)
#define ACIT_UPDATE_LUN_IOSTATUS_COUNT(pl, ios)
#define ACIT_UPDATE_ASYNC_HDR_POST_COUNT(pc, count)
#define ACIT_UPDATE_ASYNC_DATA_POST_COUNT(pc, count)
#define ACIT_UPDATE_TXULP_DB_COUNT(pc)

#define ACIT_UPDATE_LAST_IO_TICK(pc) 
#define ACIT_UPDATE_LAST_ISR_TICK(pc, pid)

#define ACIT_UPDATE_MAX_EQE_PER_ISR(pc, pid)
#define ACIT_UPDATE_FLUSH_IO_COUNTERS(pc, nf, nq)
#define ACIT_UPDATE_CQE_PER_ISR(pc, c) 

#endif

// Target Update Macros. Used in ETO calculation
#define ACIT_UPDATE_TARGET_LAST_CMDSN_TS(pc, pt) pt->stats.last_cmdsn_tick = CURRENT_TICK(pc);
#define ACIT_UPDATE_TARGET_LAST_ASYNC_TS(pc, pt) pt->stats.last_async_tick = CURRENT_TICK(pc);
#define ACIT_UPDATE_TARGET_LAST_CQERR_TS(pc, pt) pt->stats.last_cqerr_tick = CURRENT_TICK(pc);

#define ACIT_UPDATE_DRVR_MSG_SENT_COUNT(pc)           pc->stats.drvr_msg_sent++
#define ACIT_UPDATE_DRVR_MSG_RECVD_COUNT(pc, count)   {                                         \
                                                        pc->stats.drvr_msg_recvd++;             \
                                                        if(count > 1)                           \
                                                            pc->stats.drvr_msg_multiple_recvd++;\
                                                      }

typedef struct _ACIT_PARAM_PROPERTIES {
    
    const PACIT_SEXACT8 name;
    ACIT_UEXACT32       min_value;
    ACIT_UEXACT32       mem_level1_value;
    ACIT_UEXACT32       mem_level2_value;
    ACIT_UEXACT32       mem_level3_value;
    ACIT_UEXACT32       max_value;
    ACIT_UEXACT32       default_mask;    
    ACIT_UEXACT32       *value;

} ACIT_PARAM_PROPERTIES, *PACIT_PARAM_PROPERTIES;

// to fix compiler warning C4245. Define signed
// versions of -1 for all data sizes

#define ACIT_U8_MINUS_ONE   (0xff)
#define ACIT_U16_MINUS_ONE  (0xffff)
#define ACIT_U32_MINUS_ONE  (0xffffffff)
#define ACIT_U64_MINUS_ONE  (0xffffffffffffffff)

// to fix compiler warning C4127
static const ACIT_UEXACT32 acit_zero = 0;

#define LIST_INIT_HEADER(header)                              \
do {                                                          \
    ((PACIT_DC_LIST)(header))->next = (PACIT_DC_LIST)(header);\
    ((PACIT_DC_LIST)(header))->prev = (PACIT_DC_LIST)(header);\
} while(acit_zero)

#define IS_LIST_EMPTY(header) \
    (((PACIT_DC_LIST)(header))->next == ((PACIT_DC_LIST)(header)))

#define IS_LIST_NOT_EMPTY(header) \
    (!IS_LIST_EMPTY(header))

#define IS_LINK_VALID(element) \
    (!IS_LINK_INVALID(element))

#define LIST_INIT_ELEMENT(element) LIST_INIT_HEADER(element)

#define IS_LINK_INVALID(element) IS_LIST_EMPTY(element)

#define IS_LINK_VALID(element)  (!IS_LINK_INVALID(element))

#define LIST_ADDTO_HEAD(element, header)                                     \
    do {                                                                     \
    ((PACIT_DC_LIST)(element))->next       = ((PACIT_DC_LIST)(header))->next;\
    ((PACIT_DC_LIST)(element))->prev       =  (PACIT_DC_LIST)(header);       \
    ((PACIT_DC_LIST)(header))->next->prev  =  (PACIT_DC_LIST)(element);      \
    ((PACIT_DC_LIST)(header))->next        =  (PACIT_DC_LIST)(element);      \
} while(acit_zero)


#define LIST_ADDTO_TAIL(element, header)                                     \
    do {                                                                     \
    ((PACIT_DC_LIST)(element))->next       =  (PACIT_DC_LIST)(header);       \
    ((PACIT_DC_LIST)(element))->prev       = ((PACIT_DC_LIST)(header))->prev;\
    ((PACIT_DC_LIST)(header))->prev->next  =  (PACIT_DC_LIST)(element);      \
    ((PACIT_DC_LIST)(header))->prev        =  (PACIT_DC_LIST)(element);      \
} while(acit_zero)

#define LIST_REMOVE_ELEMENT(element)                                           \
do {                                                                           \
    ((PACIT_DC_LIST)(element))->prev->next = ((PACIT_DC_LIST)(element))->next; \
    ((PACIT_DC_LIST)(element))->next->prev = ((PACIT_DC_LIST)(element))->prev; \
} while(acit_zero)

#define LIST_REMOVEFROM_HEAD(element, header)                                \
    do {                                                                     \
    *((PACIT_DC_LIST*)element)               = ((PACIT_DC_LIST)header)->next;\
    (*((PACIT_DC_LIST*)element))->next->prev = (PACIT_DC_LIST)header;        \
    ((PACIT_DC_LIST)header)->next       = (*((PACIT_DC_LIST*)element))->next;\
} while(acit_zero)

#define LIST_MOVE(to_header, from_header)                                    \
    do {                                                                     \
    ((PACIT_DC_LIST)to_header)->next  =   ((PACIT_DC_LIST)from_header)->next;\
    ((PACIT_DC_LIST)to_header)->prev  =   ((PACIT_DC_LIST)from_header)->prev;\
    ((PACIT_DC_LIST)from_header)->next->prev = (PACIT_DC_LIST)to_header;     \
    ((PACIT_DC_LIST)from_header)->prev->next = (PACIT_DC_LIST)to_header;     \
    LIST_INIT_HEADER(from_header);                                           \
} while(acit_zero)

#define LIST_MOVETO_TAIL(to_header, from_header)                            \
do {                                                                        \
    ((PACIT_DC_LIST)to_header)->prev->next=((PACIT_DC_LIST)from_header)->next;\
    ((PACIT_DC_LIST)from_header)->next->prev=((PACIT_DC_LIST)to_header)->prev;\
    ((PACIT_DC_LIST)to_header)->prev =  ((PACIT_DC_LIST)from_header)->prev; \
    ((PACIT_DC_LIST)from_header)->prev->next =  ((PACIT_DC_LIST)to_header); \
    LIST_INIT_HEADER(from_header);                                          \
} while(acit_zero)

#define ACIT_COMPILETIME_ASSERT(condition) \
switch(0){case 0:case (condition):;}


/*
 * OSM function macros that redirect to the OSM call via OSM API Table
 */
#define osm_complete_io     \
    (CONTROLLER_WS(pcontroller_key)->osm_api_table.posm_complete_io)
#define osm_read_config8    \
    (CONTROLLER_WS(pcontroller_key)->osm_api_table.posm_read_config8)
#define osm_read_config16   \
    (CONTROLLER_WS(pcontroller_key)->osm_api_table.posm_read_config16)
#define osm_read_config32   \
    (CONTROLLER_WS(pcontroller_key)->osm_api_table.posm_read_config32)
#define osm_write_config8   \
    (CONTROLLER_WS(pcontroller_key)->osm_api_table.posm_write_config8)
#define osm_write_config16  \
    (CONTROLLER_WS(pcontroller_key)->osm_api_table.posm_write_config16)
#define osm_write_config32  \
    (CONTROLLER_WS(pcontroller_key)->osm_api_table.posm_write_config32)

/* ACIT TMF Function helper macros */
#define ACIT_IS_TARGET_RESET_TMF(v)             ((v == ACIT_TMF_TARGET_COLD_RESET || v == ACIT_TMF_TARGET_WARM_RESET))
#define ACIT_IS_RESET_TMF(v)                    ((v >= ACIT_TMF_LUN_RESET) &&   \
                                                 (v <= ACIT_TMF_TARGET_COLD_RESET))
#define ACIT_IS_COLD_RESET_TMF(v)               ((v == ACIT_TMF_TARGET_COLD_RESET))

/* macros to retrieve OSM function pointers from controller ws */

#define OSM_GET_CONFIG_VALUE    \
    (pcontroller->osm_api_table.posm_get_config_value)
#define OSM_COMPLETE_IO         \
    (pcontroller->osm_api_table.posm_complete_io)
#define OSM_COMPLETE_IOCTL      \
    (pcontroller->osm_api_table.posm_complete_ioctl)
#define OSM_COMPLETE_TMF        \
    (pcontroller->osm_api_table.posm_complete_tmf)
#define OSM_ASYNC_EVENT         \
    (pcontroller->osm_api_table.posm_async_event)
#define OSM_READ8               \
    (pcontroller->osm_api_table.posm_read8)
#define OSM_READ16              \
    (pcontroller->osm_api_table.posm_read16)
#define OSM_READ32              \
    (pcontroller->osm_api_table.posm_read32)
#define OSM_READ_BUFFER         \
    (pcontroller->osm_api_table.posm_read_buffer)
#define OSM_WRITE8              \
    (pcontroller->osm_api_table.posm_write8)
#define OSM_WRITE16             \
    (pcontroller->osm_api_table.posm_write16)
#define OSM_WRITE32             \
    (pcontroller->osm_api_table.posm_write32)
#define OSM_WRITE_BUFFER        \
    (pcontroller->osm_api_table.posm_write_buffer)
#define OSM_READ_CONFIG8        \
    (pcontroller->osm_api_table.posm_read_config8)
#define OSM_READ_CONFIG16       \
    (pcontroller->osm_api_table.posm_read_config16)
#define OSM_READ_CONFIG32       \
    (pcontroller->osm_api_table.posm_read_config32)
#define OSM_READ_CONFIG_BUFFER  \
    (pcontroller->osm_api_table.posm_read_config_buffer)
#define OSM_WRITE_CONFIG8       \
    (pcontroller->osm_api_table.posm_write_config8)
#define OSM_WRITE_CONFIG16      \
    (pcontroller->osm_api_table.posm_write_config16)
#define OSM_WRITE_CONFIG32      \
    (pcontroller->osm_api_table.posm_write_config32)
#define OSM_WRITE_CONFIG_BUFFER \
    (pcontroller->osm_api_table.posm_write_config_buffer)
#define OSM_GET_BUS_ADDRESS     \
    (pcontroller->osm_api_table.posm_get_bus_address)
#define OSM_DELAY_TIMER         \
    (pcontroller->osm_api_table.posm_delay_timer)
#define OSM_MAP_IO_HANDLE       \
    (pcontroller->osm_api_table.posm_map_io_handle)
#define OSM_RELEASE_IO_HANDLE   \
    (pcontroller->osm_api_table.posm_release_io_handle)
#if 0	
#define OSM_SYNCHRONIZE_RANGE   \
    (pcontroller->osm_api_table.posm_synchronize_range)
#define OSM_POLL_REGISTER       \
    (pcontroller->osm_api_table.posm_poll_register)
#endif	
#define OSM_LOG_ERROR           \
    (pcontroller->osm_api_table.posm_log_error)
#define OSM_LOG_EVENT           \
    (pcontroller->osm_api_table.posm_log_event)

/*
 * ACIT Internal Functions declared below.
 */

ACIT_UEXACT32
ap_flag_busy_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           reason);

ACIT_UEXACT32
ap_flag_ready_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           reason);

ACIT_UEXACT32
ap_flag_busy_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           reason);

ACIT_UEXACT32
ap_flag_ready_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           reason);

ACIT_UEXACT32
ap_flag_busy_lun(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           reason);

ACIT_UEXACT32
ap_flag_ready_lun(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           reason);

ACIT_UEXACT32
ap_set_target_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           state);

ACIT_UEXACT32
ap_clear_target_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           state);

ACIT_UEXACT32
ap_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request);

ACIT_VOID
ap_free_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
ap_complete_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PI_SOL_CQE              psol);

ACIT_VOID
ap_complete_io_error(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    ACIT_UEXACT32           ioh_flag);

ACIT_VOID
ap_get_inv_params(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle,
    PACIT_DC_LIST           element,
    PACIT_UEXACT16          picd,
    PACIT_UEXACT16          pcid);

ACIT_UEXACT32
ap_process_io_disposition_after_timeout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_WS             pio,
    ACIT_UEXACT32           disposition);

ACIT_UEXACT32
ap_complete_invalidate_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle);

ACIT_UEXACT32
ap_invalidate_cmd_list(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_DC_LIST           plist,
    ACIT_UEXACT32           cmd_type,
    ACIT_UEXACT32           toremove);

ACIT_UEXACT32
ap_set_invalidate_icd_status(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle,
    PACIT_DC_LIST           element,
    ACIT_UEXACT32           inv_status,
    ACIT_UEXACT32           byte_count);

ACIT_UEXACT32
ap_complete_invalidate_icds(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle);

ACIT_UEXACT32
ap_return_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_WS             pio,
    ACIT_UEXACT32           io_status,
    ACIT_UEXACT32           unlink);

ACIT_UEXACT32
ap_return_hiber_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request);

PACIT_LOGOUT_HANDLE
ap_alloc_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT32           reason_code);

ACIT_VOID
ap_free_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle);

ACIT_UEXACT32
ap_complete_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle,
    ACIT_UEXACT8            iresponse,
    ACIT_UEXACT8            timeout);


PACIT_NOP_HANDLE
ap_alloc_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    PACIT_VOID              ppdu,
    PACIT_NOP_HANDLE        poptional_handle);

ACIT_VOID
ap_free_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle);

ACIT_UEXACT32
ap_send_nopout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    PACIT_VOID              pdu,
    PACIT_NOP_HANDLE        poptional_handle);

ACIT_UEXACT32
ap_setup_ioh_for_tmf(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
ap_ready_for_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

PACIT_TMF_HANDLE
ap_alloc_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle,
    PACIT_TMF_WS            ptmf,
    ACIT_UEXACT32           tmf_function);

ACIT_UEXACT32
ap_reinit_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle);

ACIT_UEXACT32
ap_process_io_tick(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_process_tmf_tick(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_process_logout_tick(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_process_target_tick(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_prepare_vase(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_initialize_vase(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_UEXACT32          pmemory);

ACIT_VOID
ap_set_mem_const (
    PACIT_CONTROLLER_WS     pcontroller,
    const PACIT_UEXACT8     name,
    ACIT_UEXACT32           index,
    ACIT_UEXACT32           size_in_bytes,
    ACIT_UEXACT32           alignment_mask,
    ACIT_UEXACT32           category);

ACIT_VOID
ap_set_mem_variable (
    PACIT_CONTROLLER_WS     pcontroller,
    const PACIT_UEXACT8     name,
    ACIT_UEXACT32           index,
    ACIT_UEXACT32           num_elements,
    ACIT_UEXACT32           element_size,
    ACIT_UEXACT32           alignment_mask,
    ACIT_UEXACT32           category);

ACIT_UEXACT32
ap_init_all_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_init_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_init_io_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_init_logout_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_init_tmf_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_init_nop_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_init_invalidate_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_assign_sglh2_ioh(
    PACIT_CONTROLLER_WS     pcontroller);

PACIT_IO_HANDLE
ap_alloc_io_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_free_io_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle);

// No MCS support
#if 0
ACIT_STATUS
ap_init_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           connection_handle);
    
ACIT_STATUS
ap_add_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           session_handle, 
    ACIT_UEXACT32           connection_handle);
#endif

ACIT_STATUS
ap_add_target(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           session_handle, 
    ACIT_UEXACT8            domain,
    ACIT_UEXACT16           error_status,
    ACIT_UEXACT8            is_login_pending);

ACIT_STATUS
ap_check_duplicate_target(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           session_handle,
    PACIT_UEXACT32          ptarget_index);

ACIT_VOID
ap_delete_target_entry(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           target_index);

ACIT_STATUS
ap_remove_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_UEXACT32
ap_delete_all_targets(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_flush_lun_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);

ACIT_UEXACT32
ap_flush_target_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_UEXACT32
ap_flush_controller_tmfs(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_flush_queued_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);

ACIT_UEXACT32
ap_complete_discovery(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_start_abort_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request);

ACIT_UEXACT32
ap_start_lun_reset(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request);

ACIT_UEXACT32
ap_start_target_reset(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request);

ACIT_UEXACT32
ap_start_abort_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request);

ACIT_VOID
ap_return_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_REQUEST       ptmf_request,
    ACIT_UEXACT32           tmf_request);

ACIT_UEXACT32
ap_complete_abort_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT32           tmf_status);

ACIT_UEXACT32
ap_check_and_issue_abort(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
ap_complete_abort_multiple_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT32           tmf_status);

ACIT_UEXACT32
ap_complete_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT8            iresponse,
    ACIT_UEXACT8            max_cmd_sn_offset,
    ACIT_UEXACT32           exp_cmd_sn);

ACIT_UEXACT32
ap_complete_nop_in(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle);

ACIT_UEXACT32
ap_abort_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
ap_reset_lun(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);

ACIT_UEXACT32
ap_reset_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           tmf_function);

PACIT_INVALIDATE_HANDLE
ap_alloc_inv_resources(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_free_tmf_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle);

PACIT_LOGOUT_HANDLE
ap_alloc_logout_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_free_logout_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle);

PACIT_TMF_HANDLE
ap_alloc_tmf_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_free_tmf_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle);

PACIT_NOP_HANDLE
ap_alloc_nop_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_free_nop_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle);

PACIT_INVALIDATE_HANDLE
ap_alloc_inv_handle(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_free_inv_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle);

ACIT_VOID
ap_queue_io_ws(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TARGET_WS             ptarget,
    PACIT_LUN_WS                plun,
    PACIT_IO_WS                 pio);

ACIT_VOID
ap_queue_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle);

ACIT_VOID
ap_dequeue_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle);

ACIT_UEXACT32
ap_return_queued_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           io_status,
    ACIT_UEXACT32           port_num);

#if ACIT_CFG_DEBUG_BUILD

ACIT_VOID
ap_validate_iows_list(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);

ACIT_VOID
ap_validate_lists(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    PACIT_LUN_WS            plun);

ACIT_VOID
ap_print_cmd_ticks(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    PACIT_LUN_WS            plun);

#endif

ACIT_UEXACT32
ap_free_pending_handles(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           port_num);

ACIT_VOID
ap_free_queued_ws_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           free_ios);

ACIT_UEXACT32
ap_return_queued_lun_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    ACIT_UEXACT32           tmf_status,
    ACIT_UEXACT32           free_handle);

ACIT_UEXACT32
ap_return_queued_target_tmfs(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           tmf_status,
    ACIT_UEXACT32           free_handle);

ACIT_UEXACT32
ap_invalidate_pending_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_HANDLE         pio_handle,
    ACIT_UEXACT32           cri);

PACIT_TARGET_WS
ap_get_target_ws_from_session_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           session_handle);

ACIT_UEXACT32
ap_init_eto(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           eto);

ACIT_UEXACT32
ap_start_eto(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_UEXACT32
ap_stop_eto(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_VOID
ap_process_eto_expiry(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_VOID
ap_force_eto_expiry(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_UEXACT32
ap_setup_lun_for_reset(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);

ACIT_UEXACT32
ap_setup_target_for_cleanup(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT32           cleanup_flag,
    ACIT_UEXACT32           is_abortive);

ACIT_UEXACT32
ap_complete_target_recovery(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           status,
    ACIT_UEXACT32           ignore_link);

ACIT_UEXACT32
ap_complete_all_target_recovery(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_send_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT32           reason_code);

ACIT_UEXACT32
ap_process_target_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           async);

ACIT_UEXACT32
ap_log_osm_event(
    PACIT_CONTROLLER_WS     pcontroller,
    UNI_ERROR_LEVEL         level,
    UNI_SUBMODULE           submodule,
    UNI_PHASE               phase,
    ACIT_UEXACT8            addi_len,
    ACIT_UEXACT16           error_code,
    ACIT_UEXACT32           dump_data_size,
    PACIT_VOID              pdump_data,
    PACIT_UEXACT32          paddi_dwords);

ACIT_UEXACT32
ap_process_ldto_tick(      
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_process_link_down(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           port_num);
 
ACIT_UEXACT32
ap_process_link_up(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           port_num);

#define IS_TARGET_MISSING(ptarget) (ptarget->sub_state & TARGET_SUB_STATE_ETO_IN_PROGRESS)

#define IS_CONTROLLER_LDTO_EXPIRED(pc, port)    \
    (ACIT_IS_CONTROLLER_LINK_DOWN(pc, port) && !pc->ld_timeout_ticks[port])

#define IS_TARGET_ETO_EXPIRED(ptarget)  (ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_ETO_EXPIRED))

#define IS_CONTROLLER_UE(pc)   (ACIT_IS_BIT_SET(pc->busy, CONTROLLER_BUSY_UE))

//#define GET_TARGET_EXEC_TICKS(ptarget, retry_count)          \
//    ((retry_count + 1) * ptarget->last_cmd_ticks * ptarget->timeout_factor)

#define GET_TMF_TIMEOUT_TICKS(pc, pt)    (pc->params.tmf_timeout_ticks)

#define ACIT_IS_UE_RECOVERY_ENABLED(pc) ((pc->params.ue_recovery) && (pc->eh_stats.ue_recoveries < pc->params.max_ue_recoveries))

// Cleanup Reasons. This affects the IO status and TMF status returned
#define CL_REASON_UE_TOLERANCE      1
#define CL_REASON_LDTO_EXPIRY       2
#define CL_REASON_ETO_EXPIRY        4

INLINE
ACIT_UEXACT32
ap_decide_io_disposition(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_IO_WS             pio);

ACIT_VOID
ap_save_cmd_state(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           itt,
    ACIT_UEXACT32           cmd_sn);

ACIT_VOID
ap_print_cmd_state(
    PACIT_CONTROLLER_WS     pcontroller);

#if ACIT_CFG_DEBUG_BUILD

ACIT_VOID
ap_process_dbg_flags(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_dbg_dump_pending_cmds(
    PACIT_CONTROLLER_WS     pcontroller);

#endif

ACIT_STATUS
ap_get_param_num (
    PACIT_CONTROLLER_KEY    pcontroller_key, 
    const ACIT_PARAM_PROPERTIES *pp, 
    ACIT_UEXACT32           *mask,
    ACIT_UEXACT32           operating_mem_level);

ACIT_UEXACT32
ap_init_fw_config(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_determine_load(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_UEXACT32          pio_load,
    PACIT_UEXACT32          pload_factor,
    PACIT_UEXACT32          ptarget_load,
    ACIT_UEXACT32           pid);

#if 0

ACIT_UEXACT32
ap_determine_load(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_UEXACT32          pio_load,
    PACIT_UEXACT32          pload_factor,
    PACIT_UEXACT32          ptarget_factor);

#endif

ACIT_UEXACT32
ap_process_scsi_event(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    PACIT_VOID              psense_data, 
    ACIT_UEXACT16           sense_len);

ACIT_VOID
ap_update_lun_retry_param(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LUN_WS            plun);

ACIT_VOID
ap_cleanup_target(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cleanup_type);

ACIT_VOID
ap_process_ue_occurrance(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_flag_ue_recovery_complete(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
ap_complete_ue_recovery(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           recovery_status);

INLINE
ACIT_UEXACT32
ap_update_io_ticks(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun,
    PACIT_IO_WS             pio);

ACIT_VOID
ap_decide_io_disposition_in_cleanup(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle);

ACIT_VOID
ap_recalc_io_tick(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PACIT_IO_WS             pio);
