/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */


#ifndef ACIT_IFACE_H
#define ACIT_IFACE_H
/* ACIT API Table. Common to all hba instances */
extern ACIT_API_TABLE          acitapi_table;
extern struct se_param_struct se_param[];
#define SHTDWN_EVENT 0xFE
#define BE2ISCSI_OSM_API_TABLE {                                           \
        osm_api_table_version     :     {1,1},                            \
        posm_get_config_value     :     be2iscsi_get_config_value,         \
        posm_complete_io          :     be2iscsi_complete_io,              \
        posm_complete_ioctl       :     be2iscsi_complete_ioctl, 	  \
        posm_complete_tmf         :     be2iscsi_complete_tmf,             \
        /*posm_complete_lun_reset   :     be2iscsi_complete_lun_reset,*/   \
        /*posm_complete_target_reset:     be2iscsi_complete_target_reset,*/\
        posm_async_event          :     be2iscsi_async_event,              \
        posm_read8                :     be2iscsi_read8,                    \
        posm_read16               :     be2iscsi_read16,                   \
        posm_read32               :     be2iscsi_read32,                   \
        posm_read_buffer          :     be2iscsi_read_buffer,              \
        posm_write8               :     be2iscsi_write8,                   \
        posm_write16              :     be2iscsi_write16,                  \
        posm_write32              :     be2iscsi_write32,                  \
        posm_write_buffer         :     be2iscsi_write_buffer,             \
        posm_read_config8         :     be2iscsi_read_config8,             \
        posm_read_config16        :     be2iscsi_read_config16,            \
        posm_read_config32        :     be2iscsi_read_config32,            \
        /*posm_read_config_buffer   :     be2iscsi_read_config_buffer,   */\
        posm_write_config8        :     be2iscsi_write_config8,            \
        posm_write_config16       :     be2iscsi_write_config16,           \
        posm_write_config32       :     be2iscsi_write_config32,           \
        /*posm_write_config_buffer  :     be2iscsi_write_config_buffer,  */\
        posm_get_bus_address      :     be2iscsi_get_bus_address,          \
        posm_delay_timer          :     be2iscsi_delay_timer,          	\
        posm_map_io_handle        :     be2iscsi_map_io_handle,            \
        posm_release_io_handle    :     be2iscsi_release_io_handle,        \
        posm_log_error            :     be2iscsi_log_error		    \
        /*posm_synchronize_range    :     be2iscsi_synchronize_range,    */\
        /*posm_poll_register        :     be2iscsi_poll_register         */\
}

struct be2iscsi_disp_struct {
 unsigned int  error_code;
 char 	       *str;
 unsigned int  count;
 unsigned int  seq[3];
};



inline static
ACIT_STATUS
acitapi_get_api_version(
                PACIT_VERSION   pacit_api_version)
{
	BEGIN;	
        return acitapi_table.pacit_get_api_version(pacit_api_version);
}

inline static
ACIT_STATUS
acitapi_get_supported_device(
                ACIT_UEXACT32   index,
                PACIT_DEVICE    pacit_device)
{
	BEGIN;	
        return acit_get_supported_device(
                                index, pacit_device);

}

inline static
ACIT_STATUS
acitapi_check_supported_device(
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_DEVICE            pacit_device)
{
	BEGIN;	
        return acit_check_supported_device(
                                pcontroller_key, pacit_device);
}

inline static
ACIT_STATUS
acitapi_get_ws_sizes(
                PACIT_VERSION   pacit_api_version,
                PACIT_UEXACT32  pcontroller_ws_size,
                PACIT_UEXACT32  ptarget_ws_size,
                PACIT_UEXACT32  plun_ws_size,
                PACIT_UEXACT32  pio_ws_size,
                PACIT_UEXACT32  ptmf_ws_size)
{
	BEGIN;	
        return acit_get_ws_sizes( pacit_api_version,
			pcontroller_ws_size,
			ptarget_ws_size, plun_ws_size, pio_ws_size, 
			ptmf_ws_size);

}
        

inline static
ACIT_STATUS
acitapi_init_controller_ws(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_VOID              pcontroller_ws)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_init_controller_ws(
                                pcontroller_key, pcontroller_ws);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;

}

inline static
ACIT_STATUS
acitapi_init_target_ws(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_TARGET_KEY        ptarget_key,
                PACIT_VOID              ptarget_ws)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_init_target_ws(
                                pcontroller_key, ptarget_key,
                                ptarget_ws);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_init_lun_ws(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_TARGET_KEY        ptarget_key,
                PACIT_LUN_KEY           plun_key,
                PACIT_VOID              plun_ws)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_init_lun_ws(
                                pcontroller_key, ptarget_key,
                                plun_key, plun_ws);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_update_lun_info(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_TARGET_KEY        ptarget_key,
                PACIT_LUN_KEY           plun_key,
                PACIT_VOID              plun_info)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_update_lun_info(
                                pcontroller_key, ptarget_key,
                                plun_key, plun_info);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}


inline static
ACIT_STATUS
acitapi_init_io_ws(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_IO_REQUEST        pio_request,
                PACIT_IO_KEY            pio_key,
                se_target 		*ptarget,
                PACIT_VOID              pio_ws)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)&&!(ptarget->state & SE_TARGET_OFFLINE )) {
		acit_status = acitapi_table.pacit_init_io_ws(
                                pcontroller_key, pio_request,
                                pio_key, pio_ws);
	} else {
		acit_status = SHTDWN_EVENT;	
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;

}


inline static
ACIT_STATUS
acitapi_init_tmf_ws(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_TMF_REQUEST        ptmf_request,
                PACIT_TMF_KEY            ptmf_key,
                PACIT_VOID              ptmf_ws)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)) {
        	acit_status = acitapi_table.pacit_init_tmf_ws(
                                pcontroller_key, ptmf_request,
                                ptmf_key, ptmf_ws);
	} else {
		acit_status = SHTDWN_EVENT;	
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;

}
inline static
ACIT_STATUS
acitapi_set_osm_api_table(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_OSM_API_TABLE     posm_api_table,
                ACIT_UEXACT32           table_length)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_set_osm_api_table(
                                pcontroller_key, posm_api_table,
                                table_length);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_init_configurable_params(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_init_configurable_params(
                                pcontroller_key);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;

}

inline static
ACIT_STATUS
acitapi_get_mem_params(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_UEXACT32          poperating_mem_level)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_get_mem_params(
                                pcontroller_key, poperating_mem_level);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;

}

inline static
ACIT_STATUS
acitapi_set_mem_params(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                ACIT_UEXACT32           operating_mem_level)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_set_mem_params(
                                pcontroller_key, operating_mem_level);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;

}

inline static
ACIT_STATUS
acitapi_get_mem_descriptor(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_DEVICE            pacit_device,
                ACIT_UEXACT32           index,
                PACIT_MEM_DESCRIPTOR    pmem_descriptor)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_get_mem_descriptor(
                                pcontroller_key, pacit_device, index,
                                pmem_descriptor);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_set_mem_descriptor(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                ACIT_UEXACT32           index,
                PACIT_MEM_DESCRIPTOR    pmem_descriptor)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_set_mem_descriptor(
                                pcontroller_key, index,
                                pmem_descriptor);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_init_controller(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                ACIT_UEXACT32           init_type)
{
        ACIT_STATUS     acit_status;
	BEGIN;	
	spin_lock(&phba->acitapi_lock);
	acit_status = acitapi_table.pacit_init_controller(pcontroller_key, 
				init_type,
				phba->pcidev);
      	        spin_unlock(&phba->acitapi_lock);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_stop_controller(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                ACIT_UEXACT32           stop_type)
{
        ACIT_STATUS     acit_status;
	BEGIN;	
        acit_status = acitapi_table.pacit_stop_controller(pcontroller_key,
                                                            stop_type); 
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_discover_targets(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_DISCOVER_PARAM    pdiscover_param)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_discover_targets(
                                pcontroller_key, pdiscover_param);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_enable_irq(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_enable_irq(pcontroller_key);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}
        

inline static
ACIT_STATUS
acitapi_disable_irq(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        acit_status = acitapi_table.pacit_disable_irq(pcontroller_key);
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_process_irq(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)) {
        	acit_status = acitapi_table.pacit_process_irq(pcontroller_key);
	} else {
		acit_status = SHTDWN_EVENT;	
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
	END;
        return acit_status;
}


inline static
ACIT_STATUS
acitapi_process_irq_msix(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)) {
        	acit_status = acitapi_table.pacit_process_msi_irq(pcontroller_key,0);//Why is this message id required?
	} else {
		acit_status = SHTDWN_EVENT;	
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
	END;
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_start_io(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                se_target		*ptarget,
                PACIT_IO_REQUEST        pio_request)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)&&!(ptarget->state & SE_TARGET_OFFLINE )) {
        	acit_status = acitapi_table.pacit_start_io(
                                pcontroller_key, pio_request);
	} else {
		acit_status = SHTDWN_EVENT;	
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
	END;
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_start_tmf(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_TMF_REQUEST       ptmf_request)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
	BEGIN;	
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)) {
        	acit_status = acitapi_table.pacit_start_tmf(
                                pcontroller_key, ptmf_request);
	} else {
		acit_status = SHTDWN_EVENT;	
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
	END;
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_start_ioctl(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key,
                PACIT_IOCTL_REQUEST     pioctl_request)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
        pioctl_request->ioctl_flag = 0;
	BEGIN;
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)) {
        	acit_status = acitapi_table.pacit_start_ioctl(
                                pcontroller_key, pioctl_request);
	} else {
		acit_status = SHTDWN_EVENT;	
		complete(&phba->ioctl_completion);
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
	END;
        return acit_status;
}

inline static
ACIT_STATUS
acitapi_process_timer_tick(
                se_hba                *phba,
                PACIT_CONTROLLER_KEY    pcontroller_key)
{
        ACIT_U32_64     flags = 0;
        ACIT_STATUS     acit_status;
        spin_lock_irqsave(&phba->acitapi_lock, flags);
        if(!(phba->state & SE_STATE_SHUTDOWN)) {
        	acit_status = acitapi_table.pacit_process_timer_tick(pcontroller_key);
	} else {
		acit_status = SHTDWN_EVENT;	
	}
        spin_unlock_irqrestore(&phba->acitapi_lock, flags);
        return acit_status;
}


#if 0

#define acitapi_get_ws_sizes            \
        acitapi_table.pacit_get_ws_sizes

#define acitapi_init_controller_ws      \
        acitapi_table.pacit_init_controller_ws

#define acitapi_init_target_ws          \
        acitapi_table.pacit_init_target_ws

#define acitapi_init_lun_ws             \
        acitapi_table.pacit_init_lun_ws

#define acitapi_init_io_ws              \
        acitapi_table.pacit_init_io_ws

#define acitapi_set_osm_api_table       \
        acitapi_table.pacit_set_osm_api_table

#define acitapi_init_configurable_params\
        acitapi_table.pacit_init_configurable_params

#define acitapi_get_mem_descriptor          \
        acitapi_table.pacit_get_mem_descriptor

#define acitapi_set_mem_descriptor            \
        acitapi_table.pacit_set_mem_descriptor

#define acitapi_init_controller         \
        acitapi_table.pacit_init_controller

#define acitapi_get_mem_params          \
        acitapi_table.pacit_get_mem_params

#define acitapi_set_mem_params          \
        acitapi_table.pacit_set_mem_params

#define acitapi_discover_targets        \
        acitapi_table.pacit_discover_targets

#define acitapi_enable_irq              \
        acitapi_table.pacit_enable_irq

#define acitapi_disable_irq             \
        acitapi_table.pacit_disable_irq

#define acitapi_process_irq             \
        acitapi_table.pacit_process_irq

#define acitapi_process_irq_msix        \
        acitapi_table.pacit_process_msi_irq

#define acitapi_start_io                \
        acitapi_table.pacit_start_io

#define acitapi_process_timer_tick      \
        acitapi_table.pacit_process_timer_tick

#define acitapi_get_capabilities        \
        acitapi_table.pacit_get_capabilities

#endif

ACIT_UEXACT32
be2iscsi_get_config_value(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        PACIT_CONFIG_PARAM      pconfig_param);

ACIT_UEXACT32
be2iscsi_complete_io(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        PACIT_IO_REQUEST        pio_request);

ACIT_UEXACT32
be2iscsi_complete_tmf(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        PACIT_TMF_REQUEST       ptmf_request);

ACIT_UEXACT32
be2iscsi_complete_ioctl(
	PACIT_CONTROLLER_KEY	pcontroller_key,
	PACIT_IOCTL_REQUEST	pioctl_request);

ACIT_UEXACT32
be2iscsi_complete_abort_io(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        PACIT_IO_REQUEST        pio_request);

ACIT_UEXACT32
be2iscsi_complete_lun_reset(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        PACIT_IO_REQUEST        pio_request);

ACIT_UEXACT32
be2iscsi_complete_target_reset(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        PACIT_IO_REQUEST        pio_request);

ACIT_UEXACT32
be2iscsi_async_event(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           event_class,
        ACIT_UEXACT32           event_code,
        ACIT_UEXACT32           info);

ACIT_UEXACT8
be2iscsi_read8(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset);

ACIT_UEXACT16
be2iscsi_read16(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset);

ACIT_UEXACT32
be2iscsi_read32(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset);

ACIT_UEXACT32
be2iscsi_read_buffer(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset,
        PACIT_VOID              pbuffer,
        ACIT_UEXACT32           length);

ACIT_VOID
be2iscsi_write8(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset,
        ACIT_UEXACT8            value);

ACIT_VOID
be2iscsi_write16(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset,
        ACIT_UEXACT16           value);

ACIT_VOID
be2iscsi_write32(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset,
        ACIT_UEXACT32           value);

ACIT_UEXACT32
be2iscsi_write_buffer(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle,
        ACIT_UEXACT32           offset,
        PACIT_VOID              pbuffer,
        ACIT_UEXACT32           length);

ACIT_UEXACT32
be2iscsi_read_config8(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           offset,
        PACIT_UEXACT8           pbuffer);

ACIT_UEXACT32
be2iscsi_read_config16(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           offset,
        PACIT_UEXACT16          pbuffer);

ACIT_UEXACT32
be2iscsi_read_config32(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           offset,
        PACIT_UEXACT32          pbuffer);

ACIT_UEXACT32
be2iscsi_read_config_buffer(PACIT_CONTROLLER_KEY pcontroller_key);

ACIT_UEXACT32
be2iscsi_write_config8(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           offset,
        ACIT_UEXACT8            value);

ACIT_UEXACT32
be2iscsi_write_config16(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           offset,
        ACIT_UEXACT16           value);

ACIT_UEXACT32
be2iscsi_write_config32(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           offset,
        ACIT_UEXACT32           value);

ACIT_UEXACT32
be2iscsi_write_config_buffer(PACIT_CONTROLLER_KEY pcontroller_key);

ACIT_BUS_ADDRESS
be2iscsi_get_bus_address(
        PACIT_CONTROLLER_KEY    pcontroller_key,
	PACIT_VOID		prequest,
        PACIT_VOID              pvirtual_address,
	PACIT_UEXACT32		plength);

ACIT_UEXACT32
be2iscsi_delay_timer(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           time);

ACIT_UEXACT32
be2iscsi_map_io_handle(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_UEXACT32           range_index,
        ACIT_UEXACT32           length,
        ACIT_UEXACT32           attributes,
	ACIT_BUS_ADDRESS *      pbus_address,
        PACIT_VOID *            pvirtual_address,
        PACIT_REGISTER_HANDLE   pregister_handle);

ACIT_UEXACT32
be2iscsi_release_io_handle(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        ACIT_REGISTER_HANDLE    register_handle);

ACIT_UEXACT32
be2iscsi_synchronize_range(PACIT_CONTROLLER_KEY    pcontroller_key);

ACIT_UEXACT32
be2iscsi_poll_register(PACIT_CONTROLLER_KEY    pcontroller_key);


ACIT_VOID be2iscsi_log_error(PACIT_CONTROLLER_KEY controller_key, ACIT_UEXACT32 error_code);

#endif
