/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        acitapi.c

    Purpose:

        This file implements all functions that are exported by the ACIT
        library for OSMs.


--*/


#include "acitpch.h"

ACIT_VOID
acit_unitt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
acit_unitt_init_hiber_controller(
    PACIT_CONTROLLER_WS     pcontroller);
/*
 * Definition for Number of Hardware devices supported by ACIT Library
 */
#define ACIT_NUMBER_OF_DEVICES_SUPPORTED        (0x05)

/*
 * Max number of storage functions that can be on the system
 */
#define ACIT_MAX_DEVICES                        (32)

static PACIT_CONTROLLER_WS     g_pcontroller[ACIT_MAX_DEVICES] = {0};
static ACIT_UEXACT32           acit_api_type = 0;

/*
 * Static tables used by the ACIT library.
 */
static const
ACIT_API_TABLE     acit_api_table =
{
    {ACIT_API_HI_VERSION, ACIT_API_LO_VERSION},
    acit_get_api_version,
    acit_init_controller_ws,
    acit_init_target_ws,
    acit_init_lun_ws,
    acit_init_io_ws,
    acit_init_tmf_ws,
    acit_set_osm_api_table,
    acit_init_configurable_params,
    acit_set_mem_params,
    acit_get_mem_params,
    acit_get_mem_descriptor,
    acit_set_mem_descriptor,
    acit_init_controller,
    acit_stop_controller,
    acit_discover_targets,
    acit_enable_irq,
    acit_disable_irq,
    acit_process_irq,
    acit_process_msi_irq,
    acit_start_io,
    acit_start_ioctl,
    acit_start_tmf,
    acit_process_timer_tick,
    acit_update_lun_info,
    acit_get_capabilities
};

/*
 * This is the API table that will be used for crash dump
 * mode. ACIT operating mode is ACIT_OP_RESTRICTED if this
 * table is to be loaded
 */

ACIT_STATUS
acit_start_hiber_io(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request);

ACIT_STATUS
acit_process_hiber_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key);

static const
ACIT_API_TABLE     acit_min_api_table =
{
    {ACIT_API_HI_VERSION, ACIT_API_LO_VERSION},
    acit_get_api_version,
    acit_init_controller_ws,
    acit_init_target_ws,
    acit_init_lun_ws,
    acit_init_io_ws,
    acit_init_tmf_ws,
    acit_set_osm_api_table,
    acit_init_configurable_params,
    acit_set_mem_params,
    acit_get_mem_params,
    acit_get_mem_descriptor,
    acit_set_mem_descriptor,
    acit_init_controller,
    acit_stop_controller,
    acit_discover_targets,
    acit_enable_irq,
    acit_disable_irq,
    acit_process_hiber_irq,
    NULL,
    acit_start_hiber_io,
    acit_start_ioctl,
    acit_start_tmf,
    acit_process_timer_tick,
    acit_update_lun_info,
    acit_get_capabilities
};



/*
 * Updating acit_device structure as per EFS Rev0.92
 * Note: Exporting more than 1 device causes Windows 2000
 * default OS installation to fail. Slipstreaming SP4 into
 * the Windows 2000 OS CD will solve this issue
 */

static const
ACIT_DEVICE     acit_device[ACIT_NUMBER_OF_DEVICES_SUPPORTED] =
{    
  {
     {
      ACIT_API_HI_VERSION,
      ACIT_API_LO_VERSION
     },
     BE_PCI_VENDORID_STR,
     BE_PCI_DEVICEID0_STR,
     BE_PCI_VENDORID,
     BE_PCI_DEVICEID0,
     0,     // subsystem device id not used
     0,     // subsystem vendor id not used
     0,     // revision id not used
     0x0,   // rsvd
     ACIT_DEVICE_BUS_PCI
  },

  {
     {
      ACIT_API_HI_VERSION,
      ACIT_API_LO_VERSION
     },
     BE_PCI_VENDORID_STR,
     BE_PCI_DEVICEID1_STR,
     BE_PCI_VENDORID,
     BE_PCI_DEVICEID1,
     0,     // subsystem device id not used
     0,     // subsystem vendor id not used
     0,     // revision id not used
     0x0,   // rsvd
     ACIT_DEVICE_BUS_PCI
  },

  {
     {
      ACIT_API_HI_VERSION,
      ACIT_API_LO_VERSION
     },
     BE_PCI_VENDORID_STR,
     BE_PCI_DEVICEID2_STR,
     BE_PCI_VENDORID,
     BE_PCI_DEVICEID2,
     0,     // subsystem device id not used
     0,     // subsystem vendor id not used
     0,     // revision id not used
     0x0,   // rsvd
     ACIT_DEVICE_BUS_PCI
  },

  {
     {
      ACIT_API_HI_VERSION,
      ACIT_API_LO_VERSION
     },
     BE_PCI_VENDORID_STR,
     BE_PCI_DEVICEID3_STR,
     BE_PCI_VENDORID,
     BE_PCI_DEVICEID3,
     0,     // subsystem device id not used
     0,     // subsystem vendor id not used
     0,     // revision id not used
     0x0,   // rsvd
     ACIT_DEVICE_BUS_PCI
  },

  {
     {
      ACIT_API_HI_VERSION,
      ACIT_API_LO_VERSION
     },
     BE_PCI_VENDORID_STR,
     BE_PCI_DEVICEID4_STR,
     BE_PCI_VENDORID,
     BE_PCI_DEVICEID4,
     0,     // subsystem device id not used
     0,     // subsystem vendor id not used
     0,     // revision id not used
     0x0,   // rsvd
     ACIT_DEVICE_BUS_PCI
  }

};

/*++

Function:
    acit_initialize

Parameter:
    operating_mode  [in] Operating mode for ACIT library.
    pacit_api_table [in] Number of bytes that can be copied at given location.
    table_length    [in] Number of valid bytes of memory available at location identified by
                         pacit_api_table pointer.
Return:
    None

Context:
    This function can be called from Interrupt Context

Description:

    This function is used by OSMs to initialize the ACIT library with a 
    particular operation mode. ACIT library supports a normal operating mode 
    and a restricted mode. The restricted operating mode is characterised by 
    lower numbers in terms of number of targets, I/O performance and queue depth 
    parameters, but also has the benifit of low memory utilization.

    This API also returns a copy of the ACIT_API_TABLE structure back to OSMs. 
    This structure contain function pointer to all the exported ACIT library 
    function. Once OSMs obtain this table, they can make ACIT library function calls. 
    The API table is copied to the memory location identified by the pacit_api_table. 
    If the size specified by table_length is smaller than the size of the ACIT API 
    table, only the size specified is copied, leaving the OSM with an incomplete 
    API table. If the length specified is larger than the size of the ACIT API table, 
    only the valid length is copied.

--*/

ACIT_STATUS
acit_initialize(
    ACIT_UEXACT32               operating_mode,
    PACIT_API_TABLE             pacit_api_table,
    ACIT_UEXACT32               table_length)
{
    OSM_ASSERT(pacit_api_table);
    OSM_ASSERT(table_length);

    switch(operating_mode)
    {
    case ACIT_OP_MODE_NORMAL:
        OSM_MEMCPY(pacit_api_table, (PACIT_VOID)&acit_api_table, table_length);
        acit_api_type = ACIT_OP_MODE_NORMAL;
        break;
        
    case ACIT_OP_MODE_RESTRICTED:
        OSM_MEMCPY(pacit_api_table, (PACIT_VOID)&acit_min_api_table, table_length);
        acit_api_type = ACIT_OP_MODE_RESTRICTED;
        break;

    default:
        OSM_ASSERT(0);
        break;
    }

    // setup SA's debug levels
#if ACIT_CFG_DEBUG_BUILD
    sa_trace_set_level(DL_ERR | DL_WARN | DL_NOTE);
#endif
    return ACIT_STATUS_SUCCESS;
}

/*++

Function:
    acit_get_supported_device

Parameter:
    index           [in ] Zero based index for each controller.
    pacit_device    [out] Pointer where PCI identifiers are returned.

Return:
    ACIT_STATUS_SUCCESS if information about a device is being returned.
    ACIT_STATUS_NO_DEVICE if no more devices are supported.

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to build a list of devices supported by
    the ACIT library by retrieving the list of PCI attributes like DeviceID,
    VendorID, RevisionCode, SubSystemID and SubSystemVendorID. This list may
    then be used by OSMs to scan the PCI bus to look for supported
    controllers available in the system. Parameter index is provided by the
    OSM. ACIT library fills in the structure at pacit_device and returns.
    Typically, OSM calls this function in a loop starting with an index of 0
    and repeats this process incrementing the index value by one each time.
    This process is repeated until the function returns ACIT_STATUS_NO_DEVICE.

--*/

ACIT_STATUS
acit_get_supported_device(
    ACIT_UEXACT32           index,
    PACIT_DEVICE            pacit_device)
{
    OSM_ASSERT(pacit_device);

    if(index < ACIT_NUMBER_OF_DEVICES_SUPPORTED)
    {
        /*
         * Structure copy from our static table to the specified pointer
         */
        *pacit_device = acit_device[index];

        return ACIT_STATUS_SUCCESS;
    }

    return ACIT_STATUS_NO_DEVICE;
}



/*++

Function:
    acit_check_supported_device

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pacit_device    [in ] Pointer where PCI identifiers are returned.

Return:
    ACIT_STATUS_SUCCESS if the values specified in pacit_device is supported.
    ACIT_STATUS_NO_DEVICE if the values specified in pacit_device is not
        supported.

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to check if the ACIT library supports a
    given controller. This is usually used by OSMs when the host OS implements
    a PCI bus scan and invokes the OSM to handle a specific PCI device. If a
    pcontroller_key parameter of non-NULL is specified, the ACIT library may
    validate the identified controller by using the OSM API functions like
    osm_read_configXX to read the PCI configuration space of the controller.
    OSMs should only specify a non-NULL pcontroller_key after a call to
    acit_init_controller_ws has successfully completed.

--*/

ACIT_STATUS
acit_check_supported_device(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DEVICE                pacit_device)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_U32                    index;
    PCICFG_ID_F0_CSR            id;
    ACIT_STATUS                 status;

    /*
     * Cannot assert on pcontroller_key as it may not be provided.
     */
    if(pcontroller_key)
    {

        pcontroller = GET_CONTROLLER_WS(pcontroller_key);

        OSM_ASSERT(pcontroller);

        id.dw = ACIT_PCICFG0_READ (pcontroller, id);

    }
    else
    {
        /*
         * No controller key was provided, so we just check for the specified
         * device in our internal table.
         */
        id.vendorid = pacit_device->vendor_id;
        id.deviceid = pacit_device->device_id;
    }

    /* First assume it is unknown device. */
    status = ACIT_STATUS_NO_DEVICE;

    for(index = 0; index < ACIT_NUMBER_OF_DEVICES_SUPPORTED; index ++)
    {
        if(
            (id.vendorid == acit_device[index].vendor_id) &&
            (id.deviceid == acit_device[index].device_id)
            )
        {
            status = ACIT_STATUS_SUCCESS;
            break;
        }
    }

    return status;
}



/*++

Function:
    acit_get_api_version

Parameter:
    pacit_api_version [out] Pointer where API version is returned

Return:
    ACIT_STATUS_NO_DEVICE
    ACIT_STATUS_INVALID_PARAMETER

Context:
    This function can be called from Interrupt Context

Description:
    This function implements the code to report the ACIT API Version
    to the OSM. OSMs use this to ensure the interface has not changed
    un-expectedly.

--*/

ACIT_STATUS
acit_get_api_version(
    PACIT_VERSION               pacit_api_version)
{
    OSM_ASSERT(pacit_api_version);

    pacit_api_version->hi_ver = ACIT_API_HI_VERSION;
    pacit_api_version->lo_ver = ACIT_API_LO_VERSION;

    return ACIT_STATUS_SUCCESS;
}



/*++

Function:
    acit_get_ws_sizes

Parameter:
    pacit_api_version   [out] Version of ACIT library.
    pcontroller_ws_size [out] Pointer where controller WS size is returned.
    ptarget_ws_size     [out] Pointer where target WS size is returned.
    plun_ws_size        [out] Pointer where lun WS size is returned.
    pio_ws_size         [out] Pointer where I/O WS size is returned.
    ptmf_ws_size        [out] Pointer where TMF WS size is returned.

Return:
    ACIT_STATUS_SUCCESS.

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to get the workspace sizes required by the
    ACIT library. This is used by the OSMs to allocate enough memory for ACIT
    workspace for each Controller, Target, Logical Unit and for each I/O.

--*/

ACIT_STATUS
acit_get_ws_sizes(
    PACIT_VERSION               pacit_api_version,
    PACIT_UEXACT32              pcontroller_ws_size,
    PACIT_UEXACT32              ptarget_ws_size,
    PACIT_UEXACT32              plun_ws_size,
    PACIT_UEXACT32              pio_ws_size,
    PACIT_UEXACT32              ptmf_ws_size)
{
    ACIT_UEXACT32               controller_ws_size;
    ACIT_UEXACT32               target_ws_size;
    ACIT_UEXACT32               lun_ws_size;
    ACIT_UEXACT32               io_ws_size;
    ACIT_UEXACT32               tmf_ws_size;

    OSM_ASSERT(pacit_api_version);
    OSM_ASSERT(pcontroller_ws_size);
    OSM_ASSERT(ptarget_ws_size);
    OSM_ASSERT(plun_ws_size);
    OSM_ASSERT(pio_ws_size);
    OSM_ASSERT(ptmf_ws_size);

    /*
     * Update version info
     */
    acit_get_api_version(pacit_api_version);

    /*
     * Init locals
     */
    controller_ws_size  = 0;
    target_ws_size      = 0;
    lun_ws_size         = 0;
    io_ws_size          = 0;
    tmf_ws_size         = 0;

    /*
     * First add our own requirements to the total
     */
    *pcontroller_ws_size     = sizeof(ACIT_CONTROLLER_WS) + 
                               (sizeof(ACIT_MEM_DESCRIPTOR) * ACIT_MEM_MAX);

    *ptarget_ws_size         = sizeof(ACIT_TARGET_WS);
    *plun_ws_size            = sizeof(ACIT_LUN_WS);
    *pio_ws_size             = sizeof(ACIT_IO_WS);
    *ptmf_ws_size            = sizeof(ACIT_TMF_WS);

    /*
     * check if the hardware related structure sizes are of correct
     * size. Add compile time checks
     */
    OSM_ASSERT(sizeof(ISCSI_WRB) == 64);
    OSM_ASSERT(sizeof(MCC_WRB) == 256);
    OSM_ASSERT(sizeof(SGE) == 16);
    OSM_ASSERT(sizeof(ISCSI_CQE) == 16);
    OSM_ASSERT(sizeof(EQ_ENTRY) == 4);
    OSM_ASSERT(sizeof(MCC_CQ_ENTRY) == 16);

    /*
     * Get work space sizes required by iSCSI and add them to total
     */
    iscsi_get_ws_sizes( &controller_ws_size,
                        &target_ws_size,
                        &lun_ws_size,
                        &io_ws_size,
                        &tmf_ws_size);

    *pcontroller_ws_size     += controller_ws_size;
    *ptarget_ws_size         += target_ws_size;
    *plun_ws_size            += lun_ws_size;
    *pio_ws_size             += io_ws_size;
    *ptmf_ws_size            += tmf_ws_size;

    /*
     * Get work space sizes required by Management and add them to total
     */
    mgmt_get_ws_sizes(  &controller_ws_size,
                        &target_ws_size,
                        &lun_ws_size,
                        &io_ws_size,
                        &tmf_ws_size);

    *pcontroller_ws_size     += controller_ws_size;
    *ptarget_ws_size         += target_ws_size;
    *plun_ws_size            += lun_ws_size;
    *pio_ws_size             += io_ws_size;
    *ptmf_ws_size            += tmf_ws_size;

    /*
     * Get work space sizes required by HWI and add them to total
     */
    hwi_get_ws_sizes(   &controller_ws_size,
                        &target_ws_size,
                        &lun_ws_size,
                        &io_ws_size,
                        &tmf_ws_size);

    *pcontroller_ws_size     += controller_ws_size;
    *ptarget_ws_size         += target_ws_size;
    *plun_ws_size            += lun_ws_size;
    *pio_ws_size             += io_ws_size;
    *ptmf_ws_size            += tmf_ws_size;

    /*
     * Align all the workspace sizes to 8 byte
     */
    ACIT_ALIGN((*pcontroller_ws_size), ACIT_MEM_ALIGNMENT_MASK);
    ACIT_ALIGN((*ptarget_ws_size), ACIT_MEM_ALIGNMENT_MASK);
    ACIT_ALIGN((*plun_ws_size), ACIT_MEM_ALIGNMENT_MASK);
    ACIT_ALIGN((*pio_ws_size), ACIT_MEM_ALIGNMENT_MASK);
    ACIT_ALIGN((*ptmf_ws_size), ACIT_MEM_ALIGNMENT_MASK);

    /*
     * Make sure all our workspace memory size required is according to
     * our alignment requirements
     */
    OSM_ASSERT(!(*pcontroller_ws_size & ACIT_MEM_ALIGNMENT_MASK));
    OSM_ASSERT(!(*ptarget_ws_size & ACIT_MEM_ALIGNMENT_MASK));
    OSM_ASSERT(!(*plun_ws_size & ACIT_MEM_ALIGNMENT_MASK));
    OSM_ASSERT(!(*pio_ws_size & ACIT_MEM_ALIGNMENT_MASK));
    OSM_ASSERT(!(*ptmf_ws_size & ACIT_MEM_ALIGNMENT_MASK));

    return ACIT_STATUS_SUCCESS;
}



/*++

Function:
    acit_init_controller_ws

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pcontroller_ws  [in ] Pointer to the Controller workspace.

Return:
    ACIT_STATUS_SUCCESS if the workspace is successfully initialized.

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to provide the ACIT library with the pointer
    to the Controller workspace. The Controller workspace allocated by the OSM
    should be atleast the size of Controller workspace reported by the
    acit_get_ws_sizes. This function should be called to initialize the
    Controller workspace before calling other ACIT API functions that take the
    Controller Key pointer as a parameter.

--*/

ACIT_STATUS
acit_init_controller_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_VOID                  pcontroller_ws)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_UEXACT32               acit_ws_size = 0;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("\n"));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof ACIT_UEXACT8  = %u\n", (ACIT_UEXACT32)sizeof(ACIT_UEXACT8)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof ACIT_UEXACT16 = %u\n", (ACIT_UEXACT32)sizeof(ACIT_UEXACT16)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof ACIT_UEXACT32 = %u\n", (ACIT_UEXACT32)sizeof(ACIT_UEXACT32)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof ACIT_UEXACT64 = %u\n", (ACIT_UEXACT32)sizeof(ACIT_UEXACT64)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof ACIT_U32_64   = %u\n", (ACIT_UEXACT32)sizeof(ACIT_U32_64)));
    
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof PACIT_UEXACT8  = %u\n", (ACIT_UEXACT32)sizeof(PACIT_UEXACT8)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof PACIT_UEXACT16 = %u\n", (ACIT_UEXACT32)sizeof(PACIT_UEXACT16)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof PACIT_UEXACT32 = %u\n", (ACIT_UEXACT32)sizeof(PACIT_UEXACT32)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof PACIT_UEXACT64 = %u\n", (ACIT_UEXACT32)sizeof(PACIT_UEXACT64)));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sizeof PACIT_U32_64   = %u\n", (ACIT_UEXACT32)sizeof(PACIT_U32_64)));

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pcontroller_ws);
    OSM_ASSERT(!((ACIT_PTR)pcontroller_ws & ACIT_MEM_ALIGNMENT_MASK));

    /*
     * Save our own workspace pointer in the Key
     */
    pcontroller_key->acit_key = pcontroller_ws;

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    if(!pcontroller)
    {
        OSM_ASSERT(0);
        return ACIT_STATUS_INVALID_PARAMETER;
    }

    acit_ws_size = sizeof(ACIT_CONTROLLER_WS) +
                         (sizeof(ACIT_MEM_DESCRIPTOR) * ACIT_MEM_MAX);

    OSM_MEMSET(pcontroller , 0, acit_ws_size);

#if ACIT_CFG_VASE_ENABLED
    ap_prepare_vase(pcontroller);
#else
    pcontroller->vase_mem_needed = 0;
#endif

    /*
     * Initialize our Controller workspace.
     */
    pcontroller->pcontroller_key = pcontroller_key;
    pcontroller->ws_sizes.acit_controller_ws_size = acit_ws_size;
    pcontroller->mem_descriptors = (PACIT_VOID)(pcontroller + 1);

    LIST_INIT_HEADER(&pcontroller->io_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->io_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->io_queued_list);
    LIST_INIT_HEADER(&pcontroller->inv_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->inv_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->io_pending_reset_list);
    LIST_INIT_HEADER(&pcontroller->logout_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->logout_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->target_event_list);
    LIST_INIT_HEADER(&pcontroller->tmf_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->tmf_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->tmf_target_list);
    LIST_INIT_HEADER(&pcontroller->tmf_queued_list);
    LIST_INIT_HEADER(&pcontroller->nop_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->nop_handle_busy_list);

    // Set to default.  
    pcontroller->state.operating_mem_level = ACIT_MEM_LEVEL_DEFAULT;

    /*
     * Get iSCSI, Mgmt and HWI workspace requirements again so that we can save
     * it too now that we have our own workspace
     */
    iscsi_get_ws_sizes( &pcontroller->ws_sizes.iscsi_controller_ws_size,
                        &pcontroller->ws_sizes.iscsi_target_ws_size,
                        &pcontroller->ws_sizes.iscsi_lun_ws_size,
                        &pcontroller->ws_sizes.iscsi_io_ws_size,
                        &pcontroller->ws_sizes.iscsi_tmf_ws_size);
    mgmt_get_ws_sizes(  &pcontroller->ws_sizes.mgmt_controller_ws_size,
                        &pcontroller->ws_sizes.mgmt_target_ws_size,
                        &pcontroller->ws_sizes.mgmt_lun_ws_size,
                        &pcontroller->ws_sizes.mgmt_io_ws_size,
                        &pcontroller->ws_sizes.mgmt_tmf_ws_size);
    hwi_get_ws_sizes(   &pcontroller->ws_sizes.hwi_controller_ws_size,
                        &pcontroller->ws_sizes.hwi_target_ws_size,
                        &pcontroller->ws_sizes.hwi_lun_ws_size,
                        &pcontroller->ws_sizes.hwi_io_ws_size,
                        &pcontroller->ws_sizes.hwi_tmf_ws_size);

    acit_ws_size = pcontroller->ws_sizes.acit_controller_ws_size;
    /*
     * Calculate iSCSI workspace pointer and initialize it
     */
    if(pcontroller->ws_sizes.iscsi_controller_ws_size)
    {
        pcontroller->piscsi_ws = (PISCSI_CONTROLLER_WS)((PACIT_UEXACT8)pcontroller + acit_ws_size) ;
    }
    else
    {
        pcontroller->piscsi_ws = ACIT_NULL;
    }

    iscsi_init_controller_ws(pcontroller);

    /*
     * Calculate the Management workspace also. This is located beyond the
     * iSCSI workspace, so we add the iSCSI workspace size to get to it.
     */
    if(pcontroller->ws_sizes.mgmt_controller_ws_size)
    {
        pcontroller->pmgmt_ws = (PACIT_VOID)
                                (((PACIT_UEXACT8) (pcontroller) + acit_ws_size) +
                                 pcontroller->ws_sizes.iscsi_controller_ws_size);
    }
    else
    {
        pcontroller->pmgmt_ws = ACIT_NULL;
    }

    mgmt_init_controller_ws(pcontroller);

    /*
     * HWI workspace located beyond MGMT workspace..
     */
    if(pcontroller->ws_sizes.hwi_controller_ws_size)
    {
        pcontroller->phwi_ws = (PACIT_VOID)
                                (((PACIT_UEXACT8) (pcontroller) + acit_ws_size) +
                                 pcontroller->ws_sizes.iscsi_controller_ws_size +
                                 pcontroller->ws_sizes.mgmt_controller_ws_size);
    }
    else
    {
        pcontroller->phwi_ws = ACIT_NULL;
    }

    hwi_init_controller_ws(pcontroller);

    return ACIT_STATUS_SUCCESS;
}



/*++

Function:
    acit_init_target_ws

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    ptarget_key     [in ] Pointer to the target specific Key
    ptarget_ws      [in ] Pointer to the target workspace

Return:
    ACIT_STATUS_SUCCESS if the workspace is successfully initialized

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to provide the ACIT library with the pointer
    to the Target workspace. The Target workspace allocated by the OSM
    should be atleast the size of Target workspace reported by the
    acit_get_ws_sizes. This function should be called to initialize the
    Target workspace before calling other ACIT API functions that take the
    Target Key pointer as a parameter.

    Update: 05/11/09 - OSMs must not access target structure if this API
    fails. There were some cases where offload failed and we have to report
    the target as removed in the offload callback.

--*/

ACIT_STATUS
acit_init_target_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_VOID                  ptarget_ws)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget;
    ACIT_U32                    index;
    ACIT_UEXACT32               status, session_handle;
    ACIT_UEXACT16               connection_handle, cid, cq_id;
    ACIT_UEXACT16               def_hdr_id, def_data_id;
    PHYS_ADDR                   template_address = {{{0,0}}};

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(ptarget_key);
    OSM_ASSERT(ptarget_ws);
    OSM_ASSERT(!((ACIT_PTR)ptarget_ws & ACIT_MEM_ALIGNMENT_MASK));

    ptarget_key->acit_key = ptarget_ws;

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptarget = GET_TARGET_WS(ptarget_key);

    if(!ptarget)
    {
        OSM_ASSERT(0);
        return ACIT_STATUS_INVALID_PARAMETER;
    }

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(ptarget);

    OSM_MEMSET( ptarget_ws,
                0,
                sizeof(ACIT_TARGET_WS));

    ptarget->pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    /* The target workspace's target key is initialized to the
     * corresponding key which helps to find index of the target
     * used in mgmt_init_target_ws. The fn is in switch session
     * ioctl submission path with session id as input parameter
     * - JB
     */
    ptarget->ptarget_key = ptarget_key;

    LIST_INIT_HEADER(&ptarget->io_handle_busy_list);
    LIST_INIT_HEADER(&ptarget->io_queued_list);
    LIST_INIT_HEADER(&ptarget->io_pending_reset_list);
    LIST_INIT_HEADER(&ptarget->tmf_handle_busy_list);
    LIST_INIT_HEADER(&ptarget->tmf_lun_list);
    LIST_INIT_HEADER(&ptarget->nop_handle_list);
    LIST_INIT_HEADER(&ptarget->logout_handle_list);

    LIST_INIT_ELEMENT(&ptarget->event_link);
    LIST_INIT_ELEMENT(&ptarget->tmf_link);
    /*
     * Initialize timeout related parameters 
     */
    ptarget->timeout_factor = pcontroller->params.timeout_factor;
    //ptarget->last_cmd_ticks = ptarget->default_cmd_ticks;
    ptarget->target_reset_queued = ACIT_FALSE;

    /*
     * Calculate the Target workspace for iSCSI module and initialize it
     */
    if(pcontroller->ws_sizes.iscsi_target_ws_size)
    {
        ptarget->piscsi_ws = (PACIT_VOID)(ptarget + 1);
    }
    else
    {
        ptarget->piscsi_ws = ACIT_NULL;
    }

    // Before initializing iSCSI workspace, set up the target index in target ws

    for (index = 0;
        (index < BE_MAX_TARGETS) && (&pcontroller->ptarget_key[index] != ptarget->ptarget_key);
        index++);

    ptarget->target_table_index = index;

    // if this target is being logged out, do not initialize it
    if(ACIT_IS_BIT_SET(pcontroller->ptarget_table->target_entry[index].acit_tflags,
                        ACIT_TFLAG_REMOVE_PENDING))
    {
        //
        // We don't have to call ap_remove_target from here since the target remove is
        // anyway in progress.
        //
        OSM_ASSERT(0);
        return ACIT_STATUS_NO_DEVICE;
    }

    session_handle = pcontroller->ptarget_table->target_entry[index].session_handle;

    status = iscsi_init_target_ws(pcontroller, 
                                  ptarget, 
                                  &connection_handle, 
                                  &cid,
                                  &template_address);

    if(status != ISCSI_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("iscsi_init_target_ws failed\n"));
        ap_remove_target(pcontroller, ptarget);
        return ACIT_STATUS_INSUFFICIENT_RESOURCE;
    }

    /*
     * Calculate the Target workspace for Management module and initialize it
     * This is located beyond the iSCSI workspace so we add the iSCSI
     * workspace size to get to it
     */
    if(pcontroller->ws_sizes.mgmt_target_ws_size)
    {
        ptarget->pmgmt_ws = (PACIT_VOID)
                            ( ((PACIT_UEXACT8) (ptarget + 1)) +
                                 pcontroller->ws_sizes.iscsi_target_ws_size);
    }
    else
    {
        ptarget->pmgmt_ws = ACIT_NULL;
    }

    cq_id = (ACIT_UEXACT16)HWI_GET_CQ_ID(pcontroller, 0);
    def_hdr_id = (ACIT_UEXACT16)HWI_GET_DEF_HDRQ_ID(pcontroller);
    def_data_id = (ACIT_UEXACT16)HWI_GET_DEF_BUFQ_ID(pcontroller);

    if (pcontroller->state.api_type == ACIT_OP_MODE_RESTRICTED) 
        status = iscsi_init_hiber_target_ws(pcontroller, 
                                    ptarget, 
                                    session_handle,
                                    connection_handle
                                    );
    else
        status = mgmt_init_target_ws(pcontroller, 
                                    ptarget, 
                                    session_handle,
                                    connection_handle,
                                    cid,
                                    cq_id,
                                    def_hdr_id,
                                    def_data_id,
                                    &template_address,
                                    ACIT_TRUE   // synchronous
                                    );
    return status;
}



/*++

Function:
    acit_init_lun_ws

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    ptarget_key     [in ] Pointer to the Target specific Key
    plun_key        [in ] Pointer to the LUN specific Key
    plun_ws         [in ] Pointer to the LUN workspace

Return:
    ACIT_STATUS_SUCCESS if the workspace is successfully initialized

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to provide the ACIT library with the pointer
    to the LUN workspace. The LUN workspace allocated by the OSM should be
    atleast the size of LUN workspace reported by the acit_get_ws_sizes. This
    function should be called to initialize the LUN workspace before calling
    other ACIT API functions that take the LUN Key pointer as a parameter.

--*/

ACIT_STATUS
acit_init_lun_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_LUN_KEY               plun_key,
    PACIT_VOID                  plun_ws)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget;
    PACIT_LUN_WS                plun;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(ptarget_key);
    OSM_ASSERT(plun_key);
    OSM_ASSERT(!((ACIT_PTR)plun_ws & ACIT_MEM_ALIGNMENT_MASK));
    OSM_ASSERT(plun_ws);

    plun_key->acit_key = plun_ws;

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptarget = GET_TARGET_WS(ptarget_key);
    plun = GET_LUN_WS(plun_key);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p Lun 0x%p\n",
                            ptarget, plun));

    if(!plun)
    {
        OSM_ASSERT(0);
        return ACIT_STATUS_INVALID_PARAMETER;
    }

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(ptarget);
    OSM_ASSERT(plun);

    OSM_MEMSET( plun_ws,
                0,
                sizeof(ACIT_LUN_WS));

    plun->dev_type = ACIT_SCSI_DEVTYPE_UNKNOWN;
    plun->plun_key = plun_key;
    plun->counters.lun_reset_queued = ACIT_FALSE;

    // Initialize retries based on global parameters.
    plun->retry_io = pcontroller->params.io_retries;

    LIST_INIT_HEADER(&plun->io_handle_busy_list);
    LIST_INIT_HEADER(&plun->io_queued_list);
    LIST_INIT_HEADER(&plun->io_pending_reset_list);
    LIST_INIT_HEADER(&plun->tmf_handle_busy_list);
    LIST_INIT_HEADER(&plun->tmf_abort_list);

    LIST_INIT_ELEMENT(&plun->tmf_link);

    /*
     * Calculate the I/O workspace for iSCSI module and initialize it
     */
    if(pcontroller->ws_sizes.iscsi_lun_ws_size)
    {
        plun->piscsi_ws = plun + 1;
    }
    else
    {
        plun->piscsi_ws = ACIT_NULL;
    }

    iscsi_init_lun_ws(pcontroller, ptarget, plun);

    /*
     * Calculate the I/O workspace for Management module and initialize it
     * This is located beyond the iSCSI workspace so we add the iSCSI
     * workspace size to get to it
     */
    if(pcontroller->ws_sizes.mgmt_lun_ws_size)
    {
        plun->pmgmt_ws = (PACIT_VOID)
                            ( ((PACIT_UEXACT8) (plun + 1)) +
                             pcontroller->ws_sizes.iscsi_lun_ws_size);
    }
    else
    {
        plun->pmgmt_ws = ACIT_NULL;
    }

    mgmt_init_lun_ws(pcontroller, ptarget, plun);

    return ACIT_STATUS_SUCCESS;
}

/*++

Function:
    acit_init_io_ws

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pio_request     [in ] Pointer to the I/O Request structure
    pio_key         [in ] Pointer to the I/O specific Key
    pio_ws          [in ] Pointer to the I/O workspace

Return:
    ACIT_STATUS_SUCCESS if the workspace is successfully initialized

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to provide the ACIT library with the pointer
    to the I/O workspace. The I/O workspace allocated by the OSM should be
    atleast the size of I/O workspace reported by the acit_get_ws_sizes. This
    function should be called to initialize the I/O workspace before calling
    other ACIT API functions that take the I/O Key pointer as a parameter.

--*/

ACIT_STATUS
acit_init_io_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request,
    PACIT_IO_KEY                pio_key,
    PACIT_VOID                  pio_ws)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget;
    PACIT_IO_WS                 pio;    

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pio_request);
    OSM_ASSERT(pio_key);
    OSM_ASSERT(pio_ws);
    OSM_ASSERT(!((ACIT_PTR)pio_ws & ACIT_MEM_ALIGNMENT_MASK));
    OSM_ASSERT(pio_request->pio_key == pio_key);
    OSM_ASSERT(pio_request->ptarget_key);

    pio_key->acit_key = pio_ws;

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptarget     = GET_TARGET_WS(pio_request->ptarget_key);
    pio         = GET_IO_WS(pio_key);

    OSM_ASSERT(ptarget);

    if(!pio)
    {
        OSM_ASSERT(0);
        return ACIT_STATUS_INVALID_PARAMETER;
    }

    if(pcontroller->busy)
    {
        if(IS_CONTROLLER_LDTO_EXPIRED(pcontroller, 0) ||
           IS_CONTROLLER_UE(pcontroller))
        {
            // Return the same status even if there is a UE. The OSMs do not
            // have to duplicate code to fail IOs for a UE..
            return ACIT_STATUS_LINK_DOWN_TIMEOUT; 
        }
    }

    if(ptarget->busy)
    {
        if(IS_TARGET_ETO_EXPIRED(ptarget))
        {
            return ACIT_STATUS_TARGET_TIMEOUT;
        }
    }

    /*
     * Build the I/O WS structure.
     */
    pio->pio_key = pio_request->pio_key;
    pio->pio_request = pio_request;
    pio->retry_count = 0;
    pio->io_retries = 0;
    pio->original_ticks = pio_request->timeout * TICKS_PER_SEC(pcontroller);
    pio->available_ticks = pio->original_ticks;

#if ACIT_CFG_DEBUG_BUILD
    pio->internal_available_ticks = -1;
#endif

    pio->io_status = 0;
    pio->pio_handle = ACIT_NULL;
    pio->queued = 0;

    pio->bhs_len = ACIT_CMD_BHS_LEN;
    pio->bhs_pa = OSM_GET_BUS_ADDRESS(pcontroller_key,
                                      pio_request->posm_reference,
                                      &pio->cmd_bhs,
                                      &pio->bhs_len);

    /* Reset the size back to whatever is allocated on the virtual space */
    pio->bhs_len = ACIT_CMD_BHS_LEN;
    OSM_ASSERT(pio->bhs_pa.u.a64.address);
    OSM_ASSERT(pio->bhs_len);

    OSM_ASSERT(!pcontroller->ws_sizes.iscsi_io_ws_size);
    OSM_ASSERT(!pcontroller->ws_sizes.mgmt_io_ws_size);


    iscsi_init_io_ws(pcontroller, pio);

#if 0

    /*
     * Calculate the I/O workspace for iSCSI module and initialize it
     */
    if(pcontroller->ws_sizes.iscsi_io_ws_size)
    {
        pio->piscsi_ws = (PISCSI_IO_WS)(pio + 1);
    }
    else
    {
        pio->piscsi_ws = ACIT_NULL;
    }

    iscsi_init_io_ws(pcontroller, ptarget, pio);

    /*
     * Calculate the I/O workspace for Management module and initialize it
     * This is located beyond the iSCSI workspace so we add the iSCSI
     * workspace size to get to it
     */
    if(pcontroller->ws_sizes.mgmt_io_ws_size)
    {
        pio->pmgmt_ws = (PACIT_VOID)
                            ( ((PACIT_UEXACT8) (pio + 1)) +
                               pcontroller->ws_sizes.iscsi_io_ws_size);
    }
    else
    {
        pio->pmgmt_ws = ACIT_NULL;
    }

    mgmt_init_io_ws(pcontroller, pio);
#endif

    return ACIT_STATUS_SUCCESS;
}


/*++

Function:
    acit_init_tmf_ws

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    ptmf_request    [in ] Pointer to the TMF Request structure
    ptmf_key        [in ] Pointer to the TMF specific Key
    ptmf_ws         [in ] Pointer to the TMF workspace

Return:
    ACIT_STATUS_SUCCESS if the workspace is successfully initialized

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to provide the ACIT library with the pointer
    to the TMF workspace. The TMF workspace allocated by the OSM should be
    atleast the size of TMF workspace reported by the acit_get_ws_sizes. This
    function should be called to initialize the TMF workspace before calling
    other ACIT API functions that take the TMF Key pointer as a parameter.

--*/

ACIT_STATUS
acit_init_tmf_ws(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TMF_REQUEST           ptmf_request,
    PACIT_TMF_KEY               ptmf_key,
    PACIT_VOID                  ptmf_ws)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget;
    PACIT_TMF_WS                ptmf;
    ACIT_UEXACT32               port_num;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(ptmf_request);
    OSM_ASSERT(ptmf_key);
    OSM_ASSERT(ptmf_ws);
    OSM_ASSERT(!((ACIT_PTR)ptmf_ws & ACIT_MEM_ALIGNMENT_MASK));
    OSM_ASSERT(ptmf_request->ptmf_key == ptmf_key);
    OSM_ASSERT(ptmf_request->timeout);

    ptmf_key->acit_key = ptmf_ws;

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptarget     = GET_TARGET_WS(ptmf_request->ptarget_key);
    ptmf        = GET_TMF_WS(ptmf_key);

    if(!ptmf)
    {
        OSM_ASSERT(0);
        return ACIT_STATUS_INVALID_PARAMETER;
    }

    port_num = ISCSI_GET_TARGET_PORT(ptarget);

    if(pcontroller->busy)
    {
        if(IS_CONTROLLER_LDTO_EXPIRED(pcontroller, port_num) ||
           IS_CONTROLLER_UE(pcontroller))
        {
            // Return the same status even if there is a UE. The OSMs do not
            // have to duplicate code to fail IOs for a UE..
            return ACIT_STATUS_LINK_DOWN_TIMEOUT; 
        }
    }

    if(IS_TARGET_ETO_EXPIRED(ptarget))
    {
        return ACIT_STATUS_TARGET_TIMEOUT;
    }

    /*
     * Build the I/O WS structure.
     */
    ptmf->ptmf_key = ptmf_request->ptmf_key;
    ptmf->ptmf_request = ptmf_request;
    ptmf->original_ticks = ptmf_request->timeout * TICKS_PER_SEC(pcontroller);
    ptmf->available_ticks = ptmf->original_ticks;
    ptmf->tmf_status = 0;
    LIST_INIT_ELEMENT(&ptmf->controller_link);

    /*
     * Calculate the TMF workspace for iSCSI module and initialize it
     */
    if(pcontroller->ws_sizes.iscsi_tmf_ws_size)
    {
        ptmf->piscsi_ws = ptmf + 1;
    }
    else
    {
        ptmf->piscsi_ws = ACIT_NULL;
    }

    iscsi_init_tmf_ws(pcontroller, ptarget, ptmf);

    /*
     * Calculate the TMF workspace for Management module and initialize it
     * This is located beyond the iSCSI workspace so we add the iSCSI
     * workspace size to get to it
     */
    if(pcontroller->ws_sizes.mgmt_tmf_ws_size)
    {
        ptmf->pmgmt_ws = (PACIT_VOID)
                            ( ((PACIT_UEXACT8) (ptmf + 1)) +
                               pcontroller->ws_sizes.iscsi_tmf_ws_size);
    }
    else
    {
        ptmf->pmgmt_ws = ACIT_NULL;
    }

    mgmt_init_tmf_ws(pcontroller, ptmf);

    return ACIT_STATUS_SUCCESS;
}


/*++

Function:
    acit_set_osm_api_table

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    posm_api_table  [in ] Pointer to the table of OSM function pointers

Return:
    ACIT_STATUS_SUCCESS if the version is supported by ACIT library.
    ACIT_STATUS_INCORRECT_VERSION if the table version is not supported.

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to provide the OSM API table to the ACIT
    library. ACIT library copies the table content to a private storage and
    returns. This function should only be used during initialization and
    after a call to acit_init_controller_ws has been successfully made.

--*/

ACIT_STATUS
acit_set_osm_api_table(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_OSM_API_TABLE         posm_api_table,
    ACIT_UEXACT32               table_length)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_STATUS                 status;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(posm_api_table->osm_api_table_version.hi_ver ==
                ACIT_OSM_API_TABLE_HI_VERSION);

    if(posm_api_table->osm_api_table_version.hi_ver ==
                ACIT_OSM_API_TABLE_HI_VERSION)
    {

        pcontroller = GET_CONTROLLER_WS(pcontroller_key);

        OSM_MEMCPY( &pcontroller->osm_api_table,
                    posm_api_table,
                    table_length);

        status = ACIT_STATUS_SUCCESS;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_FATAL,
                    ("API table version mismatch\n"));
        status = ACIT_STATUS_INCORRECT_VERSION;
    }

    return status;
}


ACIT_STATUS
ap_init_mem_params(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32           i;
    ACIT_STATUS             status;
    PACIT_PARAMETERS        p;
    ACIT_UEXACT32           cxn = BE_MAX_CXNS;
#ifdef ENABLE_512_TARGETS
    ACIT_UEXACT32   ioctls = 64;
#else
    ACIT_UEXACT32   ioctls = 16;
#endif

    p = &pcontroller->params;

#if defined (_WINISCSI_INBOX) || defined(_WINISCSI)
    // Allow non-constant aggregate initializer
#pragma warning (disable:4204)
#endif

    {
        const ACIT_PARAM_PROPERTIES mem_pp[] = {
            /*Name,          Min,  Level1, Level2, Level3, Max,  Mask, pointer*/
            // FixmeBE2: Revisit the cmd/connection count again.. bumping it down to 128 for emulation
            {"cxn_cmds",     1,    2,     128,    256,    256,  0,    &p->num_cmd_per_cxn},
            {"cxns",         1,    1,     cxn,    cxn,    cxn,   0,   &p->num_cxn},
            {"tgts",         1,    1,     cxn,    cxn,    cxn,   0,   &p->targets_per_controller},
            {"ios_per_ctrl", 1,    7,     (1024-BE_MAX_CXNS),    (1024-BE_MAX_CXNS), BE_MAX_IOS, 0, &p->ios_per_controller},
            {"logouts",      1,    1,     64,     64,     64,   0,    &p->logouts_per_controller},
            {"tmfs",         1,    1,     16,     32,     64,   0,    &p->tmfs_per_ctrl},
            {"asyncpdus",    0,    8,     cxn,    cxn,    cxn*2,0,    &p->num_async_pdus},
            {"ioctls",       1,    8,     ioctls, (ioctls * 2), 128,0,&p->ioctls_per_ctrl},
            {"eqs",          0,    1024,  1024,   1024,   2048, 0,    &p->num_eq_entries},
            {"cqs",          0,    128,   1024,   1024,   2048, 0,    &p->num_cq_entries},
            {"defpdu_buf_sz",0,    64,    64,     64,     1024, 0,    &p->default_pdu_buf_sz},
#ifndef ENABLE_512K_IOSZ  
            {"sge",          0,    32,    32,     32,     256,  0,    &p->num_sge},
#else  
            {"sge",          0,    32,    256,    256,    256,  0,    &p->num_sge},
#endif
            // num_cq options. Default is 1 and supporting OSMs can set it to max ACIT_CFG_MAX_PARTITIONS
            {"num_cq",       1,   1,      1,      1,      ACIT_CFG_MAX_PARTITIONS, 0, &p->num_cq},

        };

#if defined (_WINISCSI_INBOX) || defined(_WINISCSI)
    // Allow non-constant aggregate initializer
#pragma warning (default:4204)
#endif

        // These parameters depend on the operating memory level.
        for (i = 0; i < sizeof (mem_pp) / sizeof (mem_pp[0]); i++) {
            status = ap_get_param_num (pcontroller->pcontroller_key,
                                       &mem_pp[i], 
                                       &pcontroller->params.default_flags,
                                       pcontroller->state.operating_mem_level);
            if (status) {
                return ACIT_STATUS_ERROR;
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
ap_init_dic_params(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32           i;
    ACIT_STATUS             status;
    PACIT_PARAMETERS        p;
    
    p = &pcontroller->params;

#if defined (_WINISCSI_INBOX) || defined(_WINISCSI)
    // Allow non-constant aggregate initializer
#pragma warning (disable:4204)
#endif

    {
        const ACIT_PARAM_PROPERTIES dic_pp[] = {
            /*Name,              Min,Level1, Level2,    Level3, Max,        Mask, pointer*/
            {"dic",              0,  1,     1,         1,      1,          0,    (PACIT_UEXACT32)&p->enable_dic},
            {"min_timer",        1,  64,    64,        0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.min_timer_val},
            {"max_timer",        1,  512,   512,       0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.max_timer_val},
            {"timer_incs",       1,  16,    16,        0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.timer_increments},
            {"max_ir",           1,  10000, 10000,     0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.max_interrupt_rate},
            {"min_ir",           1,  5500,  5500,      0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.min_interrupt_rate},
            {"iot4_dic",         1,  50000, 50000,     0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.io_threshold_for_dic},
            {"ior4_dic",         1,  5000,  5000,      0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.io_resolution_for_dic},
            {"tgt_load",         1,  8,     8,         0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.tgt_threshold_for_load},
            {"intr_res",         0,  500,   500,       0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.interrupt_resolution},
            {"intr_steps",       0,  2000,  2000,      0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.interrupt_dec_steps},
            {"io_sv",            0,  400000,400000,    0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->dic.io_safety_valve},                  
            {"im_policy",        0,  2,     2,         2,      4,          0,    (PACIT_UEXACT32)&p->dic.policy},
        };

#if defined (_WINISCSI_INBOX) || defined(_WINISCSI)
    // Allow non-constant aggregate initializer
#pragma warning (default:4204)
#endif

        // These parameters do not depend on operating memory level.  They are always configured to use level 2. The
        // user may explicitly change them from the min to max values.  Level3 is never used.
        // The Min and Max serves as ranges within which the configurable parameters values may 
        // reside
        for (i = 0; i < sizeof (dic_pp) / sizeof (dic_pp[0]); i++) {
            status = ap_get_param_num (pcontroller->pcontroller_key, 
                                       &dic_pp[i], 
                                       &pcontroller->params.default_flags, 
                                       ACIT_MEM_LEVEL_2);
            if (status) {
                return ACIT_STATUS_ERROR;
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
ap_init_non_mem_params(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32           i;
    ACIT_STATUS             status;
    PACIT_PARAMETERS        p;
    ACIT_UEXACT32           max_eto = ACIT_CFG_MAX_ETO, def_eto = ACIT_CFG_DEFAULT_ETO;
    
    p = &pcontroller->params;

#if defined (_WINISCSI_INBOX) || defined(_WINISCSI)
    // Allow non-constant aggregate initializer
#pragma warning (disable:4204)
#endif

    {
        const ACIT_PARAM_PROPERTIES non_mem_pp[] = {
            /*Name,               Min,Level1, Level2,    Level3, Max,        Mask, pointer*/
            {"l_to_t",            0,  10,       10,        0,      0xffffffff, 0,    &p->logout_timeout_ticks},
            {"mcs",               1,  1,        1,         0,      4,          0,    &p->connections_per_session},
            {"tps",               1,  2,        2,         0,      256,        0,    &p->ticks_per_second},
            {"eto",               0,  def_eto,  def_eto,   0,      max_eto,    0,    &p->eto},
            {"ldto",              0,  20,       20,        0,      MAX_LDTO,   0,    &p->ldto},
            {"dbg",               0,  0x8000000,0x8000000, 0,      0xffffffff, 0,    &p->dbg_flags},
            {"eq_timer",          8,  64,       64,        0,      2000,       0,    &p->eq_timer}, // in usec
            {"cxn_tmfs",          1,  1,        1,         0,      0xffffffff, 0,    &p->num_tmfs_per_cxn},
            {"inv_tick",          0,  2,        2,         0,      0xffffffff, 0,    &p->invalidation_ticks},
            {"tgt_def_tick",      0,  8,        8,         0,      0xffffffff, 0,    &p->tgt_default_ticks},
            {"to_factor",         0,  2,        2,         0,      0xffffffff, 0,    &p->timeout_factor},
            {"io_retries_b4_erl0",0,  3,        3,         0,      0xffffffff, 0,    &p->io_retries_b4_erl0},
            {"hc_freq",           1,  2,        2,         0,      0xffffffff, 0,    &p->health_check_freq},
            {"sync_cd_ioc",       0,  0,        0,         0,      0x1,        0,    &p->crash_dump.sync_completion},
            {"ior",               0,  1,        1,         0,      1,          0,    &p->io_retries},
            {"port",              0,  0,        0,         0,      0xffffffff, 0,    &p->hardware_port},
            {"max_cqe",           0,  64,       64,        0,      0xffffffff, 0,    &p->max_cqe_per_isr},
            {"tmf_to",            0,  10,       10,        0,      max_eto,    0,    &p->tmf_timeout_ticks},
            {"io_retries_b4_err", 0,   3,       3,         0,      0xffffffff, 0,    &p->io_retries_b4_err},
            {"dump_rst",          0,  0,        0,         0,      0x1,        0,    &p->crash_dump.dump_rst},
            {"ird",               0,  0,        0,         0,      0x1,        0,    (PACIT_UEXACT32)&p->interrupt_redirection},
            {"offload_errs",      1,  5,        5,         0,      0xffffffff, 0,    (PACIT_UEXACT32)&p->offload_failure_threshold},
            {"ue_recovery",       0,  0,        0,         0,      1,          0,    &p->ue_recovery},
            {"max_ue_reco",       1,  5,        5,         5,      0xffffffff, 0,    &p->max_ue_recoveries},
        };

#if defined (_WINISCSI_INBOX) || defined(_WINISCSI)
    // Allow non-constant aggregate initializer
#pragma warning (default:4204)
#endif

        // These parameters do not depend on operating memory level.  They are always configured to use level 2. The
        // user may explicitly change them from the min to max values.  Level3 is never used.
        // The Min and Max serves as ranges within which the configurable parameters values may 
        // reside
        for (i = 0; i < sizeof (non_mem_pp) / sizeof (non_mem_pp[0]); i++) {
            status = ap_get_param_num (pcontroller->pcontroller_key, 
                                       &non_mem_pp[i], 
                                       &pcontroller->params.default_flags, 
                                       ACIT_MEM_LEVEL_2);
            if (status) {
                return ACIT_STATUS_ERROR;
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}


/*++

Function:
    acit_init_configurable_params

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS if the parameters were initialized successfully

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to start the initialization of configurable
    parameters in the ACIT library. This function should only be used during
    initialization and after a call to acit_init_controller_ws has been made.
    ACIT library initiates a series of call back to osm_get_config_value from
    within the context of this function to obtain user or OSM specified
    configuration values for various ACIT operating parameters.

--*/

ACIT_STATUS
acit_init_configurable_params(
    PACIT_CONTROLLER_KEY        pcontroller_key)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_PARAMETERS            p;
    ACIT_STATUS                 status;
    ACIT_UEXACT32               pid;


    OSM_ASSERT(pcontroller_key);
    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    p = &pcontroller->params;

    // functionizing the below initializations since prefast on x64
    // was complaining that the stack size of >1024 bytes was too large
    status = ap_init_mem_params(pcontroller);

    if(!status)
    {
        status = ap_init_dic_params(pcontroller);
        if(!status)
        {
            status = ap_init_non_mem_params(pcontroller);
        }
    }

    if(status != ACIT_STATUS_SUCCESS)
    {
        return status;
    }

    // store it in controller ws for referencing later
    pcontroller->state.invalidation_ticks = (ACIT_UEXACT8)p->invalidation_ticks;

    // TODO -
    // Configure the defaults for some parameters based on setting of others. If these
    // dependencies exist.  This is done by defining a "mask" value in the table above,
    // so we can detect if the default value was used for the parameter.
    if(pcontroller->params.num_cxn != pcontroller->params.targets_per_controller)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Connection Count %d mismatch with Target Count %d!, %d will be used\n",
                                            pcontroller->params.num_cxn,
                                            pcontroller->params.targets_per_controller,
                                            pcontroller->params.targets_per_controller));

        pcontroller->params.num_cxn = pcontroller->params.targets_per_controller;
    }

    //
    // Reload the dic related parameters based on im value
    //
    switch (p->dic.policy)
    {
    case 0:
        // disabled
        p->enable_dic = 0;

        break;

    case 1:
        // IM_Aggressive. Suggested for lightly loaded configuration
        // upto 1-8 targets per port. Interrupt rate will be the
        // highest among all schemes
        p->dic.max_timer_val = 128;
        p->dic.io_threshold_for_dic = 100000;
        p->dic.interrupt_resolution = 1000;
        p->dic.tgt_threshold_for_load = 16;

        break;

    case 2:
        // IM_Default: nothing to do. Suggested for configurations with
        // 8 to 16 targets per port
        break;

    case 3:

        // IM_Conservative. Suggested for higher loaded configurations

        // Use the same values as default, but max cqe will be 48
        p->max_cqe_per_isr = 48;

        break;

    case 4:

        // IM_Cautious. Siggested for max configuration with 128 targets per port
        // Interrupt rate will be the lowest among all schemes

        // Use the same values as default, but max cqe will be 32
        p->max_cqe_per_isr = 32;

        break;

    }


    // This is a hack and needs to be removed. For some reason the flags have a value
    // of -1 which isn't good for debug builds
    p->dbg_flags = 0;

    // initialize those parameters that depend on the just read parameters
    p->icds_per_controller = p->ios_per_controller + p->num_cxn;
    pcontroller->dic.enabled = p->enable_dic;

    for(pid = 0; pid < p->num_cq; pid++)
    {
        pcontroller->partition[pid].dic.interrupt_rate = p->dic.max_interrupt_rate;
        pcontroller->partition[pid].dic.eq_timer = p->eq_timer;
    }

    /*
     * Initialize our memory descriptor table that specifies
     * our memory requirements.
     */
    ap_init_all_mem_descriptors(pcontroller);

    return ACIT_STATUS_SUCCESS;
}



/*++

Function:
    acit_set_mem_params

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    operating_mem_level [in] Indicates new memory level for use by ACIT.

Return:
    ACIT_STATUS_SUCCESS if the new operating_mem_level was set.
    ACIT_STATUS_INVALID_PARAMETER if an incorrect opertaing_mem_level
        value was specified.
Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to configure a memory level for ACIT
    operations. This call can only be made before calling acit_init_controller.
    Once the OSM calls acit_init_controller, this call cannot be made unless
    the ACIT library is re-initialed completely for the given controller.
    This call is used by the OSM to adjust the size of memory required by ACIT
    library before the controller is initialized. If OSMs do not explicitly
    call this function, the default memory level is set to
    ACIT_MEM_LEVEL_DEFAULT. OSMs can call this function with a
    operating_mem_level value including and in between values
    ACIT_MEM_LEVEL_MINIMUM to ACIT_MEM_LEVEL_MAXIMUM. OSMs can use the
    ACIT_MEM_LEVEL_MINIMUM to bring up a driver instance with very minimum
    memory requirement for extreme cases. The memory level is controlled
    internally by the ACIT library by reducing the supported features like
    number of targets, or number of concurrent I/Os, or number of logical
    units per target.

--*/

ACIT_STATUS
acit_set_mem_params(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               operating_mem_level)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_STATUS                 status;

    OSM_ASSERT(pcontroller_key);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    OSM_ASSERT(pcontroller);

    /*
     * Save the new operating memory level in our workspace.
     */
    if(operating_mem_level <= ACIT_MEM_LEVEL_MAXIMUM)
    {
        pcontroller->state.operating_mem_level = (ACIT_UEXACT8)operating_mem_level;

        if (pcontroller->state.operating_mem_level != ACIT_MEM_LEVEL_DEFAULT) {
            // TODO - Ideally we could se the operating_mem_level before the
            // first call to this function -- or include it as a parameter from
            // the OSM for acit_init_configurable_params to avoid doing all this
            // work twice.
            acit_init_configurable_params (pcontroller_key);
        }

        /*
         * Re-initialize our memory descriptors based on new operating level
         */
        ap_init_all_mem_descriptors(pcontroller);

        status = ACIT_STATUS_SUCCESS;
    }
    else
    {
        status = ACIT_STATUS_INVALID_PARAMETER;
    }

    return status;
}



/*++

Function:
    acit_get_mem_params

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    poperating_mem_level [out] Pointer where current memory level is returned

Return:
    ACIT_STATUS_SUCCESS.

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to obtain the current memory level for ACIT
    operations. The value returned by the ACIT library is including and in
    between values ACIT_MEM_LEVEL_MINIMUM and ACIT_MEM_LEVEL_MAXIMUM.

--*/

ACIT_STATUS
acit_get_mem_params(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_UEXACT32              poperating_mem_level)
{
    PACIT_CONTROLLER_WS         pcontroller;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(poperating_mem_level);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    OSM_ASSERT(pcontroller);

    /*
     * Save the new operating memory level in our workspace.
     */
    *poperating_mem_level = pcontroller->state.operating_mem_level;

    return ACIT_STATUS_SUCCESS;
}



/*++

Function:
    acit_get_mem_descriptor

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pacit_device    [in ] Pointer to an ACIT_DEVICE structure
    index           [in ] Zero based index to identify a memory block
    pmem_descriptor [out] Pointer to ACIT_MEM where ACIT returns
                          parameter for memory block

Return:
    ACIT_STATUS_MORE_MEMORY if ACIT library has more memory blocks to report.
    ACIT_STATUS_SUCCESS when ACIT library is reporting the last memory block.

Context:
    This function can be called from Interrupt Context

Description:
    ACIT library requires memory for internal operations. Since multiple blocks
    of memory may be required, each memory block is referred to using an index.
    This function returns memory allocation needed by ACIT library for the each
    index. The OSM must call acit_get_mem_descriptor repeatedly starting with an
    index value of zero and incrementing it with each call until ACIT library
    returns ACIT_STATUS_SUCCESS.  This indicates that no more memory is
    required by the ACIT library. acit_get_mem_descriptor also updates the category
    element of the ACIT_MEM structure indicated by pmem_descriptor
    informing the OSM about the memory category required.  This memory category
    is filled out by the ACIT indicating if memory is ACIT_MEM_CATEGORY_CACHED
    or ACIT_MEM_CATEGORY_NON_CACHED. The OSM must allocate the requested memory
    from the proper memory pool. The OSM then passes the block to the ACIT
    through a call to acit_set_mem_descriptor. If the OSM fails to allocate the
    memory, it may start over by freeing up the existing allocations and
    lowering the operating memory level by calling acit_set_mem_params.

    At the OSM's convenience, calls between acit_get_mem_descriptor and
    acit_set_mem_descriptor can be interleaved. Alternatively, all calls to
    acit_get_mem_descriptor can be handled first, followed by all calls to
    acit_set_mem_descriptor. Memory allocations become final and ACIT library
    begins to use the provided memory when acit_init_controller is called
    later.

--*/

ACIT_STATUS
acit_get_mem_descriptor(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DEVICE                pacit_device,
    ACIT_UEXACT32               index,
    PACIT_MEM_DESCRIPTOR        pmem_descriptor)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_STATUS                 status;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pmem_descriptor);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(index < pcontroller->acit_var_desc_index.acit_mem_max);

    if(index < pcontroller->acit_var_desc_index.acit_mem_max)
    {
        /*
        * Structure copy the memory descriptor.
        */
        *pmem_descriptor =
            pcontroller->mem_descriptors[index];


        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,(("MemDescriptor: Index %d Alignment 0x%x sz %d\n"),
                                            index,
                                            pcontroller->mem_descriptors[index].alignment_mask,
                                            pcontroller->mem_descriptors[index].size_in_bytes));

        /*
        * Let caller know that we need more memory or not.
        */
        if( (index + 1) == pcontroller->acit_var_desc_index.acit_mem_max)
        {
            status = ACIT_STATUS_SUCCESS;
        }
        else
        {
            status = ACIT_STATUS_MORE_MEMORY;
        }
    }
    else
    {
        status = ACIT_STATUS_ERROR;
    }

    /* No need to assert if size is zero since in hibernation case
     * few memory parameters will be assigned size of zero.
     OSM_ASSERT(pmem_descriptor->size_in_bytes);
     */
    return status;
}

/*++

Function:
    acit_set_mem_descriptor

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    index           [in ] Zero based index to identify a memory block
    pmem_descriptor [in ] Pointer to ACIT_MEM where OSM provide
                          physical and virtual address

Return:
    ACIT_STATUS_SUCCESS if the address provided is applied.
    ACIT_STATUS_INVALID_PARAMETER address provided is not applied.

Context:
    This function can be called from Interrupt Context

Description:
    After memory allocation, OSMs calls this routine to pass the virtual and
    physical address to ACIT library.  ACIT library will have exclusive
    control of the memory after it is passed to the ACIT. OSMs calls this
    routine repetitively based on the index used when calling
    acit_get_mem_descriptor.

    At the OSM's convenience, calls between acit_get_mem_descriptor and
    acit_set_mem_descriptor can be interleaved, or all the calls to
    acit_get_mem_descriptor can be handled first, followed by all calls to
    acit_set_mem_descriptor. Memory allocations become final and ACIT library begins
    to use the provided memory when acit_init_controller is called later.

--*/

ACIT_STATUS
acit_set_mem_descriptor(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               index,
    PACIT_MEM_DESCRIPTOR        pmem_descriptor)
{
    PACIT_CONTROLLER_WS         pcontroller;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pmem_descriptor);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    if (pmem_descriptor->size_in_bytes == 0)
        return ACIT_STATUS_SUCCESS;

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(pmem_descriptor->size_in_bytes);
    OSM_ASSERT(pmem_descriptor->virtual_address);
    OSM_ASSERT(pmem_descriptor->bus_address.u.a64.address);
    OSM_ASSERT(index < pcontroller->acit_var_desc_index.acit_mem_max);

    OSM_ASSERT(!(pmem_descriptor->bus_address.u.a64.address & pmem_descriptor->alignment_mask));

    /*
     * Save the returned address.
     */
    pcontroller->mem_descriptors[index].bus_address =
                                pmem_descriptor->bus_address;

    pcontroller->mem_descriptors[index].virtual_address =
                                pmem_descriptor->virtual_address;

    if(pmem_descriptor->virtual_address && 
       pmem_descriptor->size_in_bytes)
    {
        OSM_MEMSET(pmem_descriptor->virtual_address,
                   0,
                   pmem_descriptor->size_in_bytes);

    }

   OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("MemDescriptor: Index %d Alignment 0x%x PA 0x%llu VA 0x%p sz %d\n"),
                                        index,
                                        pcontroller->mem_descriptors[index].alignment_mask,
                                        pcontroller->mem_descriptors[index].bus_address.u.a64.address,
                                        pcontroller->mem_descriptors[index].virtual_address,
                                        pcontroller->mem_descriptors[index].size_in_bytes));

    return ACIT_STATUS_SUCCESS;
}

/*++

Function:
    acit_init_controller

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    init_type       [in ] Type of initialization request

Return:
    ACIT_STATUS_SUCCESS if the Controller was successfully initialized

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to cause ACIT library to initialize
    communication with the hardware function. ACIT library sets up all
    queues to start communicating with the hardware. Any initialization
    specific to the ASIC is carried out at this point.

--*/

ACIT_STATUS
acit_init_controller(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               init_type,
    PACIT_VOID                  osm_handle)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_U32                    status = 0;
    ACIT_UEXACT32               osm_status;
    ACIT_U32                    mem_index, index;
#if ACIT_CFG_VASE_ENABLED
    PACIT_VOID                  pvase_memory;
#endif

    OSM_ASSERT(pcontroller_key);
    //OSM_ASSERT(osm_handle);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    pcontroller->pcontroller_key = pcontroller_key;
    pcontroller->state.init_type = (ACIT_UEXACT8)init_type;

    // initialize global needed for debug
    for (index = 0; index < ACIT_MAX_DEVICES; index++)
    {
        if(g_pcontroller[index] == ACIT_NULL)
        {
            g_pcontroller[index] = pcontroller;
            break;
        }
    }

#if ACIT_CFG_VASE_ENABLED
    ap_initialize_vase( pcontroller, 
           pcontroller->mem_descriptors[AP_MEM_VASE].virtual_address);
#endif

    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,
                        "ACIT CONTROLLER WS",
                        pcontroller);

    /*
     * Log all the mem descriptors in memory for debug reference
     */
    for (mem_index = 0; 
            mem_index < pcontroller->acit_var_desc_index.acit_mem_max;
            mem_index ++)
    {
        DBG_LOG_FUNC_EXEC(&pcontroller->vase_master_block,
                            FUNC_acit_init_controller,
                            VASE_MODULE_ACIT,
                            mem_index,
                            (ACIT_UEXACT64)pcontroller->mem_descriptors[mem_index].virtual_address,
                            pcontroller->mem_descriptors[mem_index].size_in_bytes);
    }

    /* initialize all the available handle count to 0 */
    pcontroller->counters.io_handles_available = 0;
    pcontroller->counters.tmf_handles_available = 0;
    pcontroller->counters.logout_handles_available= 0;
    pcontroller->counters.nop_handles_available = 0;
    pcontroller->counters.inv_handles_available = 0;    
    pcontroller->counters.pending_ios = 0;
    pcontroller->counters.queued_ios = 0;
    pcontroller->counters.pending_tmfs = 0;
    pcontroller->counters.queued_tmfs = 0;
    pcontroller->counters.pending_nops = 0;
    pcontroller->counters.pending_logouts = 0;
    pcontroller->counters.pending_ioctls = 0;

    pcontroller->state.flush_queued_ios = 0;
    pcontroller->state.flush_queued_tmfs = 0;
    pcontroller->state.interrupt_enabled = 0;
    pcontroller->state.soft_reset = 0;
    pcontroller->state.api_type = (ACIT_UEXACT8)acit_api_type;

    pcontroller->busy = 0;
    pcontroller->state.arm_async_registered = 0;

    /*
     * Map the HBA memory as required.
     */

    osm_status = OSM_MAP_IO_HANDLE( pcontroller_key,
                                    0,
                                    sizeof (PCICFG0_CSRMAP),
                                    ACIT_RFLAG_LITTLE_ENDIAN,
                                    &pcontroller->bar0_pa,
                                    &pcontroller->bar0_va,
                                    &pcontroller->bar0_handle);

    if(!osm_status)
    {
        osm_status = OSM_MAP_IO_HANDLE( pcontroller_key,
                                        1,
                                        sizeof (BLADE_ENGINE_CSRMAP),
                                        ACIT_RFLAG_LITTLE_ENDIAN,
                                        &pcontroller->bar1_pa,
                                        &pcontroller->bar1_va,
                                        &pcontroller->bar1_handle);
    }

    if(!osm_status)
    {
        osm_status = OSM_MAP_IO_HANDLE( pcontroller_key,
                                        2,
                                        sizeof (PROTECTION_DOMAIN_DBMAP),
                                        ACIT_RFLAG_LITTLE_ENDIAN,
                                        &pcontroller->bar2_pa,
                                        &pcontroller->bar2_va,
                                        &pcontroller->bar2_handle);
    }

    if(osm_status)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Map IO handle failed\n"));
        return ACIT_STATUS_ERROR;
    }

    pcontroller->plogout_handle_base = (PACIT_LOGOUT_HANDLE) pcontroller->
        mem_descriptors[AP_MEM_LOGOUT_HANDLE].virtual_address;

    pcontroller->ptmf_handle_base = (PACIT_TMF_HANDLE)pcontroller->
        mem_descriptors[AP_MEM_TMF_HANDLE].virtual_address;

    pcontroller->pnop_handle_base = (PACIT_NOP_HANDLE)pcontroller->
        mem_descriptors[AP_MEM_NOP_HANDLE].virtual_address;

    pcontroller->pinv_handle_base = (PACIT_INVALIDATE_HANDLE)pcontroller->
        mem_descriptors[AP_MEM_INV_HANDLE].virtual_address;

    pcontroller->ptarget_key = (PACIT_TARGET_KEY)pcontroller->
        mem_descriptors[AP_MEM_TARGET_KEY].virtual_address;
    
    pcontroller->ptarget_table = (PACIT_TARGET_TABLE)pcontroller->
        mem_descriptors[AP_MEM_TARGET_TABLE].virtual_address;

    pcontroller->ptarget_table->total_entries = pcontroller->params.targets_per_controller;

    LIST_INIT_HEADER(&pcontroller->io_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->io_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->io_queued_list);
    LIST_INIT_HEADER(&pcontroller->inv_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->inv_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->io_pending_reset_list);
    LIST_INIT_HEADER(&pcontroller->logout_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->logout_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->target_event_list);
    LIST_INIT_HEADER(&pcontroller->tmf_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->tmf_handle_busy_list);
    LIST_INIT_HEADER(&pcontroller->tmf_target_list);
    LIST_INIT_HEADER(&pcontroller->tmf_queued_list);
    LIST_INIT_HEADER(&pcontroller->nop_handle_free_list);
    LIST_INIT_HEADER(&pcontroller->nop_handle_busy_list);

    /*
     * Initialize the Management and iSCSI modules.
     */
    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Calling mgmt_init_controller\n"));

    status = mgmt_init_controller(pcontroller);

    if(!status)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Calling iscsi_init_controller\n"));
        status = iscsi_init_controller(pcontroller);
    }

    if(!status)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Calling hwi_init_controller\n"));
        if(pcontroller->state.api_type == ACIT_OP_MODE_RESTRICTED)
            status = hwi_init_hiber_controller(pcontroller, osm_handle);
        else
            status = hwi_init_controller(pcontroller, osm_handle);        
    }

    if(!status)
    {

        //
        // Initialize all the ACIT Handles after hwi init is done. Some
        // of the counts get re-adjusted in hwi_init_controller, so initialize
        // all handles here
        //

        ap_init_io_handle(pcontroller);
        ap_init_logout_handle(pcontroller);
        ap_init_tmf_handle(pcontroller);
        ap_init_nop_handle(pcontroller);
        ap_init_invalidate_handle(pcontroller);

        if(pcontroller->state.api_type == ACIT_OP_MODE_RESTRICTED)
            acit_unitt_init_hiber_controller(pcontroller);
        else
            acit_unitt_init_controller(pcontroller);

        // Moving the link check code to after registering for async events
        // from the ARM. Turns out that if the link was switched from active
        // port to the standby port when the system is in d3 and its an iSCSI
        // boot, get link status over here detected the link was down. But
        // eventually the link up was reported asynchronously and the driver
        // happened to miss it since the async handler was registered in
        // acit discover targets. So the link detection during init will be
        // done _after_ registering for event notifications.
        // mgmt_get_link_status(pcontroller);
        //pcontroller->ld_timeout_ticks[0] = pcontroller->params.ldto * TICKS_PER_SEC(pcontroller); 

        switch(init_type)
        {

        case ACIT_INIT_TYPE_NORMAL:
            
            // Note - this will not be needed once all OSMs are
            // capable of detecting shutdown type properly. Since
            // Windows 2000 and Windows 2003 OSMs cannot differentiate
            // shutdown	from hibernation, we always ask the ARM fw
            // to save the non-persistent targets in flash. Later on
            // during a power up, for normal power up we delete these
            // non-persistent targets. This has the undersired
            // behaviour of even the non-persistent targets showing
            // up in iSCSI select between the time of reboot and
            // the powerup deleting these entries.		
            status = mgmt_delete_nonpersistent_targets(pcontroller);

        case ACIT_INIT_TYPE_POWER_MGMT:

            // re-initialize dic enabled runtime state based on config
            // parameter originally specified
            pcontroller->dic.enabled = pcontroller->params.enable_dic;
            ap_assign_sglh2_ioh(pcontroller);
            break;

        case ACIT_INIT_TYPE_HIBERNATION:
        case ACIT_INIT_TYPE_CRASHDUMP:

            // Disable dynamic interrupt coalescing for crashdump mode
            if(pcontroller->state.api_type != ACIT_OP_MODE_RESTRICTED)
                ap_assign_sglh2_ioh(pcontroller);
            pcontroller->dic.enabled = 0;
            break;

        default:

            OSM_ASSERT(0);
            break;
        }
    }

    if(status)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("be2iscsi controller 0x%p could not be started, status 0x%x\n", pcontroller, status));

        // tear down global debug variable
        for (index = 0; index < ACIT_MAX_DEVICES; index++)
        {
            if(g_pcontroller[index] == pcontroller)
            {
                g_pcontroller[index] = ACIT_NULL;
                break;
            }
        }

        return ACIT_STATUS_ERROR;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("be2iscsi controller 0x%p initialized, asic revision is A%d\n", pcontroller, pcontroller->state.asic_revision));

        if(pcontroller->state.api_type != ACIT_OP_MODE_RESTRICTED)
        {
            PBE_FUNCTION_OBJECT pfo;
            ACIT_UEXACT32 delay_multiplier;
            ACIT_UEXACT32 pid;

            pfo = hwi_get_be_pfo(pcontroller);
            delay_multiplier = be_eq_delay_to_multiplier(pfo, pcontroller->params.eq_timer);

            for(pid = 0; pid < pcontroller->params.num_cq; pid++)
            {
                mgmt_modify_eq_delay(pcontroller,
                                     pid,
                                     pcontroller->phwi_ws->partition[pid].eq.id,
                                     delay_multiplier);

                pcontroller->partition[pid].dic.eq_timer = pcontroller->params.eq_timer;
            }
        }

        return ACIT_STATUS_SUCCESS;
    }
}

/*++

Function:
    acit_stop_controller

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS if the Controller was successfully initialized

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to cause ACIT library to stop 
    the controller driver unload. OSMs call this API during shutdown
    or unload routine. ACIT library will logout of all targets 
    connected and get back all resources that are present with the hardware.

    This routine will not free the I/O mapping resources that are mapped 
    by the OSM. If needed, OSMs must explicitly free the resources that 
    were mapped by ACIT when it is no longer needed.

    This routine will also not free the memory allocated by the OSMs
    for ACIT through the memory descriptors. If OSMs need to free these 
    memory regions, they can call osm_get_mem_descriptor, and free the
    memory region described by the descriptor.

--*/

ACIT_STATUS
acit_stop_controller(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               stop_type)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_UEXACT32               status, index;
    volatile ACIT_UEXACT32      poll_count = 0;

    OSM_ASSERT((stop_type == ACIT_STOP_TYPE_POWER_MGMT) |
               (stop_type == ACIT_STOP_TYPE_NORMAL));

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    ap_flag_busy_controller(pcontroller, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS);

    // turn off dic
    pcontroller->dic.enabled = 0;

    while(pcontroller->counters.pending_ioctls && 
          (pcontroller->eh_stats.ue_detected != 1))
    {
        OSM_DELAY_TIMER(pcontroller->pcontroller_key, 500);
        acit_process_irq(pcontroller->pcontroller_key);
        OSM_DELAY_TIMER(pcontroller->pcontroller_key, 500);
        poll_count++;
        if(poll_count == 500)
        {
            // run UE detection since a UE will stall any
            // pending IOCTLs
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Waiting for %d ioctls to complete..\n", pcontroller->counters.pending_ioctls));
            hwi_check_health(pcontroller, ACIT_TRUE);
            poll_count = 0;
        }
    }

    if(pcontroller->eh_stats.ue_detected)
    {
        // Bail out now
        goto cleanup_last;
    }

    /*
     * De-register firmware events
     */
    status = mgmt_stop_controller(pcontroller, stop_type);

    OSM_ASSERT(!status);

    /*
     * Delete all targets that are connected
     */
    ap_delete_all_targets(pcontroller);

    /*
     * Lastly stop the hardware
     */
    hwi_stop_controller(pcontroller);

cleanup_last:

    ap_flag_ready_controller(pcontroller, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("be2iscsi controller 0x%p stopped successfully\n", pcontroller));

    // tear down global debug variable
    for (index = 0; index < ACIT_MAX_DEVICES; index++)
    {
        if(g_pcontroller[index] == pcontroller)
        {
            g_pcontroller[index] = ACIT_NULL;
            break;
        }
    }

    return ACIT_STATUS_SUCCESS;
}

/*++

Function:
    acit_discover_targets

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pdiscover_param [in/out] Pointer to discover parameters

Return:
    ACIT_STATUS_SUCCESS if a Target discovery was successful

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to cause ACIT library to do a discovery of
    iSCSI Target devices. During initialization, this function is called
    synchronously and only returns after the discovery operation is completed.
    After initialization, this function can be used to start a discovery
    operation asynchronously. When called after initialization, this function
    will return immediately but the discovery operation will continue in the
    background.

    OSMs should expect to see the osm_async_event function called with
    appropriate event_class and event_code to indicate if a new Target has
    been added. In the OSM provided pdiscover_param, the location indicated
    by pptarget_table pointer is filled by ACIT to point to the ACIT allocated
    ACIT_TARGET_TABLE before the first call to osm_async_event function so
    that the OSMs have the pointer to the ACIT_TARGET_TABLE structure before
    they are required to use it.

--*/

ACIT_STATUS
acit_discover_targets(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_DISCOVER_PARAM        pdiscover_param)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_U32                    status = 0, is_polled = 0;

    OSM_ASSERT(pcontroller_key);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    OSM_ASSERT(pcontroller);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_acit_discover_targets,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pdiscover_param->opcode,
                        pcontroller->ptarget_table->change_count,
                        pcontroller->ptarget_table->occupied_entries);

    *(pdiscover_param->pptarget_table) = pcontroller->ptarget_table;

    if(pcontroller->state.discovery_in_progress == ACIT_FALSE)
    {
        /*
         * Indicate that discovery operation is in process.
         */
        pcontroller->state.discovery_in_progress = ACIT_TRUE;
        pcontroller->state.discovery_type = (ACIT_UEXACT8)pdiscover_param->opcode;

        if(pdiscover_param->opcode == ACIT_DISCOVER_OPCODE_SYNCHRONOUS)
        {
            is_polled = ACIT_TRUE;
        }
        else
        {
            is_polled = ACIT_FALSE;
        }

        /*
         * Login to boot target. This IOCTL will be a NoOP for ARM fw
         * if the boot session is already established
         */

        status = mgmt_reopen_sessions(pcontroller,
                                        IOCTL_ISCSI_REOPEN_BOOT_SESSION,
                                        ACIT_U32_MINUS_ONE,   // not used
                                        0, // domain 0
                                        is_polled);

        if(pcontroller->state.operating_mem_level != ACIT_MEM_LEVEL_MINIMUM)
        {
            /*
             * Pass on this request to the Management module
             */
            status = mgmt_discover_targets(pcontroller, is_polled);

            /*
             * Now register for async event notification with ARM fw
             */
            if(pcontroller->state.arm_async_registered == 0)
            {
                hwi_register_arm_async_events(pcontroller);
                pcontroller->state.arm_async_registered = 1;
            }
        }
        else
        {
            /*
             * Report only the boot session, if found
             */
            status = mgmt_discover_boot_target(pcontroller);
        }

        if ((pcontroller->state.init_type == ACIT_INIT_TYPE_POWER_MGMT) ||
            (pcontroller->state.init_type == ACIT_INIT_TYPE_NORMAL) ||
            (pcontroller->params.targets_per_controller > 1))
        {
            status = mgmt_login_all_targets(pcontroller, is_polled);
        }

        // check link status now..
        mgmt_get_link_status(pcontroller, is_polled);
        pcontroller->ld_timeout_ticks[0] = pcontroller->params.ldto * \
                                            TICKS_PER_SEC(pcontroller); 
    }
 
    DBG_LOG_FUNC_EXIT(  &pcontroller->vase_master_block,
                        FUNC_acit_discover_targets,
                        VASE_MODULE_ACIT,
                        status,
                        pcontroller->ptarget_table->change_count,
                        pcontroller->ptarget_table->occupied_entries);

    return ACIT_STATUS_SUCCESS;
}



/*++

Function:
    acit_enable_irq

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS if IRQ was enabled

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to configure the Controller to generate
    interrupt requests. When the Controller is first initialized during
    power up, the IRQ state is disabled by default, OSM's need to call
    this function at the end of controller initialization.

--*/

ACIT_STATUS
acit_enable_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key)
{
    PACIT_CONTROLLER_WS         pcontroller;

    OSM_ASSERT(pcontroller_key);
    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    return hwi_enable_irq(pcontroller);
}



/*++

Function:
    acit_disable_irq

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS if IRQ was disabled

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to configure the Controller to disable
    interrupt generation. OSMs can use this to put the Controller in a
    disabled state during power management events.

--*/

ACIT_STATUS
acit_disable_irq(
    PACIT_CONTROLLER_KEY       pcontroller_key)
{
    PACIT_CONTROLLER_WS  pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    return hwi_disable_irq(pcontroller);
}


/*++

Function:
    acit_process_hiber_irq

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS if the IRQ was successfully processed

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs when the host OS indicates that the
    Controller has generated an interrupt. This function will verify if the
    Controller generated the interrupt and handle the condition reported by
    the Controller.

--*/

ACIT_STATUS
acit_process_hiber_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_IO_REQUEST            pio_request = ACIT_NULL;
    ACIT_U32                    status = ACIT_STATUS_UNKNOWN_INTERRUPT;
    PACIT_TARGET_WS             ptarget;
    ACIT_UEXACT32               index = 0;
    PACIT_TARGET_ENTRY          ptarget_entry = ACIT_NULL;

    OSM_ASSERT(pcontroller_key);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_acit_process_hiber_irq,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) CURRENT_TICK(pcontroller),
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            ptarget = GET_TARGET_WS(ptarget_entry[index].ptarget_key);
            pio_request = ptarget->plast_hiber_io;
            if(pio_request)
            {
                if(pio_request->ppost_routine)
                    (pio_request->ppost_routine)(pcontroller->pcontroller_key,
                                                pio_request);
                else
                    OSM_COMPLETE_IO(pcontroller->pcontroller_key,
                                    pio_request);
                status = ACIT_STATUS_SUCCESS;
                ptarget->plast_hiber_io = ACIT_NULL;
            }
            else
                status = ACIT_STATUS_UNKNOWN_INTERRUPT;
        }
    }

    DBG_LOG_FUNC_EXIT(  &pcontroller->vase_master_block,
                        FUNC_acit_start_hiber_io,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pio_request,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    return status;
}

/*++

Function:
    acit_process_irq

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS if the IRQ was successfully processed

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs when the host OS indicates that the
    Controller has generated an interrupt. This function will verify if the
    Controller generated the interrupt and handle the condition reported by
    the Controller.

--*/

ACIT_STATUS
acit_process_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_U32                    status;

    OSM_ASSERT(pcontroller_key);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    ACIT_UPDATE_LAST_ISR_TICK(pcontroller, 0);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_acit_process_irq,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) CURRENT_TICK(pcontroller),
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    status = hwi_process_irq(pcontroller);

    /*
     * Re-check the flush flag to see if any one wants to flush I/Os or TMFs
     */

    if(pcontroller->state.flush_queued_tmfs)
    {
        ap_flush_controller_tmfs(pcontroller);
    }

    if(pcontroller->state.flush_queued_ios)
    {
        ap_flush_queued_ios(pcontroller, ACIT_NULL, ACIT_NULL);
    }

    if(pcontroller->partition[0].dic.ios_per_sec > pcontroller->params.dic.io_safety_valve)
    {
        if((pcontroller->state.interrupt_enabled) &&
           (pcontroller->dic.enabled && !pcontroller->partition[0].dic.reconfigure_in_progress))
        {
            pcontroller->partition[0].dic.safety_valve_trigger++;
            hwi_check_interrupts(pcontroller, 0);
        }
    }

    // Update on-chip ring index.
    if (pcontroller->partition[0].dic.num_eq_processed > 0) 
    {
        hwi_ring_eq_db (pcontroller, 
                        pcontroller->phwi_ws->partition[0].eq.id, 
                        0, // clear interrupt
                        pcontroller->partition[0].dic.num_eq_processed, 
                        1, 
                        1);

        ACIT_UPDATE_MAX_EQE_PER_ISR(pcontroller, 0);

        pcontroller->partition[0].dic.num_eq_processed = 0;
    }
    else
    {
        //
        // ISR bit was set and yet no EQ was processed! Saw this issue on a HPDL380
        // where the EQ hadn't made it to the host memory by the time ISR was set
        // and interrupt was asserted. Workaround is to rearm the EQ with num_popped 0
        //
        if (status == ACIT_STATUS_SUCCESS)
        {
            pcontroller->partition[0].interrupt_stats.delayed_eq_dma++;

            hwi_ring_eq_db (pcontroller, 
                            pcontroller->phwi_ws->partition[0].eq.id, 
                            0, // clear interrupt
                            pcontroller->partition[0].dic.num_eq_processed, 
                            1, 
                            1);
        }
    }

    DBG_LOG_FUNC_EXIT(  &pcontroller->vase_master_block,
                        FUNC_acit_process_irq,
                        VASE_MODULE_ACIT,
                        status,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    if(status)
        return ACIT_STATUS_UNKNOWN_INTERRUPT;
    else
        return ACIT_STATUS_SUCCESS;
}


/*++

Function:
    acit_process_msi_irq

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS if the IRQ was successfully processed

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs when the host OS indicates that the
    Controller has generated an interrupt. This function will verify if the
    Controller generated the interrupt and handle the condition reported by
    the Controller.

--*/

ACIT_STATUS
acit_process_msi_irq(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    ACIT_UEXACT32               message_id)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_U32                    status;
    OSM_ASSERT(pcontroller_key);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    ACIT_UPDATE_LAST_ISR_TICK(pcontroller, message_id);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_acit_process_msi_irq,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) CURRENT_TICK(pcontroller),
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    status = hwi_process_msi_irq(pcontroller, message_id);

    /*
     * Re-check the flush flag to see if any one wants to flush I/Os or TMFs
     */

    if(pcontroller->state.flush_queued_tmfs)
    {
        ap_flush_controller_tmfs(pcontroller);
    }

    if(pcontroller->state.flush_queued_ios)
    {
        ap_flush_queued_ios(pcontroller, ACIT_NULL, ACIT_NULL);
    }

    if(pcontroller->partition[message_id].dic.ios_per_sec > pcontroller->params.dic.io_safety_valve)
    {
        if((pcontroller->state.interrupt_enabled) &&
           (pcontroller->dic.enabled && !pcontroller->partition[message_id].dic.reconfigure_in_progress))
        {
            pcontroller->partition[message_id].dic.safety_valve_trigger++;
            hwi_check_interrupts(pcontroller, message_id);
        }
    }
  
    // Update on-chip ring index.

    // since we use the clear interrupt bit in EQ,
    // update the ring irrespective of number of
    // eq entries processed
    hwi_ring_eq_db (pcontroller, 
                    pcontroller->phwi_ws->partition[message_id].eq.id, 
                    1, // clear interrupt
                    pcontroller->partition[message_id].dic.num_eq_processed, 
                    1, 
                    1);

    ACIT_UPDATE_MAX_EQE_PER_ISR(pcontroller, message_id);
        
    pcontroller->partition[message_id].dic.num_eq_processed = 0;

    DBG_LOG_FUNC_EXIT(  &pcontroller->vase_master_block,
                        FUNC_acit_process_irq,
                        VASE_MODULE_ACIT,
                        status,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    if(status)
        return ACIT_STATUS_UNKNOWN_INTERRUPT;
    else
        return ACIT_STATUS_SUCCESS;
}


/*++

Function:
    acit_start_hiber_io

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pio_request     [in ] Pointer to the ACIT_IO_REQUEST structure

Return:
    ACIT_STATUS_SUCCESS if the I/O request was successfully started

Context:
    This function can be called from Hiber or Crash dump mode.

Description:
    This function is used by OSMs to start a SCSI I/O request on a LU. OSM
    describe the I/O request using the pio_request pointer. The ACIT_IO_REQUEST
    structure contains elements that indicate the Target and LU to which this
    I/O request is to be sent. The ACIT library delivers the I/O request to
    the Controller. If a non-zero timeout value has been specified in the
    ACIT_IO_REQUEST then ACIT library will ensure that the I/O request is
    completed in the given time. If for some reason the request could not
    complete, ACIT will internally abort the I/O and complete it with an
    aborted status.

--*/

ACIT_STATUS
acit_start_hiber_io(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget;
    PACIT_LUN_WS                plun;
    PACIT_IO_WS                 pio;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pio_request);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptarget = GET_TARGET_WS(pio_request->ptarget_key);
    plun = GET_LUN_WS(pio_request->plun_key);
    pio = GET_IO_WS(pio_request->pio_key);

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(ptarget);
    OSM_ASSERT(plun);
    OSM_ASSERT(pio);

    OSM_ASSERT(ptarget->initialized);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_acit_start_hiber_io,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pio_request,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    pio->pio_handle = ACIT_NULL;

    mgmt_start_hiber_io(pcontroller, pio_request);

    DBG_LOG_FUNC_EXIT(  &pcontroller->vase_master_block,
                        FUNC_acit_start_hiber_io,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pio_request,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    return ACIT_STATUS_SUCCESS;
}


/*++

Function:
    acit_start_io

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pio_request     [in ] Pointer to the ACIT_IO_REQUEST structure

Return:
    ACIT_STATUS_SUCCESS if the I/O request was successfully started

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to start a SCSI I/O request on a LU. OSM
    describe the I/O request using the pio_request pointer. The ACIT_IO_REQUEST
    structure contains elements that indicate the Target and LU to which this
    I/O request is to be sent. The ACIT library delivers the I/O request to
    the Controller. If a non-zero timeout value has been specified in the
    ACIT_IO_REQUEST then ACIT library will ensure that the I/O request is
    completed in the given time. If for some reason the request could not
    complete, ACIT will internally abort the I/O and complete it with an
    aborted status.

--*/

ACIT_STATUS
acit_start_io(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IO_REQUEST            pio_request)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget;
    PACIT_LUN_WS                plun;
    PACIT_IO_WS                 pio;
    ACIT_UEXACT32               status;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pio_request);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptarget = GET_TARGET_WS(pio_request->ptarget_key);
    plun = GET_LUN_WS(pio_request->plun_key);
    pio = GET_IO_WS(pio_request->pio_key);

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(ptarget);
    OSM_ASSERT(plun);
    OSM_ASSERT(pio);

    ACIT_INC_IOS_SUBMITTED(pcontroller);
    ACIT_UPDATE_LAST_IO_TICK(pcontroller);

    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                IOs);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_acit_start_io,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pio_request,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    /*
     * Call ap_alloc_io_resources directly for now. Later when we begin support
     * for full-duplex operations then we can call OSM to invoke
     * ap_alloc_io_resources at this point.
     */
    status = ap_alloc_io_resources(pcontroller, pio_request);

    /*
     * Fill-up the hardware structs and submit it to H/W if we were able to get
     * all resources for this IO.
     */
    if(pio->pio_handle)
    {
        PACIT_IO_HANDLE pio_handle = pio->pio_handle;
        OSM_ASSERT(ptarget->initialized);
        hwi_fill_io_request(pcontroller->phwi_ws, pio_handle);

        // check if BHS opcode is 0
        OSM_ASSERT(pio_handle->pio->cmd_bhs.u1.iscsi_hdr[0]);
        hwi_submit_request(pcontroller, pio_handle->pwrb_handle, pio_handle->cri);
        ACIT_BIT_SET(pio_handle->flags, FLAG_IOH_IO_SUBMITTED);
    }
    else
    {
        if(status == ACIT_STATUS_INSUFFICIENT_TIMEOUT)
        {
            ap_return_io(pcontroller, 
                         pio,
                         ACIT_IO_STATUS_IO_TIMEOUT,
                         ACIT_FALSE);
        }
    }

    DBG_LOG_FUNC_EXIT(  &pcontroller->vase_master_block,
                        FUNC_acit_start_io,
                        VASE_MODULE_ACIT,
                        (ACIT_UEXACT64) pio_request,
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    return ACIT_STATUS_SUCCESS;
}



/*++

Function:
    acit_start_ioctl

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pioctl_request  [in ] Pointer to the ACIT_IOCTL_REQUEST structure

Return:
    ACIT_STATUS_SUCCESS if the IOCTL request was successfully started

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to start a IOCTL request. OSM describe the
    IOCTL request using the pioctl_request pointer.

--*/

ACIT_STATUS
acit_start_ioctl(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_IOCTL_REQUEST         pioctl_request)
{
    PACIT_CONTROLLER_WS         pcontroller;
    ACIT_UEXACT32               status;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pioctl_request);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                IOCTLs);

    if(IS_CONTROLLER_UE(pcontroller))
    {
        mgmt_return_ioctl(pcontroller, pioctl_request, MGMT_STATUS_FAILED);
        return MGMT_STATUS_FAILED;
    }

    status = mgmt_start_ioctl(pcontroller, pioctl_request);

    return status;
}


/*++

Function:
    acit_start_tmf

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pioctl_request  [in ] Pointer to the ACIT_TMF_REQUEST structure

Return:
    ACIT_STATUS_SUCCESS if the IOCTL request was successfully started

Context:
    This function can be called from Interrupt Context

Description:
    The function is used to start Task Mgmt Functions. OSM's describe the
    type of TMF request by a TMF Function. OSM's are responsible for processing
    order of TMF requests(Command Depth)

--*/
ACIT_STATUS
acit_start_tmf(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TMF_REQUEST           ptmf_request)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget = ACIT_NULL;
#if ACIT_CFG_DEBUG_BUILD
    PACIT_LUN_WS                plun = ACIT_NULL;
#endif
    PACIT_TMF_WS                ptmf;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(ptmf_request);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptmf = GET_TMF_WS(ptmf_request->ptmf_key);
    ptarget = GET_TARGET_WS_SAFE(ptmf_request->ptarget_key);

    pcontroller->eh_stats.osm_tmfs++;

    hwi_check_health(pcontroller, ACIT_TRUE);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("TMFRequest 0x%p TMFWs 0x%p Function %d\n", ptmf_request, ptmf, ptmf_request->tmf_function));

#if ACIT_CFG_DEBUG_BUILD
    plun = GET_LUN_WS_SAFE(ptmf_request->plun_key);

#if 0
    // STORPORT_DUPLEX not implemented!
    hwi_validate_pending_ios(pcontroller, ptarget, plun);

    ap_validate_lists(pcontroller, ptarget, plun);
    ap_print_cmd_ticks(pcontroller, ptarget, plun);
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("tmf %d target 0x%p lun 0x%p has %d TMFs queued\n", 
        ptmf_request->tmf_function, ptarget, plun, ptarget->counters.queued_tmfs));

#endif
#endif // ACIT_CFG_DEBUG_BUILD

    switch(ptmf_request->tmf_function)
    {
    case ACIT_TMF_ABORT_TASK:

        pcontroller->eh_stats.osm_aborts++;
        ap_start_abort_io(pcontroller, ptmf_request);
        break;

    case ACIT_TMF_LUN_RESET:

        pcontroller->eh_stats.osm_lun_resets++;
        ap_start_lun_reset(pcontroller, ptmf_request);
        break;

    case ACIT_TMF_TARGET_WARM_RESET:
    case ACIT_TMF_TARGET_COLD_RESET:

        pcontroller->eh_stats.osm_target_resets++;
        ap_start_target_reset(pcontroller, ptmf_request);
        break;

    case ACIT_TMF_ABORT_MULTIPLE_IOS:
    case ACIT_TMF_CLEAR_ACA:
    case ACIT_TMF_ABORT_TASK_SET:
    case ACIT_TMF_CLEAR_TASK_SET:
    case ACIT_TMF_TASK_REASSIGN:
    default:
        {
            ap_log_osm_event(pcontroller,
                            UNI_CFG_LOG_LEVEL_CRITICAL,
                            UNI_SUBMODULE_ACIT_GENERIC,
                            UNI_PHASE_RECOVERY,
                            1,
                            ACIT_EC_UNHANDLED_TMF_REQUEST,
                            0, ACIT_NULL,
                            (PACIT_UEXACT32)(&ptmf_request->tmf_function));

        }

        OSM_ASSERT(0);

        /*
         * Does OSMs send these?
         */

        return ACIT_STATUS_INVALID_PARAMETER;

        break;
    }

    return ACIT_STATUS_SUCCESS;
}

/*++

Function:
    acit_process_timer_tick

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key

Return:
    ACIT_STATUS_SUCCESS.

Context:
    This function can be called from Interrupt Context

Description:
    ACIT requires that OSMs call this function every 500 mili-second so that
    ACIT can complete all background processing. Apart from this function,
    the ACIT library has no other awareness of time. This is used by ACIT
    library to track pending I/O and to monitor the health of iSCSI devices
    and perform any error recovery if required.

--*/

ACIT_STATUS
acit_process_timer_tick(
    PACIT_CONTROLLER_KEY        pcontroller_key)
{
    PACIT_CONTROLLER_WS         pcontroller;

    OSM_ASSERT(pcontroller_key);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    OSM_ASSERT(pcontroller);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_acit_process_timer_tick,
                        VASE_MODULE_ACIT,
                        CURRENT_TICK(pcontroller),
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                TimerTicks);

    /*
     * Increment our current tick count to indicate that another tick occurred
     */
    pcontroller->counters.current_tick_count ++;

    /* Check for any UEs */
    if((CURRENT_TICK(pcontroller) % 
        (pcontroller->params.health_check_freq * TICKS_PER_SEC(pcontroller))) == 0)
    {
        hwi_check_health(pcontroller, ACIT_TRUE);
    }

    /* process ld timeout */
    ap_process_ldto_tick(pcontroller);

    /*
     * Process pending I/Os and see if any of them expired.
     */
    ap_process_io_tick(pcontroller);

    /*
     * Process pending TMFs if any
     */
    ap_process_tmf_tick(pcontroller);

    /*
     * Process any pending logouts and see if we havent got a response
     */
    ap_process_logout_tick(pcontroller);

    /*
     * Check if the there are any targets that have events pending. They will
     * be serviced in order, one after another
     */

    ap_process_target_tick(pcontroller);

    /*
     * For debug builds alone, poll the MGMT IOCTL pending list
     */
#if ACIT_CFG_DEBUG_BUILD
    mgmt_process_ioctl_tick(pcontroller);
    ap_process_dbg_flags(pcontroller);
#endif

    DBG_LOG_FUNC_EXIT(  &pcontroller->vase_master_block,
                        FUNC_acit_process_timer_tick,
                        VASE_MODULE_ACIT,
                        CURRENT_TICK(pcontroller),
                        pcontroller->counters.pending_ios,
                        pcontroller->busy);

    /* Check if we need to reconfigure EQ timers */
    if(pcontroller->dic.enabled)
    {
        ACIT_UEXACT32 pid;
        for(pid = 0; pid < pcontroller->params.num_cq; pid++)
        {
            if(!pcontroller->partition[pid].dic.reconfigure_in_progress)
            {
                hwi_check_interrupts(pcontroller, pid);
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}


/*++

Function:
    acit_get_capabilities

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    pcapabilities   [out] Pointer to the capabilities structure

Return:
    ACIT_STATUS_SUCCESS.

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to determine the capabilities of ACIT library
    and of the underlying layers. OSMs can then decide which features to use and
    can notify the operating system correspondingly. This API should be called
    after OSM has called acit_init_controller.
    
--*/

ACIT_STATUS
acit_get_capabilities(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_CAPABILITIES          pcapabilities)
{
    PACIT_CONTROLLER_WS         pcontroller;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pcapabilities);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);

    OSM_ASSERT(pcontroller);

    pcapabilities->num_cxn = pcontroller->params.num_cxn;
    pcapabilities->num_ios = pcontroller->params.icds_per_controller;

    if(FW_SUPPORTS_INTERRUPT_REDIRECTION(pcontroller))
    {
        pcapabilities->interrupt_redirection = 1;
    }
    else
    {
        pcapabilities->interrupt_redirection = 0;
    }


    return ACIT_STATUS_SUCCESS;
}

/*++

Function:
    acit_update_lun_info

Parameter:
    pcontroller_key [in ] Pointer to the Controller specific Key
    ptarget_key     [in ] Pointer to the Target specific Key
    plun_key        [in ] Pointer to the LUN specific Key
    plun_info       [in ] Pointer to the ACIT_LUN_INFO structure

Return:
    ACIT_STATUS_SUCCESS if the LUN info was successfully updated

Context:
    This function can be called from Interrupt Context

Description:
    This function is used by OSMs to provide the ACIT library additional
    information about the LUN. OSMs initialize the ACIT_LUN_INFO structure
    with all the fields and invoke acit_update_lun_info API. This API must
    be called after OSM has initialized the LUN via acit_init_lun_ws

--*/

ACIT_STATUS
acit_update_lun_info(
    PACIT_CONTROLLER_KEY        pcontroller_key,
    PACIT_TARGET_KEY            ptarget_key,
    PACIT_LUN_KEY               plun_key,
    PACIT_LUN_INFO              plun_info)
{
    PACIT_CONTROLLER_WS         pcontroller;
    PACIT_TARGET_WS             ptarget;
    PACIT_LUN_WS                plun;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(ptarget_key);
    OSM_ASSERT(plun_key);
    OSM_ASSERT(plun_info);

    pcontroller = GET_CONTROLLER_WS(pcontroller_key);
    ptarget = GET_TARGET_WS(ptarget_key);
    plun = GET_LUN_WS(plun_key);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Updating LUN info on Lun 0x%p\n", plun));

    if(!plun)
    {
        OSM_ASSERT(0);
        return ACIT_STATUS_INVALID_PARAMETER;
    }

    OSM_ASSERT(pcontroller);
    OSM_ASSERT(ptarget);
    OSM_ASSERT(plun);

    plun->dev_type = plun_info->device_type;

    // Initialize the retry setting based on the device type and global parameters
    ap_update_lun_retry_param(pcontroller, plun);

    return ACIT_STATUS_SUCCESS;
}


ACIT_VOID
acit_return_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request)
{
    if(pioctl_request->ppost_routine)
    {
        (pioctl_request->ppost_routine)(pcontroller->pcontroller_key, pioctl_request);   
    }
    else
    {
        OSM_COMPLETE_IOCTL(pcontroller->pcontroller_key, pioctl_request);
    }
}

ACIT_VOID
acit_unitt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{
#ifdef ACIT_UNIT_TEST
    PACIT_IO_HANDLE         pio_handle[257] = {0};
    ACIT_UEXACT32           ios = 0;
    ACIT_STATUS             status;

    for(ios = 0; ios < pcontroller->params.ios_per_controller; ios++)
    {
        pio_handle[ios] = ap_alloc_io_handle(pcontroller);
        OSM_ASSERT(pio_handle[ios]);
        pio_handle[ios]->cri = 0;
    }

    // this allocation MUST fail now
    OSM_ASSERT(pcontroller->busy & CONTROLLER_BUSY_IO_HANDLE);
    pio_handle[pcontroller->params.ios_per_controller] = ap_alloc_io_handle(pcontroller);
    OSM_ASSERT(!pio_handle[pcontroller->params.ios_per_controller]);

    for(ios = 0; ios < pcontroller->params.num_cmd_per_cxn; ios++)
    {
        status = hwi_alloc_io_resources(pcontroller, pio_handle[ios]);
        OSM_ASSERT(status == ACIT_STATUS_SUCCESS);
        OSM_ASSERT(pio_handle[ios]->psgl_handle);
        OSM_ASSERT(pio_handle[ios]->pwrb_handle);
        OSM_ASSERT(((PSGL_HANDLE)(pio_handle[ios]->psgl_handle))->sgl_index == ios);
    }
    
    OSM_ASSERT(pcontroller->busy & CONTROLLER_BUSY_IO_SGL_HANDLE);
    status = hwi_alloc_io_resources(pcontroller, 
                                    pio_handle[pcontroller->params.ios_per_controller]);

    OSM_ASSERT(status != ACIT_STATUS_SUCCESS);

    for(ios = 0; ios < pcontroller->params.num_cmd_per_cxn; ios++)
    {
        hwi_free_io_resources(pcontroller, pio_handle[ios]);
    }
    OSM_ASSERT(!pcontroller->busy & CONTROLLER_BUSY_IO_SGL_HANDLE);

    for(ios = 0; ios < pcontroller->params.ios_per_controller; ios++)
    {
        ap_free_io_handle(pcontroller, pio_handle[ios]);
    }
    OSM_ASSERT(!pcontroller->busy & CONTROLLER_BUSY_IO_HANDLE);

    mgmt_unitt_init_controller(pcontroller);

#endif
}

ACIT_VOID
acit_unitt_init_hiber_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{
#ifdef ACIT_UNIT_TEST
    PACIT_IO_HANDLE         pio_handle[257] = {0};
    ACIT_UEXACT32           ios = 0;
    ACIT_STATUS             status;

    for(ios = 0; ios < pcontroller->params.ios_per_controller; ios++)
    {
        pio_handle[ios] = ap_alloc_io_handle(pcontroller);
        OSM_ASSERT(pio_handle[ios]);
        pio_handle[ios]->cri = 0;
    }

    // this allocation MUST fail now
    OSM_ASSERT(pcontroller->busy & CONTROLLER_BUSY_IO_HANDLE);
    pio_handle[pcontroller->params.ios_per_controller] = ap_alloc_io_handle(pcontroller);
    OSM_ASSERT(!pio_handle[pcontroller->params.ios_per_controller]);

    for(ios = 0; ios < pcontroller->params.ios_per_controller; ios++)
    {
        ap_free_io_handle(pcontroller, pio_handle[ios]);
    }
    OSM_ASSERT(!pcontroller->busy & CONTROLLER_BUSY_IO_HANDLE);

    mgmt_unitt_init_controller(pcontroller);

#endif
}

ACIT_STATUS
acit_process_stat_ioctl(
    PACIT_CONTROLLER_KEY    pcontroller_key,
    PDRVR_STATS_REQUEST     pstats_hdr)
{
    PACIT_CONTROLLER_WS     pc;
    PACIT_UEXACT8           pdata_buffer;
    PACIT_PARAMETERS        pparams = ACIT_NULL;
    PACIT_FW_CONFIG         pfw_config = ACIT_NULL;
    PACIT_COUNTERS          pcounters = ACIT_NULL;
    PACIT_CONTROLLER_STATS  pcstats = ACIT_NULL;
    PACIT_CONTROLLER_INT_STATS pint_stats = ACIT_NULL;
    PACIT_CONTROLLER_DIC_COUNTERS pdic = ACIT_NULL;
    PACIT_CONTROLLER_EH_STATS peh = ACIT_NULL;
    PACIT_FW_ADDI_ATTRIBS paddi_attribs = ACIT_NULL;

    OSM_ASSERT(pcontroller_key);
    OSM_ASSERT(pstats_hdr);

    pc = GET_CONTROLLER_WS(pcontroller_key);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("opcode 0x%x buffer_len %d\n", 
                        pstats_hdr->ioctl_code,
                        pstats_hdr->buffer_len));

    pdata_buffer = (PACIT_UEXACT8)pstats_hdr + sizeof(DRVR_STATS_REQUEST);

    switch(pstats_hdr->ioctl_code) 
    {

    case IOCTL_GET_ACIT_PARAMETERS:

        pparams = (PACIT_PARAMETERS)pdata_buffer;
        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_PARAMETERS));
        OSM_MEMCPY(pparams, &pc->params, pstats_hdr->buffer_len);

        break;

    case IOCTL_GET_ACIT_COUNTERS:

        pcounters = (PACIT_COUNTERS)pdata_buffer;

        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_COUNTERS));
        OSM_MEMCPY(pcounters, &pc->counters, pstats_hdr->buffer_len);
        break;

    case IOCTL_GET_ACIT_CONTROLLER_STATS:

        pcstats = (PACIT_CONTROLLER_STATS)pdata_buffer;
        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_CONTROLLER_STATS));
        OSM_MEMCPY(pcstats, &pc->stats, pstats_hdr->buffer_len);

        // computed stats follow
        {
            PMGMT_CONTROLLER_WS pmgmt_controller;
            PMGMT_IOCTL_HANDLE pioctl_handle;
            ACIT_DC_LIST *tmp, *pioch_link;
            ACIT_UEXACT32 pending_index = 0;

            pmgmt_controller = GET_MGMT_CONTROLLER_WS(pc);

            for(pioch_link = pmgmt_controller->ioctl_handle_busy_list.next, tmp = pioch_link->next;
                pioch_link != &pmgmt_controller->ioctl_handle_busy_list;
                pioch_link = tmp, tmp = pioch_link->next) 
            {
                pioctl_handle = STRUCT_BASE(MGMT_IOCTL_HANDLE,
                                            controller_link,
                                            pioch_link);

                // copy 4 bytes of IOCTL header into stats block
                pcstats->pending_ioctl_info[pending_index] = *(PACIT_UEXACT32)pioctl_handle->pmgmt_buffer_h;
                pending_index++;

                if(pending_index >= 8)
                {
                    // no more space to store info
                    break;
                }
            }
        }

        break;

    case IOCTL_GET_ACIT_FW_CONFIG:

        pfw_config = (PACIT_FW_CONFIG)pdata_buffer;
        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_FW_CONFIG));
        OSM_MEMCPY(pfw_config, &pc->fw_config, pstats_hdr->buffer_len);

        break;

    case IOCTL_GET_ACIT_INTERRUPT_STATS:

        pint_stats = (PACIT_CONTROLLER_INT_STATS)pdata_buffer;
        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_CONTROLLER_INT_STATS));
        if(pstats_hdr->parameter1 < pc->params.num_cq)
        {
            OSM_MEMCPY(pint_stats, 
                       &pc->partition[pstats_hdr->parameter1].interrupt_stats, 
                       pstats_hdr->buffer_len);
        }

        break;

    case IOCTL_GET_ACIT_DIC_COUNTERS:

        pdic = (PACIT_CONTROLLER_DIC_COUNTERS)pdata_buffer;
        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_CONTROLLER_DIC_COUNTERS));
        if(pstats_hdr->parameter1 < pc->params.num_cq)
        {
            OSM_MEMCPY(pdic, 
                       &pc->partition[pstats_hdr->parameter1].dic, 
                       pstats_hdr->buffer_len);
        }

        break;

    case IOCTL_GET_ACIT_EH_STATS:

        peh = (PACIT_CONTROLLER_EH_STATS)pdata_buffer;
        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_CONTROLLER_EH_STATS));
        OSM_MEMCPY(peh, &pc->eh_stats, pstats_hdr->buffer_len);

        break;

    case IOCTL_GET_ACIT_FW_ADDI_ATTRIBS:

        paddi_attribs = (PACIT_FW_ADDI_ATTRIBS)pdata_buffer;
        OSM_ASSERT(pstats_hdr->buffer_len == sizeof(ACIT_FW_ADDI_ATTRIBS));
        OSM_MEMCPY(paddi_attribs, &pc->addi_attribs, pstats_hdr->buffer_len);

        break;

    default:

        OSM_ASSERT(0);
        break;
    }

    return ACIT_STATUS_SUCCESS;
}

