/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */


                                                                                
/*++

    Filename:

        sample_acitdefs.h

    Purpose:

        Sample header file provided by ACIT library. OSM developers customize
        this file to make sure basic compiler data types and intrinsic 
        functions used by ACIT library are available. ACIT library includes 
        this OSM provided header file.
        
--*/


/* #define __int64 long long */  /* for Linux systems. */

#ifndef _ACIT_DEFS_
#define _ACIT_DEFS_

/* Define 8 bit data types and their pointers */
typedef int             ACIT_S8,        *PACIT_S8;
typedef unsigned int    ACIT_U8,        *PACIT_U8;
typedef char            ACIT_SEXACT8,   *PACIT_SEXACT8;
typedef unsigned char   ACIT_UEXACT8,   *PACIT_UEXACT8;

/* Define 16 bit data types and their pointers */
typedef int             ACIT_S16,       *PACIT_S16;
typedef unsigned int    ACIT_U16,       *PACIT_U16;
typedef short           ACIT_SEXACT16,  *PACIT_SEXACT16;
typedef unsigned short  ACIT_UEXACT16,  *PACIT_UEXACT16;

/* Define 32 bit data types and their pointers */
typedef int             ACIT_S32,       *PACIT_S32;
typedef unsigned        ACIT_U32,       *PACIT_U32;
typedef int             ACIT_SEXACT32,  *PACIT_SEXACT32;
typedef unsigned int    ACIT_UEXACT32,  *PACIT_UEXACT32;

/* Define 64 bit data types and their pointers */
/*
 * 64-bit data types have been introduced in C for some time now
 * (See ISO C99 Section 6.10.8). We are using it so that we can
 * use compiler and CPU enhancements in handling 64-bit data instead
 * of handling multiple 32-bit data types in our code. This may be
 * a problem if ACIT code needs to be compiled under an old compiler
 * that does not support a native 64-bit data type (embedded platforms).
 */
typedef long long          ACIT_S64,      *PACIT_S64;
typedef unsigned long long ACIT_U64,      *PACIT_U64;
typedef long long          ACIT_SEXACT64, *PACIT_SEXACT64;
typedef unsigned long long ACIT_UEXACT64, *PACIT_UEXACT64;

typedef long            ACIT_S32_64, *PACIT_S32_64;
typedef unsigned long   ACIT_U32_64, *PACIT_U32_64;

/* Define double and float data types and their pointers */
/* Not supported: ACIT will not use these data types.
typedef double          ACIT_DOUBLE,    *PACIT_DOUBLE;
typedef float           ACIT_FLOAT,     *PACIT_FLOAT;
*/

/* Bus Address in ACIT are 64-bits always */
typedef struct _ACIT_BUS_ADDRESS32 ACIT_BUS_ADDRESS32, *PACIT_BUS_ADDRESS32;
typedef struct _ACIT_BUS_ADDRESS64 ACIT_BUS_ADDRESS64, *PACIT_BUS_ADDRESS64;
typedef struct _ACIT_BUS_ADDRESS   ACIT_BUS_ADDRESS,   *PACIT_BUS_ADDRESS;

struct _ACIT_BUS_ADDRESS32
{
    ACIT_UEXACT32       address_lo;
    ACIT_UEXACT32       address_hi;
};

struct _ACIT_BUS_ADDRESS64
{
    ACIT_UEXACT64       address;
};


struct _ACIT_BUS_ADDRESS
{
    union
    {
        ACIT_BUS_ADDRESS32  a32;
        ACIT_BUS_ADDRESS64  a64;
    } u;
};

typedef unsigned long   ACIT_PTR;  /* Size of a pointer on given platform */

typedef void            ACIT_VOID,      *PACIT_VOID;

#define ACIT_NULL       ((PACIT_VOID)0) /* Only for pointers data types */

/* Boolean definitions */
#define ACIT_TRUE       (1)
#define ACIT_FALSE      (0)

/* Compiler intrinsic functions, these may be mapped to OSM desired locations */
#define OSM_MEMSET          memset
#define OSM_MEMCPY          memcpy
#define OSM_MEMCMP          memcmp
/*CR: See if string compare is required. */
#define OSM_MALLOC          malloc  /* temporary definition. */

/* Debug support */
#if ACIT_CFG_DEBUG_BUILD

extern int acit_debug_level;

#define OSM_ASSERT(condition)               if(!(condition)) {printk("ASSERT:"#condition" %s %d\n", __FILE__, __LINE__);} 
#define OSM_ASSERT_MSG(condition, message)  if(!(condition)) {printk("ASSERT:"#message" %s %d\n",__FILE__, __LINE__);}
#define OSM_DEBUG_PRINT(severity, string) {                             \
                         if(severity & acit_debug_level) printk string; \
                     }
#define OSM_ENTER_DEBUGGER()               asm("int $3"); 
#else /* ACIT_CFG_DEBUG_BUILD */
extern int acit_debug_level;
#define OSM_ASSERT(condition) 
#define OSM_ASSERT_MSG(condition, message)

#define SET_ACIT_DEBUG_LEVEL() {acit_debug_level = 0;}
/*
 * Replace the above SET_ACIT_DEBUG_LEVEL macro with the following to
 * enable debug prints needed by microcode team
 */
/*#define SET_ACIT_DEBUG_LEVEL() {acit_debug_level = ACIT_DBGMSG_UCODE;}*/
#define OSM_DEBUG_PRINT(severity, string)

#define OSM_ENTER_DEBUGGER()
#endif /* ACIT_CFG_DEBUG_BUILD */

#ifndef _ACIT_LINUX_
static int avoid_compiler_warning(void)
{
        acit_debug_level = acit_debug_level;
        avoid_compiler_warning();
        return 1;
}
#endif

/*
 * Endian macros
 */

/* Endian byte swapping macros */
#define ACIT_SWAP16(data)   (ACIT_UEXACT16) (                               \
                            (((ACIT_UEXACT16)data & 0x00FF) << 8) |         \
                            (((ACIT_UEXACT16)data & 0xFF00) >> 8) )
#define ACIT_SWAP24(data)   (ACIT_UEXACT32) (                               \
                            (((ACIT_UEXACT32)data & 0x000000FF) << 16) |    \
                            (((ACIT_UEXACT32)data & 0x0000FF00)      ) |    \
                            (((ACIT_UEXACT32)data & 0x00FF0000) >> 16) )
#define ACIT_SWAP32(data)   (ACIT_UEXACT32) (                               \
                            (((ACIT_UEXACT32)data & 0x000000FF) << 24) |    \
                            (((ACIT_UEXACT32)data & 0x0000FF00) << 8)  |    \
                            (((ACIT_UEXACT32)data & 0x00FF0000) >> 8)  |    \
                            (((ACIT_UEXACT32)data & 0xFF000000) >> 24) )

#define ACIT_SWAP64(data)   (ACIT_U64) (                                    \
                            (((ACIT_U64)data & 0x00000000000000FF) << 56) | \
                            (((ACIT_U64)data & 0x000000000000FF00) << 40) | \
                            (((ACIT_U64)data & 0x0000000000FF0000) << 24) | \
                            (((ACIT_U64)data & 0x00000000FF000000) << 8)  | \
                            (((ACIT_U64)data & 0x000000FF00000000) >> 8)  | \
                            (((ACIT_U64)data & 0x0000FF0000000000) >> 24) | \
                            (((ACIT_U64)data & 0x00FF000000000000) >> 40) | \
                            (((ACIT_U64)data & 0xFF00000000000000) >> 56) )

#if ACIT_CFG_CPU_LITTLE_ENDIAN
/* Endian store macros */
#define ACIT_PUT_LE8(newval, oldval)       \
            newval  = (ACIT_UEXACT8)(oldval)
#define ACIT_PUT_LE16(newval, oldval)      \
            newval = (ACIT_UEXACT16)(oldval)
#define ACIT_PUT_LE32(newval, oldval)      \
            newval = (ACIT_UEXACT32)(oldval)
#define ACIT_PUT_LE64(newval, oldval)      \
            newval = (ACIT_UEXACT64)(oldval)

#define ACIT_PUT_BE8(newval, oldval)       \
            newval = (ACIT_UEXACT8)oldval
#define ACIT_PUT_BE16(newval, oldval)      \
            newval = ACIT_SWAP16(oldval)
#define ACIT_PUT_BE32(newval, oldval)      \
            newval = ACIT_SWAP32(oldval)
#define ACIT_PUT_BE64(newval, oldval)      \
            newval = ACIT_SWAP64(oldval)


/* Endian load macros */
#define ACIT_GET_LE8(plocation)       \
            *((PACIT_UEXACT8) (plocation))
#define ACIT_GET_LE16(plocation)      \
            *((PACIT_UEXACT16) (plocation))
#define ACIT_GET_LE32(plocation)      \
            *((PACIT_UEXACT32) (plocation))
#define ACIT_GET_LE64(plocation)      \
            *((PACIT_UEXACT64) (plocation))

#define ACIT_GET_BE8(plocation)       \
            *((PACIT_UEXACT8) (plocation))
#define ACIT_GET_BE16(plocation)      \
            ACIT_SWAP16(*((PACIT_UEXACT16) (plocation)))
#define ACIT_GET_BE32(plocation)      \
            ACIT_SWAP32(*((PACIT_UEXACT32) (plocation)))
#define ACIT_GET_BE64(plocation)      \
            ACIT_SWAP64(*((PACIT_UEXACT64) (plocation)))

#elif ACIT_CFG_CPU_BIG_ENDIAN

#error "Unsupported get/set endian macros"

#endif

#endif /* _ACIT_DEFS_ */

