/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//

//
// @file
//    be2_gen_check_defines.h
// 
// @brief
//      This file should only be included by the be2_gen_*.h files to verify 
//      the includer has defined appropriate abstraction macros.
//
#ifndef __be2_gen_check_defines_h__
#define __be2_gen_check_defines_h__

//
// Macros required to be defined by the caller
//
#ifndef STATIC
    #error STATIC not defined! Generally use "static".
#endif

#ifndef INLINE
    #error INLINE not defined! Generally use "inline" or "__inline".
#endif
 
#ifndef BE_GEN_ASSERT
    #error BE_GEN_ASSERT(conditional_expression) not defined!
#endif

#ifndef BE_GEN_C_ASSERT
    #error BE_GEN_C_ASSERT(constant_expression) not defined!
#endif

#ifndef BE_GEN_DEBUG_PRINT
    #error  BE_GEN_DEBUG_PRINT(string) not defined!
#endif

#ifndef BE_GEN_WRITE_CTX
    #error BE_GEN_WRITE_CTX(function_context, byte_offset, data_pointer, data_bytes) not defined!
#endif

#ifndef BE_GEN_CSR_READ_ADDR
    #error BE_GEN_CSR_READ_ADDR(function_context, address) not defined!
#endif

#ifndef BE_GEN_CSR_WRITE_ADDR
    #error BE_GEN_CSR_WRITE_ADDR(function_context, address, value) not defined!
#endif

#ifndef BE_GEN_PCI0_READ_ADDR
    #error BE_GEN_PCI0_READ_ADDR(function_context, address) not defined!
#endif

#ifndef BE_GEN_PCI0_WRITE_ADDR
    #error BE_GEN_PCI0_WRITE_ADDR(function_context, address, value) not defined!
#endif



//
// Macros that we can define if not already defined by the caller
//
#ifndef BE_GEN_STATIC_INLINE
    #define BE_GEN_STATIC_INLINE STATIC INLINE
#endif

#ifndef BE_GEN_FIELD_OFFSET
    #define BE_GEN_FIELD_OFFSET(__type__, __field__) \
        ((long)(unsigned int)&(((__type__*)0)->__field__))
#endif

#ifndef BE_GEN_SIZEOF_FIELD
    #define BE_GEN_SIZEOF_FIELD(__type__, __field__) \
        sizeof(((__type__*)0)->__field__)
#endif

#ifndef BE_GEN_NUMBER_OF_FIELD
    #define BE_GEN_NUMBER_OF_FIELD(__type__, __field__) \
        (BE_GEN_SIZEOF_FIELD(__type__, __field__)/sizeof((((__type__*)0)->__field__[0])))
#endif

#ifndef BE_GEN_TO_U16
    #define BE_GEN_TO_U16(__val__) \
        ((u16) ((__val__) & 0xFFFF))
#endif

#ifndef BE_GEN_SWAP32
    #define BE_GEN_SWAP32(__val__) \
        ((((__val__) & 0x000000FF) << 24) | \
         (((__val__) & 0x0000FF00) << 8)  | \
         (((__val__) & 0x00FF0000) >> 8)  | \
         (((__val__) & 0xFF000000) >> 24))
#endif

#ifndef BE_GEN_CSR_READ
    #define BE_GEN_CSR_READ(__fc__, __field__) \
        BE_GEN_CSR_READ_ADDR(__fc__, BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, __field__))
#endif

#ifndef BE_GEN_CSR_WRITE
    #define BE_GEN_CSR_WRITE(__fc__, __field__, __val__) \
        BE_GEN_CSR_WRITE_ADDR(__fc__, BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, __field__), __val__)
#endif

#ifndef BE_GEN_PCI0_READ
    #define BE_GEN_PCI0_READ(__fc__, __field__) \
        BE_GEN_PCI0_READ_ADDR(__fc__, BE_GEN_FIELD_OFFSET(PCICFG_PF0_CSRMAP, __field__))
#endif

#ifndef BE_GEN_PCI0_WRITE
    #define BE_GEN_PCI0_WRITE(__fc__, __field__, __val__) \
        BE_GEN_PCI0_WRITE_ADDR(__fc__, BE_GEN_FIELD_OFFSET(PCICFG_PF0_CSRMAP, __field__), __val__)
#endif

#ifndef BE_GEN_PCI_READ
    #define BE_GEN_PCI_READ(__fc__, __function__, __field__) \
	BE_GEN_PCI_READ_ADDR(__fc__, __function__, BE_GEN_FIELD_OFFSET(PCICFG## __function__ ## _CSRMAP, __field__))
#endif

#ifndef BE_GEN_PCI_WRITE
    #define BE_GEN_PCI_WRITE(__fc__, __function__, __field__, __val__) \
	BE_GEN_PCI_WRITE_ADDR(__fc__, __function__, BE_GEN_FIELD_OFFSET(PCICFG## __function__ ## _CSRMAP, __field__), __val__)
#endif

#ifndef BE_GEN_CTX_WR
    #define BE_GEN_CTX_WR(__fc__, __addr__, __ptr__, __bytes__) \
        BE2_Gen_Access_Context_Memory(__fc__, __addr__, __ptr__, __bytes__, TRUE)
#endif

#ifndef BE_GEN_CTX_RD
    #define BE_GEN_CTX_RD(__fc__, __addr__, __ptr__, __bytes__) \
        BE2_Gen_Access_Context_Memory(__fc__, __addr__, __ptr__, __bytes__, FALSE)
#endif

#endif // __be2_gen_check_defines_h__

