/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_mdio_h__
#define __be2_gen_mdio_h__

#include "be2_gen_check_defines.h"
#include "be2_gen_api_enum.h"

// Define private literals
#define BE_GEN_MDIO_ADDR_OPCODE             (0x0)
#define BE_GEN_MDIO_WRITE_OPCODE            (0x1)
#define BE_GEN_MDIO_READ_OPCODE             (0x3)
#define BE_GEN_MDIO_READINCR_OPCODE         (0x2)
#define BE_GEN_MDIO_1G_READ_OPCODE          (0x2)


STATIC INLINE
u32 BE2_Gen_MDIO_Read()
{
    u32 val;
    /* Read and return the reg_data */
    //BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_DATA, BE_GEN_GPIO_IN_MODE);
    val = BE2_Gen_GPI_Get_Data(NULL, BE_GEN_GPIO_BIT_MDIO_DATA);
    //BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_DATA, BE_GEN_GPIO_OUT_MODE);
    return val;	

}


STATIC INLINE
void BE2_Gen_MDIO_Data(u32 reg_data)
{
    /* write the reg_data */
    BE2_Gen_GPO_Set_Data(NULL, BE_GEN_GPIO_BIT_MDIO_DATA,reg_data);
}


STATIC INLINE
void BE2_Gen_MDIO_Init()
{
    BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_CLOCK, BE_GEN_GPIO_OUT_MODE);
    BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_DATA, BE_GEN_GPIO_OUT_MODE);

}


STATIC INLINE
void BE2_Gen_MDIO_Clk(u32 reg_data)
{
    /* write the reg_data */
    BE2_Gen_GPO_Set_Data(NULL, BE_GEN_GPIO_BIT_MDIO_CLOCK,reg_data);
}


STATIC INLINE
void BE2_Gen_MDIO_delay()
{
    /*delay should be atleast 160ns*/
    volatile u32 i;
    u32 delay = 0x100;
    for ( i = 0; i < delay; i++);

}


STATIC INLINE
void BE2_Gen_MDIO_Send_Bit(u32 reg_data)
{

    BE2_Gen_MDIO_Data(reg_data);
    BE2_Gen_MDIO_delay();
    BE2_Gen_MDIO_Clk(1);
    BE2_Gen_MDIO_delay();
    BE2_Gen_MDIO_Clk(0);

}


STATIC INLINE
u8 BE2_Gen_MDIO_Get_Bit()
{
    BE2_Gen_MDIO_delay();
    BE2_Gen_MDIO_Clk(1);
    BE2_Gen_MDIO_delay();
    BE2_Gen_MDIO_Clk(0);
    return (BE2_Gen_MDIO_Read()& 0xFF);

}


/* Used to send preamble , address and register (read/write)
 *
 */
STATIC INLINE
void BE2_Gen_MDIO_Send_Preamble(u32 opcode, u8 port, u8 phy_addr, u8 mdio_mode)
{

    u32 j;
    /*send 32 1's */

    for(j = 0; j < 32; j++)
    {
        BE2_Gen_MDIO_Send_Bit(1);
    }
    if(mdio_mode == BE_GEN_MDIO_OP_1G)
    {
        /*Send start bit (01) */
        BE2_Gen_MDIO_Send_Bit(0);
        BE2_Gen_MDIO_Send_Bit(1);
    }
    else
    {
        /*Send start bit (00) */
        BE2_Gen_MDIO_Send_Bit(0);
        BE2_Gen_MDIO_Send_Bit(0);
    }

    BE2_Gen_MDIO_Send_Bit(((opcode >> 1) & 0x1) != 0);
    BE2_Gen_MDIO_Send_Bit((opcode  & 0x1) != 0);
    /*Send the Port address */
    for(j = 0; j < 5; j++)
    {
        BE2_Gen_MDIO_Send_Bit((port & 0x10) != 0);
        port <<= 1;	
    }

    /* Send the device address */
    for(j = 0; j < 5; j++)
    {
        BE2_Gen_MDIO_Send_Bit((phy_addr & 0x10) != 0);
        phy_addr <<= 1;	
    }

}


STATIC INLINE 
void BE2_Gen_MDIO_Write_Addr_Data(u32 opcode, u8 port, u8 phy_addr, u32 reg_data, u8 mdio_mode)
{
    u32 j;
    BE2_Gen_MDIO_Send_Preamble(opcode, port, phy_addr, mdio_mode);
    BE2_Gen_MDIO_Init();
    /* send the Turn around -10*/
    BE2_Gen_MDIO_Send_Bit(1);
    BE2_Gen_MDIO_Send_Bit(0);
    /* Write 16 bits of register data, MSB first */
    for( j = 0; j < 16; j++)
    {
        BE2_Gen_MDIO_Send_Bit( (reg_data & 0x8000) != 0);
        reg_data <<= 1;
    }
    BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_DATA, BE_GEN_GPIO_IN_MODE);
    BE2_Gen_MDIO_Get_Bit();
    BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_DATA, BE_GEN_GPIO_OUT_MODE);
//#ifdef PROJ_QPGE
    /* needs two extra clocks */
    BE2_Gen_MDIO_Send_Bit(0);
    BE2_Gen_MDIO_Send_Bit(0);
//#endif
    return;
}



//---------------------------------------------------------------------------------------
// Function: BE2_Gen_MDIO_Read_Reg
//
//   This routine returns the value of a PHY register. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// phy_addr           - PHY address. 
// reg_addr           - Register address. 
// port               - Physical port (0 or 1). 
// return return_type - Read Data. 
//---------------------------------------------------------------------------------------
STATIC INLINE u16
BE2_Gen_MDIO_Read_Reg (
   IN void*   function_context,
   IN u8      phy_addr,
   IN u16     reg_addr,
   IN boolean port,
   IN u8      mdio_mode )
{
    u32 	j;
    u32	rdreg;

    rdreg = 0;

    BE2_Gen_MDIO_Init();
    if(mdio_mode == BE_GEN_MDIO_OP_1G)
    {
        /*send the preamble */
        BE2_Gen_MDIO_Send_Preamble(BE_GEN_MDIO_1G_READ_OPCODE, port, reg_addr, mdio_mode);
    }
    else
    {
        BE2_Gen_MDIO_Write_Addr_Data(BE_GEN_MDIO_ADDR_OPCODE, port, phy_addr, reg_addr, mdio_mode);
        /*send the preamble */
        BE2_Gen_MDIO_Send_Preamble(BE_GEN_MDIO_READ_OPCODE, port, phy_addr, mdio_mode);
    }
    /* check the Turn around bit :The PHY should be driving it to 0 */
    BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_DATA, BE_GEN_GPIO_IN_MODE);
    if(BE2_Gen_MDIO_Get_Bit() != 0)
    {
        /*PHY did not drive it low */
        for( j = 0; j < 32; j++)
        {
            BE2_Gen_MDIO_Get_Bit();
        }
    }
    else
    {
        /* Read 16 bits of register data, MSB first */
        for( j = 0; j < 16; j++)
        {
            rdreg <<= 1;
            rdreg |= BE2_Gen_MDIO_Get_Bit();
        } 
        BE2_Gen_MDIO_Get_Bit();
        BE2_Gen_MDIO_delay();
    }
    BE2_Gen_GPIO_Set_Mode(NULL, BE_GEN_GPIO_BIT_MDIO_DATA, BE_GEN_GPIO_OUT_MODE);
//#ifdef PROJ_QPGE
    /* needs two extra clocks */
    BE2_Gen_MDIO_Send_Bit(0);
    BE2_Gen_MDIO_Send_Bit(0);
//#endif
    return rdreg;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_MDIO_Write_Reg
//
//   This routine writes a PHY register. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// phy_addr           - PHY address. 
// reg_addr           - Register address. 
// port               - Physical port (0 or 1). 
// reg_data           - Data to write. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_MDIO_Write_Reg (
   IN void*   function_context,
   IN u8      phy_addr,
   IN u16     reg_addr,
   IN boolean port,
   IN u16     reg_data,
   IN u8      mdio_mode )
{
    BE2_Gen_MDIO_Init();
    if(mdio_mode == BE_GEN_MDIO_OP_10G)
    {
        BE2_Gen_MDIO_Write_Addr_Data(BE_GEN_MDIO_ADDR_OPCODE, port, phy_addr, reg_addr, mdio_mode);
        BE2_Gen_MDIO_delay();
        BE2_Gen_MDIO_Write_Addr_Data(BE_GEN_MDIO_WRITE_OPCODE, port, phy_addr, reg_data, mdio_mode);
    }
    else
    {
        BE2_Gen_MDIO_Write_Addr_Data(BE_GEN_MDIO_WRITE_OPCODE, port, reg_addr, reg_data, mdio_mode);
    }
//#ifdef PROJ_QPGE
    /* needs two extra clocks */
    BE2_Gen_MDIO_Send_Bit(0);
    BE2_Gen_MDIO_Send_Bit(0);
//#endif
    BE2_Gen_MDIO_delay();
    return;
}

// Undefine private literals
#undef BE_GEN_MDIO_ADDR_OPCODE             
#undef BE_GEN_MDIO_WRITE_OPCODE            
#undef BE_GEN_MDIO_READ_OPCODE             
#undef BE_GEN_MDIO_READINCR_OPCODE         

#endif  //__be2_gen_mdio_h__
