/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        be2iscsichrdrv.c

    Purpose:

        This file implements the different entry points that are needed
        from a character driver, the open, close and ioctl entry points.

--*/
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/pci.h>
#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include "acitdefs.h"
#include "acitscsi.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "iscsi_drvr_stats.h"
#include "acitapi.h"
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/genhd.h>
#include <linux/blkdev.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <scsi/scsi_host.h>
#include "scsi/scsi_cmnd.h"
#include "be2iscsi_common.h"
#include "be2iscsi_debug.h"
#include<linux/pci.h>
#include "be2iscsi_osm.h"
#include "acit_iface.h"
#include "version.h"
#include "scsi/scsi.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
#include "scsi/scsi_request.h"
#else
#include "scsi/scsi_eh.h"
#include "scsi/scsi_device.h"
#endif
#ifdef __VMKLNX__
#if defined(ESX_50)
#include <vmklinux_scsi.h>   // For Transport type
#include <asm/uaccess.h>
#else
#include <vmklinux26_scsi.h>   // For Transport type
#endif
#define IOCTL_CMD_VMHBA_NAME 1
#endif


extern ACIT_STATUS
acit_process_stat_ioctl(
	PACIT_CONTROLLER_KEY    pcontroller_key,
	PDRVR_STATS_REQUEST    pstats_hdr);     

/* Mutex for locking while we walk the list */
extern struct semaphore se_configuration_lock;
extern ACIT_SEXACT8 hba_count;
extern struct list_head hba_chain;
#define READ_CAPACITY_RESP_SIZE 8
#define INQUIRY_RESP_SIZE 36


ACIT_UEXACT32
se_login_completion(
	PACIT_CONTROLLER_KEY controller_key,
	PACIT_IOCTL_REQUEST pioctl_request)
{
	se_hba        *phba;
	ACIT_UEXACT32   status;
	status = ((PIOCTL_RESPONSE_HEADER)pioctl_request)->status;
	phba = pioctl_request->posm_reference;

	if (status != MGMT_STATUS_SUCCESS) {
		complete(&phba->login_completion);
	}

	return ACIT_STATUS_SUCCESS;
}

se_device *
se_find_device
(
se_hba         *phba,
se_target      *ptarget,
PACIT_UEXACT8    lun)
{
	se_device *ptemp,*ptemp0=NULL;
	list_for_each_entry(ptemp, &ptarget->lun_list, sibling_list){ 
	
		if (ptemp0 == NULL)
			ptemp0=ptemp;

		if (memcmp((char *)&ptemp->lun.u.lun_complete, lun,8)== 0)
			return ptemp;       
	}
	return ptemp0;
}


#define OSM_IS_BIT_CLR(Value, Bit)             (!((Value) & (Bit)))
se_target *
se_find_target
(
 se_hba         *phba,
 ACIT_UEXACT32    session_handle)
{
	ACIT_UEXACT32 i;
	PACIT_TARGET_ENTRY ptarget_entry;

	ptarget_entry = &phba->pacit_target_table->target_entry[0];

	for(i = 0; i < phba->pacit_target_table->total_entries; i++)
	{
		if (!(OSM_IS_BIT_CLR(ptarget_entry[i].acit_tflags,ACIT_TFLAG_IN_USE)))
		{
			if (session_handle == ptarget_entry[i].session_handle)
			{
				return ptarget_entry[i].ptarget_key->osm_key;
			}
		}
	}

	return ACIT_NULL;
}

ACIT_STATUS
se_do_io_ioctl(se_hba *phba, ACIT_UEXACT8 *cdb, PACIT_UEXACT8 lun,
		ACIT_UEXACT32 session_handle,
		PIOCTL_HEADER phdr,
		PACIT_UEXACT8 pbuffer,
		ACIT_UEXACT32 buf_size,
		ACIT_UEXACT32 min_req_size)
{

	char            *kbuf;
	ACIT_STATUS     acit_status = MGMT_STATUS_SUCCESS;
	ACIT_UEXACT32   exp_data_length;
	se_target     *ptarget;
	se_device     *pdevice;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	struct scsi_request     *sreq;
#else
	struct scsi_sense_hdr sshdr;
#endif

	exp_data_length = 
		((PIOCTL_REQUEST_HEADER)(phdr))->request_length;
	OSM_LOG_IOCTL(phba, KERN_INFO,"exp_datalen = %x, req_datalen = %x\n",
			exp_data_length, min_req_size);

	if ((cdb[0] == WRITE_12)||(cdb[0] == WRITE_10)||(cdb[0] == WRITE_6)){
		buf_size = min_req_size;
	}

	kbuf = (char *)kmalloc(buf_size, GFP_KERNEL);
	if (!kbuf){
		OSM_LOG_IOCTL(phba, KERN_ERR,"no kbuf\n");
		acit_status = MGMT_STATUS_FAILED;
		goto no_min;
	}
	memset(kbuf, 0, buf_size);

	ptarget = se_find_target(phba, session_handle);

	if (!ptarget){
		OSM_LOG_IOCTL(phba, KERN_ERR,"no ptarget\n");
		acit_status = MGMT_STATUS_INVALID_SESSION;

		goto free_buf;
	}

	pdevice = se_find_device(phba, ptarget, lun);
	if (!pdevice){
		OSM_LOG_IOCTL(phba, KERN_ERR,"no pdevice\n");
		if (cdb[0] == REPORT_LUNS){
			memset(pbuffer, 0x0, buf_size);
			acit_status = MGMT_STATUS_SUCCESS;
		}
		else		
			acit_status = MGMT_STATUS_FAILED;

		goto free_buf;
	}

#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	sreq = scsi_allocate_request(pdevice->pScsi_dev, GFP_KERNEL);
	if (!sreq){
		OSM_LOG_IOCTL(phba, KERN_ERR, "no sreq\n");
		acit_status = MGMT_STATUS_FAILED;
		goto free_buf;
	}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16) //Hoping 2.4 kernel will be covered
	switch(cdb[0]){
		case ACIT_SCSI_CMD_INQUIRY:
			sreq->sr_cmd_len = 6;
			break;
		case ACIT_SCSI_CMD_REPORTLUNS:
			sreq->sr_cmd_len = 12;
			break;
		case 0x25:
			sreq->sr_cmd_len = 10;
			break;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,5)
	if ((cdb[0] == WRITE_12)||(cdb[0] == WRITE_10)||(cdb[0] == WRITE_6))
		sreq->sr_data_direction = SCSI_DATA_WRITE;
	else
		sreq->sr_data_direction = SCSI_DATA_READ;
#else
	if ((cdb[0] == WRITE_12)||(cdb[0] == WRITE_10)||(cdb[0] == WRITE_6))
		sreq->sr_data_direction = DMA_TO_DEVICE;
	else
		sreq->sr_data_direction = DMA_FROM_DEVICE;
#endif

	scsi_wait_req(sreq, cdb, kbuf, buf_size, 8*HZ, 3);
	if (sreq->sr_result){
		acit_status = MGMT_STATUS_FAILED;
		goto cmd_failed;
	}

#else

	if ((cdb[0] == WRITE_12)||(cdb[0] == WRITE_10)||(cdb[0] == WRITE_6)){
		if (scsi_execute_req(pdevice->pScsi_dev, cdb, DMA_TO_DEVICE,
					kbuf, buf_size,&sshdr , 8*HZ,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
					3) )
#else
			3, NULL) )
#endif
			{
				acit_status = MGMT_STATUS_FAILED;
				goto cmd_failed;
			}
	}else{
		if (scsi_execute_req(pdevice->pScsi_dev, cdb, DMA_FROM_DEVICE,
					kbuf, buf_size,&sshdr , 8*HZ,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
					3) )
#else
			3, NULL) )
#endif
			{
				acit_status = MGMT_STATUS_FAILED;
				goto cmd_failed;

			}
	}
#endif

	memcpy(pbuffer, kbuf, buf_size);
cmd_failed:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
	scsi_release_request(sreq);
#endif
free_buf:
	kfree(kbuf);
	return acit_status;
no_min:
	return acit_status;
}

ACIT_STATUS se_build_ioctl_req(
		se_hba 		*phba,
		PIOCTL_HEADER 		pioctl_req, 
		PACIT_IOCTL_REQUEST 	pacit_ioctl_req,
		PACIT_VOID		pdata_buffer,
		PACIT_IOCTL_REQUEST_POST ppost_routine)
{
	unsigned int datalen;
	if (!phba || !pioctl_req || !pacit_ioctl_req)
		return ACIT_STATUS_ERROR;

	datalen = ((PIOCTL_REQUEST_HEADER)(pioctl_req))->request_length;

	memcpy(&pacit_ioctl_req->ioctl_hdr, pioctl_req, sizeof(IOCTL_HEADER));
	pacit_ioctl_req->posm_reference = phba;
	pacit_ioctl_req->ppost_routine = ppost_routine;
	pacit_ioctl_req->pdata_buffer = pdata_buffer;
	/* TODO: initialize data_length */
	pacit_ioctl_req->data_length = datalen;
	return ACIT_STATUS_SUCCESS;
}

	ACIT_UEXACT32 
se_complete_ioctl(PACIT_IOCTL_REQUEST pioctl_req)
{
	se_hba	*phba;
	phba = pioctl_req->posm_reference;

	OSM_LOG_IOCTL(phba, KERN_INFO,"Completed some internal ioctl %x\n",
			pioctl_req->ioctl_hdr.request.opcode);

	switch(pioctl_req->ioctl_hdr.request.opcode){
		case OPCODE_COMMON_GET_CNTL_ATTRIBUTES:
			{
				PMGMT_CONTROLLER_ATTRIBUTES pmca;
				pmca = 
					(PMGMT_CONTROLLER_ATTRIBUTES)pioctl_req->pdata_buffer;
				snprintf(pmca->hba_attribs.driver_version_string, 
						MGMT_MAX_DRV_VERSION_STRING,
						"%d.%d.%d.%d",BE_MAJOR,BE_MINOR,BE_BUILD,BE_BRANCH);

				pmca->hba_attribs.max_domains_supported=1;
				pmca->hba_attribs.max_cdblength = ACIT_SCSI_MAX_CDB_LENGTH;
				pmca->pci_vendor_id = phba->pcidev->vendor;
				pmca->pci_device_id = phba->pcidev->device;
				pmca->pci_sub_vendor_id = phba->pcidev->subsystem_vendor;
				pmca->pci_sub_system_id = phba->pcidev->subsystem_device;
				pmca->pci_bus_number = phba->pcidev->bus->number;
				pmca->interface_type = MGMT_HBA_IF_TYPE_INITIATOR;
				pmca->pci_device_number = PCI_SLOT(phba->pcidev->devfn);
				pmca->pci_function_number =
					PCI_FUNC(phba->pcidev->devfn);
				pmca->unique_identifier = phba->hba_unique_id;
			}	
	}
	return ACIT_STATUS_SUCCESS;
}

ACIT_SEXACT32
se_stats_ioctl(
		se_hba	*phba,
		void	*arg)
{
	PVOID pkern_arg;
	ACIT_UEXACT32 arg_len, req_len, status;
	dma_addr_t dma_addr;

	arg_len = 8;
	pkern_arg = se_pci_alloc_consistent(phba->pcidev, arg_len, &dma_addr);
	if (!pkern_arg) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"pci_alloc_consistent failed first time\n");
		return -ENOMEM;
	}

	status = copy_from_user((void *)pkern_arg, (void *)arg, arg_len);
	if (status) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"copy_from_user failed\n");
		up(&phba->ioctl_lock);
		se_pci_free_consistent(phba->pcidev, arg_len, pkern_arg, dma_addr);
		return -EFAULT;
	}

	// Save the new request length, free the old mapping and get a new one
	req_len = ((PDRVR_STATS_REQUEST)pkern_arg)->buffer_len + 
		sizeof(DRVR_STATS_REQUEST);
	se_pci_free_consistent(phba->pcidev, arg_len, pkern_arg, dma_addr);
	arg_len = req_len;
	pkern_arg = se_pci_alloc_consistent(phba->pcidev, arg_len, &dma_addr);

	status = copy_from_user((void *)pkern_arg, (void *)arg, arg_len);
	if (status) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"copy_from_user failed\n");
		se_pci_free_consistent(phba->pcidev, arg_len, pkern_arg, dma_addr);
		return -EFAULT;
	}

	status = acit_process_stat_ioctl(&phba->controller_key, pkern_arg);

	if(status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"acit failed stat ioctl\n");
		se_pci_free_consistent(phba->pcidev, arg_len, pkern_arg, dma_addr);
		return -EFAULT;
	}

	status = copy_to_user((void *)arg, (void *)pkern_arg, arg_len);
	if (status) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"copy to user failed for stat IOCTL\n");
		se_pci_free_consistent(phba->pcidev, arg_len, pkern_arg, dma_addr);
		return -EFAULT;
	}

	se_pci_free_consistent(phba->pcidev, arg_len, pkern_arg, dma_addr);
	return 0;
}


ACIT_UEXACT32
se_process_osm_ioctl(
		se_hba        *phba,
		ACIT_UEXACT32   opcode,
		ACIT_UEXACT32   subsystem,
		PACIT_VOID      pbuffer)
{
	ACIT_STATUS	status = ACIT_STATUS_SUCCESS;
	ACIT_UEXACT8	cdb[ACIT_SCSI_MAX_CDB_LENGTH] = {0};
	ACIT_STATUS	acit_status = 0;
	ACIT_UEXACT32	session_handle;
	ACIT_UEXACT8	lun[8] = {0};

	OSM_LOG_IOCTL(phba, KERN_INFO,"opcode=0x%x\n",opcode);

	if (subsystem != SUBSYSTEM_ISCSI_INI_SCSI_PASS_THRU) {
		status = ACIT_STATUS_INVALID_PARAMETER;
		return status;
	}

	switch(opcode) {
	case OPCODE_ISCSI_INI_CFG_GET_HBA_COUNT:{
			PGET_HBA_COUNT_PARAMS_OUT params;
			params = (PGET_HBA_COUNT_PARAMS_OUT)((char *)pbuffer + sizeof(IOCTL_HEADER)); 
			params->hba_count = hba_count;
		}
		((PIOCTL_RESPONSE_HEADER)pbuffer)->status = acit_status;
		break;

	case OPCODE_ISCSI_INI_BTL_MAPPING_MAX_OS_SUPPORTED_BTL:{
			PQUERY_MAX_OS_SUPPORTED_BTL_PARAMS_OUT pmaxbtl;
			pmaxbtl = (PQUERY_MAX_OS_SUPPORTED_BTL_PARAMS_OUT)((char *)pbuffer + sizeof(IOCTL_HEADER));
			put_user(255, &pmaxbtl->max_oslun_supported);
			put_user(15, &pmaxbtl->max_ostarget_supported);
			put_user(7, &pmaxbtl->max_osbus_supported);
		}
		((PIOCTL_RESPONSE_HEADER)pbuffer)->status = acit_status;
		break;		       
	     
	case OPCODE_ISCSI_INI_SCSI_PASS_THRU:{
			PSCSI_PASS_THRU_PARAMS prlun;
			PSCSI_PASS_THRU_PARAMS prlun_out;

			prlun = (PSCSI_PASS_THRU_PARAMS)((char *)pbuffer + sizeof(IOCTL_HEADER)); 
			prlun_out = (PSCSI_PASS_THRU_PARAMS)((char *)pbuffer + sizeof(IOCTL_HEADER)); 
			session_handle = prlun->request.session_handle;
			memcpy(cdb,prlun->request.cdb,ACIT_SCSI_MAX_CDB_LENGTH);	
			memcpy(lun,prlun->request.lun,MGMT_LUN_SIZE);	
			acit_status = se_do_io_ioctl(phba, cdb, lun, 
					session_handle,
					(PIOCTL_HEADER)pbuffer,
					prlun_out->response.data_buffer,
					MAX_LUN * 8 + 8, 
					prlun->request.data_transfer_size);
			if (acit_status == MGMT_STATUS_INVALID_SESSION)
				status = MGMT_STATUS_FAILED;
			((PIOCTL_RESPONSE_HEADER)pbuffer)->status = acit_status;
		     } 
		     break;          
	default:
			status = ACIT_STATUS_INVALID_PARAMETER;
			break;
			}
	return status; 
}

/*++
Function:
se_ioctl

Parameter:
inode:  [in]    pointer to inode
file:   [in]    pointer to device file
cmd:    [in]    IOCTL code
arg:    [in]    pointer to context data

Return:
	-ENXIO - bad device
	0 success

Context:
ioctl entry point

Description:
This routine is the ioctl entry point for the driver.

--*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
long
se_ioctl(
		struct file *file, 
		ACIT_UEXACT32 cmd, 
		ACIT_U32_64 arg)
#else
ACIT_SEXACT32 
se_ioctl(
		struct inode *inode, 
		struct file *file, 
		ACIT_UEXACT32 cmd, 
		ACIT_U32_64 arg)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
	ACIT_SEXACT32   	major;
	ACIT_SEXACT32   	minor;
	struct list_head        *lh;
#endif
	se_hba        		*phba= NULL;
	void 			*pkern_arg;
	unsigned int		arg_len;
	unsigned int		req_len;
	ACIT_STATUS		status;
	ACIT_IOCTL_REQUEST	*pacit_ioctl_req;
	struct completion       *pcompletion;
	PACIT_IOCTL_REQUEST_POST ppost_routine = NULL;
	int			phba_found = 0;
	dma_addr_t		dma_addr, dma_addr1;
	ACIT_U32_64     	flags = 0;

	PSET_SESSION_CONFIGURATION_PARAMS pset_session_conf;
	int eto_value = 0;
	int session_handle = 0;
	se_target *ptarget = NULL;
	se_device *pdevice = NULL;
	struct scsi_device      *psdev;


	BEGIN;        

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
	down(&se_configuration_lock);
	phba = file->private_data;
	up(&se_configuration_lock);
	if (phba == NULL) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"Host not found with private_data\n");
		return -ENXIO;
	}
	phba_found = 1;
#else
	major = MAJOR(inode->i_rdev);
	minor = MINOR(inode->i_rdev);

	down(&se_configuration_lock);

	list_for_each(lh, &hba_chain) {
		phba = list_entry(lh, se_hba, hba_queue);

#ifdef __VMKERNEL_MODULE__
		if (phba->major == major) {
#else
			if (phba->hba_unique_id == minor) {
#endif
				phba_found = 1;
				break;        /* found adapter */
			} 
		}

		up(&se_configuration_lock);

		if (!phba_found) {
			OSM_LOG_IOCTL(phba, KERN_ERR,"Host not found with major %d\n", major);
			return -ENXIO;
		}
#endif  //version 3,0,0

		/* 
		 * This will put the process to sleep if some other 
		 * ioctl is in progress.
		 */
		down(&phba->ioctl_lock);
#ifdef __VMKLNX__
		if(cmd == IOCTL_CMD_VMHBA_NAME){
			arg_len = 32;
		pkern_arg = se_pci_alloc_consistent(phba->pcidev, arg_len, &dma_addr);
		if (!pkern_arg) {
			up(&phba->ioctl_lock);
			return -ENOMEM;
		}

		status = copy_from_user((void *)pkern_arg, (void *)arg, arg_len);
		if (status) {
			OSM_LOG_IOCTL(phba, KERN_ERR,"copy_from_user failed\n");
			up(&phba->ioctl_lock);
			goto no_copy;
		}
		sprintf(pkern_arg,"%s",vmklnx_get_vmhba_name(phba->host));

		if (copy_to_user((void *)arg, (void *)pkern_arg, arg_len))
			OSM_LOG_IOCTL(phba, KERN_ERR,"copy to user failed for IOCTL_CMD_VMHBA_NAME\n");

		se_pci_free_consistent(phba->pcidev, arg_len,pkern_arg,dma_addr);

		up(&phba->ioctl_lock);
		return 0;
	}
#endif

	if(cmd == ES_MGMT_STATS_COMMAND) {

		//
		// refer to iscsi_drvr_stats.h. Note that the OSM blocks defined in
		// the header are not implemented yet. Just forward the IOCTL to
		// acit
		// 
		ACIT_SEXACT32 stat_rc;
		stat_rc = se_stats_ioctl(phba, (void *)arg);
		up(&phba->ioctl_lock);
		return stat_rc;
	} 

	arg_len = 16;
	pkern_arg = se_pci_alloc_consistent(phba->pcidev, arg_len, &dma_addr);

	if (!pkern_arg) {
		up(&phba->ioctl_lock);
		return -ENOMEM;
	}

	status = copy_from_user((void *)pkern_arg, (void *)arg, arg_len);

	if (status) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"copy_from_user failed\n");
		up(&phba->ioctl_lock);
		goto no_copy;  
	}


	OSM_LOG_IOCTL(phba, KERN_INFO,"opcode = %d, request_length = %d\n", 
			((PIOCTL_REQUEST_HEADER)pkern_arg)->opcode, 
			((PIOCTL_REQUEST_HEADER)pkern_arg)->request_length);
	req_len = ((PIOCTL_REQUEST_HEADER)pkern_arg)->request_length;
	
	se_pci_free_consistent(phba->pcidev, arg_len, 
			pkern_arg,
			dma_addr);
	arg_len = sizeof(IOCTL_HEADER) + req_len;
	/* This is based on the assumption that ioctls are not bigger than 64k */	
	ASSERT(arg_len < 128000);

 	pkern_arg = se_pci_alloc_consistent(phba->pcidev, 
			arg_len,
			&dma_addr);
	if (!pkern_arg) {
		up(&phba->ioctl_lock);
		return -ENOMEM;
	}

	status = copy_from_user((void *)pkern_arg, (void *)arg, arg_len);

	if (status) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"copy_from_user failed\n");
		up(&phba->ioctl_lock);
		goto no_copy;  
	}


	/* check if this ioctl needs to be passed down to acit */
	status = se_process_osm_ioctl(phba,
			((PIOCTL_REQUEST_HEADER)pkern_arg)->opcode, 
			((PIOCTL_REQUEST_HEADER)pkern_arg)->subsystem, 
			(PACIT_VOID)pkern_arg); 

	if (status == ACIT_STATUS_SUCCESS){
		if (copy_to_user((void *)arg, (void *)pkern_arg, arg_len))
			OSM_LOG_IOCTL(phba, KERN_ERR,"copy to user failed \n");

		se_pci_free_consistent(phba->pcidev, arg_len, 
				pkern_arg,
				dma_addr);

		up(&phba->ioctl_lock);
		return 0;
	}else if (status == MGMT_STATUS_FAILED){

		se_pci_free_consistent(phba->pcidev, arg_len, 
				pkern_arg,
				dma_addr);
		up(&phba->ioctl_lock);
		return status;
	}
	/* TODO: Handle else condition if needed */                



	pacit_ioctl_req = se_pci_alloc_consistent(phba->pcidev, sizeof(*pacit_ioctl_req), &dma_addr1);
	if (!pacit_ioctl_req) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"kmalloc failed\n");

		up(&phba->ioctl_lock);
		goto no_copy;
	}
	if(((PIOCTL_REQUEST_HEADER)pkern_arg)->opcode == OPCODE_ISCSI_INI_SESSION_SET_CONFIGURATION){

		pset_session_conf = (PSET_SESSION_CONFIGURATION_PARAMS)((char *)pkern_arg + sizeof(IOCTL_HEADER));
		eto_value = pset_session_conf->request.extended_timeout;
		session_handle = pset_session_conf->request.session_handle;
		ptarget = se_find_target(phba, session_handle);

		if(ptarget !=NULL){
			ptarget->extended_timeout = eto_value;
			list_for_each_entry(pdevice, &ptarget->lun_list, sibling_list){
				if(pdevice != NULL){
					psdev = pdevice->pScsi_dev;
					if(psdev != NULL){
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
							blk_queue_rq_timeout(psdev->request_queue, (eto_value + 30) * HZ);
#else
							psdev->timeout = (eto_value + 30) * HZ;
#endif
						}
					}

				}
			}
	}

	if (((PIOCTL_REQUEST_HEADER)pkern_arg)->opcode == OPCODE_ISCSI_INI_SESSION_LOGIN_TARGET &&
			(((PIOCTL_ISCSI_INI_SESSION_LOGIN_TARGET)pkern_arg)->params.request.session_flags & MGMT_SESSION_FLAG_DATA_SESSION)){
		spin_lock_irqsave(&phba->acitapi_lock,flags);
		phba->state |= SE_STATE_LOGIN_PENDING;
		spin_unlock_irqrestore(&phba->acitapi_lock,flags);
		ppost_routine = se_login_completion;
		phba->login_status = ACIT_STATUS_SUCCESS;
		pcompletion = &phba->login_completion;
	} else {
		pcompletion = &phba->ioctl_completion;
	}
	init_completion(pcompletion);

	/* Map the request to ACIT_IOCTL_REQUEST */ 
	status = se_build_ioctl_req(phba, 
			(PIOCTL_HEADER)pkern_arg,
			pacit_ioctl_req,
			(PACIT_VOID)
			((char *)pkern_arg + sizeof(IOCTL_HEADER)),
			ppost_routine);

	if (status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"se_build_ioctl_req failed\n");
		up(&phba->ioctl_lock);
		goto no_req;	
	}

	/* Call into ACIT */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
			pacit_ioctl_req); 

	if (status == SHTDWN_EVENT) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"acitapi_start_ioctl failed\n");
		up(&phba->ioctl_lock);
		goto no_req;
	}
	/* Wait for completion. Completion is done in osm_complete_ioctl */
	wait_for_completion(pcompletion);

	/* fill ioctl requests with osm related info */
	se_complete_ioctl(pacit_ioctl_req);
	
	if (phba->login_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"ioctl failed\n");
		phba->login_status = ACIT_STATUS_SUCCESS;
		up(&phba->ioctl_lock);
		goto no_req;
	}

	memcpy(pkern_arg, &pacit_ioctl_req->ioctl_hdr, sizeof(IOCTL_HEADER));
	/* Copy the data buffer to user space */
	if (copy_to_user((void *)arg, (void *)pkern_arg, arg_len))
		OSM_LOG_IOCTL(phba, KERN_ERR,"copy to user failed \n");
	se_pci_free_consistent(phba->pcidev, arg_len, 
			pkern_arg,
                                dma_addr);
	se_pci_free_consistent(phba->pcidev, sizeof(*pacit_ioctl_req), 
                                pacit_ioctl_req,
                                dma_addr1);

	up(&phba->ioctl_lock);
	END;
	return 0;
no_req:
	se_pci_free_consistent(phba->pcidev, sizeof(*pacit_ioctl_req), 
                                pacit_ioctl_req,
                                dma_addr1);
no_copy:
	se_pci_free_consistent(phba->pcidev, arg_len, 
                                pkern_arg,
                                dma_addr);
	END;
	return 0;
	}

	/*++
Function:
        se_open

Parameter:
        inode:  [in]    pointer to inode
        file:   [in]    pointer to device file

Return:
        -ENXIO - bad device
        0 success

Context:
        open entry point

Description:
        This routine is the open entry point called when an application
        opens the be2iscsi device node using the open system call.

--*/

ACIT_SEXACT32 
se_open(
	struct inode *inode, 
	struct file *file)
{
	se_hba        *phba = NULL;
	ACIT_SEXACT32   major;
	ACIT_SEXACT32   minor;
	struct list_head        *lh;
	int phba_found = 0;
	major = MAJOR(inode->i_rdev);
	minor = MINOR(inode->i_rdev);

	down(&se_configuration_lock);
	list_for_each(lh, &hba_chain){ 
		phba = list_entry(lh, se_hba, hba_queue);
#ifdef __VMKERNEL_MODULE__
		if (phba->major == major) {
#else
			if (phba->hba_unique_id == minor) {
#endif
				phba_found = 1;
				break;        /* found adapter */
			}
	}

	if (!phba_found) 
	{
		up(&se_configuration_lock);
		return -ENXIO;
	}

	phba->in_use++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
	file->private_data = phba;
#endif

	/*
	 * Make a call to MOD_INC_USE_COUNT so that the module system 
	 * knows not to unload it while the device is open.
	 */   
	if (phba->host->hostt->module) {
		if (!try_module_get(phba->host->hostt->module)) {
			up(&se_configuration_lock);
			return -EBUSY;
		}
	}	
	up(&se_configuration_lock);
	return 0;
}


/*++
Function:
        se_close

Parameter:
        inode:  [in]    pointer to inode
        file:   [in]    pointer to device file

Return:
        -ENXIO - bad device
        0 success

Context:
        close entry point

Description:
        This routine is the close entry point called when an application
        closes the be2iscsi device node using the close system call.

--*/
ACIT_SEXACT32 
se_close(struct inode *inode, struct file *file)
{
	se_hba        *phba = NULL;
	ACIT_SEXACT32   major;
	ACIT_SEXACT32   minor;
	struct list_head        *lh;

	int phba_found = 0; 
	major = MAJOR(inode->i_rdev);
	minor = MINOR(inode->i_rdev);

	down(&se_configuration_lock);
	list_for_each(lh, &hba_chain){ 
		phba = list_entry(lh, se_hba, hba_queue);
#ifdef __VMKERNEL_MODULE__
		if (phba->major == major) {
#else
			if (phba->hba_unique_id == minor) {
#endif
				phba_found =1;
				break;        /* found adapter */
			}
	}
	up(&se_configuration_lock);

	if (!phba_found) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"Host not found: 0x%lx\n",
				(unsigned long)major);
		return -ENXIO;
	}

	phba->in_use--;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
	file->private_data = NULL;
#endif

	/*
	 * call MOD_DEC_USE_COUNT to indicate that the module system 
	 * can now be unloaded
	 */    
	if (phba->host->hostt->module) {
		module_put(phba->host->hostt->module);
	}

	return 0;
}

struct file_operations se_fops = 
{
	.owner   = THIS_MODULE,
	.open    = se_open,
	.release = se_close,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
	.unlocked_ioctl   = se_ioctl
#else
	.ioctl   = se_ioctl
#endif
};

PACIT_IOCTL_REQUEST 
se_build_internal_ioctl(se_hba 	*phba,
			ACIT_UEXACT32	opcode,
			ACIT_UEXACT32	data_length,
			ACIT_UEXACT32	timeout,
			ACIT_UEXACT8	domain,
			ACIT_UEXACT8	type
			)
{
	PACIT_IOCTL_REQUEST pacit_ioctl_req;
	PIOCTL_HEADER pioctl_header;
#ifdef VIRT_TO_BUS_NOTSUPPORTED
	pacit_ioctl_req = phba->internal_ioctl;
#else
	pacit_ioctl_req = &phba->internal_ioctl;
#endif
	pioctl_header = &pacit_ioctl_req->ioctl_hdr;	

	if (!pioctl_header) {
		OSM_LOG_IOCTL(phba, KERN_ERR,"No pioctl_header Returning NULL \n");
		return NULL;
	}	

	memset(pacit_ioctl_req, 0, sizeof(*pacit_ioctl_req));

	/*
	 * Fill in the IOCTL header
	 */
	pioctl_header->request.opcode = opcode;

	if (1 == type) {
	    pioctl_header->request.subsystem = IOCTL_SUBSYSTEM_COMMON;
	} else if (2 == type) {
	    pioctl_header->request.subsystem = IOCTL_SUBSYSTEM_COMMON_ISCSI;
	} else {
	    pioctl_header->request.subsystem = IOCTL_SUBSYSTEM_ISCSI_INI;
	}

	pioctl_header->request.port_number = 0;
	pioctl_header->request.domain = domain;
	pioctl_header->request.request_length = data_length;
	pioctl_header->request.timeout = timeout;

	pacit_ioctl_req->posm_reference = phba;
	pacit_ioctl_req->ppost_routine = NULL;
	pacit_ioctl_req->data_length = data_length;

	/*
	 * we dont support internal ioctls with more 128k size
	 */
	ASSERT(data_length < 128000);

	phba->ioctl_buf.virtual_address = se_pci_alloc_consistent(
		phba->pcidev,
		data_length,
		(dma_addr_t *)&phba->ioctl_buf.bus_address.u.a64.address);

	if (!phba->ioctl_buf.virtual_address)
		return NULL;

	phba->ioctl_buf.size_in_bytes = data_length;
	pacit_ioctl_req->pdata_buffer = phba->ioctl_buf.virtual_address +
				sizeof(IOCTL_HEADER);

	return pacit_ioctl_req;
}
