/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   beclib_common.h

@brief
   
CVS Header Information:
   $Header: /proj/be/archive/dsn/sw/host/drivers/beclib/beclib_common.h,v 1.18 2006/04/20 23:49:03 menz Exp $
*/
#ifndef __beclib_common_h__
#define __beclib_common_h__

#include "sa.h"

// Uncomment the following define to enable a test function to configure each PCI function
// and tear it down.  This tests the BECLIB init and cleanup code without actually using 
// any of the rings (except MCC).
// #define BE_ENABLE_TEST

// Defined for high level build only
// #define _BE_HIGH_LEVEL_BUILD

// This enables the download of chaos firmware.  In the high level API the chaos
// will be configured automatically.  In the low level API, the driver must #include
// the ulpchaos_ll.h file and call the configuration function with the appropriate
// allocated memory.  
//#define BE_ENABLE_ULP_CHAOS

// This enables the compilation of the emulation code, but does not
// emulate anything by default.  The developer can add specific opcodes
// to emulate, to the be_mccemul_is_emulated function.  The ARM must be present
// during selective emulation.  This should only be used for development!!!
//#define BE_SELECTIVE_EMULATION

// Uncomment the following define to emulate ARM/MPU.
#if defined(SA_MICROSOFT) && defined(BE_EMULATION)
    #pragma message("Building beclib emulated version.")
#endif

// Enable each srcgen struct to be an array of dwords
#define SG_BMAP_UNION

#ifndef BECLIB_BE3_SUPPORT
#define BECLIB_BE3_SUPPORT
#endif

// Pack srcgen structs
#if defined (SA_MICROSOFT) 
    #ifndef SG_PRAGMA_PACK
        #define SG_PRAGMA_PACK
    #endif
#else
    #ifndef SG_ATTRIBUTE_PACK
        #define SG_ATTRIBUTE_PACK
    #endif
#endif

// Define the static inline srcgen qualifier for inline functions.
#define SG_STATIC_INLINE    STATIC INLINE

#if !defined(BE_CONFIG)
    //Uncomment for testing purposes: #error "BE_CONFIG not defined"
    // Default to configuration 0 (see fw/config/readme.txt for more info) 
    #define BE_CONFIG 0
#endif

#define BE_GEN_STATIC_INLINE STATIC INLINE
#define BE_GEN_ASSERT ASSERT

#if (defined(SA_NDIS) || defined(SA_WIN32U)) && !defined(BE_EMULATION)

    // BE2 API
    #include "be2nic.h"
    #include "be2nic_bmap.h"

    #ifdef BECLIB_BE3_SUPPORT
        // BE3 API
        #include "be3nic_v1.h"
        #include "be3nic_v1_bmap.h"
    #endif
    
    // SkyHawk API
    #include "sh_nic_v2.h"
    #include "sh_nic_v2_bmap.h"

#elif !defined(BE_EMULATION)

    #include "be2_config_sg.h"       //ID literals
    #include "be2_gen_id_ranges.h"   //ID conversion macros

    // Srcgen includes
    // srcgen array map output
    #include "all_csr.h"
    #include "host_struct.h"
    #include "hw_host_struct.h"
    #include "ioctl_top.h"
    #include "lro_tcp_context.h"
    
    // srcgen bitmap output
    #include "all_csr_bmap.h"
    #include "host_struct_bmap.h"
    #include "hw_host_struct_bmap.h"
    #include "ioctl_top_bmap.h"
    #include "lro_tcp_context_bmap.h"

    // BE3 API
    #ifdef BECLIB_BE3_SUPPORT
        #undef SG_C_ASSERT
        #include "be3nic_v1.h"
        #include "be3nic_v1_bmap.h"
    #endif
    
#endif


#if defined(BE_EMULATION) || defined(BE_SELECTIVE_EMULATION)
    #include "be2_config_sg.h"       //ID literals
    #include "be2_gen_id_ranges.h"   //ID conversion macros

    // Srcgen includes
    // srcgen array map output
    #include "all_csr.h"
    #include "host_struct.h"
    #include "hw_host_struct.h"
    #include "ioctl_top.h"
    #include "lro_tcp_context.h"
    
    // srcgen bitmap output
    #include "all_csr_bmap.h"
    #include "host_struct_bmap.h"
    #include "hw_host_struct_bmap.h"
    #include "ioctl_top_bmap.h"
    #include "lro_tcp_context_bmap.h"

    // BE3 API
    #ifdef BECLIB_BE3_SUPPORT
        #undef SG_C_ASSERT
        #include "be3nic_v1.h"
        #include "be3nic_v1_bmap.h"
    #endif

    #include "fw_impl.h"
    #include "fw_impl_bmap.h"
#endif  


#include "bestatus.h"

// Debug trace categories.
typedef enum _BE_DEBUG_LEVELS {
    DL_HW    = 0x00000100, 
    DL_IOCTL = 0x00000200, 
    
    BE_DL_MIN_VALUE = 0x100,  // this is the min value used
    BE_DL_MAX_VALUE = 0x800   // this is the higheset value used
} BE_DEBUG_LEVELS, *PBE_DEBUG_LEVELS;

SA_GLOBAL_C_ASSERT (beclib_debug_level_range, ((u32)BE_DL_MIN_VALUE > (u32)SA_DL_MAX_VALUE));

#include "beregister.h"          // Macros for register access using srcgen structs
#include "beclib_stats.h"

//  BECLIB status data type maps to the SA_STATUS
typedef SA_STATUS   BESTATUS, *PBESTATUS;


// BECLIB calling convention.  Use default currently.
#define BECALL 

// Forward declarations.
struct _BE_FUNCTION_OBJECT;
struct _BE_EQ_OBJECT;
struct _BE_CQ_OBJECT;

// Callback types.  Deprecate most of these.
typedef void (BECALL *MCC_WRB_CQE_CALLBACK)(PVOID context, BESTATUS status, MCC_WRB* optional_wrb);
typedef void (BECALL *MCC_ASYNC_EVENT_CALLBACK) (PVOID context, u32 event_code, PVOID event);

typedef BESTATUS (*EQ_CALLBACK)( struct _BE_FUNCTION_OBJECT*, struct _BE_EQ_OBJECT*, PVOID context);   
typedef BESTATUS (*CQ_CALLBACK)( struct _BE_FUNCTION_OBJECT*, struct _BE_CQ_OBJECT*, PVOID context);   

#endif // __beclib_common_h__




