/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   beclib.h

@brief

    High 

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/beclib_hl.h,v 1.25 2006/03/10 19:17:31 menz Exp $
*/
#ifndef __beclib_higher_h__
#define __beclib_higher_h__

#define BE_EXTERN_C
#define BE_FAR

#include "beclib_common.h"

SA_CPP_HEADER

//============================================================================
//                              T Y P E D E F S
//============================================================================
#if defined( BECLIB_BUILD )
    #define BE_DECLARE_HANDLE(hname,basename)  typedef struct basename * hname;
    #if defined(SA_MICROSOFT)
        #define BEAPI   __declspec(dllexport)    
    #else
        #define BEAPI   
    #endif
#else
    #define BE_DECLARE_HANDLE(hname,basename)  typedef struct _H##hname { int bogus; } * hname; 
    #if defined(SA_MICROSOFT)
        #define BEAPI   __declspec(dllimport)
    #else
        #define BEAPI
    #endif
#endif

BE_DECLARE_HANDLE( CHIPHANDLE,              _BE_CHIP_OBJECT );
BE_DECLARE_HANDLE( FUNCTIONHANDLE,          _BE_FUNCTION_OBJECT );             
BE_DECLARE_HANDLE( PDHANDLE,                _BE_PD_OBJECT );
BE_DECLARE_HANDLE( EQHANDLE,                _BE_EQ_OBJECT );
BE_DECLARE_HANDLE( CQHANDLE,                _BE_CQ_OBJECT );
BE_DECLARE_HANDLE( MCCHANDLE,               _BE_MCC_OBJECT );
BE_DECLARE_HANDLE( ETHSQHANDLE,             _BE_ETHSQ_OBJECT );
BE_DECLARE_HANDLE( ETHRQHANDLE,             _BE_ETHRQ_OBJECT );
BE_DECLARE_HANDLE( ISCSIWRBQ,               _BE_ISCSI_WRB_QUEUE_OBJECT );
BE_DECLARE_HANDLE( ISCSIDEFQ,               _BE_DEFAULT_PDU_QUEUE_OBJECT );
BE_DECLARE_HANDLE( ISCSI_CONNECTION_HANDLE, _BE_ISCSI_CONNECTION_OBJECT );
BE_DECLARE_HANDLE( TOE_DQ_HANDLE,           _BE_TOE_DQ_OBJECT );
BE_DECLARE_HANDLE( TOE_CONNECTION_HANDLE,   _BE_TOE_CONNECTION_OBJECT );
BE_DECLARE_HANDLE( IWARP_CONNECTION_HANDLE, _BE_IWARP_CONNECTION_OBJECT );
BE_DECLARE_HANDLE( FCOEWRBQ,                _BE_FCOE_WRB_QUEUE_OBJECT );
BE_DECLARE_HANDLE( FCOE_INTERFACE_HANDLE,   _BE_FCOE_INTERFACE_OBJECT );
BE_DECLARE_HANDLE( FCOEDEFQ,                _BE_DEFAULT_PDU_QUEUE_OBJECT );
BE_DECLARE_HANDLE( RXIFHANDLE,              _BE_RXIF_OBJECT );


// Callback types
typedef void (BECALL *KLIB_COMPLETION_ROUTINE)( FUNCTIONHANDLE function_object, PVOID context, BESTATUS status );

//============================================================================
//                              M A C R O S
//============================================================================

// IScsi connection flags
#define BE_TOECF_IPSEC_ENABLED              0x08    // On if this connections traffic should be passed through the IPSec module.

// QP/iWARP connection flags

typedef PVOID   HDEVICE_CONTEXT,  *PHDEVICE_CONTEXT;

// definitions
#if !defined( SOCKET ) && !defined( _WINSOCK2API_ )
    typedef struct socket* SOCKET;
    #define BE_HAVE_SOCKET
#endif

typedef BESTATUS (BECALL* BE_SOCKET_COMPLETION_CALLBACK)( SOCKET    Socket, 
                                                          BESTATUS  FinalStatus,
                                                          u32 BytesXferedOrNewSocket,
                                                          PVOID     Context );


typedef struct iovec*   PBE_IOVEC;

//============================================================================
//                      V B D   D I S P A T C H
//============================================================================

#define CHIPIF_DOWNCALL_VERSION    0x10000


//
// IScsi connection flags
//
#define BE_ISCF_HEADER_CRC_ENABLED          0x01    // On if header CRCs are enabled
#define BE_ISCF_DATA_CRC_ENABLED            0x02    // On if data CRCs are enabled.
#define BE_ISCF_SUPPORT_IMMEDIATE_DATA      0x04    // On if immediate data may be included in the SCSI Command PDU.
#define BE_ISCF_IPSEC_ENABLED               0x08    // On if this connections traffic should be passed through the IPSec module.
#define BE_ISCF_INITIAL_R2T_ENABLED         0x10    // On if this connection requires an initial R2T


//
// This structure is exchanged from the shared klib to the upper layer driver
// as part of the driver's init. If the ULD controls actual hardware then
// the ULD calls BeRegisterDevice via the 
// PCM_RESOURCE_LIST...DeviceSpecificData passed to his 'effective' Pnp
// StartRoutine. 
//
#include <sa_packon.h>

typedef struct _CHIPIF_DOWNCALL {
    //
    // Client Initializes these values prior to registration
    //

    u32                  Size;               // ! Size of this structure in bytes
    u32                  Version;            // ! Current version

    u32                  IsVm;               // ! TRUE if running on VM

    //
    // Klib Initializes all of the following values during registration
    //                                
                            
    FUNCTIONHANDLE          FunctionHandle;

    //
    // Firmware Utilities
    // 
    
    void (BECALL * BeEPReadMem) (
        IN  FUNCTIONHANDLE FunctionObject,
        IN  u32 processorID,
        IN  u32 offset,
        OUT u32 *result);

    void (BECALL * BeEPWriteMem) (
        IN  FUNCTIONHANDLE FunctionObject,
        IN  u32 processorID,
        IN  u32 offset,
        IN  u32 value);

    //
    // EQ Functions
    //
    BESTATUS (BECALL * BeEqCreate)(
            IN      FUNCTIONHANDLE      FunctionObject, 
            IN      PVOID               EqBaseVa,       // EQE buffer
            IN      u32              SizeEncoding,   // EQE Size Encoding
            IN      u32              NumEntries,     // EQ Count Encodings    
            IN      u32              Watermark,      // CEV_WMARK_* or -1
            IN      u32              TimerDelay,     // In 8us units, or -1
            OUT     EQHANDLE*           ppEqHandle );

    u32 (BECALL * BeEqGetId)(
            IN EQHANDLE EqHandle); 
        
    BESTATUS (BECALL * BeEqFree)(
            IN      EQHANDLE           EQHandle);

    void (BECALL * BeEqDelegateProcessing)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      u32              EqId );

    //void (BECALL * BeMccProcessCq) (
    //    IN MCCHANDLE FunctionObject,
    //    IN boolean rearm);

    EQ_CALLBACK (BECALL * BeEqSetCallback)( 
        IN      EQHANDLE        EqObject,
        IN      EQ_CALLBACK     Callback,
        IN      PVOID           Context );

    //
    // Mcc stuff
    //
        
            
    BESTATUS (BECALL * BeMccCreate)(
            IN      FUNCTIONHANDLE      FunctionObject, 
            IN      EQHANDLE            pEqH,
            OUT     MCCHANDLE *         ppMccHandle);

    BESTATUS (BECALL* BeMccPostWrb)(
            IN      FUNCTIONHANDLE          FuncObj,
            IN      MCC_WRB*           Wrb,
            IN      MCC_WRB_CQE_CALLBACK    CompletionCallback,
            IN      PVOID                   CompletionCallbackContext );

    BESTATUS (BECALL * BeMccFree)(
            IN      MCCHANDLE          pMCCHandle);

    //
    // Cq Stuff
    //

    BESTATUS (BECALL * BeCqCreate)(
        IN      FUNCTIONHANDLE      FunctionObject, 
        IN      PVOID               CqBaseVa, 
        IN      CQ_CNT_ENUM         NumEntries,     // count encodings
        IN      boolean             solEventEnable,
        IN      boolean             eventable,
        IN      boolean             nodelay,
        IN      boolean             wme,
        IN      u32              wmThresh,       // Watermark encodings
        IN      EQHANDLE            EQHandle,
        OUT     CQHANDLE *          CqHandle);

    u32 (BECALL * BeCqGetId)(
        IN CQHANDLE cqHandle); 

    BESTATUS (BECALL * BeCqFree)(
        IN      CQHANDLE           pCQHandle);  

    void (BECALL * BeCqDelegateProcessing)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      u32              CqId );

    CQ_CALLBACK (BECALL * BeCqSetCallback)( 
        IN      CQHANDLE        CqObject,
        IN      CQ_CALLBACK     Callback,
        IN      PVOID           Context );

    BESTATUS (BECALL* BeCqUnmapRing) (
        IN      CQHANDLE        CqHandle);

    BESTATUS (BECALL* BeCqRemapRing) (    
        IN      CQHANDLE      CqHandle,
        IN      CQ_CNT_ENUM   num_entries,     // count encodings
        IN      PVOID         cq_base_va);


    ///
    /// iScsi stuff
    //
        
    BESTATUS (BECALL * BeIscsiDqCreate)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      PVOID                   SendQueueVa,
        IN      u32                     RingSizePowerOfTwo,
        IN      u32                     rsvd,
        IN      CQHANDLE                CqHandle,
        OUT     ISCSIDEFQ*              ppIscsiDefQ);
        
    BESTATUS (BECALL * BeIscsiDqFree)(
        IN      ISCSIDEFQ               IScsiDefQ);

    u32 (BECALL * BeIscsiDqGetId)(
        IN      ISCSIDEFQ               IScsiDefQ);
    


    BESTATUS (BECALL * BeIscsiMapFrags) (
        IN      FUNCTIONHANDLE       FunctionObject,
        IN      PVOID                va,
        IN      u32                  FragPageCount);

    BESTATUS (BECALL * BeIscsiUnmapFrags) (
        IN      FUNCTIONHANDLE       FunctionObject,
        IN      u32                  FragPageCount);



    BESTATUS (BECALL * BeIscsiCreateWrbQueue) (
        IN  FUNCTIONHANDLE           FunctionObject,
        IN  PVOID                    WrbRingVa,
        IN  u32                      WrbRingBytes,
        OUT ISCSIWRBQ                *wrbq);

    BESTATUS (BECALL * BeIscsiFreeWrbQueue) (
        IN ISCSIWRBQ                 wrbq);      
    


    BESTATUS (BECALL * BeIscsiOffloadInitiateFromTcpStack)(        
        IN  ISCSIWRBQ                wrbq,
        IN  ISCSIDEFQ                defq,
        IN  CQHANDLE                 cq,
        IN  u32                      MaxSegmentLength,
        IN  u32                      FirstBurstLength,
        IN  u32                      MaxBurstLength,
        IN  u32                      IScsiFlags,
        IN  u32                      ExpStatSn,
        //IN  u64                      PduBufferPa,
        IN  SOCKET                   Socket,
        OUT ISCSI_CONNECTION_HANDLE* cxn,
        IN  MCC_WRB_CQE_CALLBACK     Callback,
        IN  PVOID                    CallbackContext);

    BESTATUS (BECALL * BeIscsiOffloadTerminate) (
        IN ISCSI_CONNECTION_HANDLE ConnectionObject,
        IN MCC_WRB_CQE_CALLBACK    Callback,
        IN PVOID                   CallbackContext );

    u32 (BECALL * BeIscsiConnectionGetCid)(
        IN      ISCSI_CONNECTION_HANDLE cxn );

    BESTATUS (BECALL * BeIscsiConnectionResourcesCreate)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      PVOID                   WrbRingVa,
        IN      u32                     WrbRingSizePowerOfTwo,
        IN      PVOID                   FragVa,
        IN      u32                     FragPageCount,
        IN      u32                     MaxSegmentLength,
        IN      u32                     FirstBurstLength,
        IN      u32                     MaxBurstLength,
        IN      u32                     IScsiFlags,
        IN      u32                     exp_stat_sn,
        IN      ISCSIDEFQ               ScsiDefaultQueue,
        IN      CQHANDLE                CqHandle,
        OUT     ISCSI_CONNECTION_HANDLE* ScsiHandle );

     //
    /// FCoE Stuff
    //
    BESTATUS (BECALL * BeFCoECreateWrbQueue) (
        IN  FUNCTIONHANDLE           FunctionObject,
        IN  PVOID                    WrbRingVa,
        IN  u32                      WrbRingBytes,
        OUT FCOEWRBQ                 *wrbq);

    BESTATUS (BECALL * BeFCoEDqCreate)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      PVOID                   SendQueueVa,
        IN      u32                     RingSizePowerOfTwo,
        IN      u32                     rsvd,
        IN      CQHANDLE                CqHandle,
        OUT     FCOEDEFQ*              	ppIscsiDefQ);
        
    BESTATUS (BECALL * BeFCoEDqFree)(
        IN      FCOEDEFQ               FCoEDefQ);

    u32 (BECALL * BeFCoEDqGetId)(
        IN      FCOEDEFQ               FCoEDefQ);

    BESTATUS (BECALL * BeFCoEFreeWrbQueue) (
        IN FCOEWRBQ                 wrbq);

    BESTATUS (BECALL * BeFCoEMapFrags) (
        IN      FUNCTIONHANDLE       FunctionObject,
        IN      PVOID                va,
        IN      u32                  FragPageCount);

    BESTATUS (BECALL * BeFCoEUnmapFrags) (
        IN      FUNCTIONHANDLE       FunctionObject,
        IN      u32                  FragPageCount);


    BESTATUS (BECALL * BeFCoEInterfaceOffload)(
        IN  FCOEWRBQ                wrbq,
        IN  CQHANDLE                 cq,
        IN  FCOE_INTERFACE_HANDLE    intf,
        IN  MCC_WRB_CQE_CALLBACK     Callback,
        IN  PVOID                    CallbackContext);

    BESTATUS (BECALL * BeFCoEOffloadTerminate) (
        IN FCOE_INTERFACE_HANDLE   InterfaceObject,
        IN MCC_WRB_CQE_CALLBACK    Callback,
        IN PVOID                   CallbackContext );

    BESTATUS (BECALL * BeFCoEInterfaceResourcesCreate) (
        IN FUNCTIONHANDLE  func_obj,
        IN PVOID           wrbq,
        IN u32             wrb_ring_size_power,
        IN PVOID           FragVa,
        IN u32             FragPageCount,
        IN u32             HdrAddrHi,
        IN u32             HdrAddrLo,
        IN CQHANDLE        CqObj,
        OUT FCOE_INTERFACE_HANDLE *ppIntfObject);

    u32 (BECALL * BeFCoEGetInterfaceCid)(
        IN      FCOE_INTERFACE_HANDLE intf);


 
    //
    /// Ethernet stuff
    //

    BESTATUS (BECALL * BeEthSqCreate)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      PVOID               SendQueueVa,
        IN      u32              RingSizePowerOfTwo,
        IN      CQHANDLE            CqHandle,
        OUT     ETHSQHANDLE*        ppEthSQ );

    u32( BECALL * BeEthSqGetRid)(
       IN       ETHSQHANDLE         EthSq );                              
    
    BESTATUS (BECALL * BeEthSqFree)(
        IN      ETHSQHANDLE         EthSQ);

    BESTATUS( BECALL * BeEthRqCreate)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      PVOID               RqBaseVa,
        IN      CQHANDLE            CqObject,
        IN      u32                 RssId,
        OUT     ETHRQHANDLE*        ppEthRq );

    u32( BECALL * BeEthRqGetRid)(
       IN       ETHRQHANDLE         EthRq );

    BESTATUS( BECALL * BeEthRqFree)(
        IN      ETHRQHANDLE         EthRq );

    BESTATUS( BECALL * BeErxFragSizeGet)(
        IN      FUNCTIONHANDLE      FunctionObject,
        OUT     u32*                FragSizeBytes,
        IN      RXIFHANDLE          rxif_object );

#if defined(BE_EMULATION)
    BESTATUS( BECALL * BeErxFragSizeSet)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      ERX_FRAG_SIZE_ENUM  FragSizeEncoded,
        OUT     u32*                FragSizeBytes,
        IN      RXIFHANDLE          rxif_object );
#endif /* defined(BE_EMULATION) */


    //
    // TOE stuff
    //

    BESTATUS (BECALL * BeToeDqCreate)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      PVOID               DefaultBufferQueueVa,
        IN      u32              RingSizePowerOfTwo,
        OUT     TOE_DQ_HANDLE*      ppToeDefaultQueue );

    BESTATUS (BECALL * BeToeDqFree )(
        OUT     TOE_DQ_HANDLE       ToeDefaultQueue );

    u32 (BECALL * BeToeDqGetId )(
        OUT     TOE_DQ_HANDLE       ToeDefaultQueue );

    BESTATUS (BECALL * BeToeConnectionResourcesCreate)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      PVOID                   SendRingVa,
        IN      u32                  SendRingSizePowerOfTwo,
        IN      PVOID                   RecvRingVa,
        IN      u32                  RecvRingSizePowerOfTwo,
        IN      u32                  Flags,
        IN      u32                  SaIndex,
        IN      TOE_DQ_HANDLE           ToeDefaultQueue,
        IN      CQHANDLE                SendCq,
        IN      CQHANDLE                RecvCq,
        OUT     TOE_CONNECTION_HANDLE*  ToeHandle );

    u32 (BECALL * BeToeConnectionGetCid)(
        IN      TOE_CONNECTION_HANDLE   ToeHandle );

    BESTATUS (BECALL * BeToeOffloadInitiate)(
        IN      TOE_CONNECTION_HANDLE       ConnectionObject,
        IN      PTEMPLATE_HEADER       TemplateHeader,
        IN      PTCP_CONTEXT           TcpState,
        IN      MCC_WRB_CQE_CALLBACK        CompletionCallback,
        IN      PVOID                       CompletionCallbackContext );
    
    BESTATUS (BECALL * BeToeOffloadUpdate)(
        IN      TOE_CONNECTION_HANDLE       ConnectionObject,
        IN      PTEMPLATE_HEADER       TemplateHeader,
        IN      PTCP_CONTEXT           TcpState,
        IN      MCC_WRB_CQE_CALLBACK        CompletionCallback,
        IN      PVOID                       CompletionCallbackContext );

    BESTATUS (BECALL * BeToeOffloadUpload)(    
        IN  TOE_CONNECTION_HANDLE        ConnectionObject,
        IN  u32                          Flags,
        IN  u32                          RstSeq,
        OUT PTCP_CONTEXT                 TcpState,
        IN  MCC_WRB_CQE_CALLBACK         CompletionCallback,
        IN  PVOID                        CompletionCallbackContext,
        IN  PVOID                        QueueContext);


    BESTATUS (BECALL * BeToeOffloadQuery)(
        IN      TOE_CONNECTION_HANDLE       ConnectionObject,
        IN OUT  PTEMPLATE_HEADER       TemplateHeader,
        IN OUT  PTCP_CONTEXT           TcpState,
        IN      MCC_WRB_CQE_CALLBACK        CompletionCallback,
        IN      PVOID                       CompletionCallbackContext );
    
    BESTATUS (BECALL * BeToeOffloadEnable)(
        IN      TOE_CONNECTION_HANDLE       ConnectionObject,
        IN      boolean                     Enable,
        IN      MCC_WRB_CQE_CALLBACK        CompletionCallback,
        IN      PVOID                       CompletionCallbackContext );

    BESTATUS (BECALL * BeToeConnectionResourcesFree)(
        IN      TOE_CONNECTION_HANDLE       ToeHandle );  

    BESTATUS (BECALL * BeToeConnectionFlush) (
        IN      TOE_CONNECTION_HANDLE       ToeHandle );

    BESTATUS (BECALL * BeToeConnectionSendReset) (
        IN      TOE_CONNECTION_HANDLE       ToeHandle );

    //
    // iWARP stuff
    //

    BESTATUS (BECALL * BeIWarpResourcesCreate)(
        IN      FUNCTIONHANDLE              FunctionObject,
        IN      PDHANDLE                    PdObject,
        IN      PVOID                       SendRingVa,
        IN      u32                      SendRingSizePowerOfTwo,
        IN      PVOID                       RecvRingVa,
        IN      u32                      RecvRingSizePowerOfTwo,
        IN      u32                      Flags,
        IN      u32                      OrrjMax,
        IN      u32                      IrrjMax,
        IN      CQHANDLE                    SendCq,
        IN      CQHANDLE                    RecvCq,
        OUT     IWARP_CONNECTION_HANDLE*    IWarpHandle );

    u32 (BECALL * BeIWarpGetCid)(
        IN      IWARP_CONNECTION_HANDLE IWarpHandle );

    void    (BECALL * BeIWarpSetSocket)(
        IN      IWARP_CONNECTION_HANDLE     ConnectionObject,
        IN      SOCKET                      Socket );
    
    BESTATUS (BECALL * BeIWarpOffloadInitiate)(
        IN      IWARP_CONNECTION_HANDLE     ConnectionObject,
        IN      KLIB_COMPLETION_ROUTINE     CompletionCallback,
        IN      PVOID                       CompletionCallbackContext );
    
    BESTATUS (BECALL * BeIWarpOffloadTerminate)(
        IN      IWARP_CONNECTION_HANDLE     ConnectionObject,
        IN      KLIB_COMPLETION_ROUTINE     CompletionCallback,
        IN      PVOID                       CompletionCallbackContext );

    BESTATUS (BECALL * BeIWarpResourcesFree)(
        IN      IWARP_CONNECTION_HANDLE IWarpHandle );  

    //
    // Rx Filter stuff  
    //
    BESTATUS (BECALL *BeMacAddressReadWrite)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      boolean             Port1,  // VM must always set to false
        IN      boolean             Mac1,   // VM must always set to false
        IN      boolean             Mgmt,
        IN      boolean             Write,
        IN      boolean             Permanent,
        IN OUT  PSA_MAC_ADDRESS     MacAddress,
        IN      MCC_WRB_CQE_CALLBACK    CompletionCallback,
        IN      PVOID                   CompletionCallbackContext);

    BESTATUS (BECALL *BeMulticastAddressSetTable)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      PVOID               McHashTable );

    BESTATUS (BECALL *BeVLanTagAddRemove)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      u16              VLanTag,
        IN      boolean             Add     );

    BESTATUS (BECALL *BeIpSecAddRemove)(
        IN   FUNCTIONHANDLE FunctionObject,
        IN   u32            IpAddressSource,
        IN   u32            IpAddressDestination,
        IN   boolean        Add);

    BESTATUS (BECALL *BeWakeOnLanSet)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      boolean             Mac1,
        IN      PVOID               Pattern,    
        IN      PVOID               PatternMask );

    BESTATUS (BECALL *BeWakeOnLanRead)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      boolean             Mac1,
        OUT     PVOID               Pattern,
        OUT     PVOID               PatternMask);

    BESTATUS (BECALL *BeWakeOnLanRemove)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      boolean             Mac1 );

    BESTATUS (BECALL *BeRssKeySet)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      u32*             RssKey );

    BESTATUS (BECALL *BeRssEntrySetCpu)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      u32              RssEntry,
        IN      u32              Cpu );
                        
    BESTATUS (BECALL *BeGetStatistic)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN OUT  u32*              pStatistic,
        IN      u32              StatIndex);
                            
    //
    // Interrupts
    //
    void (BECALL * BeInterruptEnableEq)(
        IN      EQHANDLE            EqObject,
        IN      boolean             Enable );

    void (BECALL * BeInterruptEnableAll)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      boolean             Enable );

    //
    // PD stuff for iWARP
    //
    PDHANDLE (BECALL* BePdCreateForIWarp)(
        IN      FUNCTIONHANDLE      FunctionObject );

    u32  (BECALL* BePdGetId)(
        IN      PDHANDLE            PdObject );

    u64  (BECALL* BePdGetDoorbellPage)(
        IN      PDHANDLE            PdObject );

    BESTATUS (BECALL* BePdRelease )(
        IN      PDHANDLE            PdObject );

    //
    // Flash routines
    //
    BESTATUS (BECALL* BeFlashRead)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      u32              FlashAddress,
        OUT     PVOID               Buffer,
        IN      u32              BufferLength );

    BESTATUS (BECALL* BeFlashWrite)(
        IN      FUNCTIONHANDLE      FunctionObject,
        IN      u32              FlashAddress,
        IN      PVOID               Buffer,
        IN      u32              BufferLength );
        
    //
    // Stag support
    //
#if defined(BE_EMULATION)
    BESTATUS (BECALL * BeStagAllocate)(
        IN      PDHANDLE            PdHandle,
        IN      u32              RequestCount,
        OUT     PSTAG_D        StagList,
        OUT     u32*             StagsAllocated  );

    BESTATUS (BECALL * BeStagRelease) (
        IN      PDHANDLE            PdObject,
        IN      PSTAG_D        StagList,
        IN      u32              StagCount );

    BESTATUS (BECALL * BeMemoryRegionNonSharedRegister)(
        IN      PDHANDLE                    PdHandle,
        IN      STAG_D                 Stag,
        IN      PSA_SGL                  Sgl,
        IN      PVOID                       Va,
        IN      u32                      Flags,
        IN      MTPTE_ACCESS_RIGHTS_ENUM    Access );

    BESTATUS (BECALL * BeMemoryRegionNonSharedDeregister)(
        IN      PDHANDLE                    PdObject,
        IN      STAG_D                 Stag );
#endif /* defined(BE_EMULATION) */

//////////////// Not implemented below this level
                                                      
    //
    // Vm stuff
    //
    u64 (BECALL* BeVmDomainToPhysical)(
                IN      u64*         DomainPa,
                IN OUT  u64*         Pa,
                IN      u32          AddressCount );

///////////// End not implemted ///////////////////////////////

      
    //
    // Utilities
    //
    BESTATUS (BECALL * GetSglMappingForVa)(
        IN PVOID       Va,
        IN u32       Len,
        IN PSA_SGL  *     ppSgl);

    void (BECALL * FreeSglMapping)(
        IN  PSA_SGL     pSgl );   

    BESTATUS (BECALL * UtMap) (
            IN FUNCTIONHANDLE   FunctionObject,
            IN u32              Cra,
            IN u64              Pa );

    BESTATUS (BECALL * UtUnmap) (
            IN FUNCTIONHANDLE   FunctionObject,
            IN u32              Cra );

    //
    // SeTcp Client (user) calls
    //
    SOCKET      (BECALL*  BeSocketCreate)(void);

    BESTATUS    (BECALL* BeSocketConnect)( SOCKET Socket, u32 IpAddr, u16 port,
                                           BE_SOCKET_COMPLETION_CALLBACK CompletionRoutine,
                                           PVOID    CompletionContext );

    BESTATUS    (BECALL* BeSocketGetLocalAddress)( SOCKET Socket, u32* IpAddress, u16* Port );
        
    BESTATUS    (BECALL* BeSocketBind)( SOCKET Socket, u32 IpAddress, u16 Port );

    BESTATUS    (BECALL* BeSocketListen)( SOCKET Socket, u32 Backlog );

    BESTATUS    (BECALL* BeSocketAccept)( SOCKET Socket,
                                          BE_SOCKET_COMPLETION_CALLBACK CompletionRoutine,
                                          PVOID    CompletionContext );

    BESTATUS    (BECALL* BeSocketSend)( SOCKET Socket, PBE_IOVEC BufferDescriptors, 
                                        u32 BufferCount,
                                        BE_SOCKET_COMPLETION_CALLBACK CompletionRoutine,
                                        PVOID    CompletionContext );

    BESTATUS    (BECALL* BeSocketRecv)( SOCKET Socket, 
                                        PBE_IOVEC BufferDescriptors, 
                                        u32 BufferCount,
                                        BE_SOCKET_COMPLETION_CALLBACK CompletionRoutine,
                                        PVOID    CompletionContext );

    BESTATUS    (BECALL* BeSocketOptionSet)( SOCKET Socket, u32 Level,
                                             u32 Name, PVOID Value,
                                             u32 ValueSize );

    BESTATUS    (BECALL* BeSocketOptionGet)( SOCKET Socket, u32 Level,
                                             u32 Name, PVOID Value,
                                             u32* ValueSize );

    BESTATUS    (BECALL* BeSocketShutdown)( SOCKET Socket, u32 How );    

    BESTATUS    (BECALL* BeSocketClose)( SOCKET Socket, boolean Abort );  
    
    BESTATUS    (BECALL* BeSeTcpErrorToBeStatus)( int Error );

    //
    // SeTcp Mac level calls
    //
    BESTATUS (BECALL* BeTcpBindAdapter)( 
        IN      FUNCTIONHANDLE          FunctionHandle ,
        IN      PSA_MAC_ADDRESS            MacAddress,
        IN      boolean                 Dhcp,
        IN      u32                  IpAddress,
        IN      u32                  NetMask,
        IN      u32                  Gateway    );

    BESTATUS (BECALL* BeTcpUnbindAdapter)( 
        IN      FUNCTIONHANDLE          FunctionHandle );

    BESTATUS (BECALL* BeTcpUpdateIpSettings)(
        IN      FUNCTIONHANDLE          FunctionHandle,
        IN      boolean                 Dhcp,
        IN      u32                  IpAddress,
        IN      u32                  SubnetMask,
        IN      u32                  Gateway );


    BESTATUS (BECALL* BeTcpIndicatePacket)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      PVOID                   Packet,
        IN      u32                  PacketLength );

    BESTATUS (BECALL* BeTcpSendComplete)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      PVOID                   Packet );

    void (BECALL* BeTcpForwardSegmentComplete)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      PVOID                   Segment );

    BESTATUS (BECALL* BeTcpIndicateEvent)(
        IN      FUNCTIONHANDLE          FunctionObject,
        IN      u32                  Cid,
        IN      TOE_EVENT_CODE          Event );

    void (BECALL* BeEnableInterrupts)(
        IN      FUNCTIONHANDLE          function_object);
   
} SA_PACK CHIPIF_DOWNCALL, *PCHIPIF_DOWNCALL;

#include <sa_packoff.h>

//
// This structure defines any routines in the ULD that the shared klib may
// call.  The only thing I can think of right now that will use this
// is the wsd proxy, SeTcp, and the VMM.
//
typedef struct _CHIPIF_UPCALL {
    u32                  Size;           // Size of this structure

    u32                 IsScsi;         // True if this is a network function
    HDEVICE_CONTEXT         DeviceContext;  // Passed back to ULD - *Must be unique system wide*!    

    BESTATUS (BECALL *OnVmRequest)( HDEVICE_CONTEXT DeviceContext, u32 VmId, 
                                    PVOID Payload, u32 PayloadLength );

    //
    // Calls for SeTcp
    //
    void (BECALL *MacReturnFrame)( 
        IN      HDEVICE_CONTEXT DeviceContext, 
        IN      PVOID           Frame, 
        IN      u32          Length );

    BESTATUS (BECALL *MacSendFrame)( 
        IN      HDEVICE_CONTEXT DeviceContext, 
        IN      PVOID           Buffer, 
        IN      u32          BufferLength );

    BESTATUS (BECALL* MacForwardSegment)(
        IN      HDEVICE_CONTEXT     DeviceContext,
        IN      PVOID               Segment,
        IN      u32                 SegmentLength );
   


} CHIPIF_UPCALL, *PCHIPIF_UPCALL;

//
// The miniport calls this routine as part of his initialization. The
// address for this routine is passed in as part of his resources via whatever
// routine the port wrapper calls during IRP_MN_START_DEVICE
//
BEAPI
BESTATUS
BECALL
BeRegisterDevice( 
    IN      SA_DEV*                 sa_dev,
    IN      PCHIPIF_UPCALL          Upcall,
    OUT     PCHIPIF_DOWNCALL        Downcall,
    IN      u32                  Reserved );

BEAPI
BESTATUS
BECALL
BeDeregisterDevice( 
    IN      PVOID                   DeviceContext,
    OUT     PCHIPIF_DOWNCALL        BeDispatch);

SA_CPP_TRAILER

#endif // __beclib_h__



