/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   beclib_private_hl.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/beclib_private_hl.h,v 1.31 2006/03/23 17:16:49 menz Exp $
*/
#ifndef __beclib_private_hl_h__
#define __beclib_private_hl_h__

#define BE_VM_MAX  SA_NUMBER_OF_FIELD(BLADE_ENGINE_PCI0_DBMAP, pci_func0_protection_domain_dbmap)


// ------------------------------------------------------
//  Chip Object
// ------------------------------------------------------

PBE_CHIP_OBJECT
be_chip_find_or_allocate (
    IN PBE_FUNCTION_OBJECT FunctionObject);

BESTATUS
be_chip_allocate (    
    IN      PBE_FUNCTION_OBJECT function_object,
    IN      PBE_CHIP_OBJECT    *pchip_object);

void
be_chip_free (
    IN PBE_FUNCTION_OBJECT FunctionObject);

BESTATUS
be_chip_allocate_jell (
    PBE_CHIP_OBJECT chip,
    PBE_FUNCTION_OBJECT function);

BESTATUS
be_chip_free_jell (
    PBE_FUNCTION_OBJECT function_object);


void be_chip_lock (IN PBE_CHIP_OBJECT chip_object);
void be_chip_unlock (IN PBE_CHIP_OBJECT chip_object);

void be_chip_blocking_lock (IN PBE_CHIP_OBJECT chip_object);
void be_chip_blocking_unlock (IN PBE_CHIP_OBJECT chip_object);

// ------------------------------------------------------
//  Protection Domain
// ------------------------------------------------------

PBE_PD_OBJECT
be_pd_allocate (PBE_FUNCTION_OBJECT function_object, 
                boolean is_rdma);

BESTATUS
be_pd_release (PBE_PD_OBJECT pd);

u32 be_pd_get_id (PBE_PD_OBJECT pd);

// ------------------------------------------------------
//  Function Object
// ------------------------------------------------------

BESTATUS
be_function_allocate (
    IN  SA_DEV*              sa_dev,
    IN  PCHIPIF_UPCALL       Upcall,
    OUT PBE_FUNCTION_OBJECT* ppFunctionObject);

BESTATUS
be_function_free (
    IN PBE_FUNCTION_OBJECT FunctionObject);

void
_be_function_add_pd (
    IN      PBE_FUNCTION_OBJECT     func_obj, 
    IN      PBE_PD_OBJECT           pd_object);

void
_be_function_remove_pd (
    IN      PBE_FUNCTION_OBJECT     func_obj, 
    IN      PBE_PD_OBJECT           pd_object);

STATIC INLINE
PBE_CHIP_OBJECT
be_function_get_chip (PBE_FUNCTION_OBJECT function_object) {
    FUNCTION_ASSERT (function_object);
    return function_object->parent_chip;
}

BESTATUS
be_function_allocate_template_headers (
    IN   PBE_FUNCTION_OBJECT function_object);

void
be_function_free_templates (
    PBE_FUNCTION_OBJECT function_object);

BESTATUS
be_function_allocate_zero_buffer (
    PBE_FUNCTION_OBJECT function_object);

BESTATUS
be_function_allocate_pdu_buffer (
    IN PBE_FUNCTION_OBJECT function_object);

void
be_function_free_pdu_buffer (
    IN PBE_FUNCTION_OBJECT function_object);

BESTATUS
be_function_ulpchaos_allocate (
    IN PBE_FUNCTION_OBJECT function_object);

void
be_function_free_ulpchaos (
    IN PBE_FUNCTION_OBJECT function_object);


BESTATUS
be_function_verify_page_tables (
    PBE_FUNCTION_OBJECT function_object,
    u32 min_num_free);

// ------------------------------------------------------
//  Event Queue
// ------------------------------------------------------

BESTATUS
be_eq_allocate (
    IN  PBE_FUNCTION_OBJECT FunctionObject,
    IN  PVOID               EqBaseVa,          // EQE buffer
    IN  u32                 SizeEncoding,      // EQE Size Encoding
    IN  u32                 NumEntriesEncoded, // EQ Count Encodings
    IN  u32                 Watermark,         // CEV_WMARK_* or -1
    IN  u32                 TimerDelay,        // In 8us units, or -1
    OUT PBE_EQ_OBJECT*      ppEqObject);

BESTATUS 
be_eq_free (
    IN PBE_EQ_OBJECT EqObject);

// ------------------------------------------------------
//  Completion Queue Object
// ------------------------------------------------------

BESTATUS
be_cq_object_allocate (
    IN      PBE_FUNCTION_OBJECT FunctionObject, 
    IN      PVOID               CqBaseVa, 
    IN      CQ_CNT_ENUM         NumEntriesEncoded,     // count encodings
    IN      boolean             solEventEnable,
    IN      boolean             eventable,
    IN      boolean             nodelay,
    IN      boolean             wme,
    IN      u32                 wmThresh,       // Watermark encodings
    IN      PBE_EQ_OBJECT       EqObject,
    OUT     PBE_CQ_OBJECT*      ppCqObject );

BESTATUS 
be_cq_object_free (
    IN PBE_CQ_OBJECT CqObject);

BESTATUS 
be_cq_unmap_ring(
    IN      PBE_CQ_OBJECT cq_object);

BESTATUS 
be_cq_remap_ring (
    IN      PBE_CQ_OBJECT cq_object,
    IN      CQ_CNT_ENUM   num_entries,     // count encodings
    IN      PVOID         cq_base_va );

// ------------------------------------------------------
//  MCC QUEUE
// ------------------------------------------------------

BESTATUS 
be_mcc_allocate (
    IN      PBE_FUNCTION_OBJECT function_object, 
    IN      PBE_EQ_OBJECT       eq_object, 
    //IN      PBE_MCC_OBJECT*     pmcc);
    OUT     PBE_MCC_OBJECT*     pmcc);

BESTATUS 
be_mcc_free(
    IN      PBE_MCC_OBJECT  mcc);

BESTATUS
be_mcc_delegate_process_cq ( 
    IN      PBE_FUNCTION_OBJECT     function_object,
    IN      PBE_CQ_OBJECT           mcc_cq,
    IN      PVOID                   context);

BESTATUS
be_mcc_post_wrb (
    IN PBE_FUNCTION_OBJECT function_object, 
    IN PMCC_WRB       wrb,
    IN MCC_WRB_CQE_CALLBACK    CompletionCallback,
    IN PVOID                   CompletionCallbackContext );


// ------------------------------------------------------
//  Mailbox Object
// ------------------------------------------------------

// ------------------------------------------------------
//  Default PDU Queue
// ------------------------------------------------------

BESTATUS 
pdu_default_create (
    IN      PBE_FUNCTION_OBJECT             function_object, 
    IN      PVOID                           send_queue_va, 
    IN      u32                             ring_size_power_of_two, 
    IN      u32                             rsvd, 
    IN      PBE_CQ_OBJECT                   cq_object, 
    OUT     PBE_DEFAULT_PDU_QUEUE_OBJECT*   pdefq);

BESTATUS 
be_pdu_default_free (
    IN     PBE_DEFAULT_PDU_QUEUE_OBJECT   defq);

// ------------------------------------------------------
//  iSCSI WRB ring
// ------------------------------------------------------

BESTATUS BECALL 
be_iscsi_wrbq_allocate (
    IN  BE_FUNCTION_OBJECT*         function_object,
    IN  PVOID                       va,
    IN  u32                         length,
    OUT PBE_ISCSI_WRB_QUEUE_OBJECT* wrbq);

BESTATUS BECALL
be_iscsi_wrbq_free (
    IN PBE_ISCSI_WRB_QUEUE_OBJECT wrbq); 

// ------------------------------------------------------
//  iSCSI connection
// ------------------------------------------------------

u32 
be_iscsi_cxn_get_cid (
    IN PBE_ISCSI_CONNECTION_OBJECT cxn);

BESTATUS BECALL
be_iscsi_offload_initiate_tcp_stack (
    IN  PBE_ISCSI_WRB_QUEUE_OBJECT   wrbq,
    IN  PBE_DEFAULT_PDU_QUEUE_OBJECT defq,
    IN  PBE_CQ_OBJECT                cq,
    IN  u32                          max_segment_length,
    IN  u32                          first_burst_length,
    IN  u32                          max_burst_length,
    IN  u32                          flags,
    IN  u32                          exp_stat_sn,
//    IN  u64                          pdu_pa,
    IN  SOCKET                       socket,
    OUT PBE_ISCSI_CONNECTION_OBJECT* cxn_out,
    IN  MCC_WRB_CQE_CALLBACK         callback,
    IN  PVOID                        callback_context);

BESTATUS BECALL
be_iscsi_offload_terminate (
    IN      PBE_ISCSI_CONNECTION_OBJECT cxn, 
    IN      MCC_WRB_CQE_CALLBACK        callback, 
    IN      PVOID                       callback_context);


BESTATUS
be_iscsi_map_frags (    
    PBE_FUNCTION_OBJECT function_object,    
    PVOID va,
    u32  num_pages
);

BESTATUS
be_iscsi_unmap_frags (    
    PBE_FUNCTION_OBJECT function_object,    
    u32  num_pages
);

BESTATUS 
be_iscsi_cxn_object_create(
    IN      FUNCTIONHANDLE                  FunctionObject,
    IN      PVOID                           WrbRingVa,
    IN      u32                             WrbRingSizePowerOfTwo,
    IN      PVOID                           FragVa,
    IN      u32                             FragPageCount,
    IN      u32                             MaxSegmentLength,
    IN      u32                             FirstBurstLength,
    IN      u32                             MaxBurstLength,
    IN      u32                             Flags,
    IN      u32                             exp_stat_sn,
    IN      PBE_DEFAULT_PDU_QUEUE_OBJECT    ScsiDefaultQueue,
    IN      PBE_CQ_OBJECT                   CqObject,
    OUT     PBE_ISCSI_CONNECTION_OBJECT*    ppConnectionObject );

// -----------------------------------------------------
//  FCoE WRB Queue and offload stuff
// -----------------------------------------------------
BESTATUS
be_fcoe_create_wrbq(
  IN  PBE_FUNCTION_OBJECT function_object,
  IN  PVOID va,
  IN  u32   length,
  OUT PBE_FCOE_WRB_QUEUE_OBJECT *wrbq
);

BESTATUS
be_fcoe_remove_wrbq(
  IN PBE_FCOE_WRB_QUEUE_OBJECT wrbq
);

BESTATUS
be_fcoe_map_frags (
    PBE_FUNCTION_OBJECT function_object,
    PVOID va,
    u32  num_pages
);

BESTATUS
be_fcoe_unmap_frags (
    PBE_FUNCTION_OBJECT function_object,
    u32  num_pages
);


BESTATUS
be_fcoe_interface_resources_create(
    IN PBE_FUNCTION_OBJECT       func_obj,
    IN PVOID                     wrbq,
    IN u32                       wrb_size,
    IN PVOID                     frag_va,
    IN u32                       frag_page_cnt,	
    IN u32                       hdr_addr_hi,
    IN u32                       hdr_addr_lo,
    IN PBE_CQ_OBJECT             cq,
    OUT PBE_FCOE_INTERFACE_OBJECT *intf
);

BESTATUS
be_fcoe_interface_offload(
  IN  PBE_FCOE_WRB_QUEUE_OBJECT wrbq,
  IN  PBE_CQ_OBJECT             cq,
  IN  PBE_FCOE_INTERFACE_OBJECT intf,
  IN  MCC_WRB_CQE_CALLBACK         callback,
  IN  PVOID                        callback_context
);

BESTATUS
be_fcoe_terminate_offload(
  IN  PBE_FCOE_INTERFACE_OBJECT intf,
  IN  MCC_WRB_CQE_CALLBACK         callback,
  IN  PVOID                        callback_context
);

u32
be_fcoe_get_interface_cid(
  IN  PBE_FCOE_INTERFACE_OBJECT intf
);

// ------------------------------------------------------
//  Ethernet Send Queue
// ------------------------------------------------------

BESTATUS 
be_eth_sq_allocate (
    IN      PBE_FUNCTION_OBJECT function_object, 
    IN      PVOID               send_queue_va, 
    IN      u32                 ring_size_power_of_two, 
    IN      PBE_CQ_OBJECT       cq_object, 
    OUT     PBE_ETHSQ_OBJECT*   peth_sq);

BESTATUS 
be_eth_sq_free (
    IN      PBE_ETHSQ_OBJECT        eth_sq);


// ------------------------------------------------------
//  Ethernet Receive Queue
// ------------------------------------------------------

BESTATUS
be_eth_rq_allocate (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PVOID                   rq_base_va, 
    IN      PBE_CQ_OBJECT           cq_object, 
    IN      u32                     rss_index, 
    OUT     PBE_ETHRQ_OBJECT*       peth_rq);

BESTATUS
be_eth_rq_free(
    IN      PBE_ETHRQ_OBJECT        eth_rq );

#if defined(BE_EMULATION)
BESTATUS
be_eth_rq_get_frag_size_legacy (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      ERX_FRAG_SIZE_ENUM      encoded_value,
    OUT     u32*                    frag_size_bytes);
#endif /* defined(BE_EMULATION) */

// ------------------------------------------------------
//  TOE
// ------------------------------------------------------

BESTATUS 
be_toe_connection_object_allocate (
    IN  FUNCTIONHANDLE             function_object,
    IN  PVOID                      send_ring_va,
    IN  u32                        send_ring_size_power_of_two,
    IN  PVOID                      recv_ring_va,
    IN  u32                        recv_ring_size_power_of_two,
    IN  u32                        flags,
    IN  u32                        sa_index,
    IN  PBE_TOE_DQ_OBJECT          toe_default_queue,
    IN  PBE_CQ_OBJECT              send_cq_object,
    IN  PBE_CQ_OBJECT              recv_cq_object,
    OUT PBE_TOE_CONNECTION_OBJECT* pcxn);

BESTATUS 
be_toe_connection_object_free (
    IN PBE_TOE_CONNECTION_OBJECT cxn);

BESTATUS 
be_toe_offload_initiate (
    IN PBE_TOE_CONNECTION_OBJECT cxn,
    IN PTEMPLATE_HEADER          header,
    IN PTCP_CONTEXT              tcp,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context);

BESTATUS 
be_toe_offload_update (
    IN PBE_TOE_CONNECTION_OBJECT cxn,
    IN PTEMPLATE_HEADER          header,
    IN PTCP_CONTEXT              tcp,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context);

BESTATUS 
be_toe_offload_enable (
    IN PBE_TOE_CONNECTION_OBJECT cxn,
    IN boolean                   enable,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context);

BESTATUS 
be_toe_default_receive_object_allocate (
    IN  PBE_FUNCTION_OBJECT function_object,
    IN  PVOID               recv_queue_va,
    IN  u32                 ring_size_power_of_two,
    OUT PBE_TOE_DQ_OBJECT*  pdq);

BESTATUS 
be_toe_default_receive_object_free (
    IN     PBE_TOE_DQ_OBJECT   dq);

BESTATUS 
be_tcp_sq_allocate (
    IN  PBE_FUNCTION_OBJECT       function_object,
    IN  PVOID                     send_queue_va,
    IN  u32                       ring_size_power_of_two,
    IN  PBE_CQ_OBJECT             cq_object,
    OUT PBE_TOE_SQ_OBJECT*        psq);

BESTATUS 
be_tcp_sq_free (
    IN PBE_TOE_SQ_OBJECT tcp_sq);

// ------------------------------------------------------
//  RDMA/iWARP
// ------------------------------------------------------

// ------------------------------------------------------
//  MPU Emulation
// ------------------------------------------------------


// ------------------------------------------------------
//  Flash
// ------------------------------------------------------
BESTATUS 
be_flash_read (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32                 flash_address,
    IN PVOID               buffer,
    IN u32                 buffer_length);

BESTATUS 
be_flash_write (
    IN PBE_FUNCTION_OBJECT          function_object,
    IN IOCTL_COMMON_WRITE_FLASHROM *va_for_ioctl,
    IN u64                          pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK         callback,
    IN PVOID                        callback_context,
    IN PBE_FLASHACCESS_QUEUE_CONTEXT  queue_context);

BESTATUS 
be_rxf_rss_key_set (
    IN      PBE_FUNCTION_OBJECT function_object, 
    IN      u32*             rss_key);

BESTATUS 
be_rxf_rss_entry_set_cpu (
    IN      PBE_FUNCTION_OBJECT function_object, 
    IN      u32              rss_entry, 
    IN      u32              cpu);

BESTATUS
be_rxf_get_statistic (
    IN    PBE_FUNCTION_OBJECT function_object, 
    u32*  statistic, 
    u32   stat_index);

// ------------------------------------------------------
//  Util
// ------------------------------------------------------

BESTATUS
be_get_sgl_mapping_for_va (
    IN  PVOID    va,
    IN  u32      len,
    OUT PSA_SGL* sgl);

BESTATUS
be_shared_memory_allocate (
    IN  u32                       Length,
    IN  u32                       Tag,
    IN  boolean                   NeedVa,
    IN  SA_PHYSICAL_ADDRESS*      HighestAcceptableAddress,
    OUT PBE_SHARED_MEM_DESCRIPTOR SharedMemoryDescriptor);

void
be_shared_memory_free (
    IN PBE_SHARED_MEM_DESCRIPTOR SharedMemoryDescriptor);


BESTATUS be_util_ut_map (
        IN PBE_FUNCTION_OBJECT   FunctionObject,
        IN u32                   Cra,
        IN u64                   pa);

BESTATUS be_util_ut_unmap (
        IN PBE_FUNCTION_OBJECT   FunctionObject,
        IN u32                   Cra);

// ------------------------------------------------------
//  Locking
// ------------------------------------------------------


// Acquire or release the single static global lock for BECLIB
void be_main_global_lock();
void be_main_global_unlock();


// ------------------------------------------------------
//  TCP stack to deprecate
// -----------------------------------------------------

#if defined(BE_EMULATION)
BESTATUS
SeTcpOffloadInitiate (
    IN SOCKET                  Socket,
    IN PBE_FUNCTION_OBJECT     FunctionObject,
    IN u32                     Cid,
    IN RXF_CXN_TYPE_ENUM       ConnectionType,
    IN KLIB_COMPLETION_ROUTINE CompletionCallback,
    IN PVOID                   CompletionCallbackContext);
#endif /* defined(BE_EMULATION) */


void
be_offload_connection_filter_build (
    OUT MPU_OFFLOAD_CXN_ENTRY* connection_filter,
    IN  const              TEMPLATE_HEADER*   template_header,
    IN  u8                 chute,
    IN  boolean            enable);



#endif // __beclib_private_h__


