/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#ifndef _BESTATUS_
#define _BESTATUS_

#define BE_SUCCESS                      ((SA_STATUS)0x00000000L) // ntsubauth


//
//  Status values are 32 bit values layed out as follows:
//
//   3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
//   1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
//  +---+-+-------------------------+-------------------------------+
//  |Sev|C|       Facility      |           Code        |
//  +---+-+-------------------------+-------------------------------+
//
//  where
//
//  Sev - is the severity code
//
//      00 - Success
//      01 - Informational
//      10 - Warning
//      11 - Error
//
//  C - is the Customer code flag
//
//  Facility - is the facility code
//
//  Code - is the facility's status code
//

//
//  Values are 32 bit values layed out as follows:
//
//   3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
//   1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
//  +---+-+-+-----------------------+-------------------------------+
//  |Sev|C|R|     Facility          |               Code            |
//  +---+-+-+-----------------------+-------------------------------+
//
//  where
//
//      Sev - is the severity code
//
//          00 - Success
//          01 - Informational
//          10 - Warning
//          11 - Error
//
//      C - is the Customer code flag
//
//      R - is a reserved bit
//
//      Facility - is the facility code
//
//      Code - is the facility's status code
//
//
// Define the facility codes
//
#define FACILITY_RPC_STUBS               0x3
#define FACILITY_RPC_RUNTIME             0x2
#define FACILITY_IO_ERROR_CODE           0x4
#define FACILITY_BE_ERROR_CODE           0x7


//
// Define the severity codes
//
#define STATUS_SEVERITY_WARNING          0x2
#define STATUS_SEVERITY_SUCCESS          0x0
#define STATUS_SEVERITY_INFORMATIONAL    0x1
#define STATUS_SEVERITY_ERROR            0x3


//
// MessageId: BE_PENDING
//
// MessageText:
//
//  The BladeEngine Driver call succeeded, and pended operation.
//
#define BE_PENDING                       ((SA_STATUS)0x20070001L)
#define BE_STATUS_PENDING                (BE_PENDING)

//
// MessageId: BE_NOT_OK
//
// MessageText:
//
//  An error occurred.
//
#define BE_NOT_OK                        ((SA_STATUS)0xE0070002L)

//
// MessageId: BE_STATUS_SYSTEM_RESOURCES
//
// MessageText:
//
//  Insufficient host system resources exist to complete the API.
//
#define BE_STATUS_SYSTEM_RESOURCES       ((SA_STATUS)0xE0070003L)

//
// MessageId: BE_STATUS_CHIP_RESOURCES
//
// MessageText:
//
//  Insufficient chip resources exist to complete the API.
//
#define BE_STATUS_CHIP_RESOURCES         ((SA_STATUS)0xE0070004L)

//
// MessageId: BE_STATUS_NO_RESOURCE
//
// MessageText:
//
//  Insufficient resources to complete request.
//
#define BE_STATUS_NO_RESOURCE            ((SA_STATUS)0xE0070005L)

//
// MessageId: BE_STATUS_BUSY
//
// MessageText:
//
//  Resource is currently busy.
//
#define BE_STATUS_BUSY                   ((SA_STATUS)0xE0070006L)

//
// MessageId: BE_STATUS_INVALID_PARAMETER
//
// MessageText:
//
//  Invalid Parameter in request.
//
#define BE_STATUS_INVALID_PARAMETER      ((SA_STATUS)0xE0000007L)

//
// MessageId: BE_STATUS_LIMITS_EXCEEDED
//
// MessageText:
//
//  Resource limits exceeded.
//
#define BE_STATUS_LIMITS_EXCEEDED        ((SA_STATUS)0xE0000008L)

//
// MessageId: BE_STATUS_ILLEGAL_OPERATION
//
// MessageText:
//
//  Requested operation is not permitted.
//
#define BE_STATUS_ILLEGAL_OPERATION      ((SA_STATUS)0xE0000009L)

//
// MessageId: BE_STATUS_IO_ERROR
//
// MessageText:
//
//  An IO error occured.
//
#define BE_STATUS_IO_ERROR               ((SA_STATUS)0xE000000AL)

//
// MessageId: BE_STATUS_INVALID_RESOURCE
//
// MessageText:
//
//  Requested resources does not exist.
//
#define BE_STATUS_INVALID_RESOURCE       ((SA_STATUS)0xE000000BL)

//
// MessageId: BE_STATUS_INVALID_ADDRESS
//
// MessageText:
//
//  Invalid address.
//
#define BE_STATUS_INVALID_ADDRESS        ((SA_STATUS)0xE000000CL)

//
// MessageId: BE_STATUS_NOT_SUPPORTED
//
// MessageText:
//
//  Requested operation is not supported.
//
#define BE_STATUS_NOT_SUPPORTED           ((SA_STATUS)0xE000000DL)

//  Polling MPU mailbox timed out.
#define BE_STATUS_MAILBOX_TIMEOUT         ((SA_STATUS)0xE000000EL)

//  Polling MCC cq timed out.
#define BE_STATUS_POLLING_CQ_TIMEOUT      ((SA_STATUS)0xE000000FL)

//  Polling MCC cq timed out.
#define BE_STATUS_UE_DETECTED             ((SA_STATUS)0xE0000010L)

// Semaphore cannot be acquired
#define BE_STATUS_SEMAPHORE_BUSY           ((SA_STATUS)0xE0000011L)  

// UE recovery algorithm failed.
#define BE_STATUS_RECOVERY_FAILED          ((SA_STATUS)0xE0000012L)  

// Recovery disabled during to versioning 
#define BE_STATUS_UE_RECOVERY_VERSION_MISMATCH ((SA_STATUS)0xE0000013L)  

// Software UE ignored due to unprivileged function not being allowed to cause a UE
#define BE_STATUS_SOFTWARE_UE_IGNORED      ((SA_STATUS)0xE0000014L)  

// UE recovery algorithm failed.
#define BE_STATUS_RECOVERY_DISABLED        ((SA_STATUS)0xE0000015L)  


// Pass POST errors from the MPU_EP_SEMAPHORE register (mpu.ep.ep_semaphore)
// via the lower 16-bits of the error code, when the upper 16-bit are this
// value.
// This may be returned by be_function_create_ex or
// be_function_and_chip_create_ex.
#define BE_STATUS_POST                    ((SA_STATUS)0xEE000000)

// Returns BOOLEAN value indicating a MPU POST error.
#define BE_STATUS_IS_POST_ERROR(_v)       ((SA_STATUS)(0xFFFF0000 & (_v)) == BE_STATUS_POST)

// If a POST error, this returns the ARM's POST stage field.
#define BE_STATUS_POST_STAGE(_v)          ((u32)(0x0000FFFF & (_v)))

// Warning from be_ue_register that recovery is not supported.
#define BE_STATUS_WARN_NO_RECOVERY        ((SA_STATUS)0x20070002L)


// ***************************************************************************
//                     E T H E R N E T   S T A T U S
// ***************************************************************************
//
// MessageId: BE_ETH_INIT_FAILURE
//
// MessageText:
//
//  The BladeEngine Ethernet device driver failed to initialize.
//
#define BE_ETH_INIT_FAILURE              ((SA_STATUS)0xE0070100L)

//
// MessageId: BE_ETH_TX_ERROR
//
// MessageText:
//
//  The Ethernet device driver failed to transmit a packet.
//
#define BE_ETH_TX_ERROR                  ((SA_STATUS)0xE0070101L)

//
// MessageId: BE_ETH_OFFL_FAIL
//
// MessageText:
//
//  The Ethernet device driver failed to initialize a connection.
//
#define BE_ETH_OFFL_FAIL                 ((SA_STATUS)0xE0070102L)

// ***************************************************************************
//                             R D M A  S T A T U S
// ***************************************************************************
//
// MessageId: BE_STATUS_RDMA_INIT_FAILURE
//
// MessageText:
//
//  The BladeEngine RDMA device driver failed to initialize.
//
#define BE_STATUS_RDMA_INIT_FAILURE      ((SA_STATUS)0xE0070200L)

//
// MessageId: BE_STATUS_RDMA_RESOURCE_AQUIRED
//
// MessageText:
//
//  RDMA Resources were previously allocated.  No new allocation occured.
//
#define BE_STATUS_RDMA_RESOURCE_AQUIRED  ((SA_STATUS)0x20070201L)

//
// MessageId: BE_STATUS_TCP_RESOURCE_AQUIRED
//
// MessageText:
//
//  TCP Resources were previously allocated.  No new allocation occured.
//
#define BE_STATUS_TCP_RESOURCE_AQUIRED   ((SA_STATUS)0x20070202L)

//
// MessageId: BE_STATUS_TCP_SOCKET_FAILED
//
// MessageText:
//
//  Unable to create private socket.
//
#define BE_STATUS_TCP_SOCKET_FAILED      ((SA_STATUS)0xE0070203L)

//
// MessageId: BE_STATUS_RDMA_INVALID_ADDRESS
//
// MessageText:
//
//  Invalid ip address format.
//
#define BE_STATUS_RDMA_INVALID_ADDRESS   ((SA_STATUS)0xE0070204L)

//
// MessageId: BE_STATUS_RDMA_INVALID_SOCKET
//
// MessageText:
//
//  Invalid BE socket.
//
#define BE_STATUS_RDMA_INVALID_SOCKET    ((SA_STATUS)0xE0070205L)

//
// MessageId: BE_STATUS_RDMA_BIND_FAILED
//
// MessageText:
//
//  The bind attempt failed.
//
#define BE_STATUS_RDMA_BIND_FAILED       ((SA_STATUS)0xE0070206L)

//
// MessageId: BE_STATUS_RDMA_CONNECT_FAILED
//
// MessageText:
//
//  The connect attempt failed.
//
#define BE_STATUS_RDMA_CONNECT_FAILED    ((SA_STATUS)0xE0070207L)

//
// MessageId: BE_STATUS_RDMA_LISTEN_FAILED
//
// MessageText:
//
//  The listen attempt failed.
//
#define BE_STATUS_RDMA_LISTEN_FAILED     ((SA_STATUS)0xE0070208L)

//
// MessageId: BE_STATUS_NO_QUEUED_ACCEPT
//
// MessageText:
//
//  No accept requests are queued.
//
#define BE_STATUS_NO_QUEUED_ACCEPT       ((SA_STATUS)0xE0070209L)

//
// MessageId: BE_STATUS_RDMA_OFFLOAD_FAILED
//
// MessageText:
//
//  RDMA offload failed.
//
#define BE_STATUS_RDMA_OFFLOAD_FAILED    ((SA_STATUS)0xE0070210L)

//
// MessageId: BE_STATUS_RDMA_LISTEN_TWICE
//
// MessageText:
//
//  The socket was already in a listening state.
//
#define BE_STATUS_RDMA_LISTEN_TWICE      ((SA_STATUS)0x20070211L)

// ***************************************************************************
//                     I S C S I   S T A T U S
// ***************************************************************************
//
// MessageId: BE_ISCSI_INIT_FAILURE
//
// MessageText:
//
//  The BladeEngine iSCSI device driver failed to initialize.
//
#define BE_ISCSI_INIT_FAILURE            ((SA_STATUS)0xE0070300L)

//
// MessageId: BE_ISCSI_INVALID_STATE
//
// MessageText:
//
//  The iSCSI Session State is invalid.
//
#define BE_ISCSI_INVALID_STATE           ((SA_STATUS)0xE0070301L)

//
// MessageId: BE_ISCSI_WRE_RESOURCES
//
// MessageText:
//
//  Out of WRE resources.
//
#define BE_ISCSI_WRE_RESOURCES           ((SA_STATUS)0xE0070302L)

//
// MessageId: BE_ISCSI_LOGIN_FAILURE
//
// MessageText:
//
//  iSCSI login failure.
//
#define BE_ISCSI_LOGIN_FAILURE           ((SA_STATUS)0xE0070303L)

//
// MessageId: BE_ISCSI_PDU_TIMEOUT
//
// MessageText:
//
//  iSCSI CMD PDU exceeded the SCSI timeout.
//
#define BE_ISCSI_PDU_TIMEOUT             ((SA_STATUS)0xE0070304L)

//
// MessageId: BE_ISCSI_PROTOCOL_ERR
//
// MessageText:
//
//  Invalid header format or other protocol error detected in the data stream.
//
#define BE_ISCSI_PROTOCOL_ERR            ((SA_STATUS)0xE0070305L)

//
// MessageId: BE_ISCSI_UNDEFINED
//
// MessageText:
//
//  The requested operation is undefined.
//
#define BE_ISCSI_UNDEFINED               ((SA_STATUS)0xE0070306L)

// ***************************************************************************
//                     S H A R E D   S T A T U S
// ***************************************************************************
//
// MessageId: BE_MY_INTERRUPT
//
// MessageText:
//
//  This is my interrupt.
//
#define BE_MY_INTERRUPT                  ((SA_STATUS)0x20070400L)

//
// MessageId: BE_NOT_MY_INTERRUPT
//
// MessageText:
//
//  This is not my interrupt.
//
#define BE_NOT_MY_INTERRUPT              ((SA_STATUS)0x20070401L)

//
// MessageId: BE_STATUS_VBD_INVALID_VERSION
//
// MessageText:
//
//  The device driver is not compatible with this version of the VBD.
//
#define BE_STATUS_VBD_INVALID_VERSION    ((SA_STATUS)0xE0070402L)

//
// MessageId: BE_STATUS_DOMAIN_DENIED
//
// MessageText:
//
//  The operation failed to complete due to insufficient access rights for the requesting domain.
//
#define BE_STATUS_DOMAIN_DENIED          ((SA_STATUS)0xE0070403L)

//
// MessageId: BE_STATUS_VLAN_TAG_NOT_FOUND
//
// MessageText:
//
//  The vlan tag was not found in the rxf table.
//
#define BE_STATUS_VLAN_TAG_NOT_FOUND     ((SA_STATUS)0xE0070404L)

//
// MessageId: BE_STATUS_TOO_MANY_VLAN_TAGS
//
// MessageText:
//
//  The maximum number of VLan tags have already been registered with the device.
//
#define BE_STATUS_TOO_MANY_VLAN_TAGS     ((SA_STATUS)0xE0070405L)

//
// MessageId: BE_STATUS_IPSEC_PAIR_NOT_FOUND
//
// MessageText:
//
//  The ipsec pair was not found in the rxf table.
//
#define BE_STATUS_IPSEC_PAIR_NOT_FOUND   ((SA_STATUS)0xE0070406L)

//
// MessageId: BE_STATUS_TOO_MANY_IPSEC_PAIRS
//
// MessageText:
//
//  The maximum number of ipsec pairs have already been registered with the device.
//
#define BE_STATUS_TOO_MANY_IPSEC_PAIRS   ((SA_STATUS)0xE0070407L)

//
// MessageId: BE_STATUS_TCP_INIT_FAILED
//
// MessageText:
//
//  The embedded TCP/IP stack failed to start.
//
#define BE_STATUS_TCP_INIT_FAILED        ((SA_STATUS)0xE0070408L)

//
// MessageId: BE_STATUS_TCP_NOT_STARTED
//
// MessageText:
//
//  The embedded TCP/IP stack has not been started.
//
#define BE_STATUS_TCP_NOT_STARTED        ((SA_STATUS)0xE0070409L)

//
// MessageId: BE_STATUS_ERX_FRAG_SIZE_FIXED
//
// MessageText:
//
//  The erx frag size can not be changed at this time.
//
#define BE_STATUS_ERX_FRAG_SIZE_FIXED    ((SA_STATUS)0x20070410L)

//
// MessageId: BE_STATUS_MR_INVALID_LENGTH
//
// MessageText:
//
//  The length given for a memory region is invalid or zero.
//
#define BE_STATUS_MR_INVALID_LENGTH      ((SA_STATUS)0xE0070411L)

//
// MessageId: BE_STATUS_MR_INVALID_STATE
//
// MessageText:
//
//  The operation cannot be performed on the memory region given the current state of the memory region.
//
#define BE_STATUS_MR_INVALID_STATE       ((SA_STATUS)0xE0070412L)

//
// MessageId: BE_STATUS_MR_INVALID_STAG
//
// MessageText:
//
//  The stag is not valid.
//
#define BE_STATUS_MR_INVALID_STAG        ((SA_STATUS)0xE0070413L)


//
// MessageId: BE_STATUS_
//
// MessageText:
//
//  xxxxxxxxx.
//
#define BE_STATUS_NO_MCC_WRB                 ((SA_STATUS)0xE0070414L)

#define BE_STATUS_NO_FREE_PAGE_TABLES        ((SA_STATUS)0xE0070415L)



#endif // _BESTATUS_

