/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   bsd_tcp_api_hl.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/bsd_tcp_api_hl.h,v 1.2 2006/02/10 18:02:46 menz Exp $
*/
#ifndef __bsd_tcp_api_hl_h__
#define __bsd_tcp_api_hl_h__

SA_CPP_HEADER

typedef struct _NT_ADAPTER_CONTEXT  *PNT_ADAPTER_CONTEXT;
struct sockaddr;

SOCKET
socket(
    IN int af,
    IN int type,
    IN int protocol
    );


int
accept(
    IN SOCKET s,
    OUT struct sockaddr BE_FAR * addr,
    IN OUT int BE_FAR * addrlen,
    IN OUT struct socket** NewSocket
    );

int
bind(
    IN SOCKET s,
    IN struct sockaddr * name,
    IN int namelen
    );

int
listen(
    IN SOCKET s,
    IN int backlog
    );

int
closesocket(
    IN SOCKET s
    );

int
connect(
    IN SOCKET s,
    IN const struct sockaddr BE_FAR * name,
    IN int namelen
    );

int
send( SOCKET, struct iovec* BufferDescriptors, 
      u32 BufferCount, 
      int    flags,
      u32* BytesSent );

int
sendto( struct socket* Socket, void* buf, size_t len, int flags,
        const struct sockaddr* to, unsigned int tolen,
        u32* BytesSent );

int
recv( SOCKET, struct iovec* BufferDescriptors, 
      u32 BufferCount, 
      int    flags,
      u32* BytesRecv );

int
recvfrom( struct socket* Socket, void *buf, size_t len, int flags, 
          struct sockaddr* from, unsigned int* fromlenaddr,
          u32* BytesRecv );

int setsockopt( SOCKET, int level, int name, 
                const void* val, int valsize );

int 
getsockopt( struct socket* s, int level, int name, void* val, unsigned int* avalsize );

int
getsockname( struct socket* so, struct sockaddr* asa, unsigned int* alen );

int
sys_getpeername( struct socket* so, struct sockaddr* asa, unsigned int* alen );

PNT_ADAPTER_CONTEXT 
AdapterAttach(
    IN      PVOID           ClientContext,
    IN      const char*     AdapterName,
    IN      PCHIPIF_UPCALL  Upcall,
    IN      PSA_MAC_ADDRESS    MacAddress     );

void
AdapterDetach(
    IN      PNT_ADAPTER_CONTEXT     AdapterContext );

void 
AdapterIndicatePacket(
    IN      PNT_ADAPTER_CONTEXT AdapterContext,
    IN      PVOID               Packet,
    IN      u32              PacketLength );

void
AdapterSendComplete(
    IN      PVOID                   Packet );

int 
BsdTcpAddressAssign(
    IN      const char* AdapterName,
    IN      u32      IpAddress,
    IN      u32      SubnetMask,
    IN      u32      Gateway );

BESTATUS
BsdTcpErrorToBeStatus(
    IN      int       Error );

SOCKET
BsdTcpSocket(void);

SOCKET 
BsdSocketGeneric(
    int af,
    int type,
    int protocol );

BESTATUS
BsdTcpSocketClose(
    IN      SOCKET      Socket,
    IN      boolean     Abort );

BESTATUS
BsdTcpSocketBind(
    IN      SOCKET          Socket,
    IN      u32          IpAddress,
    IN      u16          Port );

BESTATUS
BsdTcpSocketConnect(
    IN      SOCKET      Socket,
    IN      u32      IpAddress,
    IN      u16      Port );

BESTATUS
BsdTcpSocketSend(
    IN      SOCKET          Socket,
    IN      struct iovec*   BufferDescriptors,
    IN      u32          BufferCount,
    IN      u32*         BytesXfered );

BESTATUS
BsdTcpSocketRecv(
    IN      SOCKET          Socket,
    IN      struct iovec*   BufferDescriptors,
    IN      u32          BufferCount,
    IN      u32*         BytesXfered );

BESTATUS
BsdTcpSocketListen(
    IN      SOCKET          Socket,
    IN      u32          Backlog );

BESTATUS
BsdTcpSocketAccept(
    IN      SOCKET          Socket,
    OUT     SOCKET*         NewSocket );

BESTATUS
BsdTcpSocketGetLocalAddress(
    IN      SOCKET          Socket,
    IN      u32*         IpAddress,
    IN      u16*         Port );

BESTATUS
BsdTcpSocketOptionSet(
    IN      SOCKET          Socket,
    IN      u32          Level,
    IN      u32          Name,
    IN      PVOID           Value,
    IN      u32          ValueSize );

BESTATUS
BsdTcpSocketOptionGet(
    IN      SOCKET          Socket,
    IN      u32          Level,
    IN      u32          Name,
    IN      PVOID           Value,
    IN      u32*         ValueSize );

BESTATUS
BsdTcpOffloadBuildTemplateHeader(
    IN      SOCKET                  Socket,
    IN OUT  PTEMPLATE_HEADER   TemplateHeader );


////////////////// ASYNC routines ////////////////////////////

BESTATUS
BsdTcpSocketConnectAsync(
    IN      SOCKET                          Socket,
    IN      u32                          IpAddress,
    IN      u16                          Port,
    IN      BE_SOCKET_COMPLETION_CALLBACK   CompletionRoutine,
    IN      PVOID                           CompletionContext );

BESTATUS
BsdTcpSocketAcceptAsync(
    IN      SOCKET                          Socket,
    IN      BE_SOCKET_COMPLETION_CALLBACK   CompletionRoutine,
    IN      PVOID                           CompletionContext );

BESTATUS
BsdTcpSocketAcceptAsync(
    IN      SOCKET                          Socket,
    IN      BE_SOCKET_COMPLETION_CALLBACK   CompletionRoutine,
    IN      PVOID                           CompletionContext );

BESTATUS
BsdTcpSocketSendAsync(
    IN      SOCKET                          Socket,
    IN      struct iovec*                   BufferDescriptors,
    IN      u32                          BufferCount,
    IN      BE_SOCKET_COMPLETION_CALLBACK   CompletionRoutine,
    IN      PVOID                           CompletionContext );

BESTATUS
BsdTcpSocketRecvAsync(
    IN      SOCKET                          Socket,
    IN      struct iovec*                   BufferDescriptors,
    IN      u32                          BufferCount,
    IN      BE_SOCKET_COMPLETION_CALLBACK   CompletionRoutine,
    IN      PVOID                           CompletionContext );


BESTATUS
BsdTcpBindAdapter(
    IN      PBE_FUNCTION_OBJECT     FunctionObject,
    IN      PSA_MAC_ADDRESS            MacAddress,
    IN      boolean                 Dhcp,
    IN      u32                  IpAddress,
    IN      u32                  NetMask,
    IN      u32                  Gateway    );

BESTATUS
BsdTcpUnbindAdapter(
    IN      PBE_FUNCTION_OBJECT     FunctionObject );

BESTATUS 
BsdTcpUpdateIpSettings(
    IN      PBE_FUNCTION_OBJECT     FunctionObject,
    IN      boolean                 Dhcp,
    IN      u32                  IpAddress,
    IN      u32                  SubnetMask,
    IN      u32                  Gateway );

BESTATUS
BsdTcpIndicatePacket(
    IN      PBE_FUNCTION_OBJECT     FunctionObject,
    IN      PVOID                   Packet,
    IN      u32                  PacketLength );

BESTATUS
BsdTcpSendComplete(
    IN      PBE_FUNCTION_OBJECT     FunctionObject,
    IN      PVOID                   Packet );

void
BsdTcpForwardSegmentComplete(
    IN      PBE_FUNCTION_OBJECT     FunctionObject,
    IN      PVOID                   Segment );

BESTATUS 
BsdTcpIndicateEvent(
    IN      FUNCTIONHANDLE          FunctionObject,
    IN      u32                  Cid,
    IN      TOE_EVENT_CODE          Event );


SA_CPP_TRAILER
#endif // __bsd_tcp_api_hl_h__

