/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#ifndef __ue_tiny_log_h__
#define __ue_tiny_log_h__

typedef enum {

    UE_TLOG_CODE_MIN = 200,
    UE_TLOG_CODE_MAX = 250,

    
    UE_TLOG_SEMAPHORE_TIMEOUT               = 201,
    UE_TLOG_RECOVERY_DURING_LOAD            = 202,
    UE_TLOG_RECOVERY_FAILED_DURING_LOAD     = 203,
    UE_TLOG_SCRATCHPAD_WRITE                = 206,
    UE_TLOG_INIT_RECOVERY                   = 207,
    UE_TLOG_MASTER                          = 208,
    UE_TLOG_RESET_SUCCESS                   = 209,
    UE_TLOG_RESET_FAIL                      = 210,
    UE_TLOG_SLAVE_COOP                      = 211,
    UE_TLOG_RECOVERY_START                  = 212,
    UE_TLOG_SOFTWARE_UE                     = 214,
    UE_TLOG_SOFTWARE_UE_IGNORED             = 215,
    UE_TLOG_FAT_TIMEOUT                     = 216,
    UE_TLOG_IDLE                            = 218,
    UE_TLOG_START                           = 219,
    UE_TLOG_VERIFY_RECOVERY                 = 220,
    UE_TLOG_WAIT_FAT                        = 222,
    UE_TLOG_RESET                           = 223,   
    UE_TLOG_ABORTED                         = 225,
    UE_TLOG_DONE                            = 226,
    UE_TLOG_VERSION_MISMATCH                = 227,
    UE_TLOG_SEMAPHORE_STOLEN                = 228,
    UE_TLOG_STEAL_SEMAPHORE                 = 229,
    UE_TLOG_RESET_ERROR                     = 230,
    UE_TLOG_MASTER_RELINQUISH               = 232,
    UE_TLOG_SLAVE_TIMEOUT                   = 234,
    UE_TLOG_SOFTWARE_UE_DISABLED            = 235,
    UE_TLOG_MASTER_TIMEOUT                  = 236,
    UE_TLOG_POST_RESET                      = 237,
  
} UE_TLOG_CODES;

STATIC INLINE void
be_ue_tlog_add_names (const char* tcodes[], u32 max_codes)
{
    #define UE_TLOG_ASSIGN(_code, _str)     \
        if (_code < max_codes) {    \
            tcodes[_code] = _str;   \
        }                           \

    
    UE_TLOG_ASSIGN (UE_TLOG_SEMAPHORE_TIMEOUT               , "Semaphore timeout.");
    UE_TLOG_ASSIGN (UE_TLOG_RECOVERY_DURING_LOAD            , "Recovery in progress during driver load."); 
    UE_TLOG_ASSIGN (UE_TLOG_RECOVERY_FAILED_DURING_LOAD     , "Recovery failed during driver load."); 
    UE_TLOG_ASSIGN (UE_TLOG_SCRATCHPAD_WRITE                , "Scratchpad write. Value:0x%I64x"); 
    UE_TLOG_ASSIGN (UE_TLOG_INIT_RECOVERY                   , "Recovered during registration.");
    UE_TLOG_ASSIGN (UE_TLOG_MASTER                          , "Elected master.");
    UE_TLOG_ASSIGN (UE_TLOG_RESET_SUCCESS                   , "PCI reset succeeded.");
    UE_TLOG_ASSIGN (UE_TLOG_RESET_FAIL                      , "PCI reset failed.");
    UE_TLOG_ASSIGN (UE_TLOG_SLAVE_COOP                      , "Slave stole master role. PCI Reset skipped. Scratchpad:0x%I64x");
    UE_TLOG_ASSIGN (UE_TLOG_RECOVERY_START                  , "Recovery started. UE:0x%I64x");
    UE_TLOG_ASSIGN (UE_TLOG_SOFTWARE_UE                     , "Software UE. Optional code:0x%I64x");
    UE_TLOG_ASSIGN (UE_TLOG_SOFTWARE_UE_IGNORED             , "Software UE ignored for VM. Optional code:0x%I64x");
    UE_TLOG_ASSIGN (UE_TLOG_FAT_TIMEOUT                     , "Non-fatal FAT timeout.");
    UE_TLOG_ASSIGN (UE_TLOG_IDLE                            , "Recovery State Idle.");
    UE_TLOG_ASSIGN (UE_TLOG_START                           , "Recovery State Start.");
    UE_TLOG_ASSIGN (UE_TLOG_VERIFY_RECOVERY                 , "Recovery State Verify Recovery. Time(us):%I64d");
    UE_TLOG_ASSIGN (UE_TLOG_WAIT_FAT                        , "Recovery State Wait for FAT. Time(us):%I64d");
    UE_TLOG_ASSIGN (UE_TLOG_RESET                           , "Recovery State PCI Reset. Time(us):%I64d");    
    UE_TLOG_ASSIGN (UE_TLOG_POST_RESET                      , "Recovery State After Reset Wait. Time(us):%I64d");
    UE_TLOG_ASSIGN (UE_TLOG_ABORTED                         , "Recovery State Aborted due to failure. Time(us):%I64d");
    UE_TLOG_ASSIGN (UE_TLOG_DONE                            , "Recovery State Success. Time(us):%I64d");
    UE_TLOG_ASSIGN (UE_TLOG_VERSION_MISMATCH                , "Recovery disabled due to version mismatch. Version:0x%I64x");
    UE_TLOG_ASSIGN (UE_TLOG_SEMAPHORE_STOLEN                , "Semaphore stolen. Value:0x%I64x");             
    UE_TLOG_ASSIGN (UE_TLOG_RESET_ERROR                     , "PCI Reset sequence error or timeout.");    
    UE_TLOG_ASSIGN (UE_TLOG_MASTER_RELINQUISH               , "Relinquish master after failure.");
    UE_TLOG_ASSIGN (UE_TLOG_SLAVE_TIMEOUT                   , "Recovery State Slave Wait. Time(us):%I64d");
    UE_TLOG_ASSIGN (UE_TLOG_MASTER_TIMEOUT                  , "Recovery State Master Wait. Time(us):%I64d");
    UE_TLOG_ASSIGN (UE_TLOG_SOFTWARE_UE_DISABLED            , "Software UE disabled. Optional code:0x%I64x");
    

}

#endif

