/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
// Autogenerated file. Do not edit!
// srcgen version: 0146
#ifndef __headers_enum_h__
#define __headers_enum_h__
#include "setypes.h"
#include "defines_enum.h"


#ifndef SG_STATIC_INLINE
   #define SG_STATIC_INLINE static inline
#endif

// --- ETH_TYPE_ENUM ---
// Ethernet type field. 
#define ETH_TYPE_MAX_LENGTH             (1500)    // Maximum length for IEEE802.3 Field.
#define ETH_TYPE_IP                     (2048)    // DOD Internet Protocol (IP)
#define ETH_TYPE_ARP                    (2054)    // Address Resolution Protocol (ARP)
#define ETH_TYPE_RARP                   (32821)   // Reverse Address Resolution Protocol
                                                  // (RARP)
#define ETH_TYPE_VLAN                   (33024)   
#define ETH_TYPE_IPX                    (33079)   // Novell (old) NetWare IPX (ECONFIG E
                                                  // option)
#define ETH_TYPE_LLDP                   (35020)   // Link Layer Discovery Protocol
#define ETH_TYPE_FCoE                   (35078)   // Fibre Channel Over Ethernet
#define ETH_TYPE_FIP                    (35092)   // FCoE Initialization Protocol

typedef enum {
   ENUM_ETH_TYPE_MAX_LENGTH = 0x5dcUL,
   ENUM_ETH_TYPE_IP = 0x800UL,
   ENUM_ETH_TYPE_ARP = 0x806UL,
   ENUM_ETH_TYPE_RARP = 0x8035UL,
   ENUM_ETH_TYPE_VLAN = 0x8100UL,
   ENUM_ETH_TYPE_IPX = 0x8137UL,
   ENUM_ETH_TYPE_LLDP = 0x88ccUL,
   ENUM_ETH_TYPE_FCoE = 0x8906UL,
   ENUM_ETH_TYPE_FIP = 0x8914UL
} ETH_TYPE_ENUM;

// --- IP_VERSION_ENUM ---
#define IPV4                            (4)       
#define IPV6                            (6)       

typedef enum {
   ENUM_IPV4 = 0x4UL,
   ENUM_IPV6 = 0x6UL
} IP_VERSION_ENUM;

// --- IP_PROTOCOL_ENUM ---
#define IP_PROTOCOL_TCP                 (6)       
#define IP_PROTOCOL_UDP                 (17)      

typedef enum {
   ENUM_IP_PROTOCOL_TCP = 0x6UL,
   ENUM_IP_PROTOCOL_UDP = 0x11UL
} IP_PROTOCOL_ENUM;

// --- TCP_STATE ---
#define TCPS_CLOSED                     (0)       // Closed
#define TCPS_LISTEN                     (1)       // listening for connection
#define TCPS_SYN_SENT                   (2)       // have sent SYN
#define TCPS_SYN_RECEIVED               (3)       // have sent and received SYN, waiting
                                                  // for ACK
#define TCPS_ESTABLISHED                (4)       // established
#define TCPS_CLOSE_WAIT                 (5)       // received FIN, waiting for application
                                                  // close
#define TCPS_FIN_WAIT_1                 (6)       // have closed, sent FIN, waiting for
                                                  // ACK and FIN
#define TCPS_CLOSING                    (7)       // simultaneous close, waiting for ACK
#define TCPS_LAST_ACK                   (8)       // received FIN have closed, waiting for
                                                  // ACK
#define TCPS_FIN_WAIT_2                 (9)       // have closed, awaiting FIN
#define TCPS_TIME_WAIT                  (10)      // 2MSL, wait state after active close
#define TCP_NSTATES                     (11)      // Number of TCP states

typedef enum {
   ENUM_TCPS_CLOSED = 0x0UL,
   ENUM_TCPS_LISTEN = 0x1UL,
   ENUM_TCPS_SYN_SENT = 0x2UL,
   ENUM_TCPS_SYN_RECEIVED = 0x3UL,
   ENUM_TCPS_ESTABLISHED = 0x4UL,
   ENUM_TCPS_CLOSE_WAIT = 0x5UL,
   ENUM_TCPS_FIN_WAIT_1 = 0x6UL,
   ENUM_TCPS_CLOSING = 0x7UL,
   ENUM_TCPS_LAST_ACK = 0x8UL,
   ENUM_TCPS_FIN_WAIT_2 = 0x9UL,
   ENUM_TCPS_TIME_WAIT = 0xaUL,
   ENUM_TCP_NSTATES = 0xbUL
} TCP_STATE;

// --- TH_FLAGS ---
#define TH_FIN                          (1)       
#define TH_SYN                          (2)       
#define TH_RST                          (4)       
#define TH_PUSH                         (8)       
#define TH_ACK                          (16)      
#define TH_URG                          (32)      

typedef enum {
   ENUM_TH_FIN = 0x1UL,
   ENUM_TH_SYN = 0x2UL,
   ENUM_TH_RST = 0x4UL,
   ENUM_TH_PUSH = 0x8UL,
   ENUM_TH_ACK = 0x10UL,
   ENUM_TH_URG = 0x20UL
} TH_FLAGS;

// --- TCP_FLAGS ---
#define TF_SENTFIN                      (1)       // Have sent FIN
#define TF_USE_TSTMP                    (2)       // Use timestamp option
#define TF_FLUSH_PENDING                (4)       // TxULP needs to schedule a flush
#define TF_RXMT_PENDING                 (8)       // Retransmit bits needs to be set for
                                                  // TCP scheduler
#define TF_IN_RXMT                      (16)      // We are in retransmit
#define TF_DELEGATED_DATA               (32)      // Firmware needs to handle delegated
                                                  // data
#define TF_SACK_ENABLE                  (64)      // Set when SACK is negotiated on a
                                                  // connection
#define TF_TPOST_UPLOAD_DONE            (128)     // Upon reception of flush completions,
                                                  // if TF_TPRE_UPLOAD_DONE is not set in
                                                  // rx_flags, the driver should poll
                                                  // t_flags for TF_TPOST_UPLOAD_DONE
                                                  // before it can consider the upload
                                                  // complete 
#define TF_SEND_PROBE                   (256)     // Indicates that txulp should initiate
                                                  // a window probe 
#define TF_SENT_PROBE                   (512)     // Indicates that a window probe has
                                                  // been sent 
#define TF_TPOST_FLUSH_RCVD             (1024)    // Indicates that tpost received the
                                                  // upload flush from txulp 
#define TF_TPOST_FLUSH_DONE             (2048)    // Indicates that tpost is done with a
                                                  // flush 
#define TF_TX_LAST                      (2048)    // No space in t_flags for any more
                                                  // flags

typedef enum {
   ENUM_TF_SENTFIN = 0x1UL,
   ENUM_TF_USE_TSTMP = 0x2UL,
   ENUM_TF_FLUSH_PENDING = 0x4UL,
   ENUM_TF_RXMT_PENDING = 0x8UL,
   ENUM_TF_IN_RXMT = 0x10UL,
   ENUM_TF_DELEGATED_DATA = 0x20UL,
   ENUM_TF_SACK_ENABLE = 0x40UL,
   ENUM_TF_TPOST_UPLOAD_DONE = 0x80UL,
   ENUM_TF_SEND_PROBE = 0x100UL,
   ENUM_TF_SENT_PROBE = 0x200UL,
   ENUM_TF_TPOST_FLUSH_RCVD = 0x400UL,
   ENUM_TF_TPOST_FLUSH_DONE = 0x800UL,
   ENUM_TF_TX_LAST = 0x800UL
} TCP_FLAGS;

// --- TCP_RX_FLAGS ---
#define TF_RXMT                         (1)       // Received retransmit IPC from TPRE
#define TF_FASTRXMT                     (2)       // Received fast retransmit IPC from
                                                  // TPRE
#define TF_OOO_PENDING                  (4)       // Receive out of order pending
#define TF_OOO_FIN                      (8)       // Receive out of order finish
#define TF_DROP_OOR_ACK                 (16)      // Firmware does not ack an unacceptable
                                                  // ack
#define TF_TT_REPLAY                    (32)      // Indicates that the TT replay is
                                                  // ongoing
#define TF_SACK_SEEN                    (64)      // SACK blocks are seen on this
                                                  // connection. Process the packet in
                                                  // slow path.
#define TF_TPRE_UPLOAD_DONE             (128)     // Upon reception of flush completions,
                                                  // if TF_TPRE_UPLOAD_DONE is not set in
                                                  // rx_flags, the driver should poll
                                                  // t_flags for TF_TPOST_UPLOAD_DONE
                                                  // before it can consider the upload
                                                  // complete 
#define TF_RX_LAST                      (128)     // No space in rx_flags for any more
                                                  // flags

typedef enum {
   ENUM_TF_RXMT = 0x1UL,
   ENUM_TF_FASTRXMT = 0x2UL,
   ENUM_TF_OOO_PENDING = 0x4UL,
   ENUM_TF_OOO_FIN = 0x8UL,
   ENUM_TF_DROP_OOR_ACK = 0x10UL,
   ENUM_TF_TT_REPLAY = 0x20UL,
   ENUM_TF_SACK_SEEN = 0x40UL,
   ENUM_TF_TPRE_UPLOAD_DONE = 0x80UL,
   ENUM_TF_RX_LAST = 0x80UL
} TCP_RX_FLAGS;

// --- TCP_TIMER_IDS ---
#define TCPT_REXMT                      (1)       // Retransmit
#define TCPT_PERSIST                    (2)       // Persist
#define TCPT_KEEP                       (3)       // Keep Alive
#define TCPT_DELACK                     (4)       // Delayed ACK

typedef enum {
   ENUM_TCPT_REXMT = 0x1UL,
   ENUM_TCPT_PERSIST = 0x2UL,
   ENUM_TCPT_KEEP = 0x3UL,
   ENUM_TCPT_DELACK = 0x4UL
} TCP_TIMER_IDS;

// --- BE_TCPOPT ---
// TCP Option Values 
#define BE_TCPOPT_EOL                   (0)       
#define BE_TCPOPT_NOP                   (1)       
#define BE_TCPOPT_MAXSEG                (2)       
#define BE_TCPOPT_WINDOW                (3)       
#define BE_TCPOPT_SACKPERMITTED         (4)       
#define BE_TCPOPT_SACK                  (5)       
#define BE_TCPOPT_TIMESTAMP             (8)       

typedef enum {
   ENUM_BE_TCPOPT_EOL = 0x0UL,
   ENUM_BE_TCPOPT_NOP = 0x1UL,
   ENUM_BE_TCPOPT_MAXSEG = 0x2UL,
   ENUM_BE_TCPOPT_WINDOW = 0x3UL,
   ENUM_BE_TCPOPT_SACKPERMITTED = 0x4UL,
   ENUM_BE_TCPOPT_SACK = 0x5UL,
   ENUM_BE_TCPOPT_TIMESTAMP = 0x8UL
} BE_TCPOPT;

// --- BE_TCPOLEN ---
// TCP Option lengths 
#define BE_TCPOLEN_SACKPERMITTED        (2)       
#define BE_TCPOLEN_WINDOW               (3)       
#define BE_TCPOLEN_MAXSEG               (4)       
#define BE_TCPOLEN_SACK_SINGLE          (10)      
#define BE_TCPOLEN_TIMESTAMP            (10)      

typedef enum {
   ENUM_BE_TCPOLEN_SACKPERMITTED = 0x2UL,
   ENUM_BE_TCPOLEN_WINDOW = 0x3UL,
   ENUM_BE_TCPOLEN_MAXSEG = 0x4UL,
   ENUM_BE_TCPOLEN_SACK_SINGLE = 0xaUL,
   ENUM_BE_TCPOLEN_TIMESTAMP = 0xaUL
} BE_TCPOLEN;

// --- ISCSI_OPCODE ---
// ISCSI Message Opcode Definitions Vendor Specific Initiator 0x1c-0x1e Vendor Specific
// Target 0x3c-0x3e 
#define ISCSI_OPCODE_NOP_OUT            (0)       
#define ISCSI_OPCODE_SCSI_CMD           (1)       
#define ISCSI_OPCODE_SCSI_TASK_MGMT     (2)       
#define ISCSI_OPCODE_LOGIN              (3)       
#define ISCSI_OPCODE_TEXT               (4)       
#define ISCSI_OPCODE_SCSI_DATA_OUT      (5)       
#define ISCSI_OPCODE_LOGOUT             (6)       
#define ISCSI_OPCODE_SNACK              (16)      
#define ISCSI_OPCODE_NOP_IN             (32)      
#define ISCSI_OPCODE_NOPIN              (32)      
#define ISCSI_OPCODE_SCSI_RESPONSE      (33)      
#define ISCSI_OPCODE_RESPONSE           (33)      
#define ISCSI_OPCODE_SCSI_TASK_MGMT_RESPONSE (34) 
#define ISCSI_OPCODE_LOGIN_RESPONSE     (35)      
#define ISCSI_OPCODE_TEXT_RESPONSE      (36)      
#define ISCSI_OPCODE_SCSI_DATA_IN       (37)      
#define ISCSI_OPCODE_DATAIN             (37)      
#define ISCSI_OPCODE_LOGOUT_RESPONSE    (38)      
#define ISCSI_OPCODE_R2T                (49)      
#define ISCSI_OPCODE_RTT                (49)      
#define ISCSI_OPCODE_ASYNC_MSG          (50)      
#define ISCSI_OPCODE_REJECT             (63)      

typedef enum {
   ENUM_ISCSI_OPCODE_NOP_OUT = 0x0UL,
   ENUM_ISCSI_OPCODE_SCSI_CMD = 0x1UL,
   ENUM_ISCSI_OPCODE_SCSI_TASK_MGMT = 0x2UL,
   ENUM_ISCSI_OPCODE_LOGIN = 0x3UL,
   ENUM_ISCSI_OPCODE_TEXT = 0x4UL,
   ENUM_ISCSI_OPCODE_SCSI_DATA_OUT = 0x5UL,
   ENUM_ISCSI_OPCODE_LOGOUT = 0x6UL,
   ENUM_ISCSI_OPCODE_SNACK = 0x10UL,
   ENUM_ISCSI_OPCODE_NOP_IN = 0x20UL,
   ENUM_ISCSI_OPCODE_NOPIN = 0x20UL,
   ENUM_ISCSI_OPCODE_SCSI_RESPONSE = 0x21UL,
   ENUM_ISCSI_OPCODE_RESPONSE = 0x21UL,
   ENUM_ISCSI_OPCODE_SCSI_TASK_MGMT_RESPONSE = 0x22UL,
   ENUM_ISCSI_OPCODE_LOGIN_RESPONSE = 0x23UL,
   ENUM_ISCSI_OPCODE_TEXT_RESPONSE = 0x24UL,
   ENUM_ISCSI_OPCODE_SCSI_DATA_IN = 0x25UL,
   ENUM_ISCSI_OPCODE_DATAIN = 0x25UL,
   ENUM_ISCSI_OPCODE_LOGOUT_RESPONSE = 0x26UL,
   ENUM_ISCSI_OPCODE_R2T = 0x31UL,
   ENUM_ISCSI_OPCODE_RTT = 0x31UL,
   ENUM_ISCSI_OPCODE_ASYNC_MSG = 0x32UL,
   ENUM_ISCSI_OPCODE_REJECT = 0x3fUL
} ISCSI_OPCODE;

// --- FCOE_SOF ---
// FCoE encodings for SOF
#define FCOE_SOFf                       (40)      
#define FCOE_SOFi2                      (45)      
#define FCOE_SOFi3                      (46)      
#define FCOE_SOFn2                      (53)      
#define FCOE_SOFn3                      (54)      

typedef enum {
   ENUM_FCOE_SOFf = 0x28UL,
   ENUM_FCOE_SOFi2 = 0x2dUL,
   ENUM_FCOE_SOFi3 = 0x2eUL,
   ENUM_FCOE_SOFn2 = 0x35UL,
   ENUM_FCOE_SOFn3 = 0x36UL
} FCOE_SOF;

// --- FCOE_EOF ---
// FCoE encodings for EOF
#define FCOE_EOFn                       (65)      
#define FCOE_EOFt                       (66)      
#define FCOE_EOFni                      (73)      
#define FCOE_EOFa                       (80)      

typedef enum {
   ENUM_FCOE_EOFn = 0x41UL,
   ENUM_FCOE_EOFt = 0x42UL,
   ENUM_FCOE_EOFni = 0x49UL,
   ENUM_FCOE_EOFa = 0x50UL
} FCOE_EOF;

// --- FC_RESP_CODE_ENUM ---
// FCP response codes
#define TASK_MGMT_FUN_COMPLETE          (0)       // Task Management function complete
#define FCP_BURST_LEN_MISMATCH          (1)       // FCP_DATA length different than
                                                  // FCP_BURST_LEN
#define FCP_CMND_INVALID                (2)       // FCP_CMND fields invalid
#define FCP_PARAM_MISMATCH              (3)       // FCP_DATA parameter mismatch with
                                                  // FCP_DATA_RO
#define TASK_MGMT_REJECTED              (4)       // Task Management function rejected
#define TASK_MGMT_FAILED                (5)       // Task Management function failed
#define TASK_MGMT_BAD_LUN               (9)       // Task Management function incorrect
                                                  // logical unit number

typedef enum {
   ENUM_TASK_MGMT_FUN_COMPLETE = 0x0UL,
   ENUM_FCP_BURST_LEN_MISMATCH = 0x1UL,
   ENUM_FCP_CMND_INVALID = 0x2UL,
   ENUM_FCP_PARAM_MISMATCH = 0x3UL,
   ENUM_TASK_MGMT_REJECTED = 0x4UL,
   ENUM_TASK_MGMT_FAILED = 0x5UL,
   ENUM_TASK_MGMT_BAD_LUN = 0x9UL
} FC_RESP_CODE_ENUM;

// --- FC_ELS_COMMAND_CODE ---
// FC-LS-2 ELS command codes
#define FC_ELS_CMD_LS_RJT               (1)       // Link Service Reject
#define FC_ELS_CMD_LS_ACC               (2)       // Link Service Accept
#define FC_ELS_CMD_RRQ                  (18)      // Reinstate Recovery Qualifier Request
                                                  // Sequence
#define FC_ELS_CMD_REC                  (19)      // Read Exchange Concise Request
                                                  // Sequence

typedef enum {
   ENUM_FC_ELS_CMD_LS_RJT = 0x1UL,
   ENUM_FC_ELS_CMD_LS_ACC = 0x2UL,
   ENUM_FC_ELS_CMD_RRQ = 0x12UL,
   ENUM_FC_ELS_CMD_REC = 0x13UL
} FC_ELS_COMMAND_CODE;

// --- FC_FC4_COMMAND_CODE ---
// FCP-4 Link Services command codes
#define FC_FCP_LS_RJT                   (1)       // Link Service Reject
#define FC_FCP_LS_ACC                   (2)       // Link Service Accept
#define FC_FCP_LS_SRR                   (20)      // Sequence Retransmission Request

typedef enum {
   ENUM_FC_FCP_LS_RJT = 0x1UL,
   ENUM_FC_FCP_LS_ACC = 0x2UL,
   ENUM_FC_FCP_LS_SRR = 0x14UL
} FC_FC4_COMMAND_CODE;

// --- FC_RJT_REASON_CODE ---
// FC-FS-3/FC-LS-2/FCP-3 link services reject reason codes.
#define FC_RJT_INVALID_COMMAND_CODE     (1)       // The Command code in the Sequence
                                                  // being rejected is invalid.
#define FC_RJT_LOGICAL_ERROR            (3)       // The request identified by the Command
                                                  // code is invalid or logically
                                                  // inconsistent for the conditions
                                                  // present.
#define FC_RJT_LOGICAL_BUSY             (5)       // The Basic Link Service is logically
                                                  // busy and unable to process the
                                                  // request at this time.
#define FC_RJT_PROTOCOL_ERROR           (7)       // This indicates that an error has been
                                                  // detected that violates the rules of
                                                  // FC-2 protocol that are not specified
                                                  // by other error codes.
#define FC_RJT_UNABLE_TO_PERFORM_COMMAND_REQUEST (9) // The Recipient of a Link Service
                                                     // command is unable to perform the
                                                     // request at this time.
#define FC_RJT_COMMAND_NOT_SUPPORTED    (11)      // The Recipient of a Link Service
                                                  // command does not support the command
                                                  // requested.
#define FC_RJT_COMMAND_ALREADY_IN_PROGRESS (14)   // Command is already in progress.
#define FC_RJT_VENDOR_SPECIFIC_ERROR    (255)     // The Vendor specific error bits may be
                                                  // used to specify vendor unique reason
                                                  // codes.

typedef enum {
   ENUM_FC_RJT_INVALID_COMMAND_CODE = 0x1UL,
   ENUM_FC_RJT_LOGICAL_ERROR = 0x3UL,
   ENUM_FC_RJT_LOGICAL_BUSY = 0x5UL,
   ENUM_FC_RJT_PROTOCOL_ERROR = 0x7UL,
   ENUM_FC_RJT_UNABLE_TO_PERFORM_COMMAND_REQUEST = 0x9UL,
   ENUM_FC_RJT_COMMAND_NOT_SUPPORTED = 0xbUL,
   ENUM_FC_RJT_COMMAND_ALREADY_IN_PROGRESS = 0xeUL,
   ENUM_FC_RJT_VENDOR_SPECIFIC_ERROR = 0xffUL
} FC_RJT_REASON_CODE;

// --- FC_BA_RJT_REASON_CODE_EXPLANATION ---
// FC-FS-3 BA_RJT basic link services reason code explanations
#define FC_BA_RJT_NO_ADDL_EXPL          (0)       // No additional explanation
#define FC_BA_RJT_INV_ID_COMBO          (3)       // Invalid OX_ID-RX_ID combination
#define FC_BA_RJT_SEQ_ABORTED           (5)       // Sequence aborted, no sequence
                                                  // information provided

typedef enum {
   ENUM_FC_BA_RJT_NO_ADDL_EXPL = 0x0UL,
   ENUM_FC_BA_RJT_INV_ID_COMBO = 0x3UL,
   ENUM_FC_BA_RJT_SEQ_ABORTED = 0x5UL
} FC_BA_RJT_REASON_CODE_EXPLANATION;

// --- FC_LS_RJT_REASON_CODE_EXPLANATION ---
// FC-LS-2/FCP-3 extended link services reject reason code explanations
#define FC_RJT_NO_ADDL_EXPL             (0)       // No additional explanation
#define FC_RJT_SVC_PARM_ERR_OPTIONS     (1)       // Service Parm error - Options
#define FC_RJT_SVC_PARM_ERR_INIT_CTL    (3)       // Service Parm error - Initiator Ctl
#define FC_RJT_SVC_PARM_ERR_RCPNT_CTL   (5)       // Service Parm error - Recipient Ctl
#define FC_RJT_SVC_PARM_ERR_FLD_SIZE    (7)       // Service Parm error - Receive
                                                  // Data_Field Size
#define FC_RJT_SVC_PARM_ERR_CONC_SEQ    (9)       // Service Parm error - Concurrent Seq
#define FC_RJT_SVC_PARM_ERR_CREDIT      (11)      // Service Parm error - Credit
#define FC_RJT_INV_N_F_PORT_NAME        (13)      // Invalid N_Port/F_Port_Name
#define FC_RJT_INV_NODE_FABRIC_NAME     (14)      // Invalid node/Fabric Name
#define FC_RJT_INV_SRV_PARMS            (15)      // Invalid Common Service Parameters
#define FC_RJT_INV_ASSOC_HDR            (17)      // Invalid Association_Header
#define FC_RJT_ASSOC_HDR_REQ            (19)      // Association_Header required
#define FC_RJT_INV_ORIG_S_ID            (21)      // Invalid Originator S_ID
#define FC_RJT_INV_ID_COMBO             (23)      // Invalid OX_ID-RX_ID combination
#define FC_RJT_CMD_IN_PROGRESS          (25)      // Command (request) already in progress
#define FC_RJT_N_PORT_LOGIN_REQ         (30)      // N_Port Login required
#define FC_RJT_INV_N_PORT_ID            (31)      // Invalid N_Port_ID
#define FC_RJT_INSUFF_RESOURCES         (41)      // Insufficient resources to support
                                                  // Login
#define FC_RJT_CANT_SUPPLY_DATA         (42)      // Unable to supply requested data
#define FC_RJT_RQST_UNSUPPORTED         (44)      // Request not supported
#define FC_RJT_INV_PYLD_LEN             (45)      // Invalid Payload length
#define FC_RJT_NO_ALIAS_IDS_AVAIL       (48)      // No Alias_IDs available for this
                                                  // Alias_ID Type
#define FC_RJT_ALIAS_NOT_ACT_NO_RES     (49)      // Alias_ID not activated (no resources
                                                  // available)
#define FC_RJT_ALIAS_NOT_ACT_INV        (50)      // Alias_ID not activated (invalid
                                                  // Alias_ID)
#define FC_RJT_ALIAS_NOT_DEACT_INV      (51)      // Alias_ID not deactivated (doesn't
                                                  // exist)
#define FC_RJT_ALIAS_NOT_DEACT_RES      (52)      // Alias_ID not deactivated (resource
                                                  // problem)
#define FC_RJT_SVC_PARM_CONFLICT        (53)      // Service Parameter conflict
#define FC_RJT_INV_ALIAS_TOKEN          (54)      // Invalid Alias_Token
#define FC_RJT_UNSUPP_ALIAS_TOKEN       (55)      // Unsupported Alias_Token
#define FC_RJT_CANT_FORM_ALIAS_GRP      (56)      // Alias Group cannot be formed
#define FC_RJT_INV_PORT_NODE_NAME       (68)      // Invalid Port/Node_Name
#define FC_RJT_LOGIN_EXT_UNSUPP         (70)      // Login Extension not supported
#define FC_RJT_AUTH_REQ                 (72)      // Authentication required (see FC-SP)
#define FC_RJT_SCAN_VAL_NOT_ALLOWED     (80)      // Periodic Scan Value not allowed
#define FC_RJT_SANNING_UNSUPP           (81)      // Periodic Scanning not supported

typedef enum {
   ENUM_FC_RJT_NO_ADDL_EXPL = 0x0UL,
   ENUM_FC_RJT_SVC_PARM_ERR_OPTIONS = 0x1UL,
   ENUM_FC_RJT_SVC_PARM_ERR_INIT_CTL = 0x3UL,
   ENUM_FC_RJT_SVC_PARM_ERR_RCPNT_CTL = 0x5UL,
   ENUM_FC_RJT_SVC_PARM_ERR_FLD_SIZE = 0x7UL,
   ENUM_FC_RJT_SVC_PARM_ERR_CONC_SEQ = 0x9UL,
   ENUM_FC_RJT_SVC_PARM_ERR_CREDIT = 0xbUL,
   ENUM_FC_RJT_INV_N_F_PORT_NAME = 0xdUL,
   ENUM_FC_RJT_INV_NODE_FABRIC_NAME = 0xeUL,
   ENUM_FC_RJT_INV_SRV_PARMS = 0xfUL,
   ENUM_FC_RJT_INV_ASSOC_HDR = 0x11UL,
   ENUM_FC_RJT_ASSOC_HDR_REQ = 0x13UL,
   ENUM_FC_RJT_INV_ORIG_S_ID = 0x15UL,
   ENUM_FC_RJT_INV_ID_COMBO = 0x17UL,
   ENUM_FC_RJT_CMD_IN_PROGRESS = 0x19UL,
   ENUM_FC_RJT_N_PORT_LOGIN_REQ = 0x1eUL,
   ENUM_FC_RJT_INV_N_PORT_ID = 0x1fUL,
   ENUM_FC_RJT_INSUFF_RESOURCES = 0x29UL,
   ENUM_FC_RJT_CANT_SUPPLY_DATA = 0x2aUL,
   ENUM_FC_RJT_RQST_UNSUPPORTED = 0x2cUL,
   ENUM_FC_RJT_INV_PYLD_LEN = 0x2dUL,
   ENUM_FC_RJT_NO_ALIAS_IDS_AVAIL = 0x30UL,
   ENUM_FC_RJT_ALIAS_NOT_ACT_NO_RES = 0x31UL,
   ENUM_FC_RJT_ALIAS_NOT_ACT_INV = 0x32UL,
   ENUM_FC_RJT_ALIAS_NOT_DEACT_INV = 0x33UL,
   ENUM_FC_RJT_ALIAS_NOT_DEACT_RES = 0x34UL,
   ENUM_FC_RJT_SVC_PARM_CONFLICT = 0x35UL,
   ENUM_FC_RJT_INV_ALIAS_TOKEN = 0x36UL,
   ENUM_FC_RJT_UNSUPP_ALIAS_TOKEN = 0x37UL,
   ENUM_FC_RJT_CANT_FORM_ALIAS_GRP = 0x38UL,
   ENUM_FC_RJT_INV_PORT_NODE_NAME = 0x44UL,
   ENUM_FC_RJT_LOGIN_EXT_UNSUPP = 0x46UL,
   ENUM_FC_RJT_AUTH_REQ = 0x48UL,
   ENUM_FC_RJT_SCAN_VAL_NOT_ALLOWED = 0x50UL,
   ENUM_FC_RJT_SANNING_UNSUPP = 0x51UL
} FC_LS_RJT_REASON_CODE_EXPLANATION;

// --- FC_E_STAT ---
// FC E_STAT field
#define FC_E_STAT_EXCHANGE_COMPLETE     (536870912) // Exchange is complete
#define FC_E_STAT_HOLD_SEQ_INITIATIVE   (1073741824) // Responder holds sequence
                                                     // initiative

typedef enum {
   ENUM_FC_E_STAT_EXCHANGE_COMPLETE = 0x20000000UL,
   ENUM_FC_E_STAT_HOLD_SEQ_INITIATIVE = 0x40000000UL
} FC_E_STAT;

// --- FIP_OPERATION_CODE ---
// FIP operation codes
#define FIP_OP_CODE_DISC                (1)       // Discovery
#define FIP_OP_CODE_VLI                 (2)       // Virtual Link Instantiation
#define FIP_OP_CODE_MISC                (3)       // Miscellaneous

typedef enum {
   ENUM_FIP_OP_CODE_DISC = 0x1UL,
   ENUM_FIP_OP_CODE_VLI = 0x2UL,
   ENUM_FIP_OP_CODE_MISC = 0x3UL
} FIP_OPERATION_CODE;

// --- FIP_SUBCODE ---
// FIP subcodes
#define FIP_SUBCODE_DISC_SOLICITATION   (1)       // Discovery Solicitation
#define FIP_SUBCODE_VLI_REQUEST         (1)       // Virtual Link Instantiation Request
#define FIP_SUBCODE_KEEP_ALIVE          (1)       // FIP Keep Alive
#define FIP_SUBCODE_DISC_ADVERTISEMENT  (2)       // Discovery Advertisement
#define FIP_SUBCODE_VLI_REPLY           (2)       // Virtual Link Instantiation Reply
#define FIP_SUBCODE_CLR_VIRT_LINKS      (2)       // FIP Clear Virtual Links

typedef enum {
   ENUM_FIP_SUBCODE_DISC_SOLICITATION = 0x1UL,
   ENUM_FIP_SUBCODE_VLI_REQUEST = 0x1UL,
   ENUM_FIP_SUBCODE_KEEP_ALIVE = 0x1UL,
   ENUM_FIP_SUBCODE_DISC_ADVERTISEMENT = 0x2UL,
   ENUM_FIP_SUBCODE_VLI_REPLY = 0x2UL,
   ENUM_FIP_SUBCODE_CLR_VIRT_LINKS = 0x2UL
} FIP_SUBCODE;

// --- FIP_DESCRIPTOR_TYPE ---
// FIP descriptor types
#define FIP_DESC_PRIORITY               (1)       // Priority
#define FIP_DESC_MAC_ADDRESS            (2)       // MAC address
#define FIP_DESC_FC_MAP                 (3)       // FC-MAP
#define FIP_DESC_NAME_IDENTIFIER        (4)       // Name_Identifier
#define FIP_DESC_FABRIC                 (5)       // Fabric
#define FIP_DESC_MAX_FCOE_SIZE          (6)       // Max FCoE Size
#define FIP_DESC_FLOGI                  (7)       // FLOGI
#define FIP_DESC_FDISC_NPIV             (8)       // FDISC_NPIV
#define FIP_DESC_LOGO                   (9)       // LOGO
#define FIP_DESC_ELP                    (10)      // ELP
#define FIP_DESC_VN_PORT_ID             (11)      // VN_Port Identification
#define FIP_DESC_FKA_ADV_PERIOD         (12)      // FKA_ADV_Period
#define FIP_DESC_VENDOR_ID              (13)      // Vendor_ID
#define FIP_DESC_MAX_CRITICAL           (127)     // Maximum critical descriptor

typedef enum {
   ENUM_FIP_DESC_PRIORITY = 0x1UL,
   ENUM_FIP_DESC_MAC_ADDRESS = 0x2UL,
   ENUM_FIP_DESC_FC_MAP = 0x3UL,
   ENUM_FIP_DESC_NAME_IDENTIFIER = 0x4UL,
   ENUM_FIP_DESC_FABRIC = 0x5UL,
   ENUM_FIP_DESC_MAX_FCOE_SIZE = 0x6UL,
   ENUM_FIP_DESC_FLOGI = 0x7UL,
   ENUM_FIP_DESC_FDISC_NPIV = 0x8UL,
   ENUM_FIP_DESC_LOGO = 0x9UL,
   ENUM_FIP_DESC_ELP = 0xaUL,
   ENUM_FIP_DESC_VN_PORT_ID = 0xbUL,
   ENUM_FIP_DESC_FKA_ADV_PERIOD = 0xcUL,
   ENUM_FIP_DESC_VENDOR_ID = 0xdUL,
   ENUM_FIP_DESC_MAX_CRITICAL = 0x7fUL
} FIP_DESCRIPTOR_TYPE;

#endif // __headers_enum_h__

