/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   iscsi_drvr_stats.h

@brief
   
*/
#ifndef __iscsi_drvr_stats_h__
#define __iscsi_drvr_stats_h__


// statistics IOCTLs
#define IOCTL_GET_OSM_PARAMETERS            (0x00000001)
#define IOCTL_GET_OSM_STATS                 (0x00000002)
#define IOCTL_GET_OSM_EH_STATS              (0x00000003)
#define IOCTL_GET_OSM_TARGET_COUNT          (0x00000004)
#define IOCTL_GET_OSM_TARGET_COUNTERS       (0x00000005)
#define IOCTL_GET_OSM_TARGET_EH_STATS       (0x00000006)

#define IOCTL_GET_ACIT_PARAMETERS           (0x00010000)
#define IOCTL_GET_ACIT_FW_CONFIG            (0x00020000)
#define IOCTL_GET_ACIT_COUNTERS             (0x00030000)
#define IOCTL_GET_ACIT_CONTROLLER_STATS     (0x00040000)
#define IOCTL_GET_ACIT_INTERRUPT_STATS      (0x00050000)
#define IOCTL_GET_ACIT_DIC_COUNTERS         (0x00060000)
#define IOCTL_GET_ACIT_EH_STATS             (0x00070000)
#define IOCTL_GET_ACIT_TARGET_COUNTERS      (0x00080000)
#define IOCTL_GET_ACIT_TARGET_EH_STATS      (0x00090000)
#define IOCTL_GET_ACIT_FW_ADDI_ATTRIBS      (0x000A0000)

// Signature for lxstats on Linux
#define ES_MGMT_STATS_COMMAND               (0xbe00)

//
// request format. The application has to fill the structure in
// this format
//
typedef struct _DRVR_STATS_REQUEST
{
    ACIT_UEXACT32       ioctl_code;
    ACIT_UEXACT32       buffer_len;
    ACIT_UEXACT32       parameter1;
    ACIT_UEXACT32       parameter2;

}DRVR_STATS_REQUEST, *PDRVR_STATS_REQUEST;


typedef struct _WINOSM_PARAMETERS_BLOCK
{
    // WinOSM defined parameters
    ACIT_UEXACT32       max_targets;
    ACIT_UEXACT32       max_luns;
    ACIT_UEXACT32       ticks_per_sec;
    ACIT_UEXACT32       path_first;
    ACIT_UEXACT32       max_ios;
    ACIT_UEXACT32       num_sge;
    ACIT_UEXACT32       max_io_size;
    ACIT_UEXACT32       lun_queue_depth;
    ACIT_UEXACT32       sync_mode_fd;
    ACIT_UEXACT32       target_reset;
    ACIT_UEXACT32       wmi_immediate_data;
    ACIT_UEXACT32       wmi_initial_r2t;
    ACIT_UEXACT32       wmi_fbl;
    ACIT_UEXACT32       wmi_mbl;
    ACIT_UEXACT32       wmi_max_recv_segment;
    ACIT_UEXACT32       acit_controller_ws_size;
    ACIT_UEXACT32       acit_target_ws_size;
    ACIT_UEXACT32       acit_lun_ws_size;
    ACIT_UEXACT32       acit_io_ws_size;
    ACIT_UEXACT32       acit_tmf_ws_size;
    ACIT_UEXACT32       srb_extension_size;
    ACIT_UEXACT32       lu_extension_size;
    ACIT_UEXACT32       device_extension_size;
    ACIT_UEXACT32       uncached_extension_size;
    ACIT_UEXACT32       large_io;
    ACIT_UEXACT32       interrupt_redirection;
    ACIT_UEXACT32       concurrent_channels;

} WINOSM_PARAMETERS_BLOCK, *PWINOSM_PARAMETERS_BLOCK;

#define WMI_UNKNOWN_GUID    22

typedef struct _WINOSM_STATS_BLOCK
{
    ACIT_UEXACT32       timer_ticks;
    ACIT_UEXACT32       cmds_in_controller;
    ACIT_UEXACT32       max_cmds_in_controller;

    ACIT_UEXACT32       wmi_requests;
    ACIT_UEXACT32       wmi_ioctl_success;
    ACIT_UEXACT32       wmi_unexpected_error;
    ACIT_UEXACT32       wmi_processing_error;
    ACIT_UEXACT32       wmi_invalid_requests;
    ACIT_UEXACT32       wmi_data_overruns;
    ACIT_UEXACT32       wmi_insufficient_buffer;
    ACIT_UEXACT32       wmi_invalid_device;
    ACIT_UEXACT32       wmi_unknown_error;

    ACIT_UEXACT32       io_success;
    ACIT_UEXACT32       io_timeout;
    ACIT_UEXACT32       io_device_timeout;
    ACIT_UEXACT32       io_aborted;
    ACIT_UEXACT32       io_error;
    ACIT_UEXACT32       io_unknown_error;

    ACIT_UEXACT32       buildio_err_timeout;
    ACIT_UEXACT32       buildio_err_sel_timeout;
    ACIT_UEXACT32       buildio_err_unknown;

    ACIT_UEXACT32       unknown_buildio_srb;
    ACIT_UEXACT32       unknown_queueio_srb;

    ACIT_UEXACT32       unload_type_shutdown;
    ACIT_UEXACT32       unload_type_power_management;
    ACIT_UEXACT32       unload_type_unknown;
    ACIT_UEXACT32       unload_type_uninitialized;
    ACIT_UEXACT32       deferred_ioctls;

    ACIT_UEXACT32       wmitype[WMI_UNKNOWN_GUID+1];

} WINOSM_STATS_BLOCK, *PWINOSM_STATS_BLOCK;

typedef struct _WINOSM_EH_STATS_BLOCK
{
    ACIT_UEXACT32       aborts;
    ACIT_UEXACT32       abort_success;
    ACIT_UEXACT32       abort_failed;
    ACIT_UEXACT32       abort_error;

    ACIT_UEXACT32       lun_resets;
    ACIT_UEXACT32       lun_reset_success;
    ACIT_UEXACT32       lun_reset_error;

    ACIT_UEXACT32       target_resets;
    ACIT_UEXACT32       target_reset_success;
    ACIT_UEXACT32       target_reset_error;
 
    ACIT_UEXACT32       bus_resets;
    ACIT_UEXACT32       bus_reset_failures;
 
} WINOSM_EH_STATS_BLOCK, *PWINOSM_EH_STATS_BLOCK;

typedef struct _ACIT_PARAMETERS
{
    ACIT_UEXACT32       default_flags;
    ACIT_UEXACT32       targets_per_controller;
    ACIT_UEXACT32       connections_per_session;
    ACIT_UEXACT32       ios_per_controller;
    ACIT_UEXACT32       icds_per_controller;
    
    ACIT_UEXACT32       logouts_per_controller;
    ACIT_UEXACT32       logout_timeout_ticks;
    ACIT_UEXACT32       ticks_per_second;
    ACIT_UEXACT32       tmfs_per_ctrl;
    ACIT_UEXACT32       tmf_timeout_ticks;

    ACIT_UEXACT32       eto;
    ACIT_UEXACT32       ldto;

    ACIT_UEXACT32       dbg_flags;  
    ACIT_UEXACT32       num_tmfs_per_cxn;
    ACIT_UEXACT32       num_cmd_per_cxn;    
    ACIT_UEXACT32       num_cxn;

    ACIT_UEXACT32       ioctls_per_ctrl;
    ACIT_UEXACT32       num_eq_entries;
    ACIT_UEXACT32       num_cq_entries;
    ACIT_UEXACT32       eq_timer;

    /*
     * These are calculated from other params. They're here
     * for debug purposes
     */
    ACIT_UEXACT32       num_mcc_pages;
    ACIT_UEXACT32       num_mcc_cq_pages;
    ACIT_UEXACT32       num_cq_pages;
    ACIT_UEXACT32       num_eq_pages;

    ACIT_UEXACT32       num_hdr_pages;
#ifdef BE_EMULATION
    ACIT_UEXACT32       num_pt_pages;
    ACIT_UEXACT32       num_jell_pages;
    ACIT_UEXACT32       num_r2tarray_pages;

#ifdef BE_RXIF_CREATE
    ACIT_UEXACT8        mac[2][6];
#endif
#endif // BE_EMULATION

    ACIT_UEXACT32       num_async_pdu_buf_pages;
    ACIT_UEXACT32       num_async_pdu_buf_sgl_pages;

    //ACIT_UEXACT32       num_async_pdu_buf_cq_pages;
    //ACIT_UEXACT32       num_async_pdu_hdr_pages;
    //ACIT_UEXACT32       num_async_pdu_hdr_sgl_pages;
    //ACIT_UEXACT32       num_async_pdu_hdr_cq_pages;

    ACIT_UEXACT32       default_pdu_buf_sz;
    ACIT_UEXACT32       num_sge;

    ACIT_UEXACT32       invalidation_ticks;
    ACIT_UEXACT32       tgt_default_ticks;
    ACIT_UEXACT32       timeout_factor;

    ACIT_UEXACT32       io_retries_b4_erl0;
    ACIT_UEXACT32       health_check_freq;

    ACIT_UEXACT32       hardware_port;
    ACIT_UEXACT32       num_async_pdus;
    ACIT_UEXACT32       max_cqe_per_isr;
    ACIT_UEXACT32       io_retries;
    ACIT_UEXACT32       io_retries_b4_err;
    ACIT_UEXACT32       num_cq;

    ACIT_UEXACT32       enable_dic;

    struct {

        ACIT_UEXACT32       policy;
        ACIT_UEXACT32       min_timer_val;
        ACIT_UEXACT32       max_timer_val;
        ACIT_UEXACT32       timer_increments;
        ACIT_SEXACT32       min_interrupt_rate;
        ACIT_SEXACT32       max_interrupt_rate;
        ACIT_UEXACT32       io_threshold_for_dic;
        ACIT_UEXACT32       io_resolution_for_dic;
        ACIT_UEXACT32       interrupt_resolution;
        ACIT_UEXACT32       tgt_threshold_for_load;
        ACIT_UEXACT32       interrupt_dec_steps;
        ACIT_UEXACT32       io_safety_valve;

    } dic;

    struct {

        ACIT_UEXACT32       sync_completion;
        ACIT_UEXACT32       dump_rst;

    } crash_dump;

    ACIT_UEXACT32       interrupt_redirection;
    ACIT_UEXACT32       offload_failure_threshold;
    ACIT_UEXACT32       ue_recovery;
    ACIT_UEXACT32       max_ue_recoveries;

} ACIT_PARAMETERS, *PACIT_PARAMETERS;

typedef struct _ACIT_FW_CONFIG {

    // group together since they are used most frequently
    // for cid to cri conversion
    ACIT_UEXACT32       iscsi_cid_start;
    ACIT_UEXACT32       phys_port;

    ACIT_UEXACT32       isr_offset;
    ACIT_UEXACT32       iscsi_icd_start;
    ACIT_UEXACT32       iscsi_cid_count;
    ACIT_UEXACT32       iscsi_icd_count;
    ACIT_UEXACT32       function_mode;
    ACIT_UEXACT32       pci_function;
    ACIT_UEXACT32       be_config_number;

} ACIT_FW_CONFIG, *PACIT_FW_CONFIG;

typedef struct _ACIT_FW_ADDI_ATTRIBS {

    ACIT_UEXACT16       ipl_file_number;
    ACIT_UEXACT8        ipl_file_version;
    ACIT_UEXACT8        interrupt_redirection;

} ACIT_FW_ADDI_ATTRIBS, *PACIT_FW_ADDI_ATTRIBS;

typedef struct _ACIT_COUNTERS {

    ACIT_UEXACT32       current_tick_count;
    ACIT_UEXACT16       io_handles_available;
    ACIT_UEXACT16       tmf_handles_available;
    ACIT_UEXACT16       logout_handles_available;
    ACIT_UEXACT16       nop_handles_available;
    ACIT_UEXACT16       inv_handles_available;
    
    ACIT_UEXACT16       pending_ios;        /* Total IOs in ACIT */
    ACIT_UEXACT16       queued_ios;         /* IOs in io_queued_list */
    ACIT_UEXACT16       pending_tmfs;
    ACIT_UEXACT16       tmf_queued_targets;
    ACIT_UEXACT16       queued_tmfs;
    ACIT_UEXACT16       pending_nops;
    ACIT_UEXACT16       pending_logouts;
    ACIT_UEXACT16       pending_ioctls;

    ACIT_UEXACT16       num_io_invs;
    ACIT_UEXACT16       num_targets_in_recovery;

    ACIT_UEXACT32       delayed_ioctls;
    ACIT_UEXACT32       current_num_targets;

} ACIT_COUNTERS, *PACIT_COUNTERS;

typedef struct _ACIT_CONTROLLER_STATS {

    ACIT_UEXACT32       reset_counters;

    ACIT_UEXACT32       io_status[10];
    ACIT_UEXACT64       ios_submitted;
    ACIT_UEXACT64       ios_completed;
    ACIT_UEXACT32       max_queue_depth;

    ACIT_UEXACT32       data_session_logins;
    ACIT_UEXACT32       data_session_drvr_logouts;
    ACIT_UEXACT32       data_session_armfw_logouts;

    ACIT_UEXACT32       async_hdr_posted;
    ACIT_UEXACT32       async_data_posted;
    ACIT_UEXACT32       txulp_db;

    ACIT_UEXACT32       max_num_queued;
    ACIT_UEXACT32       max_num_flushed;

    ACIT_UEXACT32       last_io_tick;
    //ACIT_UEXACT32       last_isr_tick;

    ACIT_UEXACT32       max_cqe_per_isr;
    ACIT_UEXACT32       max_eqe_per_isr;

    ACIT_UEXACT32       drvr_msg_sent;
    ACIT_UEXACT32       drvr_msg_recvd;
    ACIT_UEXACT32       drvr_msg_multiple_recvd;

    ACIT_UEXACT32       async_scsi_recvd;

    ACIT_UEXACT32       untagged_cmd;

    ACIT_UEXACT32       target_busy;
    ACIT_UEXACT32       target_ready;
    ACIT_UEXACT32       controller_busy;
    ACIT_UEXACT32       controller_ready;

    ACIT_UEXACT32       ioctls_submitted;
    ACIT_UEXACT32       ioctls_completed;

    ACIT_UEXACT32       async_event_new_target;
    ACIT_UEXACT32       async_event_new_iscsi_cxn;
    ACIT_UEXACT32       async_event_new_tcp_cxn;
    ACIT_UEXACT32       async_event_link_down;
    ACIT_UEXACT32       async_event_link_up;
    ACIT_UEXACT32       async_event_unknown;
    ACIT_UEXACT32       async_event_ignored;
    ACIT_UEXACT32       async_event_iscsi_cos;
    //
    // this stat has to be computed at runtime when the
    // application issues the IOCTLs
    //
    ACIT_UEXACT32       pending_ioctl_info[8];

    ACIT_UEXACT32       reopen_all_sessions;

} ACIT_CONTROLLER_STATS, *PACIT_CONTROLLER_STATS;

typedef struct _ACIT_CONTROLLER_INT_STATS {

    ACIT_UEXACT32       unknown_ints;
    ACIT_UEXACT32       unknown_eqes;
    ACIT_UEXACT32       unknown_cqes;
    ACIT_UEXACT32       unknown_cmds;
    ACIT_UEXACT32       unknown_cmd_errors;
    ACIT_UEXACT32		delayed_eq_dma;

    ACIT_UEXACT32       cqe_gt64;
    ACIT_UEXACT32       cqe_gt32;
    ACIT_UEXACT32       cqe_gt16;
    ACIT_UEXACT32       cqe_lt16;

    ACIT_UEXACT32       last_isr_tick;

    ACIT_UEXACT32       max_cqe_per_isr;
    ACIT_UEXACT32       max_eqe_per_isr;

    ACIT_UEXACT32       eq_maj_ints;
    ACIT_UEXACT32       cq_ints;
    ACIT_UEXACT32       dft_pdu_ints;
    ACIT_UEXACT32       drvr_msg_ints;
    ACIT_UEXACT32       iscsi_cmd_ints;
    ACIT_UEXACT32       iscsi_cq_processed;

} ACIT_CONTROLLER_INT_STATS, *PACIT_CONTROLLER_INT_STATS;

typedef struct _ACIT_CONTROLLER_DIC_COUNTERS {

    ACIT_UEXACT32       enabled;    // this is a run-time state
    ACIT_UEXACT32       reconfigure_in_progress;

    ACIT_UEXACT32       interrupts;
    ACIT_UEXACT32       max_interrupts_per_sec;

    ACIT_UEXACT32       ios_per_sec;
    ACIT_UEXACT32       num_eq_processed;

    ACIT_UEXACT32       last_ios_per_sec;
    ACIT_UEXACT32       max_ios_per_sec;
    ACIT_UEXACT32       max_overload;

    ACIT_UEXACT32       last_interrupt_count;
    ACIT_UEXACT32       interrupts_per_sec;

    ACIT_UEXACT32       eq_timer;
    ACIT_SEXACT32       interrupt_rate;
    ACIT_UEXACT32       current_load_factor;
    ACIT_UEXACT32       current_target_load;

    ACIT_UEXACT32       safety_valve_trigger;
    ACIT_UEXACT32       eq_runs[32];

} ACIT_CONTROLLER_DIC_COUNTERS, *PACIT_CONTROLLER_DIC_COUNTERS;


typedef struct _ACIT_CONTROLLER_EH_STATS {

    ACIT_UEXACT32       target_recoveries;

    ACIT_UEXACT32       io_timeouts;
    ACIT_UEXACT32       tmf_timeouts;
    ACIT_UEXACT32       tmf_internal_timeouts;
    ACIT_UEXACT32       tmf_errors;
    ACIT_UEXACT32       ld_timeouts;

    ACIT_UEXACT32       io_retry_count;
    ACIT_UEXACT32       tmfs_issued;
    ACIT_UEXACT32       osm_tmfs;
    ACIT_UEXACT32       osm_aborts;
    ACIT_UEXACT32       osm_lun_resets;
    ACIT_UEXACT32       osm_target_resets;

    ACIT_UEXACT32       ssn_failure_count;
    ACIT_UEXACT32       ssn_logout_normals;
    ACIT_UEXACT32       ssn_logout_others;
    ACIT_UEXACT32       last_ssn_failure_type;
    ACIT_UEXACT32       ssn_digest_errors;
    ACIT_UEXACT32       ssn_cxn_timeout_errors;
    ACIT_UEXACT32       ssn_format_errors;

    // status from command invalidation
    ACIT_UEXACT32       inv_sts_success;
    ACIT_UEXACT32       inv_sts_ioctl_err;
    ACIT_UEXACT32       inv_sts_ioctl_timeout;
    ACIT_UEXACT32       inv_sts_not_found;
    ACIT_UEXACT32       inv_sts_err;

    ACIT_UEXACT32       io_busy_during_lun_rst_timeout;
    ACIT_UEXACT32       io_queued_during_lun_rst_timeout;
    ACIT_UEXACT32       io_rstpend_during_lun_rst_timeout;

    ACIT_UEXACT32       io_busy_during_tgt_rst_timeout;
    ACIT_UEXACT32       io_queued_during_tgt_rst_timeout;
    ACIT_UEXACT32       io_rstpend_during_tgt_rst_timeout;

    ACIT_UEXACT32       tmf_err_stats[9];

    ACIT_UEXACT32       invs_deferred;
    ACIT_UEXACT32       invs_deferred_no_ioctls;
    ACIT_UEXACT32       invs_deferred_no_invh;

    ACIT_UEXACT32       cq_errors_invalid;
    ACIT_UEXACT32       offload_failures;
    ACIT_UEXACT32       ping_wait_timeouts;

    ACIT_UEXACT32       eto_expiry;
    ACIT_UEXACT32       cxn_invalidations;
    ACIT_UEXACT32       cxn_uploads;

    ACIT_UEXACT32       ioretry_exceeded;
    ACIT_UEXACT32       rejects;
    ACIT_UEXACT32       async_logouts;
    ACIT_UEXACT32       async_reopens;
    ACIT_UEXACT32       async_param_renego;
    ACIT_UEXACT32       io_status_errors;
    ACIT_UEXACT32       unknown_defpdu;
    ACIT_UEXACT32       reopen_mismatch;

    ACIT_UEXACT32       login_failure_arm_fw;
    ACIT_UEXACT32       login_failure_driver;
    ACIT_UEXACT32       logout_failure_io_busy;
    ACIT_UEXACT32       logout_failure_recovery;
    ACIT_UEXACT32       logout_failure_armfw_session;

    ACIT_UEXACT32       ue_low;
    ACIT_UEXACT32       ue_hi;
    ACIT_UEXACT32       ue_fw_line;
    ACIT_UEXACT32       ue_detected;
    ACIT_UEXACT32       ue_recoveries;

    ACIT_UEXACT16       cq_errors[40];

    ACIT_UEXACT32       tmf_abort_timeouts;
    ACIT_UEXACT32       tmf_lunrst_timeouts;
    ACIT_UEXACT32       tmf_tgtrst_timeouts;

    ACIT_UEXACT32       tmf_abort_internal_timeouts;
    ACIT_UEXACT32       tmf_lunrst_internal_timeouts;
    ACIT_UEXACT32       tmf_tgtrst_internal_timeouts;

    ACIT_UEXACT32       inv_over_time_limit;
    ACIT_UEXACT32       max_inv_ticks_over_limit;

} ACIT_CONTROLLER_EH_STATS, *PACIT_CONTROLLER_EH_STATS;    /* Error Handler statistics */

#endif

