/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_mem_linuxk.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/linuxk/sa_mem_linuxk.h,v 1.5 2006/02/28 01:13:44 menz Exp $
*/
#ifndef __sa_mem_linuxk_h__
#define __sa_mem_linuxk_h__

#include <asm/io.h>

typedef struct _SA_SGL {
    u32 length;
    PVOID va;
    u64 pa;
} SA_SGL, *PSA_SGL;

STATIC INLINE
u32
sa_sgl_get_byte_count(
    IN PSA_SGL sgl)
{
    ASSERT (sgl);
    return  sgl->length;
}

STATIC INLINE u32
sa_sgl_get_page_count (PSA_SGL sgl) {
    ASSERT (sgl);
    return SA_PAGES_SPANNED (sgl->va, sgl->length);
}                      

STATIC INLINE PVOID
sa_sgl_get_base_va(
    IN PSA_SGL sgl )
{
    ASSERT (sgl);
    return  sgl->va;
}

STATIC INLINE u32
sa_sgl_get_offset (
    PSA_SGL sgl)
{
    ASSERT (sgl);
    return SA_PAGE_OFFSET (sgl->va);
}


STATIC INLINE 
SA_PHYSICAL_ADDRESS
sa_sgl_get_page (
    IN  PSA_SGL  sgl, 
    IN  u32 page_index)
{    
    SA_PHYSICAL_ADDRESS pa;    

    ASSERT (sgl);
    ASSERT (page_index < sa_sgl_get_page_count (sgl));

#ifdef VIRT_TO_BUS_NOTSUPPORTED
    /* on some platforms (xscale + linux)
     * virt_to_bus() is not supported
     * drivers should use pci_alloc_consistent()
     * routines to get the DMA address.
     *
     * under such systems the sgl->pa is returned
     * NOTE: this scheme only works for SA_SGL with 
     * pa and va already setup, currently mailbox-mode-only
     * is supported on such platforms and avoiding this usage of
     * virt_to_bus() is good enough.
     */
//     ASSERT (page_index == 0);
    ASSERT (sgl->pa);
    pa = sgl->pa;
 
    pa += (u64) page_index * SA_PAGE_SIZE;    // offset to correct page (this is still a valid va)
 
    ASSERT (SA_PAGE_ALIGN (SA_U64_TO_PTR (pa)) == SA_U64_TO_PTR (pa));  // must return page aligned address.

    return pa;
#else
    // Just call virt_to_bus for the correct VA.
    if (page_index == 0) {
#ifdef __VMKERNEL_MODULE__
        pa = (sgl->va == NULL) ? 0 : virt_to_phys(sgl->va);
#else
        pa = virt_to_bus (sgl->va); 
#endif
        pa &= ~(SA_PAGE_SIZE-1);    // next lower pa
    } else {
        u64 va = SA_PTR_TO_U64 (sgl->va) & ~(SA_PAGE_SIZE-1); // next lower va from start
        va += (u64) page_index * SA_PAGE_SIZE;                     // offset to correct page (this is still a valid va)
#ifdef __VMKERNEL_MODULE__
        pa = (va == NULL) ? 0 : virt_to_phys(va);
#else
        pa = virt_to_bus (SA_U64_TO_PTR(va));
#endif
    }

    ASSERT (pa);
    ASSERT (SA_PAGE_ALIGN (SA_U64_TO_PTR (pa)) == SA_U64_TO_PTR (pa));  // must return page aligned address.
#endif // VIRT_TO_BUS_NOTSUPPORTED
    return pa;
}

#endif // __sa_mem_linuxk_h__


