/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_network_packet.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_network_packet.h,v 1.4 2005/11/16 18:30:05 menz Exp $
*/
#ifndef __sa_network_packet_h__
#define __sa_network_packet_h__

#include "system/sa_system.h"
SA_CPP_HEADER

//* Definitions for header flags.
#define SA_TCP_FLAG_FIN    0x01
#define SA_TCP_FLAG_SYN    0x02
#define SA_TCP_FLAG_RST    0x04
#define SA_TCP_FLAG_PUSH   0x08
#define SA_TCP_FLAG_ACK    0x10
#define SA_TCP_FLAG_URG    0x20
#define SA_TCP_FLAG_ECE    0x40
#define SA_TCP_FLAG_CWR    0x80
#define SA_TCP_FLAGS_ECN   (SA_TCP_FLAG_ECE | SA_TCP_FLAG_CWR)
#define SA_TCP_FLAGS_ALL   (SA_TCP_FLAG_FIN | SA_TCP_FLAG_SYN | SA_TCP_FLAG_RST | SA_TCP_FLAG_ACK | SA_TCP_FLAG_URG)

//* Definitions for IP tos bits.
#define SA_IP_TOS_ECT0    0x01
#define SA_IP_TOS_ECT1    0x02
#define SA_IP_TOS_ECN     (SA_IP_TOS_ECT0 | SA_IP_TOS_ECT1)

#define SA_IP_VER_FLAG 0xf0 // Mask for the IP version
#define SA_IP_VERSION_4  0x40 // Ipv4 will have the (verlen & SA_IP_VER_FLAG) == SA_IP_VERSION_4

// Defines for the IPv4 protocol field and the IPv6 next_header field.
// The extension headers have comments showing the order they should be placed in a packet.
#define SA_PROTOCOL_HOPBYHOP      0   /* Extension Header, order=1 */
#define SA_PROTOCOL_ICMP          1   /* Upper Layer Header */
#define SA_PROTOCOL_IGMP          2   /* Upper Layer Header */
#define SA_PROTOCOL_TCP           6   /* Upper Layer Header */
#define SA_PROTOCOL_CUSTOM        8   /* Upper Layer Header */
#define SA_PROTOCOL_UDP           17  /* Upper Layer Header */
#define SA_PROTOCOL_IPV6          41  /* Upper Layer Header */
#define SA_PROTOCOL_IPV6_ROUTING  43  /* Extension Header, order=3 */
#define SA_PROTOCOL_IPV6_FRAGMENT 44  /* Extension Header, order=4 */
#define SA_PROTOCOL_ESP           50  /* Extension Header, order=6 */
#define SA_PROTOCOL_AH            51  /* Extension Header, order=5 */
#define SA_PROTOCOL_IPV6_ICMP     58  /* Upper Layer Header */
#define SA_PROTOCOL_IPV6_NO_NEXT  59  /* No more headers, order=last */
#define SA_PROTOCOL_IPV6_DESTOPTS 60  /* Extension Header, with routing order=2, without order=7 */
#define SA_PROTOCOL_MOBILITY      135 /* Extension Header, order=8 */

// Defines for the ether-type field in ethernet header (or snap).
#define SA_MIN_ETYPE    0x600           // Minimum valid Ethertype
#define SA_ETHTYPE_VLAN 0x8100
#define SA_ETHTYPE_IP   0x0800
#define SA_ETHTYPE_IPV4 0x0800
#define SA_ETHTYPE_IPV6 0x86DD
#define SA_ETHTYPE_ARP  0x0806
#define SA_ETHTYPE_SE   0x0808

// IPv4 Flag definitions
#define SA_IPV4_FLAG_MORE_FRAGMENTS  (0x1) // set=more fragments, clear=last fragment
#define SA_IPV4_FLAG_DONT_FRAGMENT   (0x2) // set=don't fragment, clear=may fragment
#define SA_IPV4_FLAG_RSVD            (0x4) // Should never be set.

// IPv4 options
#define SA_MAX_IP_OPTION_BYTES (40)
#define SA_IPOPTION_EOL            (0x00)
#define SA_IPOPTION_NOP            (0x01)
#define SA_IPOPTION_RR             (0x07)
#define SA_IPOPTION_RTRALT         (0x14)
#define SA_IPOPTION_TS             (0x44)
#define SA_IPOPTION_SECURITY       (0x82)
#define SA_IPOPTION_LSRR           (0x83)
#define SA_IPOPTION_LSRR_E         (0x84)
#define SA_IPOPTION_SATID          (0x88)
#define SA_IPOPTION_SSRR           (0x89)  

// TCP Options
#define MAX_TCP_OPTION_BYTES (40)
#define SA_TCPOPT_KIND_EOL                   (0x00)     // End of list (RFC793)
#define SA_TCPOPT_KIND_NOP                   (0x01)     // no operation (RFC793)
#define SA_TCPOPT_KIND_MSS                   (0x02)     // max segment size (RFC793)
#define SA_TCPOPT_KIND_WSCALE                (0x03)     // window scale factor (RFC1072)
#define SA_TCPOPT_KIND_SACKPERMITTED         (0x04)     // selective ack ok (RFC1072)
#define SA_TCPOPT_KIND_SACK                  (0x05)     // selective ack (RFC1072) 
#define SA_TCPOPT_KIND_ECHO                  (0x06)     // echo (RFC1072)
#define SA_TCPOPT_KIND_ECHOREPLY             (0x07)     // echo (RFC1072)
#define SA_TCPOPT_KIND_TIMESTAMP             (0x08)     // timestamps (RFC1323)
#define SA_TCPOPT_KIND_PARTIALORDERPERMITTED (0x09)     // partial order permitted (RFC 1693)
#define SA_TCPOPT_KIND_PARTIALORDERPROFILE   (0x0a)     // partial order service profile (RFC 1693)
#define SA_TCPOPT_KIND_CC                    (0x0b)     // T/TCP CC options (RFC1644)
#define SA_TCPOPT_KIND_CCNEW                 (0x0c)     // T/TCP CC options (RFC1644)
#define SA_TCPOPT_KIND_CCECHO                (0x0d)     // T/TCP CC options (RFC1644)    
#define SA_TCPOPT_KIND_ALT_CKSM_REQUEST      (0x0e)     // Alternate checksum request (RFC1146)
#define SA_TCPOPT_KIND_ALT_CKSM_DATA         (0x0f)     // Alternate checksum data (RFC1146)
#define SA_TCPOPT_KIND_TRAILER_CKSM          (0x12)     // Trailer checksum
#define SA_TCPOPT_KIND_MD5_SIGNATURE         (0x13)     // MD5 Signature (RFC2385)

// Defines for the ARP packet hwtype.
#define SA_ARP_HWTYPE_ETHERNET  1
#define SA_ARP_HWTYPE_IEEE802   6
#define SA_ARP_HWTYPE_IEEE1394  29

// Defines for the ARP packet opcode.
#define SA_ARP_OPCODE_REQUEST   1 // ARP request
#define SA_ARP_OPCODE_REPLY     2 // ARP reply
#define SA_RARP_OPCODE_REQUEST  3 // RARP request
#define SA_RARP_OPCODE_REPLY    4 // RARP reply

// Ethernet packet sizes.
#define SA_ETHERNET_HEADER_SIZE             (sizeof (SA_ENET_HEADER))
#define SA_ETHERNET_MAX_DATA_SIZE           (1500)
#define SA_ETHERNET_MAX_PACKET_SIZE         (SA_ETHERNET_HEADER_SIZE + SA_ETHERNET_MAX_DATA_SIZE)
#define SA_ETHERNET_MIN_PACKET_SIZE         (60)

// Snap header defines.
#define SA_SNAP_SAP            170             // 0xAA
#define SA_SNAP_UI             3

// ICMP types
#define SA_ICMP_TYPE_DEST_UNREACHABLE                         (1)   // Destination unreachable. RFC 2463, RFC 4443 
#define SA_ICMP_TYPE_PACKET_TOO_BIG                           (2)   // Packet too big. RFC 2463, RFC 4443 
#define SA_ICMP_TYPE_TIME_EXCEEDED                            (3)   // Time exceeded. RFC 2463, RFC 4443 
#define SA_ICMP_TYPE_PARAM_PROBLEM                            (4)   // Parameter problem. RFC 2463, RFC 4443 
#define SA_ICMP_TYPE_PRIVATE                                  (101) // Private experimentation. RFC 4443 
#define SA_ICMP_TYPE_ERROR_MSG                                (127) // reserved for expansion of ICMPv6 error messages. RFC 4443 
#define SA_ICMP_TYPE_ECHO_REQUEST                             (128) // Echo request. RFC 2463, RFC 4443 
#define SA_ICMP_TYPE_ECHO_REPLY                               (129) // Echo reply. RFC 2463, RFC 4443 
#define SA_ICMP_TYPE_GROUP_MEMBERSHIP_QUERY                   (130) // Group Membership Query. RFC 2710 
#define SA_ICMP_TYPE_GROUP_MEMBERSHIP_REPORT                  (131) // Group Membership Report. RFC 2710 
#define SA_ICMP_TYPE_GROUP_MEMBERSHIP_REDUCTION               (132) // Group Membership Reduction. RFC 2710 
#define SA_ICMP_TYPE_ROUTER_SOLICITATION                      (133) // Router Solicitation. RFC 2461 
#define SA_ICMP_TYPE_ROUTER_ADVERTISEMENT                     (134) // Router Advertisement. RFC 2461 
#define SA_ICMP_TYPE_NEIGHBOR_SOLICITATION                    (135) // Neighbor Solicitation. RFC 2461 
#define SA_ICMP_TYPE_NEIGHBOR_ADVERTISEMENT                   (136) // Neighbor Advertisement. RFC 2461 
#define SA_ICMP_TYPE_REDIRECT                                 (137) // Redirect. RFC 2461 
#define SA_ICMP_TYPE_ROUTER_RENUMBERING                       (138) // Router Renumbering. RFC 2894 
#define SA_ICMP_TYPE_NODE_INFO_QUERY                          (139) // ICMP Node Information Query. RFC 4620 
#define SA_ICMP_TYPE_NODE_INFO_RESPONSE                       (140) // ICMP Node Information Response. RFC 4620 
#define SA_ICMP_TYPE_INVERSE_NEIGHBOR_DISCOVERY_SOLICITATION  (141) // Inverse Neighbor Discovery Solicitation Message. RFC 3122 
#define SA_ICMP_TYPE_INVERSE_NEIGHBOR_DISCOVERY_ADVERTISEMENT (142) // Inverse Neighbor Discovery Advertisement Message. RFC 3122 
#define SA_ICMP_TYPE_MLDV2_MULTICAST_LISTENER_REPORT          (143) // MLDv2 Multicast Listener Report. RFC 3810 
#define SA_ICMP_TYPE_HAAD_REQUEST                             (144) // Home Agent Address Discovery Request Message. RFC 3775 
#define SA_ICMP_TYPE_HAAD_REPLY                               (145) // Home Agent Address Discovery Reply Message RFC 3775 
#define SA_ICMP_TYPE_MOBILE_PREFIX_SOLICITATION               (146) // Mobile Prefix Solicitation. RFC 3775 
#define SA_ICMP_TYPE_MOBILE_PREFIX_ADVERTISEMENT              (147) // Mobile Prefix Advertisement RFC 3775 
#define SA_ICMP_TYPE_CERTIFICATION_PATH_SOLICITATION          (148) // Certification Path Solicitation. RFC 3971 
#define SA_ICMP_TYPE_CERTIFICATION_PATH_ADVERTISEMENT         (149) // Certification Path Advertisement. RFC 3971 
#define SA_ICMP_TYPE_EXPERIMENTAL_MOBILITY                    (150) // Experimental mobility protocols. RFC 4065 
#define SA_ICMP_TYPE_MRD_ADVERTISEMENT                        (151) // MRD, Multicast Router Advertisement. RFC 4286 
#define SA_ICMP_TYPE_MRD_SOLICITATION                         (152) // MRD, Multicast Router Solicitation. RFC 4286 
#define SA_ICMP_TYPE_MRD_TERMINATION                          (153) // MRD, Multicast Router Termination. RFC 4286 
#define SA_ICMP_TYPE_FMIPV6                                   (154) // FMIPv6 messages. RFC 5568 
#define SA_ICMP_TYPE_PRIVATE2                                 (201) // Private experimentation. RFC 4443 

// ICMP Unreachable codes
#define SA_ICMP_CODE_NOROUTE            (0) //  No route to destination
#define SA_ICMP_CODE_PROHIBITED         (1) //  Communication with destination administratively prohibited
#define SA_ICMP_CODE_SOURCE_ADDR_SCOPE  (2) //  Beyond scope of source address
#define SA_ICMP_CODE_ADDR_UNREACHABLE   (3) //  Address unreachable
#define SA_ICMP_CODE_PORT_UNREACHABLE   (4) //  Port unreachable
#define SA_ICMP_CODE_SOURCE_ADDR_POLICY (5) //  Source address failed ingress/egress policy
#define SA_ICMP_CODE_REJECT_ROUTE       (6) //  Reject route to destination


void sa_network_packet_dump (u8 *pkt);

void sa_network_packet_debug (PVOID packet, u32 length);

void
sa_network_packet_validate(
    u8 *Data,                   // ptr to contiguous data
    u32 Size,                   // size of packet
    u16 *IPChecksum,            // place to put ip checksum
    u16 *TCPChecksum,           // place to put tcp checksum
    u16 *UDPChecksum,           // place to put udp checksum
    u32 *bcast,                 // place to put detection of bcast frame
    u32 *mcast,                 // multicast?
    u32 *ipv4type,              // ip frame?
    u32 *ipv6type,              // ip frame?
    u32 *tcptype,               // tcp frame?
    u32 *udptype,               // udp frame?
    u32 *tcpcksm,               // tcp checksum done?
    u32 *udpcksm,               // udp checksum done?
    u32 *isipFrag);             // is a single ip Frag?
#include "system/sa_packon.h" // byte packing

    // Define an ip address type.
    typedef u32 SA_IPV4, *PSA_IPV4;

    // Define a MAC address type.
    #define SA_MAC_ADDRESS_SIZE (6)
    typedef struct _SA_MAC_ADDRESS 
    { 
        u8 bytes[SA_MAC_ADDRESS_SIZE]; 

    } SA_PACK SA_MAC_ADDRESS, *PSA_MAC_ADDRESS;

    SA_GLOBAL_C_ASSERT (mac, sizeof (SA_MAC_ADDRESS) == SA_MAC_ADDRESS_SIZE);

    typedef struct _SA_MAC64
    { 
        union {
            SA_MAC_ADDRESS mac;
            u8  bytes[8];
            u64 quad;
        };

    } SA_PACK SA_MAC64, *PSA_MAC64;

    SA_GLOBAL_C_ASSERT (mac64, sizeof (SA_MAC64) == 8);


    //* Structure of an Ethernet header.
    typedef struct  _SA_ENET_HEADER {
        u8       eh_daddr[6];
        u8       eh_saddr[6];
        u16      eh_type; // see defines SA_ETHTYPE above
    } SA_PACK  SA_ENET_HEADER, *PSA_ENET_HEADER;

    typedef struct _SA_CUSTOM_HEADER {
        u8 ese_daddr[6];
        u8 ese_saddr[6];
        u16 ese_type;
    } SA_PACK   SA_CUSTOM_HEAER, *PSA_CUSTOM_HEADER;

    SA_GLOBAL_C_ASSERT (eth, sizeof (SA_ENET_HEADER) == 14);

    typedef struct _SA_VLAN_TAG {
/* on xscale + linux platform this structure
 * was not getting padded correctly to 2 Bytes.
 * causing the SA_GLOBAL_C_ASSERT  to fail the
 * compilation.
 * since we don't have access to the platform
 * further debugging/investigation could not be
 * done.
 * since this struct is not really used, BE_SKIP_SA_NTWK_TAG
 * is used as a workaround the compilation issue.
 */
#ifdef BE_SKIP_SA_NTWK_TAG
        u16 tag;
#else
        union {
            u16 tag;
            struct {
                u16 id : 12;
                u16 canonical_format : 1;
                u16 priority : 3;
            };
        };
#endif
    } SA_PACK SA_VLAN_TAG, *PSA_VLAN_TAG;

    SA_GLOBAL_C_ASSERT (vlan_tag, sizeof (SA_VLAN_TAG) == 2);

    typedef struct _SA_VLAN_TYPE_TAG {
        u16 type; // SA_ETHTYPE_VLAN
        SA_VLAN_TAG tag;
    } SA_PACK SA_VLAN_TYPE_TAG, *PSA_VLAN_TYPE_TAG;

    SA_GLOBAL_C_ASSERT (vlan_type_tag, sizeof (SA_VLAN_TYPE_TAG) == 4);

    typedef struct _SA_VLAN_TAG_TYPE {
        SA_VLAN_TAG tag;
        u16 type; // see defines SA_ETHTYPE above
    } SA_PACK SA_VLAN_TAG_TYPE, *PSA_VLAN_TAG_TYPE;

    SA_GLOBAL_C_ASSERT (vlan_tag_type, sizeof (SA_VLAN_TAG_TYPE) == 4);

    // Structure of a SNAP header.
    typedef struct _SA_SNAP_HEADER {
        u8       sh_dsap; // see defines SA_SNAP above
        u8       sh_ssap;
        u8       sh_ctl;
        u8       sh_protid[3];
        u16      sh_etype; // see defines SA_ETHTYPE above
    } SA_PACK  SA_SNAP_HEADER, *PSA_SNAP_HEADER;

    SA_GLOBAL_C_ASSERT (snap, sizeof (SA_SNAP_HEADER) == 8);

    // IP Header format.
    typedef struct _SA_IP_HEADER {
        u8      iph_verlen;             // Version and length.
        u8      iph_tos;                // DiffServe, ECN bits
        u16     iph_length;             // Total length of datagram.
        u16     iph_id;                 // Identification.
        u16     iph_offset;             // Flags and fragment offset.
        u8      iph_ttl;                // Time to live.
        u8      iph_protocol;           // Protocol. (see defines SA_PROTOCOL_ above.)
        u16     iph_xsum;               // Header checksum.
        SA_IPV4 iph_src;                // Source address.
        SA_IPV4 iph_dest;               // Destination address.
    } SA_PACK  SA_IP_HEADER, *PSA_IP_HEADER;

    SA_GLOBAL_C_ASSERT (ip, sizeof (SA_IP_HEADER) == 20);

    typedef struct _SA_IP_OPTION {
        u8 option_code;
        u8 length;
        u8 *option_data;
    } SA_PACK SA_IP_OPTION, *PSA_IP_OPTION;

    // IPV6 address
    typedef union _SA_IPV6_ADDRESS {
        u8   addr8[16];
        u16  addr16[8];
        u32  addr32[4];
    } SA_PACK SA_IPV6_ADDRESS, *PSA_IPV6_ADDRESS;
    SA_GLOBAL_C_ASSERT (ipv6_address_size, sizeof (SA_IPV6_ADDRESS) == 16);



    typedef struct _SA_IPV6_HEADER {
        u32 version_class_flow; /* 4 bits ver, 8 bits class, 20 bits flow.  see functions
                                   sa_ipv6_header_flow, sa_ipv6_header_class, sa_ipv6_header_version, etc*/
        u16 payload_length;     /* payload length */
        u8 next_header;         /* next header. see defines SA_PROTOCOL_ above */
        u8 hop_limit;           /* hop limit */
        SA_IPV6_ADDRESS src;    /* source address */
        SA_IPV6_ADDRESS dst;    /* destination address */
    } SA_PACK SA_IPV6_HEADER, *PSA_IPV6_HEADER;
    SA_GLOBAL_C_ASSERT (ipv6_header_size, sizeof (SA_IPV6_HEADER) == 40);

    typedef struct _SA_TCP_HEADER {
        u16     tcp_src;            // Source port.
        u16     tcp_dest;           // Destination port.
        u32     tcp_seq;            // Sequence number.
        u32     tcp_ack;            // Ack number.
        u8      tcp_offs;
        u8      tcp_flags;          // Flags and data offset.
        u16     tcp_window;         // Window offered.
        u16     tcp_xsum;           // Checksum.
        u16     tcp_urgent;         // Urgent pointer.
    } SA_PACK  SA_TCP_HEADER, *PSA_TCP_HEADER;

    SA_GLOBAL_C_ASSERT (tcp, sizeof (SA_TCP_HEADER) == 20);

    // Basic format.  All options start with a type.
    typedef struct _SA_TCPOPT {
        u8 kind; // see defines SA_TCPOPT_       
    } SA_PACK SA_TCPOPT, *PSA_TCPOPT;

    SA_GLOBAL_C_ASSERT (tcpopt, sizeof (SA_TCPOPT) == 1);

    typedef struct _SA_TCPOPT_EOL {
        u8 kind;        
    } SA_PACK SA_TCPOPT_EOL, *PSA_TCPOPT_EOL;
    SA_GLOBAL_C_ASSERT (tcpopt_eol, sizeof (SA_TCPOPT_EOL) == 1);

    typedef struct _SA_TCPOPT_NOP {
        u8 kind;        
    } SA_PACK SA_TCPOPT_NOP, *PSA_TCPOPT_NOP;
    SA_GLOBAL_C_ASSERT (tcpopt_nop, sizeof (SA_TCPOPT_NOP) == 1);

    typedef struct _SA_TCPOPT_MSS {
        u8 kind;        
        u8 len;
        u16 mss;
    } SA_PACK SA_TCPOPT_MSS, *PSA_TCPOPT_MSS;
    SA_GLOBAL_C_ASSERT (tcpopt_mss, sizeof (SA_TCPOPT_MSS) == 4);

    typedef struct _SA_TCPOPT_WSCALE {
        u8 kind;        
        u8 len;
        u8 scale;
    } SA_PACK SA_TCPOPT_WSCALE, *PSA_TCPOPT_WSCALE;
    SA_GLOBAL_C_ASSERT (tcpopt_wscale, sizeof (SA_TCPOPT_WSCALE) == 3);

    typedef struct _SA_TCPOPT_SACKPERMITTED {
        u8 kind;  
        u8 len;
    } SA_PACK SA_TCPOPT_SACKPERMITTED, *PSA_TCPOPT_SACKPERMITTED;
    SA_GLOBAL_C_ASSERT (tcpopt_sackpermitted, sizeof (SA_TCPOPT_SACKPERMITTED) == 2);

    typedef struct _SA_TCPOPT_SACK_BLOCK {
        u32 left_edge;
        u32 right_edge;
    } SA_PACK SA_TCPOPT_SACK_BLOCK, *PSA_TCPOPT_SACK_BLOCK;    

    typedef struct _SA_TCPOPT_SACK {
        u8 kind;  
        u8 len;
        SA_TCPOPT_SACK_BLOCK sack_block[1];    // Variable number of blocks.
    } SA_PACK SA_TCPOPT_SACK, *PSA_TCPOPT_SACK;        
    SA_GLOBAL_C_ASSERT (tcpopt_sack, sizeof (SA_TCPOPT_SACK) == 10);

    typedef struct _SA_TCPOPT_ECHO {
        u8 kind;  
        u8 len;
        u32 info;
    } SA_PACK SA_TCPOPT_ECHO, *PSA_TCPOPT_ECHO;
    SA_GLOBAL_C_ASSERT (tcpopt_echo, sizeof (SA_TCPOPT_ECHO) == 6);

    typedef struct _SA_TCPOPT_ECHOREPLY {
        u8 kind;  
        u8 len;
        u32 info;
    } SA_PACK SA_TCPOPT_ECHOREPLY, *PSA_TCPOPT_ECHOREPLY;
    SA_GLOBAL_C_ASSERT (tcpopt_echo_reply, sizeof (SA_TCPOPT_ECHOREPLY) == 6);

    typedef struct _SA_TCPOPT_TIMESTAMP {
        u8 kind;
        u8 len;
        u32 ts_value;
        u32 ts_echo_reply;
    } SA_PACK SA_TCPOPT_TIMESTAMP, *PSA_TCPOPT_TIMESTAMP;
    SA_GLOBAL_C_ASSERT (tcpopt_echo_timestamp, sizeof (SA_TCPOPT_TIMESTAMP) == 10);

    typedef struct _SA_UDP_HEADER {
        u16 src_port;
        u16 dst_port;
        u16 len;
        u16 xsum;
    } SA_PACK SA_UDP_HEADER, *PSA_UDP_HEADER;

    SA_GLOBAL_C_ASSERT (udp, sizeof (SA_UDP_HEADER) == 8);

    //
    // IPV6 Options
    //

    // hop-by-hop, dest options
    typedef struct _SA_IPV6_OPTIONS_HEADER {
        u8 next_header;
        u8 header_extension_length; // 8 byte units, not counting first 8
    } SA_PACK SA_IPV6_OPTIONS_HEADER, *PSA_IPV6_OPTIONS_HEADER;

    // routing header
    typedef struct _SA_IPV6_ROUTING_HEADER {
        u8 next_header;
        u8 header_extension_length; // 8 byte units, not counting first 8 
        u8 routing_type;
        u8 segments_left;
        u32 reserved;
    } SA_PACK SA_IPV6_ROUTING_HEADER, *PSA_IPV6_ROUTING_HEADER;

    // fragment header
    typedef struct _SA_IPV6_FRAGMENT_HEADER {
        u8 next_header;
        u8 reserved;
        u16 offset_flag;
        u32 identifier;
    } SA_PACK SA_IPV6_FRAGMENT_HEADER, *PSA_IPV6_FRAGMENT_HEADER;

    // generic extension header
    typedef struct _SA_IPV6_EXTENSION_HEADER {
        u8   next_header;
        u8   header_extension_length; // 8 byte units, not counting first 8
    } SA_PACK SA_IPV6_EXTENSION_HEADER, *PSA_IPV6_EXTENSION_HEADER;

    typedef struct _SA_IPV6_OPTION_HEADER {
        u8   type;
        u8   data_length;    // in bytes, not counting 2 for header
    } SA_PACK SA_IPV6_OPTION_HEADER, *PSA_IPV6_OPTION_HEADER;

    typedef struct _SA_IPV6_AH_HEADER {
        u8  next_header;
        u8  payload_len;
        u16 reserved;
        u32 spi;
        u32 seq;
    } SA_PACK SA_IPV6_AH_HEADER, *PSA_IPV6_AH_HEADER;

    typedef struct _SA_IPV6_ESP_HEADER {
        u32   spi;
        u32   seq;
    } SA_PACK SA_IPV6_ESP_HEADER, *PSA_IPV6_ESP_HEADER;

    typedef struct _SA_IPV6_ESP_TRAILER {
        u8 pad_length;
        u8 next_header;
    } SA_PACK SA_IPV6_ESP_TRAILER, *PSA_IPV6_ESP_TRAILER;    


    //
    // ARP header
    //
    typedef struct _SA_ARP_HEADER {

	    u16 hwtype;	// hardware type. see SA_ARP_HWTYPE
	    u16 prtype; // protocol type. 
	    u8  hwlen;  // hardware address length
	    u8  prlen;  // protocol address length
	    u16 op;	    // ARP operation (see SA_ARP_OPCODE above)

        union {
            // This is the typical format for MAC/IP addresses.
            // Assuming 6 byte hw and 4 byte protocol addresses.
            struct {
                SA_MAC_ADDRESS send_mac;   // sender's physical hardware address
                u32 send_ip;               // sender's protocol address (IP addr.)
                SA_MAC_ADDRESS target_mac; // target's physical hardware address
                u32 target_ip;             // target's protocol address (IP)
            } SA_PACK ethernet;
        };

    } SA_PACK SA_ARP_HEADER, *PSA_ARP_HEADER;

    SA_GLOBAL_C_ASSERT (arp_size, sizeof (SA_ARP_HEADER) == 28);


    typedef struct _SA_NETBIOS {
        u32 length; // 24 bits, first byte is always zero
                    // Length of all remaining headers and data (not including this header).
    } SA_PACK SA_NETBIOS, *PSA_NETBIOS;

    typedef struct _SA_SMB_HEADER {
        
        u8 protocol[4];     // 0xFF, "SMB"
        u8 command;
        u32 status;
        u8 flags;
        u16 flags2;
        u16 pid_high;
        u8 signature[8];
        u16 rsvd2;
        u16 tid; // tree id
        u16 pid; // process id
        u16 uid; // user id
        u16 mid; // multiplex id
        
    } SA_PACK SA_SMB_HEADER, *PSA_SMB_HEADER;

    SA_GLOBAL_C_ASSERT (smb_header_size, sizeof (SA_SMB_HEADER) == 32);

    // smb command 0x2e
    typedef struct _SA_SMB_READ_ANDX_RESPONSE {
        u8 word_count;
        u8 andXcmd;
        u8 rsvd;
        u16 andXoffset;
        u16 remaining;
        u16 compaction;
        u16 rsvd2;
        u16 data_len_low;
        u16 data_offset;
        u32 data_len_high;
        u8 rsvd3[6];
        u16 byte_count;
    } SA_PACK SA_SMB_READ_ANDX_RESPONSE, *PSA_SMB_READ_ANDX_RESPONSE;

    SA_GLOBAL_C_ASSERT (smb_read_andx_response_size, sizeof (SA_SMB_READ_ANDX_RESPONSE) == 27);

    // smb command 0x2f
    typedef struct _SA_SMB_WRITE_ANDX_RESPONSE {
        u8 word_count;
        u8 andXcmd;
        u8 rsvd;
        u16 andXoffset;
        u16 fid;
        u32 offset;
        u32 rsvd2;
        u16 write_mode;
        u16 remaining;
        u16 data_len_high;
        u16 data_len_low;
        u16 data_offset;
        u32 high_offset;
        u16 byte_count;
        //u8 padding; // byte_count - data_len_low (?)
        
    } SA_PACK SA_SMB_WRITE_ANDX_RESPONSE, *PSA_SMB_WRITE_ANDX_RESPONSE;

    SA_GLOBAL_C_ASSERT (smb_write_andx_response_size, sizeof (SA_SMB_WRITE_ANDX_RESPONSE) == 31);

    typedef struct _SA_ICMP_HEADER {
        u8 type;
        u8 code;
        u16 checksum;
    } SA_PACK SA_ICMP_HEADER, *PSA_ICMP_HEADER;

    SA_GLOBAL_C_ASSERT (icmp_size, sizeof (SA_ICMP_HEADER) == 4);

    typedef struct _SA_ICMP_ECHO {
        SA_ICMP_HEADER header;
        u16 id;
        u16 sequence;                
    } SA_PACK SA_ICMP_ECHO, *PSA_ICMP_ECHO;
    SA_GLOBAL_C_ASSERT (icmp_echo_size, sizeof (SA_ICMP_ECHO) == 8);

    typedef struct _SA_ICMP_PACKET_TOO_BIG {
        SA_ICMP_HEADER header;
        u32 mtu;
    } SA_PACK SA_ICMP_PACKET_TOO_BIG, *PSA_ICMP_PACKET_TOO_BIG;
    SA_GLOBAL_C_ASSERT (icmp_packet_too_big_size, sizeof (SA_ICMP_PACKET_TOO_BIG) == 8);
    

#include "system/sa_packoff.h"  // end of byte packing

// This structure holds the results of parsing a network packet.
typedef struct _SA_PARSED_PACKET {
    
    struct {
        u32 truncated : 1; // Set to 1 if packet is prematurely truncated during parse

        u32 broadcast : 1; 
        u32 multicast : 1;
        u32 vlan : 1;
        u32 snap : 1;
        u32 ipv4 : 1;
        u32 ipv6 : 1;
        u32 arp : 1;
        u32 tcp : 1;
        u32 udp : 1;
        u32 icmp : 1;
        u32 icmp_echo : 1;

        u32 bad_ip_length : 1;
        
    } flags;

    u16 type;
    u32 bytes;

    SA_ENET_HEADER *eth;
    SA_VLAN_TAG_TYPE *vlan;
    SA_SNAP_HEADER *snap;
    union {
        SA_IP_HEADER *ipv4;
        SA_IPV6_HEADER *ipv6;
        SA_ARP_HEADER *arp;
    };
    union {
        SA_ICMP_HEADER *icmp;
        SA_ICMP_ECHO *icmp_echo;
        SA_TCP_HEADER *tcp;
        SA_UDP_HEADER *udp;
    };

    PVOID ulp; // Pointer directly after the last parsed header.
    
} SA_PARSED_PACKET, *PSA_PARSED_PACKET;

// Returns the IP header length in bytes.
STATIC FORCEINLINE u32
sa_ip_header_length (const SA_IP_HEADER *ip) {
    return (ip->iph_verlen & 0x0F) << 2;
}

// Returns the IP flags.  Use SA_IPV4_FLAG_* defines above to check them.
STATIC FORCEINLINE u32
sa_ip_header_flags (const SA_IP_HEADER *ip) {
    return (u32)(SA_SWAP16 (ip->iph_offset) >> 13);
}

STATIC FORCEINLINE u32
sa_ip_header_offset (const SA_IP_HEADER *ip) {
    return (u32)(SA_SWAP16 (ip->iph_offset) & 0x1FFF);
}

// Returns TRUE if equal
STATIC FORCEINLINE boolean
sa_ipv6addr_equal (const SA_IPV6_ADDRESS *left, const SA_IPV6_ADDRESS *right) {

    boolean equal = 
        (left->addr32[0] == right->addr32[0]) &&
        (left->addr32[1] == right->addr32[1]) &&
        (left->addr32[2] == right->addr32[2]) &&
        (left->addr32[3] == right->addr32[3]);
    
    return equal;
}

STATIC FORCEINLINE u32
sa_tcp_header_length (const SA_TCP_HEADER *tcp) {
    return (u32)(tcp->tcp_offs & 0xF0) >> 2;
}

STATIC FORCEINLINE boolean
sa_mac_is_multicast (const SA_MAC_ADDRESS *mac) {
    return (boolean)(mac->bytes[0] & 0x1);
}

STATIC FORCEINLINE boolean
sa_mac_is_locally_administered (const SA_MAC_ADDRESS *mac) {
    return (boolean)(mac->bytes[0] & 0x2);
}

STATIC FORCEINLINE boolean
sa_mac_is_broadcast (const SA_MAC_ADDRESS *mac) {
    return (boolean) (
            (mac->bytes[0] == 0xFF) &&
            (mac->bytes[1] == 0xFF) &&
            (mac->bytes[2] == 0xff) &&
            (mac->bytes[3] == 0xff) &&
            (mac->bytes[4] == 0xff) &&
            (mac->bytes[5] == 0xff)
            );
}

STATIC FORCEINLINE boolean
sa_mac_is_zero (const SA_MAC_ADDRESS *mac) {
    return (boolean) (
            (mac->bytes[0] == 0) &&
            (mac->bytes[1] == 0) &&
            (mac->bytes[2] == 0) &&
            (mac->bytes[3] == 0) &&
            (mac->bytes[4] == 0) &&
            (mac->bytes[5] == 0));
}

STATIC FORCEINLINE void
sa_mac_zero (SA_MAC_ADDRESS *mac)
{
    mac->bytes[0] = 0;
    mac->bytes[1] = 0;
    mac->bytes[2] = 0;
    mac->bytes[3] = 0;
    mac->bytes[4] = 0;
    mac->bytes[5] = 0;
}

STATIC FORCEINLINE void
sa_mac_copy (SA_MAC_ADDRESS *dest, const SA_MAC_ADDRESS *src) {
    dest->bytes[0] = src->bytes[0];
    dest->bytes[1] = src->bytes[1];
    dest->bytes[2] = src->bytes[2];
    dest->bytes[3] = src->bytes[3];
    dest->bytes[4] = src->bytes[4];
    dest->bytes[5] = src->bytes[5];
}

// Returns TRUE if equal
STATIC FORCEINLINE boolean
sa_mac_equal (const SA_MAC_ADDRESS *left, const SA_MAC_ADDRESS *right) {

    boolean equal = 
        (left->bytes[0] == right->bytes[0]) &&
        (left->bytes[1] == right->bytes[1]) &&
        (left->bytes[2] == right->bytes[2]) &&
        (left->bytes[3] == right->bytes[3]) &&
        (left->bytes[4] == right->bytes[4]) &&
        (left->bytes[5] == right->bytes[5]);
    
    return equal;
}

STATIC FORCEINLINE boolean
sa_mac64_is_multicast (const SA_MAC64 *mac) {
    ASSERT (mac->bytes[6] == 0);
    ASSERT (mac->bytes[7] == 0);
    return sa_mac_is_multicast (&mac->mac);
}

STATIC FORCEINLINE boolean
sa_mac64_is_locally_administered (const SA_MAC64 *mac) {
    ASSERT (mac->bytes[6] == 0);
    ASSERT (mac->bytes[7] == 0);
    return sa_mac_is_locally_administered (&mac->mac);
}

STATIC FORCEINLINE boolean
sa_mac64_is_broadcast (const SA_MAC64 *mac) {
    ASSERT (mac->bytes[6] == 0);
    ASSERT (mac->bytes[7] == 0);
    return (boolean) (mac->quad == 0x0000FFFFFFFFFFFFULL);          
}

STATIC FORCEINLINE boolean
sa_mac64_is_zero (const SA_MAC64 *mac) {
    ASSERT (mac->bytes[6] == 0);
    ASSERT (mac->bytes[7] == 0);
    return (boolean) (mac->quad == 0);      
}

STATIC FORCEINLINE void
sa_mac64_zero (SA_MAC64 *mac)
{
    mac->quad = 0ULL;
}

STATIC FORCEINLINE void
sa_mac64_copy (SA_MAC64 *dest, const SA_MAC64 *src) {
    ASSERT (src->bytes[6] == 0);
    ASSERT (src->bytes[7] == 0);
    dest->quad = src->quad;
}

STATIC FORCEINLINE void
sa_mac64_from_mac (SA_MAC64 *dest, const SA_MAC_ADDRESS *src) {
    dest->quad = 0;
    dest->bytes[0] = src->bytes[0];
    dest->bytes[1] = src->bytes[1];
    dest->bytes[2] = src->bytes[2];
    dest->bytes[3] = src->bytes[3];
    dest->bytes[4] = src->bytes[4];
    dest->bytes[5] = src->bytes[5];
}

STATIC FORCEINLINE void
sa_mac_from_mac64 (SA_MAC_ADDRESS *dest, const SA_MAC64 *src) {
    ASSERT (src->bytes[6] == 0);
    ASSERT (src->bytes[7] == 0);
    
    dest->bytes[0] = src->bytes[0];
    dest->bytes[1] = src->bytes[1];
    dest->bytes[2] = src->bytes[2];
    dest->bytes[3] = src->bytes[3];
    dest->bytes[4] = src->bytes[4];
    dest->bytes[5] = src->bytes[5];
}

// Returns TRUE if equal
STATIC FORCEINLINE boolean
sa_mac64_equal (const SA_MAC64 *left, const SA_MAC64 *right) {
    ASSERT (left->bytes[6] == 0);
    ASSERT (left->bytes[7] == 0);

    ASSERT (right->bytes[6] == 0);
    ASSERT (right->bytes[7] == 0);

    return (boolean)(left->quad == right->quad);
}


STATIC FORCEINLINE
u32 sa_ipv6_header_version (SA_IPV6_HEADER *ip) {
    return (ip->version_class_flow & 0xF0) >> 4;
}

STATIC FORCEINLINE
void sa_ipv6_header_set_version (SA_IPV6_HEADER *ip, u32 ver) {
    ip->version_class_flow = (ip->version_class_flow & ~0xF0) | ((ver&0xF)<<4);
}

STATIC FORCEINLINE
u32 sa_ipv6_header_class (SA_IPV6_HEADER *ip) {
    u32 v = SA_SWAP32 (ip->version_class_flow);
    v = (v & 0x0FF00000) >> 20;
    return v;
}

STATIC FORCEINLINE
void sa_ipv6_header_set_class (SA_IPV6_HEADER *ip, u32 class_value) {
    u32 v = SA_SWAP32 (ip->version_class_flow);
    v = (v & ~0x0FF00000) | ((class_value&0xFF) << 20);
    ip->version_class_flow = SA_SWAP32 (v);
}

STATIC FORCEINLINE
u32 sa_ipv6_header_flow (SA_IPV6_HEADER *ip) {
    u32 v = SA_SWAP32 (ip->version_class_flow);
    v = (v & 0xFFFFF);
    return v;
}

STATIC FORCEINLINE
void sa_ipv6_header_set_flow (SA_IPV6_HEADER *ip, u32 flow) {
    u32 v = SA_SWAP32 (ip->version_class_flow);
    v = (v & ~0xFFFFF) | (flow&0xFFFFF);
    ip->version_class_flow = SA_SWAP32 (v);
}

void sa_print_packet (
    u8* packet,
    u32 length,
    SA_OUTPUT *out);


SA_CPP_TRAILER
#endif // __sa_network_packet_h__

