/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_bits.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_bits.h,v 1.2 2006/02/25 23:26:53 menz Exp $
*/
#ifndef __sa_bits_h__
#define __sa_bits_h__

#include "sa_types.h"
#include "sa_assert.h"
SA_CPP_HEADER

// Bit flags
#define SA_BIT_SET(_pvalue_, _bit_)                (*(_pvalue_) |= (_bit_))
#define SA_BIT_CLR(_pvalue_, _bit_)                (*(_pvalue_) &= ~(_bit_))

#define SA_IS_BIT_SET(_value_, _bit_)             (((_value_) & (_bit_)) == (_bit_))
#define SA_IS_BIT_CLR(_value_, _bit_)             (!SA_IS_BIT_SET(_value_,_bit_))

// Checks for truncation of the value when assigning to a bitfield that
// may have fewer bits.  e.g. 
// struct { u32 hi : 1; } me;  
// SA_SET_BITFIELD(me.hi, 2);
#define SA_SET_BITFIELD(_bitfield_, _value_)  \
    (_bitfield_) = (_value_);                 \
    ASSERT ((_bitfield_) == (_value_))        \

STATIC
FORCEINLINE
u32 sa_bit_range (u32 value, u32 bit_offset, u32 num_bits) {
    ASSERT (bit_offset <= sizeof(value)*8);
    ASSERT (num_bits <= sizeof(value)*8);
    return (value >> bit_offset) & ((1UL << num_bits) - 1);
}

STATIC
FORCEINLINE
u64 sa_bit_range64 (u64 value, u32 bit_offset, u32 num_bits) {
    ASSERT (bit_offset <= sizeof(value)*8);
    ASSERT (num_bits <= sizeof(value)*8);
    return (value >> bit_offset) & ((1ULL << num_bits) - 1);
}

STATIC
FORCEINLINE
u32 sa_upper_bits (u32 value, u32 num_bits) {    
    ASSERT (num_bits <= sizeof(value)*8);
    return sa_bit_range (value, sizeof(value)*8 - num_bits, num_bits);
}

STATIC
FORCEINLINE
u64 sa_upper_bits64 (u64 value, u32 num_bits) {
    ASSERT (num_bits <= sizeof(value)*8);
    return sa_bit_range64 (value, sizeof(value)*8 - num_bits, num_bits);
}

STATIC
FORCEINLINE
u32 sa_lower_bits (u32 value, u32 num_bits) {
    ASSERT (num_bits <= sizeof(value)*8);
    return sa_bit_range (value, 0, num_bits);
}

STATIC
FORCEINLINE
u64 sa_lower_bits64 (u64 value, u32 num_bits) {
    ASSERT (num_bits <= sizeof(value)*8);
    return sa_bit_range64 (value, 0, num_bits);
}

SA_CPP_TRAILER    
#endif // __sa_bits_h__


