/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_types.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_types.h,v 1.11 2006/03/13 19:27:31 menz Exp $
*/
#ifndef __sa_types_h__
#define __sa_types_h__

#include "sa_platform.h"

SA_CPP_HEADER

// Common floating point types.
#ifndef SA_KERNEL
    typedef float  f32;
    typedef double f64;
#endif

#if defined(SA_MICROSOFT)

    // Unsigned integers
    typedef unsigned char      u8;
    typedef unsigned short     u16;
    typedef unsigned long      u32;
    typedef unsigned __int64   u64;

    // Signed integers.
    typedef signed char        i8;
    typedef signed short       i16;
    typedef signed long        i32;
    typedef signed __int64     i64;

    // Narrow and wide characters.
    typedef char               c8;
    typedef wchar_t            c16;

    // Native size_t for interfacing to libraries.
    typedef size_t SA_SIZE_T, *PSA_SIZE_T;


    // Platform dependent boolean, where an expression
    // like (x == y) results in this type without casting.
    //#if defined(SA_CPP)        
        //typedef bool boolean; // already defined as u8
    //#elif defined(SA_KERNEL)
    #if !defined(SA_WIN_USER)       
        typedef int boolean;        
    #elif !defined(__RPCNDR_H__)
        typedef int boolean;
    #endif

#elif defined(SA_DOS)

    // Unsigned integers
    typedef unsigned char      u8;
    typedef unsigned short     u16;
    typedef unsigned int       u32;
    typedef unsigned long long u64;

    // Signed integers.
    typedef signed char        i8;
    typedef signed short       i16;
    typedef signed long        i32;
    typedef signed long long   i64;

    // Narrow and wide characters.
    typedef char               c8;
    typedef u16                c16;

    // Native size_t for interfacing to libraries.
    typedef size_t SA_SIZE_T, *PSA_SIZE_T;

    // Platform dependent boolean, where an expression
    // like (x == y) results in this type without casting.
    #if defined(SA_CPP)
        typedef bool boolean;
    #else
        typedef int boolean;
    #endif

#elif defined(SA_LINUX)

    #if defined(SA_LINUX_USER)
        // Unsigned integers
        typedef unsigned char      u8;
        typedef unsigned short     u16;
        typedef unsigned int       u32;
        typedef unsigned long long u64;
    #endif
	
    // Signed integers.
    typedef signed char        i8;
    typedef signed short       i16;
    typedef signed int         i32;
    typedef signed long long   i64;

    // Narrow and wide characters.
    typedef char               c8;
    typedef u16                c16;

    // Native size_t for interfacing to libraries.
    typedef size_t SA_SIZE_T, *PSA_SIZE_T;

    // Platform dependent boolean, where an expression
    // like (x == y) results in this type without casting.
    #if defined(SA_CPP)
        typedef bool boolean;
    #else
        typedef int boolean;
    #endif

#endif



// Common pointer to a void notation.
typedef void* PVOID;

// Explicitly sized booleans -- useful for packing a struct while still 
// indicating that the value is treated as TRUE or FALSE.
typedef u8  bool8;
typedef u16 bool16;
typedef u32 bool32;

// Define boolean types if necessary.
#ifndef TRUE
    #define TRUE (1)
#endif
#ifndef FALSE
    #define FALSE (0)
#endif


// Physical address definition
typedef u64 SA_PHYSICAL_ADDRESS, *PSA_PHYSICAL_ADDRESS;

// Define NULL appropriately for C++ versus C.    
#if defined(SA_CPP)    
    #if !defined(NULL)
        #define NULL            0
    #endif

#else
    #if !defined(NULL)
        #define NULL            ((void *) 0)
    #endif
#endif


// Define function types.
#if defined(SA_MICROSOFT)
    
    #define SA_EXTERN_C     extern "C"     
    #define STATIC          static
    #define INLINE          __inline
    #define FORCEINLINE     __forceinline
    #define SA_CDECL        __cdecl
    #define SA_STDCALL      __stdcall
    #define STDCALL         __stdcall    
    #define SA_FASTCALL     __fastcall
    #ifndef FASTCALL
        #define FASTCALL    __fastcall
    #endif
    
    #define IN
    #define OUT
    #define INOUT

#else    

    #define SA_EXTERN_C extern "C"
  #ifdef INLINE
  #undef INLINE
  #endif
    #define INLINE      inline
    #define STATIC      static
    #define FORCEINLINE inline
    #define SA_STDCALL
    #define STDCALL
    #define SA_CDECL
    #define SA_FASTCALL
    #ifndef FASTCALL
        #define FASTCALL
    #endif
    #define IN
    #define OUT
    #define INOUT
#endif


SA_CPP_TRAILER

#endif // __sa_types_h__

