/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   chap.cpp

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_chap.c,v 1.3 2005/07/30 22:11:16 menz Exp $
*/

#include "sa_chap.h"
#include "sa_mem.h"
#include "sa_md5.h"
#include "sa_rand.h"
#include "sa_text.h"
#include "sa_time.h"

void
sa_chap_generate_challenge (PSA_CHAP_CONTEXT chap, u32 len)
{
    u8 *ptr = (u8*)chap->challenge;
    u32 i;

    u32 seed_a = (u32)sa_timestamp ();    
    u32 seed_b = SA_PTR_TO_U32 (chap);  // TODO - consider using current thread
    u32 seed_c = 0xcafebabe;

    sa_rand_seed (seed_a, seed_b, seed_c);

    chap->challengeLength = len;

    // Random 32-bit words
    for (i = 0; i<len/4; i++) {
        seed_c = sa_rand ();
        *ptr++ = (u8)(seed_c >> 16);
        *ptr++ = (u8)(seed_c >> 24);        
        *ptr++ = (u8)(seed_c >> 8);
        *ptr++ = (u8)(seed_c);
    }

    seed_c = sa_rand ();
    for (i = 0; i < (len & 3); i++) {
        *ptr++ = (u8)(seed_c >> (8*i));
    }
}


void
sa_chap_calculate_response (PSA_CHAP_CONTEXT chap, const c8 * secret)
{
    SA_MD5_CONTEXT ctx;
    u32 secretLen = sa_strlen (secret);

    sa_md5_init (&ctx);        
    sa_md5_update (&ctx, &chap->id, 1);                           // ID byte
	sa_md5_update (&ctx, (const u8*)secret, secretLen);      // shared secret
	sa_md5_update (&ctx, chap->challenge, chap->challengeLength); // challenge

    chap->responseLength = 16;
	sa_md5_final (chap->response, &ctx);

    return;
}

#if !defined(SA_OMIT_TEST_CODE)

    static const c8* sa_test_challenge =
        "0xd7511ab57b7f74916b20833725003b1f73b5c8aee0608ad80854357ea3824eef03dfa35bced2e02"
        "dc7e6b9e5f184f4279e9aaede400a4341612b20e49f88372a51460d910e15856a092a2f0ffde967b"
        "541db4a1b079fc65d2b86f43de4fe483406e27873ea2b026bc79a5341f49cb1ab242330dd111136a"
        "64e1ad297c0df76c7de6e4a2dae866a023da494d766b8d4e62335d7c07213dd321bd73fc959b21c2"
        "a25250d4035da9b4a84118a22cadf9110971e21347635ac72148a3cb3cfce217755f6c644a71859b"
        "7f2996e93d0efe49a99cebe8163c348a934ef64f946799da40a8a0ad3d2aaa5ff459784e94a922b8"
        "73ee19ee06364c0d079f1df5f5c79e3200a5cd6df72218eb4d5cf9878d5f11194b20b25e1605f9f7"
        "c790e37210f9f415eca0b66fcb188e8e692ebc221d0103934aafdf07fb73d7cb5ca1151741dfc1f7"
        "690cb9d4573c8241ece0ff9ecac26776ad21dfbd782ca06adf117ffeca148218707f2548b1f223a4"
        "8b3dd3c549a822a9255438351e4ae0ff90d8c5411ed19c68fda15c2a1d987542c08baa99f4253c95"
        "828677d924f8959fe71f53ae9c0d2fd854856d6ddf4827f45431bb23c4b81615205fab31501b1611"
        "48291cb4820e20ce49f62818076439474c9cf09e85a79a4fe48f1552627770515ba23d02658a79c0"
        "b33cc79a06d4ae07663a7b17ef71165a6385308ec7b38e8ada0e5a1e6dc6618af1bca0a2838b44c4"
        "c18cd74df807d6de8c88e15450b489bff232a368c17bec96796b846f69baa6804aa72793cb0624f1"
        "d112401e686ed739b67cc6726079f719cd9d85960296175fcfab328ad284a75ee28a6e32eb891ec0"
        "1515942469e2923e8761620eba2828fe356c47f82036e0d1131a9fbb7f7af87ed1ce4c4a832c7118"
        "9da374c0126e4a734f8e95012271cd208f44edddc93b99c2e39fcc8a14fcca79b26261f3950c1162"
        "881a81e14f4f3669f1a3c0920e97d7dbbeefce31982c6079ee5a7333f6706fb5dd2e"
        "6baaa2843c7715df8ca3ddc4fb41095af7c829065b328e356232ee70730408455df5a62307b8ab3e"
        "f34311366114f0d0db6b5dd27cb64042067b2e30fcad73c9310f2ffa12e4184ccbaa7ec0df804532"
        "fb5c1d88875660af36f7e26d41dfc05d2e8ace9dd941ee1d9989a88fd6789cef4a15b80fbfcceb7f"
        "36cdf8e1586d6453d1ddd92879bb8c18a5fb8612b26961294a5608024241ef778baad7290350f1d4"
        "90c8ef91f73e7a43a5a28f6f5ee776c331dfba9cd609d0d4dd3bd639ca7c5909edd6352d337630ee"
        "b9fb9feef51eacdd8f87c582afb2d57264f8dd75b696c43301594f2b49006acfdf509bf84839a361"
        "74651bd8445b4bae2e292cc46333672ef76b16c18842dcff69ecebfecdf240bca0537e6b46deb1a7"
        "2497b2d915de2e31b602e039e4222686b0a5584f7c74f4476a6d3cb4520af";

    static const c8* sa_test_response = "0xAB656533B5F0AE541C5B1A92889AD8DE";


    typedef struct _SA_CHAP_TEST {
        SA_CHAP_CONTEXT ctx;        
        c8 responseStr[SA_MAX_CHAP_RESPONSE*2 + 3];
    } SA_CHAP_TEST, *PSA_CHAP_TEST;

    boolean
    sa_chap_test_internal (const c8* challenge, 
                           const c8 *secret, 
                           const c8 *expectedResponse, 
                           u8 id) 
    {
        SA_CHAP_TEST *test;
        boolean pass;

        // +3 holds "0x" and null
        test = (SA_CHAP_TEST*)sa_malloc (sizeof (SA_CHAP_TEST), SA_MALLOC_NORMAL, 0);
        if (!test) {
            return FALSE;
        }
        sa_memset (test, 0, sizeof (*test));

        test->ctx.id = id;
        test->ctx.challengeLength = sa_text_hex_str_to_byte_array (challenge, 
                                                                   sa_strlen (challenge), 
                                                                   test->ctx.challenge, 
                                                                   sizeof (test->ctx.challenge));
        ASSERT (test->ctx.challengeLength > 0);
        
        // Compute response
        sa_chap_calculate_response (&test->ctx, secret);

        // Convert response to text
        sa_text_byte_array_to_hex_str (test->responseStr, 
                                       sizeof (test->responseStr), 
                                       test->ctx.response, 
                                       test->ctx.responseLength);

        // Compare expected result
        pass = (sa_memcmp (test->responseStr, expectedResponse, sizeof (test->responseStr)) == 0);
        ASSERT (pass);        

        sa_free (test);
        return pass;
    }

    boolean
    sa_chap_test ()
    {
        return sa_chap_test_internal (sa_test_challenge, "george.enz.com", sa_test_response, 1);
    }

#endif

