/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

// Combination of 3 tausworthe generators -- assumes 32-bit integers
//
// Source: Pierre L'Ecuyer, "Maximally Equidistributed Combined Tausworthe
//   Generators". Mathematics of Computation, vol.65, no.213(1996), pp203--213.
//
// Special thanks to Joushua Grass (of Microsoft) for pointing out an error
//   in the initialization routine "rand_seed()" 


#include "sa_rand.h"

typedef struct _SA_RAND_GLOBALS {

    u32 s1;
    u32 s2;
    u32 s3;

    u32 mask1;
    u32 mask2;
    u32 mask3;

    u32 shft1;
    u32 shft2;
    u32 shft3;

    u32 k1;
    u32 k2;
    u32 k3;

    u32 q1;
    u32 q2;
    u32 q3;

    u32 p1;
    u32 p2;
    u32 p3;

} SA_RAND_GLOBALS, *PSA_RAND_GLOBALS;

// The seeds.
static SA_RAND_GLOBALS sa_rand_globals  = 
{
    390451501, // s1
    613566701, // s2
    858993401, // s3

    0, // mask1
    0, // mask2
    0, // mask3

    0, // shft1
    0, // shft2
    0, // shft3

    31, // k1
    29, // k2
    28, // k3

    13, // q1
    2,  // q2
    3,  // q3

    12, // p1
    4,  // p2
    17  // p3
};

// This is another set of parameters:
// static int q1=3, q2=2, q3=13, p1=20, p2=16, p3=7;

u32 sa_rand ()
{   
   u32 b;
  
   b  = ((sa_rand_globals.s1 << sa_rand_globals.q1)^sa_rand_globals.s1) >> sa_rand_globals.shft1;
   sa_rand_globals.s1 = ((sa_rand_globals.s1 & sa_rand_globals.mask1) << sa_rand_globals.p1) ^ b;
   b  = ((sa_rand_globals.s2 << sa_rand_globals.q2) ^ sa_rand_globals.s2) >> sa_rand_globals.shft2;
   sa_rand_globals.s2 = ((sa_rand_globals.s2 & sa_rand_globals.mask2) << sa_rand_globals.p2) ^ b;
   b  = ((sa_rand_globals.s3 << sa_rand_globals.q3) ^ sa_rand_globals.s3) >> sa_rand_globals.shft3;
   sa_rand_globals.s3 = ((sa_rand_globals.s3 & sa_rand_globals.mask3) << sa_rand_globals.p3) ^ b;
   return (sa_rand_globals.s1 ^ sa_rand_globals.s2 ^ sa_rand_globals.s3);
}

void sa_rand_seed( u32 seed_a, u32  seed_b, u32 seed_c )
{
   const u32 x = 4294967295UL;

   sa_rand_globals.shft1 = sa_rand_globals.k1-sa_rand_globals.p1;
   sa_rand_globals.shft2 = sa_rand_globals.k2-sa_rand_globals.p2;
   sa_rand_globals.shft3 = sa_rand_globals.k3-sa_rand_globals.p3;
   sa_rand_globals.mask1 = x << (32-sa_rand_globals.k1);
   sa_rand_globals.mask2 = x << (32-sa_rand_globals.k2);
   sa_rand_globals.mask3 = x << (32-sa_rand_globals.k3);
   if (seed_a > (1UL << (32-sa_rand_globals.k1))) sa_rand_globals.s1 = seed_a;
   if (seed_b > (1UL << (32-sa_rand_globals.k2))) sa_rand_globals.s2 = seed_b;
   if (seed_c > (1UL << (32-sa_rand_globals.k3))) sa_rand_globals.s3 = seed_c;
   sa_rand();
}



